"""
Colorful bash output.
"""

import os
from sys import stdout, stderr, platform
from time import strftime

from .notifier import Notifier


DEBUG_MODE = False


# Terminal color functions

WINDOWS = (platform == 'win32')

COLOR_CODE = {
    'WHI':"\033[1;37m",
    'LGY':"\033[0;37m",
    'GRY':"\033[1;30m",
    'BLK':"\033[0;30m",
    'RED':"\033[0;31m",
    'LRD':"\033[1;31m",
    'GRN':"\033[0;32m",
    'LGN':"\033[1;32m",
    'BRN':"\033[0;33m",
    'YLW':"\033[1;33m",
    'BLU':"\033[0;34m",
    'LBL':"\033[1;34m",
    'PUR':"\033[0;35m",
    'PNK':"\033[1;35m",
    'CYN':"\033[0;36m",
    'LCN':"\033[1;36m",
    'UNC':"\033[0m"
}

def _color_format(c, text):
    try:
        code = COLOR_CODE[c]
    except KeyError:
        code = COLOR_CODE['BLU']
    return code + text + COLOR_CODE['UNC']

def white(s):       return _color_format('WHI', s)
def lightgray(s):   return _color_format('LGY', s)
def gray(s):        return _color_format('GRY', s)
def black(s):       return _color_format('BLK', s)
def red(s):         return _color_format('RED', s)
def lightred(s):    return _color_format('LRD', s)
def green(s):       return _color_format('GRN', s)
def lightgreen(s):  return _color_format('LGN', s)
def brown(s):       return _color_format('BRN', s)
def yellow(s):      return _color_format('YLW', s)
def blue(s):        return _color_format('BLU', s)
def lightblue(s):   return _color_format('LBL', s)
def purple(s):      return _color_format('PUR', s)
def pink(s):        return _color_format('PNK', s)
def cyan(s):        return _color_format('CYN', s)
def lightcyan(s):   return _color_format('LCN', s)

COL_FUNC = {
    'white': white,
    'lightgray': lightgray,
    'gray': gray,
    'black': black,
    'red': red,
    'lightred': lightred,
    'green': green,
    'lightgreen': lightgreen,
    'brown': brown,
    'yellow': yellow,
    'blue': blue,
    'lightblue': lightblue,
    'purple': purple,
    'pink': pink,
    'cyan': cyan,
    'lightcyan': lightcyan,
}
COLORS = list(COL_FUNC.keys())
COLORS.sort()

def show_colors():
    for c in COLORS:
        print(COL_FUNC[c]('this is {}'.format(c)))

def first_cap(s):
    if len(s):
        return s[0].upper() + s[1:]
    return ''


# Main ConsolePrinter class

class ConsolePrinter(object):

    """
    A callable console printer with color, log files, and notifications.
    """

    def __init__(self, prefix='Message', prefix_color='cyan',
        message_color='lightgray', outputfile=None, timestamp=True):
        """Create a callable console color printing object.

        Keyword arguments:
        prefix -- set the default prefix string for all messages
        prefix_color, message_color -- color names for prefix and message text
        outputfile -- optional file name where messages should be echo'd
        timestamp -- print timestamps in `outputfile` if specified
        """
        self._prefix = prefix

        if WINDOWS:
            self._pref = self._msgf = str
        else:
            self._pref = COL_FUNC[prefix_color]
            self._msgf = COL_FUNC[message_color]

        self._outputfile = self._fd = None
        self.set_outputfile(outputfile)
        self.set_timestamps(timestamp)

        self._notifier = None

    def _isopen(self):
        return self._fd is not None and not self._fd.closed

    def openfile(self, quiet=False):
        """Open the currently set output file."""
        if self._outputfile is None:
            self('No output file has been set', error=True)
            return
        if self._isopen():
            return
        try:
            self._fd = open(self._outputfile, 'a', 1)
        except IOError:
            self._outputfile = self._fd = None
            self('IOError: Could not open: %s' % self._outputfile, error=True)
        else:
            if not quiet:
                self('Opened: %s' % self._outputfile)

    def set_outputfile(self, newfile):
        """Set the path to a new output file for logging messages."""
        if newfile == self._outputfile and self._isopen():
            return
        self.closefile()
        self._outputfile = newfile
        self._fd = None
        if self._outputfile is not None:
            self.openfile()

    def set_timestamps(self, active):
        """Set timestamping on/off for log files."""
        self._timestamp = bool(active)

    def closefile(self, quiet=False):
        """Close the current output file."""
        if not self._isopen():
            return
        self._fd.close()
        self._fd = None
        if not quiet:
            self('Closed: %s' % self._outputfile)

    def removefile(self):
        """Delete the current output file."""
        self.closefile()
        os.unlink(self._outputfile)
        self('Deleted: %s' % self._outputfile)

    def hline(self, ch='_', length=40):
        """Print a horizontal rule line."""
        self(ch * length, hideprefix=True)

    def __call__(self, *msg, **fmt):
        """Display a message with color prefix and multi-line indenting.

        Arguments:
        *msg -- a required string message with optional substitutions

        Keyword arguments:
        prefix -- override default prefix string
        hideprefix -- make prefix invisible but preserve indent
        debug -- specify where this is a debug message
        error -- specify whether this is an error message
        popup -- display the message with a system popup notification
        **fmt -- remaining kwargs provide formating substitutions
        """
        msg, args = msg[0], msg[1:]
        prefix = fmt.pop('prefix', self._prefix)
        hideprefix = fmt.pop('hideprefix', False)
        debug = fmt.pop('debug', False)
        error = fmt.pop('error', False)
        popup = fmt.pop('popup', False)
        msg = msg.format(*args, **fmt)

        # Construct the actual prefix to display
        if debug:
            prefix = 'debug'
        if hideprefix:
            pre = ' ' * (len(prefix) + 2)
        else:
            pre = prefix + ': '

        # Make the output a little more civilized
        if not debug:
            msg = first_cap(msg.strip())
            pre = first_cap(pre)

        # Console color print with prefix and indentation
        if error:
            pref = self._pref
            msgf = red
            console = stderr
        elif debug:
            pref = msgf = lightcyan
            console = stderr
        else:
            pref = self._pref
            msgf = self._msgf
            console = stdout

        pre_len = len(pre)
        lines = msg.split('\n')
        console.write(pref(pre) + msgf(lines[0].rstrip()) + '\n')
        for line in lines[1:]:
            console.write(' ' * pre_len + msgf(line.rstrip()) + '\n')
        console.flush()

        # Timestamped file print if available
        if self._isopen():
            if self._timestamp:
                fmt = '%H:%M:%S %m-%d-%y'
                self._fd.write('[ %s ]  ' % strftime(fmt))
            if error:
                self._fd.write('Error: ')
            self._fd.write(msg + '\n')
            self._fd.flush()

        # Request a system popup notification
        if popup:
            if self._notifier is None:
                self._notifier = Notifier(prog=self._prefix, echo=False)
            if error:
                self._notifier.notify(msg, pre[:-2], "Error")
            else:
                self._notifier.notify(msg, pre[:-2])

    def printf(self, s, color=None):
        """Raw flushed color output to the console."""
        colf = color is None and self._pref or COL_FUNC[color]
        if WINDOWS:
            stdout.write(s)
        else:
            stdout.write(colf(s))
        stdout.flush()

    def box(self, filled=True, color=None):
        """Draw a Unicode box glyph to the console."""
        self.printf('\u25a1\u25a0'[filled], color=color)

    def newline(self):
        """Insert a newline."""
        self.printf('\n')


# Convenience functions

Logger = ConsolePrinter(prefix='log')

def log(*args, **kwargs):
    Logger(*args, **kwargs)

def debug(*args, **kwargs):
    if DEBUG_MODE:
        kwargs.update(debug=True)
        Logger(*args, **kwargs)

def printf(s, c='green'):
    Logger.printf(s, color=c)
