"""
Write/clone tables/arrays/groups to/between HDF files.
"""

import os

import tables as tb

from pouty import log


class OverwriteDisallowed(Exception):
    pass


def new_table(f, where, name, description, force=False, **kwargs):
    """Add `force` keyword to tables.File.create_table().

    Arguments:
    force -- force erasure of existing table without asking
    h5file -- alternate file in which to create the array

    Returns `tables.Table` node.
    """
    try:
        table = f.get_node(where, name=name)
    except tb.NoSuchNodeError:
        pass
    else:
        if force:
            log('new_table: Erasing {} table', table._v_pathname)
            do_erase = 'y'
        else:
            do_erase = input(
                'new_table: Erase %s table? (y/N) ' % table._v_pathname)
        if not do_erase.lower().strip().startswith('y'):
            raise OverwriteDisallowed('%s table already exists' %
                    table._v_pathname)
        f.remove_node(table)

    kwargs.update(description=description)
    return f.create_table(where, name, **kwargs)

def new_array(f, where, name, x, overwrite=True, **kwargs):
    """Add `overwrite` keyword to tables.File.create_array().

    Arguments:
    overwrite -- automatically remove array if it already exists
    h5file -- alternate file in which to create the array

    Returns `tables.Array` node.
    """
    try:
        array = f.get_node(where, name=name)
    except tb.NoSuchNodeError:
        pass
    else:
        if not overwrite:
            raise OverwriteDisallowed('%s array already exists' %
                    array._v_pathname)
        f.remove_node(array)

    return f.create_array(where, name, obj=x, **kwargs)

def new_dataframe(pdf, where, name, x, overwrite=True, **kwargs):
    """Add `overwrite` keyword to tables.File.create_array().

    Arguments:
    pdf -- pandas file store
    overwrite -- automatically remove array if it already exists
    h5file -- alternate file in which to create the array

    Returns `tables.Group` node where the dataframe is stored.
    """
    key = os.path.join(where, name)
    try:
        df = pdf.get_node(key)
    except KeyError:
        pass
    else:
        if not overwrite:
            raise OverwriteDisallowed('%s dataframe already exists' %
                    df._v_pathname)

    pdf[key] = x
    grp = pdf.get_node(key)
    grp._v_attrs['title'] = kwargs.pop('title', 'Dataframe')
    return grp

def new_group(f, where, name, **kwargs):
    """Enforce `createparents=True` in creating a new group."""
    kwargs.update(createparents=True)

    try:
        group = f.create_group(where, name, **kwargs)
    except tb.NodeError:
        group = f.get_node(where, name=name)
        log('{} already exists', group._v_pathname, error=True)
    return group

def clone_node(node, destfile, parent=None, name=None):
    """Clone a node from one file to another file."""
    name = node._v_name if name is None else name
    parent = node._v_parent._v_pathname if parent is None else parent
    destpath = os.path.join(parent, name)

    if isinstance(node, tb.Array):
        log('Cloning array: {} to {}', node._v_pathname, destpath)
        arr = clone = new_array(destfile, parent, name, node.read(),
            createparents=True, title=node.title)
        for k in node._v_attrs._v_attrnames:
            setattr(arr._v_attrs, k, node._v_attrs[k])
    elif isinstance(node, tb.Table):
        log('Cloning table: {} to {}', node._v_pathname, destpath)
        tbl = clone = new_table(destfile, parent, name, node.description,
            createparents=True, title=node.title)
        row = tbl.row
        for record in node.iterrows():
            for col in node.colnames:
                row[col] = record[col]
            row.append()
        for k in node._v_attrs._v_attrnames:
            setattr(tbl._v_attrs, k, node._v_attrs[k])
        tbl.flush()

    return clone

def clone_subtree(srcfile, tree, destfile, destroot='/', classname=None):
    """Clone a subtree within one file into a different (possibly new) file."""
    if type(srcfile) is str:
        srcfile = tb.open_file(srcfile, 'r')

    if not srcfile.isopen:
        raise IOError('Source file is not open')

    if type(destfile) is str:
        destfile = tb.open_file(destfile, 'a')

    if not (destfile.isopen and destfile._iswritable()):
        raise IOError('Destination file is not open and writable')

    try:
        srctree = srcfile.get_node(tree)
    except tb.NoSuchNodeError as e:
        log('Source group does not exist: {}', tree, error=True)
        return

    for node in srcfile.walk_nodes(srctree, classname=classname):
        if isinstance(node, tb.Group):
            continue
        destpath = os.path.join(destroot, node._v_pathname[1:])
        parent, name = os.path.split(destpath)
        clone_node(node, destfile, parent=parent, name=name)

    log('Finished cloning subtree: {}', srctree._v_pathname)
    return destfile.get_node(destroot)
