"""
Collection of useful decorators.
"""

from decorator import decorator

import numpy as np

from pouty import debug

from .arrays import datahash


class lazyprop(object):

    """
    Method decorator for one-time conversion of result to class attribute

    From: http://stackoverflow.com/a/6849299
    """

    def __init__(self, func):
        self.func = func
        self.func_name = func.__name__
        debug('lazyprop: initializing {}', self.func_name)

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.func(obj)
        setattr(obj, self.func_name, value)
        debug('lazyprop: setting {} on {}', self.func_name, obj)
        return value


@decorator
def memoize(f, *args, **kwds):
    if not hasattr(f, 'cache'):
        debug('memoize: creating cache on {}', f.__name__)
        f.cache = {}
    if kwds:
        key = args, frozenset(iter(kwds.items()))
    else:
        key = args
    cache = f.cache
    if key in cache:
        debug('memoize: retrieving {} from {}', key, f.__name__)
        return cache[key]
    cache[key] = result = f(*args, **kwds)
    debug('memoize: caching {} in {}', key, f.__name__)
    return result

@decorator
def datamemoize(f, *args, **kwds):
    """Hash any array arguments."""
    if not hasattr(f, 'cache'):
        debug('datamemoize: creating cache on {}', f.__name__)
        f.cache = {}
    arrays = []
    notarrays = []
    for arg in args:
        if isinstance(arg, np.ndarray):
            arrays.append(arg)
        else:
            notarrays.append(arg)
    debug('datamemoize: hashing {} arrays', len(arrays))
    arrayhash = datahash(*arrays)
    if kwds:
        key = arrayhash, tuple(notarrays), frozenset(iter(kwds.items()))
    else:
        key = arrayhash, tuple(notarrays)
    cache = f.cache
    if key in cache:
        debug('datamemoize: retrieving {} from {}', key, f.__name__)
        return cache[key]
    cache[key] = result = f(*args, **kwds)
    debug('datamemoize: caching {} in {}', key, f.__name__)
    return result
