#!/usr/bin/env python

"""
Draw a network graph mapping recording sites to anatomical area.
"""

import os
import sys
from collections import Counter

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from pouty import log

from spc import PROJECT_ROOT, store


def main(levels):
    # Read the site and area data from the h5 file
    f = store.get(readonly=True)
    recs = f.root.recordings
    sites = list(map(lambda s: s.decode(), recs.col('site')))
    areas = recs.col('area')

    # Calculate frequency level cutoffs for each site/area pair
    pairs = Counter(zip(sites, areas))
    cutoffs = np.percentile(list(pairs.values()), np.linspace(0, 100, levels+1))
    edge_groups = [[] for _ in range(levels)]

    # Create the graph
    G = nx.Graph()

    for (site, area), count in pairs.items():
        G.add_edge(site, area, weight=count)
        for i in range(levels):
            if count <= cutoffs[i+1]:
                edge_groups[i].append((site, area))
                break

    # Create node lists
    area_nodes = list(np.unique(areas))
    site_nodes = list(np.unique(sites))

    # Layout and draw the graph nodes
    plt.figure(figsize=(12,12))
    pos = nx.shell_layout(G, nlist=[area_nodes, site_nodes], scale=1.6)
    nx.draw_networkx_edges(G, pos)

    for level in range(levels):
        nx.draw_networkx_edges(G, pos, edgelist=edge_groups[level],
                width=2*(level+1), alpha=0.3)

    nx.draw_networkx_labels(G, pos, font_size=9, font_color='b')

    # Save the image
    fn = os.path.join(PROJECT_ROOT, 'site-area-network-%d.png' % levels)
    plt.axis('off')
    plt.savefig(fn, dpi=150)
    log('Saved graph: {}', fn, prefix='AreaSiteNetwork')

    store.close()
    return 0


if __name__ == "__main__":
    n_classes = 5
    if len(sys.argv) > 1:
        n_classes = int(sys.argv[1])

    sys.exit(main(n_classes))
