#!/usr/bin/env python

"""
Import cell data from each's rat cellstructs file.
"""

import os
import sys
from collections import namedtuple

import tables as tb
import numpy as np
import scipy.io as sio

from lco import store
from lco.output import log
from lco.lib.data import new_table, recordings_desc
from lco.lib.datasets import find_rat_paths
from lco.lib.parsers import parse_ttc
from lco.lib.anatomy import area_name

CELL_STRUCTS_FN = 'cellstructs.mat'
CellStruct = namedtuple("CellStruct", "subfolder ttc cell_id site area_code")


def parse_cellstruct(cell):
    """Extract cell recording metadata from cell struct entry."""
    return CellStruct(*(str(val[0]) for val in cell[1:]))

def main():
    """Traverse cell structs files to generate a table of cell metadata."""
    f = store.get(readonly=False)
    sessions_table = f.root.sessions
    rats = np.unique(sessions_table.col('rat'))
    rat_paths = find_rat_paths()

    # Create recordings table
    recordings_table = new_table('/', 'recordings', recordings_desc,
            title='Cell-session recordings')
    rec = recordings_table.row

    # Cache for the cluster data
    ttc_cache = {}

    i_cell = 0
    for session in sessions_table.iterrows():
        rat = session['rat']
        csfile = os.path.join(rat_paths[rat], CELL_STRUCTS_FN)

        # Retrieve cellstructs data from cache
        if rat in ttc_cache:
            ttc = ttc_cache[rat]
        else:
            log('Loading {}', csfile)
            cellstructs = sio.loadmat(csfile)
            ttc = ttc_cache[rat] = cellstructs['ttc']
        N_cells = len(ttc)

        for i, cellinfo in enumerate(ttc):
            cell = parse_cellstruct(cellinfo)
            if not session['path'].endswith(cell.subfolder):
                continue

            log('Saving: Rat {}, Day {}, Cell {} [#{}]', rat, session['day'],
                    i+1, i_cell)
            tetrode, cluster = parse_ttc(cell.ttc)

            # Write row to recordings table
            rec['id'] = i_cell
            rec['session_id'] = session['id']
            rec['tetrode'] = tetrode
            rec['cluster'] = cluster
            rec['ttc'] = cell.ttc
            rec['cell_id'] = cell.cell_id
            rec['site'] = cell.site
            rec['area'] = int(cell.area_code)
            rec.append()

            i_cell += 1

        recordings_table.flush()
    store.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())
