#!/usr/bin/env python
# coding: utf-8

"""
Import LFP theta-phase data for each recording session into the h5 file.
"""

import os
import sys

import numpy as np
import scipy.io as sio
import tables as tb

from lco import store
from lco.output import log
from lco.lib.data import new_array


def load_file(session, stem):
    path = os.path.join(session['path'], '%s.mat' % stem)
    try:
        data = sio.loadmat(path)
    except ValueError as e:
        log('Failed to load {}: {}', path, e.message, error=True)
        return None
    return data


def main(overwrite=False):
    """Load LFP theta-phase data from LFPphase.mat files."""
    f = store.get(readonly=False)
    sessions = f.root.sessions

    for session in sessions.iterrows():
        group = os.path.join(session['group'], 'LFP')
        if not overwrite:
            try:
                node = f.get_node(group, 'TS')
            except tb.NoSuchNodeError:
                pass
            else:
                continue

        log('Loading LFP data: Rat {}, Day {}, {}',
            session['rat'], session['day'], session['comment'].title())

        phase = load_file(session, 'LFPphase')
        if phase is None:
            continue

        phase = {k:phase[v].squeeze() for k,v in [
                ('TS', 'CSCts'), ('thetaphase', 'Aphase')
            ]}

        log('→ Importing {} samples...', phase['TS'].size)
        for key in phase:
            arr = new_array(group, key, phase[key], createparents=True)

        if session['id'] % 5 == 0:
            store.flush()

    store.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())
