#!/usr/bin/env python

"""
Import cell data from each's rat cellstructs file.
"""

import os
import sys

import numpy as np
import scipy.io as sio

from lco import store
from lco.output import log
from lco.lib.data import new_array

USE_RAWSPIKES = True
speed_keys = ('still', 'slow', 'fast')


def main():
    """Traverse spike train files and import data into the h5 file."""
    f = store.get(readonly=False)
    sessions = f.root.sessions
    recs = f.root.recordings

    spikes_fn = 'rawspikes.mat' if USE_RAWSPIKES else 'allspikes.mat'

    for session in sessions.iterrows():
        log('Processing: Rat {}, Day {}', session['rat'], session['day'])
        spike_file = os.path.join(session['path'], spikes_fn)
        spikes = sio.loadmat(spike_file)
        group = os.path.join(session['group'], 'spikes')

        for cell in recs.where('session_id==%d' % session['id']):
            ttc = cell['ttc']

            totspks = 0
            if USE_RAWSPIKES:
                log('[{}] Importing raw spikes: {} to {}', cell['id'], ttc,
                        group)
                if ttc not in spikes:
                    log('Missing spike data for {}', ttc, error=True)
                    spikes[ttc] = np.array([])

                data = np.atleast_1d(spikes[ttc].squeeze())
                for speed in speed_keys:
                    cell['%s_spikes' % speed] = 0  # no pre-filtered spikes
                totspks = data.size

                arr = new_array(group, ttc, data, createparents=True)
            else:
                log('[{}] Importing speed-filtered spikes: {} to {}',
                        cell['id'], ttc, group)
                for speed in speed_keys:
                    key = '{speed}_{ttc}'.format(**locals())
                    if key not in spikes:
                        log('Missing spike data for {}', key, error=True)
                        spikes[key] = np.array([])

                    data = np.atleast_1d(spikes[key].squeeze())
                    cell['%s_spikes' % speed] = data.size
                    totspks += data.size

                    name = '{ttc}_{speed}'.format(**locals())
                    arr = new_array(group, name, data, createparents=True)

            cell['missing'] = (totspks == 0)
            cell.update()

        if session['id'] % 5 == 0:
            f.flush()

    store.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())
