#!/usr/bin/env python
"""
Render the data from the model simulation.
"""

import os

import numpy as np
from numpy import pi as PI
import matplotlib.pyplot as plt

from lco import PROJECT_ROOT, log
from lco.io import store
from lco.lib import motion, cluster, time, parsers
from lco.lib.session import RecordingSession
from lco.tools import maps, kernels, path, spike, adaptive
from lco.lib.data import Arenas


############## Set up output directories ###############

posterdir = os.path.join(PROJECT_ROOT, 'poster')
datadir = '/Users/joe/source/lco/rate_model/data'


############## Plotting state and functions ###############

cur = 0
f = None
axlist = None
interactive = False

def new_figure(rows, cols, figsize):
    global f, axlist, cur
    cur = 0
    f, axlist = plt.subplots(nrows=rows, ncols=cols, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        raise ValueError, 'out of axis (cur == %d)' % cur
    ax = axlist[cur]
    cur += 1
    return ax

def save_and_close(stem='misc/panel', pad=0.4, dpi=300):
    plt.tight_layout(pad=pad)
    stem = os.path.join(posterdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = path.uniquify(stem, ext='png', fmt='%s-%02d')
    plt.savefig(fn, dpi=dpi)
    log('Saved: {}', fn)
    if not interactive:
        plt.close()

plt.ioff()


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(40.0, 83.0, title, size='small', ha='center', va='bottom')

def plot_spatial_map(M, label='', **kwargs):
    ax = next_axis()
    im = maps.plot_map(M, ax=ax, **kwargs)
    quicktitle(ax, label)

def load(name):
    fn = os.path.join(datadir, '%s-trace.npy' % name)
    arr = np.load(fn)
    log('{} : {}', name, str(arr.shape))
    return arr


############## Script for the generation of many maps ###############

X = load('X')
Y = load('Y')
X = 100 * (X - X.min()) / (X.max() - X.min())
Y = 100 * (Y - Y.min()) / (Y.max() - Y.min())

BVC = load('BVC')
f_LCO = load('LCO_f')
phase_LCO = load('LCO_phase')
LVC = load('LVC')
W_VL = load('W_VL')
VCO = load('VCO')

VCO_nonoise = load('VCO-sigma0.00-alpha0.00')
VCO_noise = load('VCO-sigma0.02-alpha0.00')
VCO_feednack = load('VCO-sigma0.02-alpha0.12')
grids = load('grids')

session = RecordingSession.get(77)
averager_map = adaptive.AdaptiveAveragerMap(session)
phase_map = adaptive.AdaptivePhasemap(session)

np.random.seed(12345)

mapargs = dict(mask_color='0.9', cmin=0.0)


# BVC responses

new_figure(1, 8, (14, 2.5))
ix = np.random.permutation(BVC.shape[0])[:8]

for n, i in enumerate(ix):
    log('Computing BVC {}...', i)
    R = averager_map(X, Y, BVC[i])
    plot_spatial_map(R, label='BVC %d' % (n+1), **mapargs)

save_and_close('model-render/bvc-examples')


# LVC responses

new_figure(1, 8, (14, 2.5))
ix = np.random.permutation(LVC.shape[0])[:8]

for n, i in enumerate(ix):
    log('Computing LVC {}...', i)
    R = averager_map(X, Y, LVC[i])
    plot_spatial_map(R, label='LVC %d' % (n+1), **mapargs)

save_and_close('model-render/lvc-examples')


# LCO responses

new_figure(2, 8, (14, 4.5))
ix = np.random.permutation(f_LCO.shape[0])[:8]

pmapargs = mapargs.copy()
pmapargs.update(cmin=-PI, cmax=PI, cmap='hsv')

for n, i in enumerate(ix):
    log('Computing LCO {}...', i)
    R = averager_map(X, Y, f_LCO[i])
    P = phase_map(X, Y, phase_LCO[i])
    plot_spatial_map(R, label='LCO %d - Rate' % (n+1), **mapargs)
    plot_spatial_map(P[0], label='LCO %d - Phase' % (n+1), **pmapargs)

save_and_close('model-render/lco-pairs')


# LCO inputs and VCO responses

log('Computing LCO inputs to VCOs...')
I_LCO = np.dot(W_VL.T, f_LCO)

new_figure(3, 3, (6.75, 6.75))

for i in xrange(9):
    log('Computing I_LCO {}...', i)
    R = averager_map(X, Y, I_LCO[i])
    plot_spatial_map(R, label='I_LCO %d' % (i+1), **mapargs)

save_and_close('model-render/lco-input', pad=0.7)

new_figure(3, 3, (6.75, 6.75))

for i in xrange(9):
    log('Computing VCO {}...', i)
    R = averager_map(X, Y, VCO[i])
    plot_spatial_map(R, label='VCO %d' % (i+1), **mapargs)

save_and_close('model-render/vco-layer-exact', pad=0.7)


# Noise and feedback layers

conds = [(VCO_nonoise, 'nonoise'), (VCO_noise, 'noise'),
        (VCO_feednack, 'feedback')]

for VCOx, label in conds:
    new_figure(3, 3, (6.75, 6.75))

    for i in xrange(9):
        log('Computing VCO {} {}...', label, i)
        R = averager_map(X, Y, VCOx[i])
        plot_spatial_map(R, label='VCO %d' % (i+1), **mapargs)

    save_and_close('model-render/vco-%s' % label, pad=0.7)


# Grid cell layer

new_figure(3, 3, (6.75, 6.75))

for i in xrange(9):
    log('Computing grids {}...', i)
    R = averager_map(X, Y, grids[i])
    plot_spatial_map(R, label='Grid %d' % (i+1), **mapargs)

save_and_close('model-render/grid-cells', pad=0.7)


############## Draw anything left back to the screen ###############

plt.ion()
if interactive:
    plt.draw()
    plt.show()

