#!/usr/bin/env python
"""
Demonstration of phase stability over time.
"""

import os

import numpy as np
from numpy import pi as PI
import matplotlib.pyplot as plt

from lco import PROJECT_ROOT, log
from lco.io import store
from lco.lib import motion, cluster, time, parsers
from lco.lib.session import RecordingSession
from lco.tools import maps, kernels, path, spike
from lco.lib.data import Arenas


############## Set up output directories ###############

posterdir = os.path.join(PROJECT_ROOT, 'poster')


############## Plotting state and functions ###############

cur = 0
fignum = 100
figsize = (14, 14)
rows, cols = (5, 5)
f = None
axlist = None

def new_figure():
    global f, axlist, cur, fignum
    cur = 0
    plt.close(fignum)
    #fignum += 1
    f, axlist = plt.subplots(nrows=rows, ncols=cols, num=fignum, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        new_figure()
    ax = axlist[cur]
    cur += 1
    return ax

def save_and_close(stem='misc/panel', dpi=300):
    plt.figure(fignum)
    plt.tight_layout(pad=0.4)
    stem = os.path.join(posterdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = path.uniquify(stem, ext='png', fmt='%s-%03d')
    f.savefig(fn, dpi=dpi)
    log('Saved: {}', fn)
    plt.close(fignum)

plt.ioff()
f = new_figure()  # open a new figure


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(40.0, 83.0, title, size='small', ha='center', va='bottom')

def plot_spatial_map(M, label='', **kwargs):
    ax = next_axis()
    im = maps.plot_map(M, ax=ax, **kwargs)
    quicktitle(ax, label)


############## Script for the generation of many maps ###############

N_WINDOWS = rows * cols - 2

# Generate list of random recording samples
df = store.get(readonly=True)
recs = df.root.recordings
stable = df.root.sessions

# Get random sample from cell recordings from the cylinder arena
log('Building dict of cylinder sessions and valid clusters...')
cylinder_cells = []
for s in stable.where('(rat==10)&(day==8)'):
    for r in recs.where('session_id==%d' % s['id']):
        cylinder_cells.append(int(r['id']))

cylinder_cells = np.atleast_1d(cylinder_cells)
sample_cells = cylinder_cells
log('Sampling {} cells from {} total in the cylinder', sample_cells.size,
        cylinder_cells.size)

mapargs = dict(cbar=True,
        cbargs={'pad':0.04,'fraction':0.1,'aspect':13,'shrink':0.75},
        mask_color='0.9')

def run_stability_demonstration(s, ttc, cell_label):
    """Generate plots comparing standard and adaptive rate maps."""
    log('Phase stability: Session #{}, Cell {}', s.id, 'tt%d_c%d' % ttc)

    R = s.ratemap(ttc, adaptive=True)
    P = s.phasemap(ttc, adaptive=True)

    mapargs.update(cbar=True)

    plot_spatial_map(R, cmin=0.0, label='Rate (%s)' % label, **mapargs)
    plot_spatial_map(P, cmin=-PI, cmax=PI, satnorm=True,
            label='Phase (%s)' % label, **mapargs)

    mapargs.update(cbar=False)

    duration = s.duration
    wins = np.linspace(s.tstart, s.tend, N_WINDOWS + 1)

    for i in xrange(N_WINDOWS):
        tlim = wins[i], wins[i+1]

        Pwin = s.phasemap(ttc, adaptive=True, tlim=tlim)
        plot_spatial_map(Pwin, satnorm=True, cmin=-PI, cmax=PI,
                label='t = %gs' % np.mean(tlim), **mapargs)

    save_and_close('phase-stability/selected-cells')

for i in sample_cells:
    cell = recs[i]
    ttc = parsers.parse_ttc(cell)
    if ttc not in [(28, 1), (10, 3), (6, 5), (32,1)]:
        continue
    session = RecordingSession.get(cell['session_id'])
    if ttc not in session.clusters:
        log("Warning: cell {} not in session (#{})", ttc, session.id, error=True)
        continue
    #elif session.clusters[ttc].spikes.size / session.duration < 5:
        #log("Skipping cell with low average firing rate...", error=True)
        #continue
    label = 'Rat {}, {}'.format(session.rat, cell['ttc'])
    run_stability_demonstration(session, ttc, label)


############## Draw anything left back to the screen ###############

#plt.tight_layout(pad=0.4)
plt.ion()
#plt.draw()
#plt.show()
