#!/usr/bin/env python
"""
Phase map examples that illustrate their construction.
"""

import os

import numpy as np
from numpy import pi as PI
import matplotlib.pyplot as plt

from lco import PROJECT_ROOT, log
from lco.io import store
from lco.lib import motion, cluster, time, parsers
from lco.lib.session import RecordingSession
from lco.tools import maps, kernels, path, spikes
from lco.lib.data import Arenas


############## Set up output directories ###############

posterdir = os.path.join(PROJECT_ROOT, 'poster')


############## Plotting state and functions ###############

cur = 0
fignum = 100
figsize = (14, 5.5)
rows, cols = (2, 5)
f = None
axlist = None

def new_figure():
    global f, axlist, cur, fignum
    cur = 0
    plt.close(fignum)
    f, axlist = plt.subplots(nrows=rows, ncols=cols, num=fignum, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        new_figure()
    ax = axlist[cur]
    cur += 1
    return ax

def save_and_close(stem='misc/panel', dpi=300):
    plt.figure(fignum)
    plt.tight_layout(pad=0.4)
    stem = os.path.join(posterdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = path.uniquify(stem, ext='png', fmt='%s-%03d')
    f.savefig(fn, dpi=dpi)
    log('Saved: {}', fn)
    plt.close(fignum)

plt.ioff()
f = new_figure()  # open a new figure


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(40.0, 83.0, title, size='small', ha='center', va='bottom')

def plot_spike_autocorrelation(tb, lag, bins=127, label=''):
    ax = next_axis()
    c, lags = spikes.acorr(tb, maxlag=lag, bins=bins)
    c[np.argmax(c)] = 0.0  # kill zero
    ph = spikes.plot_correlogram((c, lags), is_corr=True, ax=ax)
    ax.set(xlabel='Lag (s)', ylabel='Spikes', yticklabels=[])
    ax.set_axis_on()
    quicktitle(ax, label)

def plot_spatial_map(M, label='', **kwargs):
    ax = next_axis()
    im = maps.plot_map(M, ax=ax, **kwargs)
    quicktitle(ax, label)


############## Script for the generation of many maps ###############

SAMPLES = 250

# Generate list of random recording samples
df = store.get(readonly=True)
recs = df.root.recordings
stable = df.root.sessions

# Get random sample from cell recordings from the cylinder arena
log('Building dict of cylinder sessions and valid clusters...')
cylinder_cells = []
for s in stable.where('arena==%d' % Arenas.circle):
    for r in recs.where('session_id==%d' % s['id']):
        cylinder_cells.append(int(r['id']))

np.random.seed(234)
cylinder_cells = np.atleast_1d(cylinder_cells)
sample_cells = cylinder_cells[np.random.permutation(cylinder_cells.size)[:SAMPLES]]
log('Sampling {} cells from {} total in the cylinder', SAMPLES,
        cylinder_cells.size)


def run_phasemap_comparisons(s, ttc, cell_label):
    """Generate plots comparing standard and adaptive rate maps."""
    log('Phasemap comparison: Session #{}, Cell {}', s.id, 'tt%d_c%d' % ttc)
    Pb = s.phasemap(ttc, adaptive=False, mask_occ=1.5, bins=15)
    Pa = s.phasemap(ttc, adaptive=True)

    st = s.clusters[ttc].fast_spikes
    bix = spikes.burst_onset(st)
    tb = st[bix][:12000]

    mapargs = dict(cbar=True,
            cbargs={'pad':0.04,'fraction':0.1,'aspect':13,'shrink':0.75},
            mask_color='0.9')

    for lag, P, label in [(0.5, Pb, '15 Bins'), (0.05, Pa, 'Adaptive')]:
        log('Computing %g-lag autocorrelation for %s...' % (lag, cell_label))
        plot_spike_autocorrelation(tb, lag, label=cell_label)

        plot_spatial_map(P[0], cmap='hsv', cmin=-PI, cmax=PI,
                label='Phase angle', **mapargs)

        plot_spatial_map(P[1], cmap='gray', cmin=0, cmax=1,
                label='Vector length', **mapargs)

        plot_spatial_map(P, cmin=-PI, cmax=PI,
                label='Length * Angle', **mapargs)

        plot_spatial_map(P, satnorm=True, cmin=-PI, cmax=PI,
                label='Normalized (%s)' % label, **mapargs)

    save_and_close('phasemap-methods/comparison')

for i in sample_cells:
    cell = recs[i]
    ttc = parsers.parse_ttc(cell)
    session = RecordingSession.get(cell['session_id'])
    if ttc not in session.clusters:
        log("Warning: cell {} not in session (#{})", ttc, session.id, error=True)
        continue
    elif session.clusters[ttc].spikes.size / session.duration < 5:
        log("Skipping cell with low average firing rate...", error=True)
        continue
    label = 'Rat {}, {}'.format(session.rat, cell['ttc'])
    run_phasemap_comparisons(session, ttc, label)


############## Draw anything left back to the screen ###############

#plt.tight_layout(pad=0.4)
plt.ion()
#plt.draw()
#plt.show()
