#!/usr/bin/env python
"""
Phase and rate map examples.
"""

import os

import numpy as np
from numpy import pi as PI
import matplotlib.pyplot as plt

from lco import PROJECT_ROOT, log
from lco.io import store
from lco.lib import motion, cluster, time, parsers
from lco.lib.session import RecordingSession
from lco.tools import maps, kernels, path, spike
from lco.lib.data import Arenas


############## Set up output directories ###############

posterdir = os.path.join(PROJECT_ROOT, 'poster')


############## Plotting state and functions ###############

cur = 0
fignum = 100
figsize = (8.95, 4.4375)
rows, cols = (1, 2)
f = None
axlist = None

def new_figure():
    global f, axlist, cur, fignum
    cur = 0
    plt.close(fignum)
    f, axlist = plt.subplots(nrows=rows, ncols=cols, num=fignum, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        new_figure()
    ax = axlist[cur]
    cur += 1
    return ax

def save_and_close(stem='misc/panel', dpi=300):
    plt.figure(fignum)
    plt.tight_layout(pad=0.4)
    stem = os.path.join(posterdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = path.uniquify(stem, ext='png', fmt='%s-%03d')
    f.savefig(fn, dpi=dpi)
    log('Saved: {}', fn)
    plt.close(fignum)

plt.ioff()
f = new_figure()  # open a new figure


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(40.0, 83.0, title, size='small', ha='center', va='bottom')

def plot_spatial_map(M, label='', **kwargs):
    ax = next_axis()
    im = maps.plot_map(M, ax=ax, **kwargs)
    quicktitle(ax, label)


############## Script for the generation of many maps ###############

SAMPLES = 50

# Generate list of random recording samples
df = store.get(readonly=True)
recs = df.root.recordings
stable = df.root.sessions

# Get random sample from cell recordings from the cylinder arena
log('Building dict of cylinder sessions and valid clusters...')
cylinder_cells = []
for s in stable.where('arena==%d' % Arenas.circle):
    for r in recs.where('session_id==%d' % s['id']):
        cylinder_cells.append(int(r['id']))

np.random.seed(127)
cylinder_cells = np.atleast_1d(cylinder_cells)
#sample_cells = cylinder_cells[np.random.permutation(cylinder_cells.size)[:SAMPLES]]
sample_cells = cylinder_cells
log('Sampling {} cells from {} total in the cylinder', sample_cells.size,
        cylinder_cells.size)


def run_phasemap_comparisons(s, ttc, cell_label):
    """Generate plots comparing standard and adaptive rate maps."""
    log('Phasemap comparison: Session #{}, Cell {}', s.id, 'tt%d_c%d' % ttc)
    R = s.ratemap(ttc, adaptive=True)
    P = s.phasemap(ttc, adaptive=True)

    mapargs = dict(cbar=True,
            cbargs={'pad':0.04,'fraction':0.1,'aspect':13,'shrink':0.75},
            mask_color='0.9')

    plot_spatial_map(R, cmin=0.0, label='Rate (%s)' % label, **mapargs)

    plot_spatial_map(P, satnorm=True, cmin=-PI, cmax=PI,
            label='Phase (%s)' % label, **mapargs)

    save_and_close('phaserate-examples/cell')

for i in sample_cells:
    cell = recs[i]
    ttc = parsers.parse_ttc(cell)
    session = RecordingSession.get(cell['session_id'])
    if ttc not in session.clusters:
        log("Warning: cell {} not in session (#{})", ttc, session.id, error=True)
        continue
    elif session.clusters[ttc].spikes.size / session.duration < 5:
        log("Skipping cell with low average firing rate...", error=True)
        continue
    label = 'Rat {}, {}'.format(session.rat, cell['ttc'])
    run_phasemap_comparisons(session, ttc, label)


############## Draw anything left back to the screen ###############

#plt.tight_layout(pad=0.4)
plt.ion()
#plt.draw()
#plt.show()

