#!/usr/bin/env python

"""
Development tests for adaptive binning algorithm
"""

import os

import numpy as np
import matplotlib.pyplot as plt

from lco import PROJECT_ROOT, log
from lco.io import store
from lco.lib import motion, cluster, time, parsers
from lco.lib.session import RecordingSession
from lco.tools import maps, kernels, paths
from lco.lib.data import Arenas


############## Set up output directories ###############

posterdir = os.path.join(PROJECT_ROOT, 'comps-update')


############## Plotting state and functions ###############

cur = 0
fignum = 100
figsize = (14, 3.5)
rows, cols = (1, 5)
f = None
axlist = None

def new_figure():
    global f, axlist, cur, fignum
    cur = 0
    plt.close(fignum)
    f, axlist = plt.subplots(nrows=rows, ncols=cols,
            sharex=True, sharey=True, num=fignum, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        new_figure()
    ax = axlist[cur]
    cur += 1
    return ax

def save_and_close(stem='misc/panel', dpi=300):
    plt.figure(fignum)
    plt.tight_layout(pad=0.4)
    stem = os.path.join(posterdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = paths.uniquify(stem, ext='png', fmt='%s-%03d')
    f.savefig(fn, dpi=dpi)
    log('Saved: {}', fn)
    plt.close(fignum)

plt.ioff()
plt.switch_backend('agg')
f = new_figure()  # open a new figure


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(40.0, 83.0, title, size='small', ha='center', va='bottom')

def plot_spike_trajectory(xs, ys, xp, yp, label=''):
    ax = next_axis()
    ax.plot(0.8 * xp, 0.8 * yp, 'k-', alpha=0.4, lw=0.4, zorder=-1)
    ax.scatter(0.8 * xs, 0.8 * ys, s=6, c='r', linewidths=(0,), alpha=0.08)
    ax.set(xlim=(0,80), ylim=(0,80), xticks=(0,40,80), yticks=(0,40,80),
            xlabel='X (cm)', ylabel='Y (cm)')
    ax.axis('equal')
    quicktitle(ax, label)

def plot_spatial_map(M, label='', **kwargs):
    ax = next_axis()
    im = maps.plot_map(M, ax=ax, **kwargs)
    quicktitle(ax, label)


############## Script for the generation of many maps ###############

SAMPLES = 10

# Generate list of random recording samples
df = store.get(readonly=True)
recs = df.root.recordings
stable = df.root.sessions

# Get random sample from cell recordings from the cylinder arena
log('Building dict of cylinder sessions and valid clusters...')
cylinder_cells = []
for s in stable.where('arena==%d' % Arenas.circle):
    for r in recs.where('session_id==%d' % s['id']):
        cylinder_cells.append(int(r['id']))

np.random.seed(1024)
cylinder_cells = np.atleast_1d(cylinder_cells)
sample_cells = cylinder_cells[np.random.permutation(cylinder_cells.size)[:SAMPLES]]
log('Sampling {} cells from {} total in the cylinder', SAMPLES,
        cylinder_cells.size)


def run_ratemap_comparisons(s, ttc, cell_label):
    """Generate plots comparing standard and adaptive rate maps."""
    log('Ratemap comparison: Session #{}, Cell {}', s.id, 'tt%d_c%d' % ttc)
    R15 = s.ratemap(ttc, adaptive=False, mask_occ=1.5, bins=15)
    R30 = s.ratemap(ttc, adaptive=False, mask_occ=0.5, bins=30)
    R45 = s.ratemap(ttc, adaptive=False, mask_occ=0.25, bins=45)

    Ra = s.ratemap(ttc, adaptive=True)

    spikes = s.clusters[ttc].spikes
    xp, yp = s.motion.x, s.motion.y
    xs, ys = s.motion.F('x', spikes), s.motion.F('y', spikes)

    plot_spike_trajectory(xs, ys, xp, yp, label=cell_label)

    mapargs = dict(cbar=True,
            cbargs={'pad':0.0,'fraction':0.1,'aspect':13,'shrink':0.5},
            cmin=0.0, mask_color='0.9')

    plot_spatial_map(R15, label='15 Bins', **mapargs)
    plot_spatial_map(R30, label='30 Bins', **mapargs)
    plot_spatial_map(R45, label='45 Bins', **mapargs)
    plot_spatial_map(Ra, label='Adaptive Kernel', **mapargs)

    save_and_close('kernel-fix')

for i in sample_cells:
    cell = recs[i]
    ttc = parsers.parse_ttc(cell)
    session = RecordingSession.get(cell['session_id'])
    label = '[{}] Rat {}, Cell {}'.format(session.id, session.rat, cell['ttc'])
    run_ratemap_comparisons(session, ttc, label)
