"""
Visualize essential data panels for all theta cells.
"""

from scipy.stats import pearsonr
import numpy as np
import scipy as sp
from matplotlib.ticker import LinearLocator, MaxNLocator
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
sns.reset_orig()

from roto import reports, axis, plots, radians, arrays, circstats

from .sync_model import IBI_MIN
from .. import context, store
from ..lib import session, labels, theta, cell
from ..lib.data import Arenas
from ..lib.lfp import PHASE_MIN, PHASE_MAX
from ..tools import maps, binned, spikes


class ThetaCellCatalog(context.SPCAnalysis):

    """
    Generate essential data plots for all theta-related cells.
    """

    def script(self, cell_id_range=None):
        self.c.cell_id_range = cell_id_range

        self.compute_spatial_maps(adaptive=True)
        self.compute_spatial_map_time_windows(adaptive=True)
        self.spike_trajectory_data()
        self.temporal_autocorr()
        self.phase_distribution()
        self.theta_modulation_index()
        self.direction_maps()
        self.direction_modulation_index()
        self.speed_maps()
        self.speed_modulation_index_and_regression()
        self.rate_phase_correlation()
        self.generate_catalog(levels=12)

    def saveplots(self):
        self.plot_cell_histos()
        self.plot_rate_slope_scatter()
        self.plot_COM_corr_scatter()
        self.save_figures(closeafter=True)

    @context.step
    def compute_spatial_maps(self, adaptive=True, overwrite=False):
        """Compute and store firing-rate and firing-phase maps."""
        self.get_datafile(False)
        if self.c.cell_id_range is None:
            arena = 'arena==%d' % Arenas.circle
            cell_rows = cell.where_session(arena)
        else:
            df = store.get(True)
            recs = df.root.recordings
            start, stop = self.c.cell_id_range
            cell_rows = recs.iterrows(start=start, stop=stop)

        cell_ids = []
        for C in cell_rows:
            c_id = C['id']
            tag = labels.session_cell(c_id)
            S = cell.session_for(c_id)

            rpath = 'ratemaps/id%04d' % C['id']
            if not self.has_node(rpath) or overwrite:
                R = S.ratemap(C, adaptive=adaptive)
                self.save_array(rpath, R, title='Ratemap: %s' % tag,
                    cell_id=c_id)

            ppath = 'phasemaps/id%04d' % C['id']
            if not self.has_node(ppath) or overwrite:
                P = S.phasemap(C, adaptive=adaptive)
                self.save_array(ppath, P, title='Phasemap: %s' % tag,
                    cell_id=c_id)
            cell_ids.append(C['id'])

        self.save_array('cell_ids', np.array(cell_ids))

    @context.step
    def compute_burstphase_maps(self, adaptive=True, overwrite=False):
        """Compute and store burst-phase maps."""
        self.get_datafile(False)
        if self.c.cell_id_range is None:
            arena = 'arena==%d' % Arenas.circle
            cell_rows = cell.where_session(arena)
        else:
            df = store.get(True)
            recs = df.root.recordings
            start, stop = self.c.cell_id_range
            cell_rows = recs.iterrows(start=start, stop=stop)

        cell_ids = []
        for C in cell_rows:
            c_id = C['id']
            tag = labels.session_cell(c_id)
            S = cell.session_for(c_id)

            bpath = 'burstmaps/id%04d' % C['id']
            if not self.has_node(bpath) or overwrite:
                B = S.phasemap(C, bursts=IBI_MIN, adaptive=adaptive)
                self.save_array(bpath, B, title='Burstmap: %s' % tag,
                    cell_id=c_id)
            cell_ids.append(C['id'])

    @context.step
    def compute_spatial_map_time_windows(self, nwindows=5, adaptive=True,
        overwrite=False):
        """Compute spatial maps across time windows."""
        df = store.get(True)
        self['time_windows'] = nwindows

        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')

        for c_id in cell_ids:
            tag = labels.session_cell(c_id)
            C = df.root.recordings[c_id]
            S = session.get(C['session_id'])
            tedges = np.linspace(S.tstart, S.tend, nwindows + 1)
            for i in range(nwindows):
                tlim = [tedges[i:i+2]]

                rpath = 'ratemaps_time/id%04d_%02d' % (c_id, i)
                if not self.has_node(rpath) or overwrite:
                    Rt = S.ratemap(C, intervals=tlim, adaptive=adaptive)
                    self.save_array(rpath, Rt,
                        title='Ratemap Window %d: %s' % (i+1, tag),
                        cell_id=c_id)

                ppath = 'phasemaps_time/id%04d_%02d' % (c_id, i)
                if not self.has_node(ppath) or overwrite:
                    Pt = S.phasemap(C, intervals=tlim, adaptive=adaptive)
                    self.save_array(ppath, Pt,
                        title='Phasemap Window %d: %s' % (i+1, tag),
                        cell_id=c_id)

    @context.step
    def spike_trajectory_data(self, bins=128):
        """Compute distributions of spikes and trajectory data."""
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')
        recs = store.get(True).root.recordings

        for c_id in cell_ids.read():
            tag = labels.session_cell(c_id)
            C = recs[c_id]
            S = cell.session_for(c_id)
            mdata = S.motion

            fastix = mdata.fast_index
            H_occ = binned.occmap(mdata.x[fastix], mdata.y[fastix],
                bins=bins, min_occ=0.0)

            st = cell.spikes(c_id, 'fast')
            xs = mdata.F('x', st)
            ys = mdata.F('y', st)
            H_spk = binned.spikemap(xs, ys, bins=bins)

            self.save_array('occmaps/id%04d' % c_id, H_occ,
                title='Occupancy: %s' % tag, cell_id=c_id)
            self.save_array('spikemaps/id%04d' % c_id, H_spk,
                title='Spikes: %s' % tag, cell_id=c_id)

    @context.step
    def temporal_autocorr(self, lag=0.25, bins=128, overwrite=False):
        """Compute temporal autocorrelograms."""
        self['maxlag'] = lag
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids').read()
        df = store.get(True)

        for c_id in cell_ids:
            tag = labels.session_cell(c_id)
            corrname = 'autocorr/id%04d' % c_id
            if self.has_node(corrname) and not overwrite:
                self.out('Found: {}', tag)
                continue
            C = df.root.recordings[c_id]
            S = cell.session_for(c_id)
            st = cell.spikes(c_id, 'fast')
            C, lags = spikes.acorr(st, maxlag=lag, bins=bins,
                parallel=st.size>2e4)
            self.save_array(corrname, C,
                title='Temporal Autocorrelation: %s' % tag, cell_id=c_id)
        self.save_array('autocorr/lags', lags, title='Time Lags')

    @context.step
    def phase_distribution(self, bins=binned.DEFAULT_PHASE_BINS):
        """Compute phase distribution."""
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids').read()
        df = store.get(True)
        edges = np.linspace(-np.pi, np.pi, bins + 1)
        self.save_array('phasedist/centers', (edges[:-1] + edges[1:]) / 2,
            title='Phase Bins')

        for c_id in cell_ids:
            tag = labels.session_cell(c_id)
            C = df.root.recordings[c_id]
            S = cell.session_for(c_id)
            st = cell.spikes(c_id, 'fast')
            sphase = S.lfp.F('phase', st)
            H = np.histogram(sphase, bins=edges)[0]
            Hg = arrays.circblur(H, radians=2*np.pi/36)
            self.save_array('phasedist/id%04d' % c_id, H,
                title='Phase Distribution: %s' % tag, cell_id=c_id)
            self.save_array('phasedist/id%04d_smooth' % c_id, Hg,
                title='Phase Distribution: %s' % tag, cell_id=c_id)

    @context.step
    def theta_modulation_index(self):
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids').read()
        index = np.empty(len(cell_ids))

        for i,c_id in enumerate(cell_ids):
            Hg = self.get_node('phasedist/id%04d_smooth' % c_id).read()
            index[i] = Hg.ptp() / Hg.max()
        self.save_array('cellprops/theta_mod_index', index,
            title='Theta Modulation Index')

    @context.step
    def direction_maps(self, blur=10.0):
        """Compute direction firing-rate maps."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        for c_id in cell_ids:
            tag = labels.session_cell(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)

            dmap, angles = S.dirmap(C)
            dmap_blur = arrays.circblur(dmap, degrees=10.0)

            self.save_array('dirmaps/id%04d' % c_id, dmap,
                    title='Direction Map: %s' % tag)
            self.save_array('dirmaps/id%04d_smooth' % c_id, dmap_blur,
                    title='Direction Map (Smoothed): %s' % tag)

        self.save_array('dirmaps/angles', angles,
                title='Direction Map Angles')

    @context.step
    def direction_modulation_index(self):
        """Compute direction modulation index."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        dirmod = np.empty(cell_ids.shape)

        for i, c_id in enumerate(cell_ids):
            dmap = self.read_array('dirmaps/id%04d_smooth' % c_id)
            dirmod[i] = dmap.ptp() / dmap.max()

        self.save_array('cellprops/direction_mod_index', dirmod,
                title='Direction Modulation Index')

    @context.step
    def speed_maps(self):
        """Compute speed firing-rate maps."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        for c_id in cell_ids:
            tag = labels.session_cell(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            smap, speeds = S.speedmap(C)

            self.save_array('speedmaps/id%04d' % c_id, smap,
                    title='Speed Map: %s' % tag)

        self.save_array('speedmaps/speeds', speeds,
                title='Speed Map Speeds (cm/s)')

    @context.step
    def speed_modulation_index_and_regression(self):
        """Compute speed modulation index."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        speedmod = np.empty(cell_ids.shape)
        speedreg = np.zeros(cell_ids.shape + (5,))
        speeds = self.read_array('speedmaps/speeds')

        for i, c_id in enumerate(cell_ids):
            s = self.read_array('speedmaps/id%04d' % c_id)
            speedmod[i] = (np.nanmax(s) - np.nanmin(s)) / np.nanmax(s)
            valid = np.logical_not(np.isnan(s))
            speedreg[i] = sp.stats.linregress(speeds[valid], s[valid])

        self.save_array('cellprops/sr_slope', speedreg[:,0])
        self.save_array('cellprops/sr_int', speedreg[:,1])
        self.save_array('cellprops/sr_r', speedreg[:,2])
        self.save_array('cellprops/speed_mod_index', speedmod,
                title='Speed Modulation Index')

    @context.step
    def rate_phase_correlation(self):
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids').read()
        rp_slope = np.zeros(len(cell_ids))
        rp_int = np.zeros_like(rp_slope)
        rp_r = np.zeros_like(rp_slope)

        for i,c_id in enumerate(cell_ids):
            tag = labels.session_cell(c_id)
            R = self.get_node('ratemaps/id%04d' % c_id).read()
            P = self.get_node('phasemaps/id%04d' % c_id).read()[0]
            valid = np.logical_and(np.isfinite(R), np.isfinite(P))
            if not valid.any():
                self.out('No valid pixels: {}', tag, error=True)
                continue
            rp_slope[i], rp_int[i], rp_r[i], _ = theta.phaseregress(R, P)
            ratephase = np.c_[R[valid], P[valid]].T
            self.save_array('ratephase/id%04d' % c_id, ratephase,
                title='Rate-phase pixel data: {}'.format(tag))

        self.save_array('cellprops/rp_slope', rp_slope)
        self.save_array('cellprops/rp_int', rp_int)
        self.save_array('cellprops/rp_r', rp_r)

    @context.step
    def generate_catalog(self, levels=18):
        report = reports.PageReport(self.path('catalog.pdf'))
        cell_ids = self.get_node('cell_ids')
        lags = self.read_array('autocorr/lags')
        phase_bins = self.read_array('phasedist/centers')
        angles = self.read_array('dirmaps/angles')
        speeds = self.read_array('speedmaps/speeds')

        cp = self.get_datafile().root.cellprops

        theta_index = cp.theta_mod_index.read()
        rp_slope = cp.rp_slope.read()
        rp_int = cp.rp_int.read()
        rp_r = cp.rp_r.read()

        dir_index = cp.direction_mod_index.read()
        speed_index = cp.speed_mod_index.read()
        sr_slope = cp.sr_slope.read()
        sr_int = cp.sr_int.read()
        sr_r =cp.sr_r.read()
        slim = np.array([binned.DEFAULT_SPEED_MIN, binned.DEFAULT_SPEED_MAX])

        radfmt = radians.RadianFormatter(inline=False)
        sixticks = LinearLocator(numticks=6)
        maxfivebins = MaxNLocator(prune='both', nbins=5)
        mpl.rc('xtick', labelsize=7.0)
        mpl.rc('ytick', labelsize=7.0)
        mpl.rc('axes', titlesize='xx-small')
        mpl.rc('axes', labelsize='xx-small')

        for i,c_id in enumerate(cell_ids):

            fig = report.page
            tag = labels.session_cell_id(c_id)
            fig.suptitle(tag)
            self.out('Rendering {}', tag)

            loc = LinearLocator(numticks=3)
            cbargs = {'pad':0.01, 'fraction':0.07, 'format':'%.1f', 'ticks':loc}
            pcbargs = cbargs.copy()
            pcbargs.update({'format':radfmt})
            mkw = dict(cbar=True, cbargs=cbargs)
            pmkw = dict(cbar=True, cbargs=pcbargs)

            occmap = self.get_node('occmaps/id%04d' % c_id)
            ax = fig.add_subplot(651)
            maps.plot(occmap.read(), cmap='gray_r', **mkw)
            ax.set_title('Occupancy')
            ax.set_axis_off()

            spkmap = self.get_node('spikemaps/id%04d' % c_id)
            ax = fig.add_subplot(652)
            maps.plot(spkmap.read(), cmap='gray_r', **mkw)
            ax.set_title('Spikes')
            ax.set_axis_off()

            ratemap = self.get_node('ratemaps', 'id%04d' % c_id)
            ax = fig.add_subplot(656)
            maps.plot(ratemap.read(), levels=levels, cmin=0.0, **mkw)
            ax.set_title('Rate')
            ax.set_axis_off()

            phasemap = self.get_node('phasemaps', 'id%04d' % c_id)
            P = phasemap.read()
            if levels is None:
                plevels = None
                pmkw.update(cmap='hsv')
            else:
                plevels = sns.hls_palette(levels, s=1.0, l=0.55)
            ax = fig.add_subplot(657)
            maps.plot(P[0], circular=True, cmin=PHASE_MIN, cmax=PHASE_MAX,
                levels=plevels, **pmkw)
            ax.set_title('Avg. Phase')
            ax.set_axis_off()

            ax = fig.add_subplot(658)
            maps.plot(P[1], cmin=0.0, cmap='gray', **mkw)
            ax.set_title('Vector Length')
            ax.set_axis_off()

            ax = fig.add_subplot(659)
            maps.plot(P, levels=plevels, satnorm=True, **pmkw)
            ax.set_title('Phase')
            ax.set_axis_off()

            ax = fig.add_subplot(6,5,10)
            maps.phasedisc(res=256, levels=plevels, inline=True, zoom=0.75,
                labelsize='x-small')

            start = 11
            nwins = self.c.time_windows
            for j in range(nwins):
                ax = fig.add_subplot(6, 5, start + j)
                Rt = self.get_node('ratemaps_time/id%04d_%02d' % (c_id, j))
                maps.plot(Rt.read(), levels=levels, cmin=0.0, **mkw)
                ax.set_title('Time %d-%d%%' % (100 * j / nwins,
                    100 * (j + 1) / nwins))
                ax.set_axis_off()

            start += 5
            for j in range(nwins):
                ax = fig.add_subplot(6, 5, start + j)
                Pt = self.get_node('phasemaps_time/id%04d_%02d' % (c_id, j))
                maps.plot(Pt.read(), levels=plevels, satnorm=True, **pmkw)
                ax.set_title('Time %d-%d%%' % (100 * j / nwins,
                    100 * (j + 1) / nwins))
                ax.set_axis_off()

            start += 5
            Ct = self.read_array('autocorr/id%04d' % c_id)
            Ct = np.ma.masked_where(lags == 0, Ct)
            ax = fig.add_subplot(6, 5, start)
            spikes.corrplot((Ct, lags), ax=ax, colors='b', style='verts',
                alpha=0.6)
            axis.despine(ax)
            ax.set(xlim=(-self.c.maxlag, self.c.maxlag), xlabel='Lag (s)',
                ylim=(0, 1.2 * np.max(Ct[:int(Ct.size/2)])), ylabel='count')

            start += 1
            Cp = self.read_array('phasedist/id%04d' % c_id)
            Cps = self.read_array('phasedist/id%04d_smooth' % c_id)
            ax = fig.add_subplot(6, 5, start)
            ax.plot(phase_bins, Cp, 'r-', lw=1, alpha=0.6)
            ax.plot(phase_bins, Cps, 'b-', lw=2, alpha=0.6)
            ax.set_title('Theta Index = {:.2f}'.format(theta_index[i]), va='top')
            axis.despine(ax)
            ax.xaxis.set_major_formatter(radfmt)
            ax.set_xticks([-np.pi, -np.pi/2, 0, np.pi/2, np.pi])
            ax.set(xlim=(-np.pi, np.pi), xlabel='Phase (rad)',
                ylim=(0, 1.2 * np.max(Cp)))

            start += 1
            D = self.read_array('dirmaps/id%04d' % c_id)
            Ds = self.read_array('dirmaps/id%04d_smooth' % c_id)
            ax = fig.add_subplot(6, 5, start)
            ax.plot(angles, D, 'r-', lw=1, alpha=0.6)
            ax.plot(angles, Ds, 'b-', lw=2, alpha=0.6)
            ax.set_title('Dir. Index = {:.2f}'.format(dir_index[i]), va='top')
            axis.despine(ax)
            ax.xaxis.set_major_formatter(radfmt)
            ax.set_xticks([0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi])
            ax.set(xlim=(0, 2*np.pi), xlabel='Direction (rad)', ylabel='rate')
            dax = ax

            start += 1
            smap = self.read_array('speedmaps/id%04d' % c_id)
            ax = fig.add_subplot(6, 5, start)
            ax.plot(speeds, smap, 'k-o', mec='k', mew=1.5, ms=2, lw=1.5)
            plt.plot(slim, sr_int[i] + sr_slope[i] * slim, 'r-',
                lw=2, alpha=0.6)
            ax.set_title('Speed r = {:.3f}'.format(sr_r[i]), va='top')
            axis.despine(ax)
            ax.set(xlim=slim, xlabel='Speed (cm/s)')
            ratemax = max(1.0, np.nanmax(smap), np.nanmax(D))
            ax.xaxis.set_major_locator(sixticks)
            ax.set_ylim(0, 1.2 * ratemax)
            dax.set_ylim(0, 1.2 * ratemax)

            start += 1
            ratephase = self.get_node('ratephase/id%04d' % c_id).read()
            if ratephase.size == 0:
                report.save_page()
                continue
            ax = fig.add_subplot(6, 5, start)
            rate, phase = ratephase
            rlim = np.array([rate.min(), rate.max()])
            rate = np.tile(rate, 2)
            phase = np.r_[phase, phase + 2*np.pi]
            plots.heatmap(rate, phase, ax=ax, cmap='gray_r')
            def _plot_regline(m):
                plt.plot(rlim, m*2*np.pi + rp_int[i] + rp_slope[i]*rlim, 'r-',
                    lw=2, alpha=0.6)
            for cycle in range(-3,4):
                _plot_regline(cycle)
            axis.despine(ax)
            ax.set(xlabel='Rate')
            ax.set_yticks([0, np.pi, 2*np.pi])
            ax.set_ylim(-np.pi, 3*np.pi)
            ax.yaxis.set_major_formatter(radfmt)
            ax.xaxis.set_major_locator(maxfivebins)
            ax.set_title('Rate-phase r = {:.3f}'.format(rp_r[i]))

            report.save_page()

        report.close()
        mpl.rc_file_defaults()

    @context.step
    def cell_properties(self):
        """Calculate various cell properties."""
        cell_ids = self.get_node('cell_ids').read()
        recs = store.get(True).root.recordings
        N = len(cell_ids)

        rats = np.zeros((N,), 'i')
        avg_rate = np.zeros((N,), 'd')
        rate_info = np.zeros_like(avg_rate)
        com = np.zeros((N,2), 'd')

        for i, c_id in enumerate(cell_ids):
            C = recs[c_id]
            self.out('Processing {}', labels.session_cell(c_id))

            S = cell.session_for(c_id)

            rats[i] = S.attrs['rat']
            rate_info[i] = S.spatial_info(C)

            fast_spikes = cell.spikes(c_id, 'fast')
            avg_rate[i] = fast_spikes.size / S.duration

            xs = S.motion.F('x', fast_spikes)
            ys = S.motion.F('y', fast_spikes)
            com[i] = xs.mean(), ys.mean()

        self.save_array('cellprops/rats', rats)
        self.save_array('cellprops/rate_info', rate_info)
        self.save_array('cellprops/avg_rate', avg_rate)
        self.save_array('cellprops/com', com)

    def _hist_filled(self, ax, data, bins=64, c='b', **kwds):
        ax.hist(data, bins=bins, zorder=-1, alpha=0.3, color=c,
            histtype='stepfilled', **kwds)
        ax.hist(data, bins=bins, linewidth=0.75, color='k', histtype='step',
            **kwds)

    def _despine(self, ax):
        ax.tick_params(top=False, right=False)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

    def plot_cell_histos(self):
        fig = self.figure('figures/histos', figsize=(9,7))
        nrows, ncols = 3, 2

        ax = fig.add_subplot(nrows, ncols, 1)
        I = self.get_node('cellprops/rate_info').read()
        self._hist_filled(ax, I, range=(0, I.max()))
        self._despine(ax)
        ax.set_ylabel('Spatial info.')

        ax = fig.add_subplot(nrows, ncols, 2)
        rmean = self.get_node('cellprops/avg_rate').read()
        self._hist_filled(ax, rmean, range=(0, 20))
        self._despine(ax)
        ax.set_ylabel('Avg. firing rate')

        ax = fig.add_subplot(nrows, ncols, 3)
        COM = self.get_node('cellprops/com').read()
        COMx = COM[:,0]
        COMy = COM[:,1]
        self._hist_filled(ax, COMx, range=(0,100), c='b')
        self._hist_filled(ax, COMy, range=(0,100), c='g')
        self._despine(ax)
        ax.set_ylabel('Spike COM (x,y)')

        ax = fig.add_subplot(nrows, ncols, 4)
        theta_index = self.get_node('cellprops/theta_mod_index').read()
        self._hist_filled(ax, theta_index)
        self._despine(ax)
        ax.set_ylabel('Theta mod.')

        ax = fig.add_subplot(nrows, ncols, 5)
        slope = self.get_node('cellprops/rp_slope').read()
        self._hist_filled(ax, slope)
        self._despine(ax)
        ax.set_ylabel('Phase-rate slope')

        ax = fig.add_subplot(nrows, ncols, 6)
        r = self.get_node('cellprops/rp_r').read()
        r = r[np.isfinite(r)]
        rm = np.ma.masked_where(np.logical_not(np.isfinite(r)), r)
        self._hist_filled(ax, r)
        self._despine(ax)
        ax.set_ylabel('Phase-rate corr.')

        self.close_datafile()

    def plot_rate_slope_scatter(self):
        rmean = self.get_node('cellprops/avg_rate').read()
        slope = self.get_node('cellprops/rp_slope').read()

        fig = self.figure('figures/rate-corr', figsize=(8,8))
        ax = fig.add_subplot(111)
        ax.scatter(rmean, slope, marker='o', s=9, edgecolor='k',
            facecolor='none', linewidths=0.7, alpha=0.7)
        ax.axvline(c='k', ls=':')
        ax.axhline(c='k', ls=':')
        ax.set(xlabel='Avg. firing rate', ylabel='Phase-rate slope')
        self._despine(ax)
        self.close_datafile()

    def plot_COM_corr_scatter(self):
        COM = self.get_node('cellprops/com').read()
        COMx = COM[:,0]
        COMy = COM[:,1]
        slope = self.get_node('cellprops/rp_slope').read()

        fig = self.figure('figures/COM-corr', figsize=(8,8))
        ax = fig.add_subplot(111)
        ax.scatter(COMx, COMy, c=slope, cmap='RdBu', s=21,
            linewidths=0.7, alpha=1.0, edgecolor='k')
        ax.set(xlabel='X (norm. pos.)', ylabel='Y (norm. pos.)')
        self._despine(ax)
        self.close_datafile()
