"""
Characterize a number of scalar functional properties for cells.
"""

import numpy as np
import scipy as sp
import pandas as pd
import tables as tb
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import ticker
from ipyparallel import interactive

from roto import arrays, circstats, plots, axis, radians, reports
from roto.parallel import queue_status_progress_bar

from .. import context, store
from ..lib import session, cell, labels, data, theta
from ..tools import binned, spikes, maps

from . import directional
from . import curl_velocity


class CellProperties(context.SPCAnalysis):

    """
    Compute a complete set of cell properties for phase/rate analysis.
    """

    def script(self, cell_id_range=None):
        self.setup(cell_id_range=cell_id_range)

        # self.compute_rate_estimates()
        self.compute_rate_estimates_parallel()
        self.compute_spike_phases()
        self.theta_modulation_index()
        self.temporal_phase_regression()
        self.collect_rate_phase_pixels()
        self.spatial_phase_regression()
        self.generate_regression_scatter_report()
        self.average_firing_rate()
        self.average_firing_phase()
        self.average_firing_direction()
        self.spatial_firing_rates()
        # self.spatial_info_and_pvalue()
        self.spatial_info_and_pvalue_parallel()
        self.generate_spatial_info_report()
        self.phase_info_and_pvalue()
        self.generate_phase_info_report()
        self.theta_freq_and_rhythmicity()
        self.generate_rhythmicity_report()
        # self.directional_info_and_pvalue()
        self.directional_info_and_pvalue_parallel()
        self.direction_maps_and_modulation_index()
        self.generate_directional_info_report()
        # self.speed_info_and_pvalue()
        self.speed_info_and_pvalue_parallel()
        self.speed_modulation()
        self.generate_speed_modulation_report()
        # self.import_allocentricity_indexes()
        # self.import_curl_modulation_indexes()
        #self.create_properties_table()

    @context.step
    def setup(self, cell_id_range=None):
        self.c.cell_id_range = cell_id_range
        cell_ids = []
        session_ids = []
        rats = []
        for C in cell.where_session('arena==%d' % data.Arenas.circle):
            c_id = C['id']
            if self.c.cell_id_range is not None:
                if c_id < self.c.cell_id_range[0] or \
                    c_id >= self.c.cell_id_range[1]:
                    continue

            cell_ids.append(c_id)
            s_id = C['session_id']
            S = session.record(s_id)
            session_ids.append(s_id)
            rats.append(S['rat'])

        self.save_array('cell_ids', np.asarray(cell_ids))
        self.save_array('rats', rats)
        self.save_array('session_ids', session_ids)

    def cleanup(self):
        store.close()
        self.close_parallel_client()
        self.close_datafile()

    @context.step
    def compute_rate_estimates(self, kernel_width=0.1):
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        for c_id in cell_ids:
            tag, st = cell.tag_and_spikes(c_id, speed='fast')
            S = cell.session_for(c_id)
            r_hat = spikes.FiringRateEstimate(st, S.duration,
                width=kernel_width)
            f_hat_spikes = r_hat(st)
            self.save_array('rate_estimates/id%04d' % c_id, f_hat_spikes,
                title=tag, cell_id=c_id)

    @context.step
    def compute_rate_estimates_parallel(self, kernel_width=0.1):
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        N = len(cell_ids)

        rc = self.get_parallel_client()
        dview = rc[:]
        dview.block = True
        dview.execute('from spc.lib import cell')
        dview.execute('from spc.tools import spikes')

        @interactive
        def _rate_estimates(c_id, width):
            st = cell.spikes(c_id, speed='fast')
            S = cell.session_for(c_id)
            r_hat = spikes.FiringRateEstimate(st, S.duration, width=width)
            return r_hat(st)

        res = dview.map_async(_rate_estimates, cell_ids,
            np.repeat(kernel_width, N))
        queue_status_progress_bar(lview)
        f_hat = res.get()
        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            self.save_array('rate_estimates/id%04d' % c_id, f_hat[i],
                title=tag, cell_id=c_id)

    @context.step
    def compute_spike_phases(self):
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        for c_id in cell_ids:
            tag, st = cell.tag_and_spikes(c_id, speed='fast')
            S = cell.session_for(c_id)
            phase = S.lfp.F('phase', st)
            self.save_array('spike_phase/id%04d' % c_id, phase, title=tag,
                cell_id=c_id)

    @context.step
    def theta_modulation_index(self, bins=binned.DEFAULT_PHASE_BINS):
        """Compute smoothed phase distributions to calculate theta index."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        edges = np.linspace(-np.pi, np.pi, bins + 1)
        I = np.empty(len(cell_ids))

        for i, c_id in enumerate(cell_ids):
            phase = self.read_array('spike_phase/id%04d' % c_id)
            H = arrays.circblur(np.histogram(phase, bins=edges)[0],
                radians=2*np.pi/bins)
            I[i] = H.ptp() / H.max()

        self.save_array('theta_mod_index', I, title='Theta Modulation Index')

    @context.step
    def temporal_phase_regression(self):
        """Compute rate-phase regression across time."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        corr = np.empty((len(cell_ids), 4))

        for i, c_id in enumerate(cell_ids):
            rate = self.read_array('rate_estimates/id%04d' % c_id)
            phase = self.read_array('spike_phase/id%04d' % c_id)
            corr[i] = theta.phaseregress(rate, phase)
            self._out.printf('\u25a0', color='green')
        self._out.printf('\n')

        self.save_array('rp_tempcorr', corr,
            title='Rate-phase Temporal Correlation: slope, phi0, rho')

    @context.step
    def collect_rate_phase_pixels(self):
        """Save valid spatial pixels for rate-phase correlations."""
        self.get_datafile(False)
        maindf = store.get(True)
        cell_ids = self.read_array('cell_ids')

        for c_id in cell_ids:
            tag = labels.session_cell_id(c_id)
            R = maindf.get_node('/ratemaps/id%04d' % c_id).read()
            P = maindf.get_node('/phasemaps/id%04d' % c_id).read()
            valid = np.logical_and(np.isfinite(R), np.isfinite(P[0]))

            self.save_array('pixels/rate/id%04d' % c_id, R[valid],
                title='Ratemap pixels: %s' % tag, cell_id=c_id)
            self.save_array('pixels/phase/id%04d' % c_id, P[0][valid],
                title='Phasemap pixels: %s' % tag, cell_id=c_id)
            self.save_array('pixels/coherence/id%04d' % c_id, P[1][valid],
                title='Coherence pixels: %s' % tag, cell_id=c_id)

    @context.step
    def spatial_phase_regression(self):
        """Compute rate-phase regression across spatial map pixels."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        pcorr = np.zeros((len(cell_ids), 4)) + np.nan
        ccorr = np.zeros((len(cell_ids), 5)) + np.nan

        for i, c_id in enumerate(cell_ids):
            R = self.read_array('pixels/rate/id%04d' % c_id)
            P = self.read_array('pixels/phase/id%04d' % c_id)
            C = self.read_array('pixels/coherence/id%04d' % c_id)

            if R.size == 0:
                continue

            pcorr[i] = theta.phaseregress(R, P)
            ccorr[i] = sp.stats.linregress(R, C)

            self._out.printf('\u25a0', color='green')
        self._out.printf('\n')

        self.save_array('rp_corr', pcorr,
            title='Rate-phase Spatial Correlation: slope, phi0, rho')
        self.save_array('rc_corr', ccorr,
            title='Rate-coherence Spatial Correlation: slope, intercept, r, p')

    @context.step
    def total_spatial_phase_shift(self):
        """Compute total shift in firing phase across spatial firing rates."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        pcorr = self.read_array('rp_corr')
        shift = np.zeros(len(cell_ids)) + np.nan

        for i, c_id in enumerate(cell_ids):
            R = self.read_array('pixels/rate/id%04d' % c_id)

            if R.size == 0:
                continue

            slope = pcorr[i,0]
            shift[i] = slope * R.ptp()

            self._out.printf('\u25a0', color='green')
        self._out.printf('\n')

        self.save_array('phase_shift', shift,
            title='Total Phase Shift across Range of Firing Rates')

    @context.step
    def generate_regression_scatter_report(self):
        """Make a report showing scatter data and resulting correlations."""
        cell_ids = self.read_array('cell_ids')
        tcorr = self.get_node('rp_tempcorr')
        pcorr = self.get_node('rp_corr')
        ccorr = self.get_node('rc_corr')
        N = 3 * np.isfinite(pcorr[:,0]).sum()

        report = reports.AxesReport(self.path('regressions'), rows=6, cols=3,
                xnorm=False, ynorm=False, naxes=N, numloc='bottom right',
                numpad=0.02, width=8.0, height=10.5)
        report.set_label_fmt(size='x-small')

        radfmt = radians.RadianFormatter(inline=False)
        mpl.rc_file_defaults()
        mpl.rc('xtick', direction='out', labelsize=7.0)
        mpl.rc('ytick', direction='out', labelsize=7.0)
        mpl.rc('axes', titlesize='xx-small')
        mpl.rc('axes', labelsize='xx-small')

        def _plot_phasereg(m, x, sl, phi):
            plt.plot(x, m*2*np.pi + phi + sl*x, 'r-', lw=2, alpha=0.6)

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            self.out('Plotting data: {}', tag)
            Rt = self.read_array('rate_estimates/id%04d' % c_id)
            Pt = self.read_array('spike_phase/id%04d' % c_id)
            R = self.read_array('pixels/rate/id%04d' % c_id)
            P = self.read_array('pixels/phase/id%04d' % c_id)
            C = self.read_array('pixels/coherence/id%04d' % c_id)
            if R.size == 0:
                self.out('No pixels for regression...', error=True)
                continue

            # Temporal rate-phase correlation
            ax = report.next_axes()
            report.page.set_tight_layout(True)
            rlim = np.array([Rt.min(), Rt.max()])
            rate = np.tile(Rt, 2)
            phase = np.r_[Pt, Pt + 2*np.pi]
            plots.heatmap(rate, phase, ax=ax, cmap='gray_r')
            sl, phi, r, p = tcorr[i]
            for cycle in range(-5, 5):
                _plot_phasereg(cycle, rlim, sl, phi)
            axis.despine(ax)
            ax.set_yticks([0, np.pi, 2*np.pi])
            ax.set_ylim(-np.pi, 3*np.pi)
            ax.yaxis.set_major_formatter(radfmt)
            ax.set_title('{}\nr = {:.2f}, p < {}'.format(tag, r, round(p, 6)))

            # Spatial rate-phase correlation
            ax = report.next_axes()
            rlim = np.array([R.min(), R.max()])
            rate = np.tile(R, 2)
            phase = np.r_[P, P + 2*np.pi]
            plots.heatmap(rate, phase, ax=ax, cmap='gray_r')
            sl, phi, r, p = pcorr[i]
            for cycle in range(-5, 5):
                _plot_phasereg(cycle, rlim, sl, phi)
            axis.despine(ax)
            if report.lastrow:
                ax.set(xlabel='Rate')
            ax.set_yticks([0, np.pi, 2*np.pi])
            ax.set_ylim(-np.pi, 3*np.pi)
            ax.yaxis.set_major_formatter(radfmt)
            ax.set_title('r = {:.2f}, p < {}'.format(r, round(p, 6)))

            # Spatial rate-coherence correlation
            ax = report.next_axes()
            rlim = np.array([R.min(), R.max()])
            plots.heatmap(R, C, ax=ax, cmap='gray_r')
            sl, phi, r, p, _ = ccorr[i]
            _plot_phasereg(0, rlim, sl, phi)
            axis.despine(ax)
            ax.set_ylabel('coherence')
            ax.set_title('r = {:.2f}, p < {}'.format(r, round(p, 6)))

        report.close()
        mpl.rc_file_defaults()

    @context.step
    def average_firing_rate(self):
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')
        rate = np.empty(len(cell_ids))

        for i, c_id in enumerate(cell_ids):
            N = cell.spikes(c_id, speed='fast').size
            mdata = cell.session_for(c_id).motion
            dur = np.diff(mdata.fast_intervals, axis=1).sum()
            rate[i] = N / dur
            self._out.printf('\u25a0', color='green')
        self._out.printf('\n')

        self.save_array('avg_rate', rate)

    @context.step
    def average_firing_phase(self):
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')
        avg = np.empty(len(cell_ids))
        coh = np.empty_like(avg)

        for i, c_id in enumerate(cell_ids):
            phase = self.read_array('spike_phase/id%04d' % c_id)
            avg[i], coh[i] = circstats.mean_vector(phase)
            self._out.printf('\u25a0', color='green')
        self._out.printf('\n')

        self.save_array('phase_avg', avg)
        self.save_array('phase_coh', coh)

    @context.step
    def average_firing_direction(self):
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')
        avg = np.empty(len(cell_ids))
        coh = np.empty_like(avg)

        for i, c_id in enumerate(cell_ids):
            st = cell.spikes(c_id, speed='fast')
            S = cell.session_for(c_id)
            ds = S.motion.F('md', st)
            avg[i], coh[i] = circstats.mean_vector(ds)
            self._out.printf('\u25a0', color='green')
        self._out.printf('\n')

        self.save_array('direction_avg', avg)
        self.save_array('direction_coh', coh)

    @context.step
    def spatial_firing_rates(self):
        """Calculate the min/avg/max spatial firing rates."""
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')
        rates = np.zeros((len(cell_ids), 3))

        for i, c_id in enumerate(cell_ids):
            pixels = self.read_array('pixels/rate/id%04d' % c_id)
            if pixels.size:
                rates[i] = pixels.min(), pixels.mean(), pixels.max()
                self._out.printf('\u25a0', color='green')
            else:
                self._out.printf('\u25a0', color='red')
        self._out.printf('\n')

        self.save_array('spatial_rates', rates)

    @context.step
    def spatial_info_and_pvalue(self):
        """Compute the Skaggs information and p-value."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        Ip = np.empty((cell_ids.size, 2))

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            self.out('Computing {}', tag)
            S = cell.session_for(c_id)
            C = cell.record(c_id)
            Ip[i] = S.spatial_info(C, pvalue=True)

        self.save_array('skaggs_info_pvalue', Ip,
            title='Skaggs Spatial Information and P-value')

    @context.step
    def spatial_info_and_pvalue_parallel(self):
        """(Parellelized) Compute the Skaggs information and p-value."""
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')

        rc = self.get_parallel_client()
        dview = rc[:]
        dview.execute('from spc.lib import cell', block=True)
        lview = rc.load_balanced_view()

        @interactive
        def _spatial_info(cell_id):
            S = cell.session_for(cell_id)
            C = cell.record(cell_id)
            I, p = S.spatial_info(C, pvalue=True)
            return I, p

        res = lview.map_async(_spatial_info, cell_ids)
        queue_status_progress_bar(lview)
        info = np.asarray(res.get())
        self.save_array('skaggs_info_pvalue', info,
            title='Skaggs Spatial Information and P-value')

    @context.step
    def generate_spatial_info_report(self):
        """Make a report of binned spatial maps and information values."""
        cell_ids = self.read_array('cell_ids')
        Ip = self.get_node('skaggs_info_pvalue')
        N = len(cell_ids)

        report = reports.AxesReport(self.path('spatialinfo'), rows=6, cols=5,
                xnorm=False, ynorm=False, naxes=N, numloc='bottom right',
                numpad=0.02, width=8.0, height=10.5)

        for i, c_id in enumerate(cell_ids):
            tag = labels.cell_id(c_id)
            S = cell.session_for(c_id)
            R = S.ratemap(cell.record(c_id), min_occ=binned.INFO_MIN_OCC,
                    adaptive=False)
            info, pval = Ip[i]

            ax = report.next_axes()
            maps.plot(R, mask_color='0.8', cmin=0.0)
            ax.set_axis_off()
            title = 'Rat {}, {}\n{:.3f} bits/spk, P<{:.3f}'.format(S.rat, tag,
                    info, pval)
            ax.set_title(title, size='xx-small')

            if report.lastrow:
                axis.xlabel(ax, 'X', size='small')
            if report.firstcol:
                axis.ylabel(ax, 'Y', size='small')

        report.close()

    @context.step
    def phase_info_and_pvalue(self):
        """Compute the phase-space mutual information and p-value."""
        self.get_datafile(False)
        cell_ids = self.get_node('cell_ids')

        rc = self.get_parallel_client()
        dview = rc[:]
        dview.execute('from spc.lib import cell', block=True)
        lview = rc.load_balanced_view()

        @interactive
        def _phase_info(cell_id):
            S = cell.session_for(cell_id)
            C = cell.record(cell_id)
            I, p = S.phase_space_info(C, pvalue=True)
            return I, p

        res = lview.map_async(_phase_info, cell_ids)
        queue_status_progress_bar(lview)
        info = np.asarray(res.get())
        self.save_array('phase_info_pvalue', info,
            title='Phase-Space Mutual Information and P-value')

    @context.step
    def generate_phase_info_report(self):
        """Make a report of phase maps and information values."""
        cell_ids = self.read_array('cell_ids')
        Ip = self.get_node('phase_info_pvalue')
        df = store.get(True)
        N = len(cell_ids)

        report = reports.AxesReport(self.path('phaseinfo'), rows=6, cols=5,
                xnorm=False, ynorm=False, naxes=N, numloc='bottom right',
                numpad=0.02, width=8.0, height=10.5)

        for i, c_id in enumerate(cell_ids):
            tag = labels.cell_id(c_id)
            S = cell.session_for(c_id)
            P = df.get_node('/phasemaps', 'id%04d' % c_id).read()
            info, pval = Ip[i]

            ax = report.next_axes()
            maps.plot(P, levels=12, mask_color='0.8', satnorm=True)
            ax.set_axis_off()
            title = 'Rat {}, {}\n{:.3f} bits/spk, P<{:.3f}'.format(S.rat, tag,
                    info, pval)
            ax.set_title(title, size='xx-small')

            if report.lastrow:
                axis.xlabel(ax, 'X', size='small')
            if report.firstcol:
                axis.ylabel(ax, 'Y', size='small')

        report.close()

    @context.step
    def theta_freq_and_rhythmicity(self):
        """Compute estimates of burst frequency and theta rhythmicity."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        df = store.get(True)
        lags = df.get_node('/theta/autocorr/lags').read()
        bins = lags.size
        maxlag = spikes.maxlag_from_lags(lags)
        self.out('Using {} bins, maxlag = {} seconds', bins, maxlag)

        F = np.empty(cell_ids.size)
        R = np.empty_like(F)
        for i, c_id in enumerate(cell_ids):
            corr = df.get_node('/theta/autocorr', 'id%04d' % c_id).read()
            F_est = theta.burst_frequency_estimate((corr, lags), maxlag=maxlag,
                bins=bins)
            if F_est == -1: # check higher prior frequency if failed
                F_est = theta.burst_frequency_estimate((corr, lags),
                    maxlag=maxlag, bins=bins, f0=10.0, f_theta_min=9.0,
                    f_theta_max=11.0)
            F[i] = F_est

            col = F_est == -1 and 'red' or 'green'
            self._out.printf('\u25a0', color=col)

            R_est = theta.rhythmicity_estimate((corr, lags), maxlag=maxlag,
                bins=bins)
            R[i] = R_est

            col = R_est == -1 and 'red' or 'green'
            self._out.printf('\u25a1 ', color=col)
        self._out.printf('\n')

        self.save_array('theta_burst_freq', F,
            title='Estimated Average Theta Burst Frequency')
        self.save_array('theta_rhythmicity', R,
            title='Estimated Average Theta Rhythmicity')

    @context.step
    def generate_rhythmicity_report(self):
        """Make a report of autocorrelograms and frequency/rhythmicity."""
        cell_ids = self.read_array('cell_ids')
        freq = self.get_node('theta_burst_freq')
        rhythmicity = self.get_node('theta_rhythmicity')
        df = store.get(True)
        lags = df.root.theta.autocorr.lags.read()
        maxlag = spikes.maxlag_from_lags(lags)
        N = len(cell_ids)

        mpl.rc('xtick', labelsize=8.0)
        mpl.rc('axes', labelsize='small')
        mpl.rc('axes', titlesize='xx-small')

        report = reports.AxesReport(self.path('rhythmicity'), rows=5, cols=4,
                xnorm=False, ynorm=False, naxes=N, numloc='bottom right',
                numpad=0.02, width=8.0, height=10.5)

        for i, c_id in enumerate(cell_ids):
            tag = labels.cell_id(c_id)
            S = cell.session_for(c_id)
            C = df.get_node('/theta/autocorr', 'id%04d' % c_id).read()
            C = np.ma.masked_where(lags == 0, C)
            f = freq[i]
            rh = rhythmicity[i]

            ax = report.next_axes()
            spikes.corrplot((C, lags), ax=ax, colors='b', style='verts',
                alpha=0.6)
            axis.despine(ax, left=False)
            ax.set(xlim=(-maxlag, maxlag),
                    ylim=(0, 1.2 * np.max(C[:int(C.size/2)])))
            ax.set_title('Rat {}, {}\nf = {:.3f} Hz, Rh = {:.3f}'.format(
                    S.rat, tag, f, rh))

            if report.lastrow:
                ax.set_xlabel('Lag (s)')
            else:
                ax.set_xticklabels([])
            if report.firstcol:
                ax.set_ylabel('Count')
            ax.set_yticklabels([])

        report.close()
        mpl.rc_file_defaults()

    @context.step
    def directional_info_and_pvalue(self):
        """Directional information rate and p-value."""
        df = self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        info = np.empty((cell_ids.size, 2))

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            self.out('Computing: {}', tag)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            info[i] = S.directional_info(C, pvalue=True)

        self.save_array('directional_info_pvalue', info,
            title='Spike Directional Information and P-value')

    @context.step
    def directional_info_and_pvalue_parallel(self):
        """(Parallelized) Directional information rate and p-value."""
        df = self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        rc = self.get_parallel_client()
        dview = rc[:]
        dview.execute('from spc.lib import cell', block=True)
        lview = rc.load_balanced_view()

        @interactive
        def _directional_info(cell_id):
            S = cell.session_for(cell_id)
            C = cell.record(cell_id)
            I, p = S.directional_info(C, pvalue=True)
            return I, p

        res = lview.map_async(_directional_info, cell_ids)
        queue_status_progress_bar(lview)
        info = np.asarray(res.get())
        self.save_array('directional_info_pvalue', info,
            title='Spike Directional Information and P-value')

    @context.step
    def direction_maps_and_modulation_index(self):
        """Compute directional maps and modulation index."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        dirmod = np.empty(cell_ids.shape)

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            D, angles = S.dirmap(C)
            Ds = arrays.circblur(D, degrees=10.0)
            dirmod[i] = Ds.ptp() / Ds.max()
            self.save_array('dirmaps/id%04d' % c_id, D,
                    title='Directional Map: {}'.format(tag))
            self.save_array('dirmaps/id%04d_smooth' % c_id, Ds,
                    title='Smoothed Directional Map: {}'.format(tag))

        self.save_array('direction_mod_index', dirmod,
                title='Direction Modulation Index')
        self.save_array('dirmaps/angles', angles,
                title='Direction Map Angles')

    @context.step
    def generate_directional_info_report(self):
        """Make a report of direction maps and information rates."""
        cell_ids = self.read_array('cell_ids')
        dirinfo = self.read_array('directional_info_pvalue')
        dir_avg = self.read_array('direction_avg')
        dir_coh = self.read_array('direction_coh')
        dir_mod = self.read_array('direction_mod_index')
        angles = self.read_array('dirmaps/angles')
        N = len(cell_ids)

        mpl.rc('xtick', labelsize=7.0)
        mpl.rc('ytick', labelsize=7.0)
        mpl.rc('axes', labelsize='small')
        mpl.rc('axes', titlesize='xx-small')

        radfmt = radians.RadianFormatter(inline=True)
        report = reports.AxesReport(self.path('direction'), rows=5, cols=4,
                xnorm=False, ynorm=False, naxes=N, numloc='bottom right',
                numpad=0.02, width=8.0, height=10.5)

        for i, c_id in enumerate(cell_ids):
            tag = labels.rat_cell_id(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            D = self.read_array('dirmaps/id%04d' % c_id)
            Ds = self.read_array('dirmaps/id%04d_smooth' % c_id)
            avg = dir_avg[i]
            coh = dir_coh[i]
            index = dir_mod[i]
            I, p = dirinfo[i]

            ax = report.next_axes()
            ax.plot(angles, D, 'r-', lw=1, alpha=0.6)
            ax.plot(angles, Ds, 'b-', lw=2, alpha=0.6)
            ax.set_title('{}\nI = {:.3f}, p < {:g}, D = {:.2f}'.format(
                tag, I, p, index))
            axis.despine(ax)
            ax.xaxis.set_major_formatter(radfmt)
            ax.set_xticks([0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi])
            ax.set(xlim=(0, 2*np.pi), ylim=(0, 1.2 * np.max(D)))

            if report.firstcol:
                ax.set_ylabel('Rate')
            if report.lastrow:
                ax.set_xlabel('Direction (rad)')
            else:
                ax.set_xticklabels([])

        report.close()
        mpl.rc_file_defaults()

    @context.step
    def speed_info_and_pvalue(self):
        """Compute speed information rate and p-values."""
        df = self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        Ip = np.empty((cell_ids.size, 2))

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            self.out('Computing {}', tag)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            Ip[i] = S.speed_info(C, pvalue=True)

        self.save_array('speed_info_pvalue', Ip,
                title='Speed Information Rate and P-value')

    @context.step
    def speed_info_and_pvalue_parallel(self):
        """(Parallelized) Speed information rate and p-value."""
        df = self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        rc = self.get_parallel_client()
        dview = rc[:]
        dview.execute('from spc.lib import cell', block=True)
        lview = rc.load_balanced_view()

        @interactive
        def _speed_info(cell_id):
            S = cell.session_for(cell_id)
            C = cell.record(cell_id)
            I, p = S.speed_info(C, pvalue=True)
            return I, p

        res = lview.map_async(_speed_info, cell_ids)
        queue_status_progress_bar(lview)
        info = np.asarray(res.get())
        self.save_array('speed_info_pvalue', info,
            title='Speed Information Rate and P-value')

    @context.step
    def speed_modulation(self):
        """Compute speed modulation index and regression."""
        df = self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        speedmod = np.empty(cell_ids.shape)
        sreg = np.empty(cell_ids.shape + (5,))

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            smap, speeds = S.speedmap(C)
            speedmod[i] = (np.nanmax(smap) - np.nanmin(smap)) / np.nanmax(smap)
            valid = np.logical_not(np.isnan(smap))
            sreg[i] = sp.stats.linregress(speeds[valid], smap[valid])
            self.save_array('speedmaps/id%04d' % c_id, smap,
                    title='Speed Map: {}'.format(tag))

        self.save_array('speedmaps/speeds', speeds, title='Speed Map Speeds')
        self.save_array('speed_mod_index', speedmod,
                title='Speed Modulation Index')
        self.save_array('speed_regress', sreg,
                title='Speed Tuning Curve Regression')

    @context.step
    def generate_speed_modulation_report(self):
        """Make a report of speed tuning curves and regressions."""
        cell_ids = self.read_array('cell_ids')
        speeds = self.read_array('speedmaps/speeds')
        regression = self.get_node('speed_regress')
        sr_slope = regression[:,0]
        sr_int = regression[:,1]
        sr_r = regression[:,2]
        Ip = self.get_node('speed_info_pvalue')
        slim = np.array([binned.DEFAULT_SPEED_MIN, binned.DEFAULT_SPEED_MAX])
        N = len(cell_ids)

        speedloc = ticker.LinearLocator(numticks=6)
        mpl.rc('xtick', labelsize=8.0)
        mpl.rc('ytick', labelsize=8.0)
        mpl.rc('axes', labelsize='small')
        mpl.rc('axes', titlesize='xx-small')

        report = reports.AxesReport(self.path('speed'), rows=5, cols=4,
                xnorm=False, ynorm=False, naxes=N, numloc='bottom right',
                numpad=0.02, width=8.0, height=10.5)

        for i, c_id in enumerate(cell_ids):
            tag = labels.rat_cell_id(c_id)
            smap = self.read_array('speedmaps/id%04d' % c_id)

            ax = report.next_axes()
            ax.plot(speeds, smap, 'k-o', mec='k', mew=1.5, ms=2, lw=1.5)
            plt.plot(slim, sr_int[i] + sr_slope[i] * slim, 'r-',
                lw=2, alpha=0.6)
            ax.set_title('{}\nI={:.3f}, p={:.3f}, r={:.2f}'.format(
                tag, Ip[i,0], Ip[i,1], sr_r[i]), va='center')
            axis.despine(ax)
            ratemax = max(1.0, np.nanmax(smap))
            ax.set(xlim=slim, ylim=(0, 1.2 * ratemax))
            ax.xaxis.set_major_locator(speedloc)

            if report.firstcol:
                ax.set_ylabel('Firing rate')
            if report.lastrow:
                ax.set_xlabel('Speed (cm/s)')
            else:
                ax.set_xticklabels([])

        report.close()
        mpl.rc_file_defaults()

    @context.step
    def import_allocentricity_indexes(self, dirmaps_rundir):
        """Import allocentricity indexes from directional maps analysis."""
        dirmaps = directional.MovementDirectionMaps.load(dirmaps_rundir)
        df = dirmaps.get_datafile()

        ralloc = df.root.properties.rate_allocentricity
        self.save_array('rate_allocentricity', ralloc,
                title='Directional Ratemap Allocentricity Index')

        palloc = df.root.properties.phase_allocentricity
        self.save_array('phase_allocentricity', palloc,
                title='Directional Phasemap Allocentricity Index')

        dirmaps.close_datafile()

    @context.step
    def import_curl_modulation_indexes(self, curlmod_rundir, grid=2):
        """Import curl modulation data from directional velocity analysis."""
        curlmod = curl_velocity.CurlModulation.load(curlmod_rundir)
        modtable = curlmod.get_node('curl_modulation')

        curlfree_angle = modtable.col('angle_mean_%d' % grid)
        self.save_array('curlfree_angle', curlfree_angle,
                title='Curlfree Modulation Average Angle')

        curlfree_index = 1 - modtable.col('angle_sd_%d' % grid) / np.sqrt(2)
        self.save_array('curlfree_index', curlfree_index,
                title='Curlfree Modulation Index')

        curlmod.close_datafile()

    @context.step
    def create_properties_table(self):
        """Collate all computed properties into a data table."""
        df = self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        N = len(cell_ids)

        coldata = [
            ('cell_ids', None, 'id'),
            ('session_ids', None, 'session_id'),
            ('rats', None, 'rat'),
            ('avg_rate', None, 'avg_rate'),
            ('phase_avg', None, 'phase_avg'),
            ('phase_coh', None, 'phase_coh'),
            ('spatial_rates', 0, 'ratemap_min'),
            ('spatial_rates', 1, 'ratemap_avg'),
            ('spatial_rates', 2, 'ratemap_max'),
            ('rp_tempcorr', 0, 'Ct_rp_sl'),
            ('rp_tempcorr', 1, 'Ct_rp_phi'),
            ('rp_tempcorr', 2, 'Ct_rp_r'),
            ('rp_tempcorr', 3, 'Ct_rp_p'),
            ('phase_shift', None, 'rp_shift'),
            ('rp_corr', 0, 'C_rp_sl'),
            ('rp_corr', 1, 'C_rp_phi'),
            ('rp_corr', 2, 'C_rp_r'),
            ('rp_corr', 3, 'C_rp_p'),
            ('rc_corr', 0, 'C_rc_sl'),
            ('rc_corr', 1, 'C_rc_int'),
            ('rc_corr', 2, 'C_rc_r'),
            ('rc_corr', 3, 'C_rc_p'),
            ('skaggs_info_pvalue', 0, 'I_rate'),
            ('skaggs_info_pvalue', 1, 'I_rate_p'),
            ('phase_info_pvalue', 0, 'I_phase'),
            ('phase_info_pvalue', 1, 'I_phase_p'),
            ('curlfree_angle', None, 'curl_angle'),
            ('curlfree_index', None, 'curlfree'),
            ('direction_avg', None, 'dir_avg'),
            ('direction_coh', None, 'dir_coh'),
            ('directional_info_pvalue', 0, 'I_dir'),
            ('directional_info_pvalue', 1, 'I_dir_p'),
            ('direction_mod_index', None, 'dir_mod'),
            ('rate_allocentricity', None, 'rate_alloc'),
            ('phase_allocentricity', None, 'phase_alloc'),
            ('speed_info_pvalue', 0, 'I_speed'),
            ('speed_info_pvalue', 1, 'I_speed_p'),
            ('speed_mod_index', None, 'speed_mod'),
            ('speed_regress', 0, 'C_sr_sl'),
            ('speed_regress', 1, 'C_sr_int'),
            ('speed_regress', 2, 'C_sr_r'),
            ('speed_regress', 3, 'C_sr_p'),
            ('theta_burst_freq', None, 'f_hat'),
            ('theta_mod_index', None, 'theta_mod'),
            ('theta_rhythmicity', None, 'theta_rhythm')
        ]

        props_descr = {
            'id': tb.UInt32Col(pos=0),
            'session_id': tb.UInt32Col(pos=1),
            'rat': tb.UInt8Col(pos=2)
        }

        props_descr.update({
            d[2]: tb.FloatCol(pos=i) for i, d in enumerate(coldata) if i > 2
        })

        table = self.create_table('tables/properties', props_descr,
            rundir=self._rundir)

        [table.row.append() for _ in range(N)]  # init empty table
        table.flush()

        for arrname, index, colname in coldata:
            self.out('Saving {}[{}] → {}', arrname, index, colname)
            if not self.has_node(arrname):
                values = np.zeros(N)
            elif index is None:
                values = self.read_array(arrname)
            else:
                values = self.get_node(arrname)[:,index]
            table.modify_column(colname=colname, column=values)
