"""
Assess dependence of firing rate on velocity direction.
"""

from matplotlib.ticker import LinearLocator
from sklearn import linear_model
import matplotlib.pyplot as plt
import numpy as np
import tables as tb

from roto.axis import quicktitle, xlabel, ylabel
from roto.circstats import mean_angle as cmean, std as cstd
from roto.misc import outer_pairs as pairs
from roto.plots import heatmap
from roto.reports import AxesReport
from roto.stats import integer_hist

from .. import store
from ..context import SPCAnalysis, step
from ..lib import cell, session, labels, motion
from ..lib.data import Arenas
from ..tools import maps


DEFAULT_TIME_BIN_SIZE = 0.160
DEFAULT_TIME_MARGIN = 60.0


class CurlModulation(SPCAnalysis):

    """
    Analysis of spatially localized directional dependence of firing rate.
    """

    def script(self, grid_sizes=[2], reports=True, **kwds):
        if not self.has_node('cell_ids'):
            self.setup(**kwds)

        self.create_binned_motion_data()
        self.create_binned_spiking_data()
        self.full_linear_model()
        for grid in grid_sizes:
            self.partitioned_linear_model(grid=grid)
            if reports:
                self.generate_curl_report(grid=grid)

    @step
    def setup(self, bin_dt=DEFAULT_TIME_BIN_SIZE, margin=DEFAULT_TIME_MARGIN,
        cell_id_range=None):
        if cell_id_range is None:
            cell_rows = cell.where_session('arena==%d' % Arenas.circle)
        else:
            cell_rows = store.get().root.recordings.iterrows(*cell_id_range)

        cell_ids = []
        session_ids = set()
        for C in cell_rows:
            cell_ids.append(C['id'])
            session_ids.add(C['session_id'])

        self.save_array('cell_ids', cell_ids)
        self.save_array('session_ids', np.array(sorted(session_ids)))

        self['N'] = len(cell_ids)
        self['margin'] = margin
        self['delta'] = bin_dt

    def cleanup(self):
        store.close()

    @step
    def create_binned_motion_data(self):
        """Load and time-bin the spike and motion data for all cells."""
        self.get_datafile(False)
        for s_id in self.read_array('session_ids'):
            sdata = session.get(s_id)
            mdata = sdata.motion
            running = sdata.motion_filter()

            # Set bin edges based on margin and bin dt
            start = sdata.tstart + self.c.margin
            end = sdata.tend - self.c.margin + self.c.delta
            edges = np.arange(start, end, self.c.delta)

            # Get running bins
            fast = mdata.speed_filter(edges)
            running = np.c_[fast[:-1], fast[1:]].all(axis=1)
            bins = np.c_[edges[:-1], edges[1:]][running]
            t_centers = bins.mean(axis=1)

            # Interpolate bin-centered motion data
            X_ = mdata.F('x', t_centers)
            Y_ = mdata.F('y', t_centers)
            Vx_ = mdata.F('x_speed_cm', t_centers)
            Vy_ = mdata.F('y_speed_cm', t_centers)
            S_ = mdata.F('speed_cm', t_centers)

            datalist = [ (edges, 'edges'), (running, 'running'), (X_, 'X'),
                         (Y_, 'Y'), (Vx_, 'Vx'), (Vy_, 'Vy'), (S_, 'S') ]

            for arr, label in datalist:
                self.save_array('motion/session_%03d/%s' % (s_id, label), arr)

    @step
    def create_binned_spiking_data(self):
        """Load spike trains to create binned spike-count data."""
        self.get_datafile(False)
        for c_id in self.read_array('cell_ids'):
            sdata = cell.session_for(c_id)
            mdata = sdata.motion
            cdata = sdata.get_cluster(cell.record(c_id))

            # Load bin edges
            where = 'motion/session_%03d' % sdata.attrs['id']
            edges = self.read_array(where, 'edges')
            running = self.read_array(where, 'running')

            # Count spikes in running bins
            spk_count = np.histogram(cdata.spikes, bins=edges)[0]
            K = spk_count[running]

            self.save_array('spikes/cell_%04d' % c_id, K)

    @staticmethod
    def spike_count_prediction(yhat, y):
        """Exponentiate and normalize the predicted spike-count values."""
        Khat = np.expm1(yhat)
        Khat = np.clip(np.round(Khat), 0, np.inf).astype('int64')
        return Khat

    @step
    def full_linear_model(self):
        """Train GLM on time-binned motion and spike-count data."""
        xy = np.empty((self.c.N, 2), 'd')

        self['p'] = 5
        model = linear_model.LinearRegression()
        K0 = np.empty(self.c.N, 'd')
        beta = np.empty((self.c.N, self.c.p), 'd')

        for i, c_id in enumerate(self.read_array('cell_ids')):
            s_id = cell.session_for(c_id).attrs['id']
            where = '/motion/session_%03d' % s_id

            X = np.c_[ self.read_array(where, 'X'),
                       self.read_array(where, 'Y'),
                       self.read_array(where, 'Vx'),
                       self.read_array(where, 'Vy'),
                       self.read_array(where, 'S') ]

            # y = np.log1p(self.read_array('spikes/cell_%04d' % c_id))
            y = self.read_array('spikes/cell_%04d' % c_id)

            model.fit(X, y)

            xy[i] = X[:,0].mean(), X[:,1].mean()
            K0[i] = model.intercept_
            beta[i] = model.coef_

        self.save_array('full_XY', xy)
        self.save_array('K0_full', K0)
        self.save_array('beta_full', beta)

    @step
    def partitioned_linear_model(self, grid=2):
        """Run the linear model on equal-sized partitions of space."""
        Nparts = grid ** 2
        partxy = np.zeros((self.c.N, 2, Nparts), 'd')

        model = linear_model.LinearRegression()
        K0 = np.empty((self.c.N, Nparts), 'd')
        beta = np.empty((self.c.N, self.c.p, Nparts), 'd')

        avg_velocity = np.zeros((self.c.N, 2, Nparts), 'd')
        dirmean = np.zeros(self.c.N, 'd')
        dirstd = np.zeros(self.c.N, 'd')

        for i, c_id in enumerate(self.read_array('cell_ids')):
            self.out('Running partition model for {}...', labels.cell_id(c_id))
            s_id = cell.session_for(c_id).attrs['id']
            where = '/motion/session_%03d' % s_id

            X_, Y_, Vx_, Vy_, S_ = tuple(self.read_array(where, name)
                    for name in ('X', 'Y', 'Vx', 'Vy', 'S'))
            X = np.c_[X_, Y_, Vx_, Vy_, S_]
            Nsamples = X.shape[0]

            # y = np.log1p(self.read_array('spikes/cell_%04d' % c_id))
            y = self.read_array('spikes/cell_%04d' % c_id)

            Xpart = np.linspace(X[:,0].min(), X[:,0].max() + 1, grid + 1)
            Ypart = np.linspace(X[:,1].min(), X[:,1].max() + 1, grid + 1)
            self.save_array('%s/grid_%d' % (where, grid), np.c_[Xpart, Ypart].T)

            for c, (j, k) in enumerate(pairs(range(grid), range(grid))):
                partix = (X_ >= Xpart[j]) * (X_ < Xpart[j+1]) * \
                         (Y_ >= Ypart[k]) * (Y_ < Ypart[k+1])

                model.fit(X[partix], y[partix])

                partxy[i, :, c] = X_[partix].mean(), Y_[partix].mean()
                K0[i, c] = model.intercept_
                beta[i, :, c] = model.coef_

                avg_velocity[i, :, c] = Vx_[partix].mean(), Vy_[partix].mean()

                self.box()
            self.newline()

            # Compute circular mean/SD of directional weights
            angles = np.arctan2(beta[i,3], beta[i,2])
            radius = np.hypot(beta[i,2], beta[i,3])
            dmean = np.rad2deg(cmean(angles, weights=radius))
            dirmean[i] = dmean + 360.0 if dmean < 0 else dmean
            dirstd[i] = cstd(angles)

        self.save_array('partition_XY_%d' % grid, partxy)
        self.save_array('K0_partition_%d' % grid, K0)
        self.save_array('beta_partition_%d' % grid, beta)
        self.save_array('avg_velocity_%d' % grid, avg_velocity)
        self.save_array('curl_modulation_%d' % grid, np.c_[dirmean, dirstd])

    @step
    def generate_curl_report(self, grid=2):
        """Create a report that shows the curl coefficients across the
        spatial partitions.
        """
        df = store.get()

        centers = self.read_array('partition_XY_%d' % grid)
        V_avg = self.read_array('avg_velocity_%d' % grid)
        beta = self.read_array('beta_partition_%d' % grid)
        curl = self.read_array('curl_modulation_%d' % grid)
        beta_full = self.read_array('beta_full')

        loc = LinearLocator(numticks=3)
        report = AxesReport(self.path('curl-report-%d' % grid),
                    naxes=3*self.c.N, cols=3, rows=5, xnorm=False, ynorm=False)

        for i, c_id in enumerate(self.read_array('cell_ids')):

            self.out('Plotting results for {}...', labels.session_cell(c_id))
            tag = labels.rat_cell_id(c_id)
            s_id = cell.session_for(c_id).attrs['id']
            where = '/motion/session_%03d' % s_id
            Xpart, Ypart = self.read_array(where, 'grid_%d' % grid)

            # Plot the cell's precomputed ratemap
            ax = report.next_axes()
            R = df.get_node('/ratemaps', 'id%04d' % c_id)
            im, cb = maps.plot(R, ax=ax, extent=[0,100,0,100], cmin=0,
                        cbar=True, cbargs={'ticks':loc, 'format':'%.1f'},
                        mask_color='0.9')
            cb.ax.tick_params(labelsize=8)
            ax.set_axis_off()
            if report.firstrow:
                quicktitle(ax, 'Firing ratemap\n{}'.format(tag))
            else:
                quicktitle(ax, '{}'.format(tag))
            if report.firstcol:
                ylabel(ax, 'Y')
            if report.lastrow:
                xlabel(ax, 'X')

            for _x in Xpart[1:-1]:
                ax.axvline(x=_x, lw=0.8, color='w', alpha=0.5)
            for _y in Ypart[1:-1]:
                ax.axhline(y=_y, lw=0.8, color='w', alpha=0.5)

            # Draw the directional-velocity coefficients as a vector field
            # alongside the speed coefficients on a scale average-velocity
            # vector for the corresponding partition
            ax = report.next_axes()

            Vbar = np.hypot(V_avg[i,0], V_avg[i,1])
            beta_S = V_avg[i] * (beta[i,4] / Vbar)

            ax.quiver(centers[i,0], centers[i,1], beta_S[0], beta_S[1],
                    scale=0.2, color='c')
            ax.quiver(centers[i,0], centers[i,1], beta[i,2], beta[i,3],
                    scale=0.2)
            ax.axis('scaled')
            ax.axis([0, 100, 0, 100])
            ax.set_axis_off()
            if report.firstrow:
                quicktitle(ax, 'Velocity weights\n{0}x{0} partition'.format(grid))
            if report.lastrow:
                xlabel(ax, 'X')

            for _x in Xpart[1:-1]:
                ax.axvline(x=_x, lw=0.8, ls=':', color='0.5', alpha=0.5)
            for _y in Ypart[1:-1]:
                ax.axhline(y=_y, lw=0.8, ls=':', color='0.5', alpha=0.5)

            # Plot the coefficients together for comparison
            ax = report.next_axes()
            ax.quiver(Xpart.mean(), Ypart.mean(), beta[i,2], beta[i,3],
                    scale=0.2, alpha=0.8)
            ax.quiver(Xpart.mean(), Ypart.mean(), beta_full[i,2],
                    beta_full[i,3], scale=0.2, color='r', alpha=0.8)
            ax.axis('scaled')
            ax.axis([0, 100, 0, 100])
            ax.set_axis_off()
            if report.firstrow:
                quicktitle(ax,
                    'Velocity weights\n{:.1f} $\pm$ {:.2f}$\degree$'.format(
                    curl[i,0], curl[i,1]))
            else:
                quicktitle(ax, '{:.1f} $\pm$ {:.2f}$\degree$'.format(
                    curl[i,0], curl[i,1]))
            if report.lastrow:
                xlabel(ax, 'X')

            for _x in Xpart[1:-1]:
                ax.axvline(x=_x, lw=0.8, ls=':', color='0.5', alpha=0.5)
            for _y in Ypart[1:-1]:
                ax.axhline(y=_y, lw=0.8, ls=':', color='0.5', alpha=0.5)

        report.close()

    @step
    def generate_curl_modulation_table(self, grids=[2]):
        """Store the curl modulation data in a data table."""

        desc = {'cell_id': tb.Int16Col(pos=0)}
        desc.update({'angle_mean_%d' % grid: tb.FloatCol() for grid in grids})
        desc.update({'angle_sd_%d' % grid: tb.FloatCol() for grid in grids})

        modtable = self.create_table('curl_modulation', desc,
                title='Velocity Modulation by Space')
        row = modtable.row

        # Initialize the table with cell-count rows
        [row.append() for i in range(self.c.N)]
        modtable.flush()

        cell_ids = self.read_array('cell_ids')
        modtable.modify_column(column=cell_ids, colname='cell_id')

        for grid in grids:
            curlmod = self.get_node('curl_modulation_%d' % grid)

            modtable.modify_column(
                    column=curlmod[:,0],
                    colname='angle_mean_%d' % grid)
            modtable.modify_column(
                    column=curlmod[:,1],
                    colname='angle_sd_%d' % grid)

            self.out('Saved {0}x{0} grid data.', grid)

        self.flush_datafile()

    @step
    def plot_curl_modulation_sigma(self, grids=[2]):
        """Plot distribution of curl modulation variance."""
        f = self.figure('csd_distros', figsize=(6,6))
        modtable = self.get_node('curl_modulation')
        dataset = [modtable.col('angle_sd_%d' % grid) for grid in grids]
        plt.violinplot(dataset, positions=grids, widths=0.7, showmedians=True,
                points=128)
        ax = plt.gca()
        xlabel(ax, 'Grid partition')
        ylabel(ax, 'Angular S.D. (a.u.)')
        quicktitle(ax, 'Directional Modulation Spread Across Spatial Partition')

    @step
    def plot_phase_coder_comparison(self, grids=[2], min_I_phase=0.1):
        """Plot spatial phase information against curl modulation."""
        df = store.get()
        cellprops = df.root.tables.properties
        modtable = self.get_node('curl_modulation')

        spc = np.logical_and(cellprops.col('I_phase') >= min_I_phase,
                             cellprops.col('I_phase_p') <= 0.01)
        rest = np.logical_not(spc)

        f = self.figure('phase_coding_curlmod', figsize=(8,6))
        for i, grid in enumerate(grids):
            ax = plt.subplot(1, len(grids), i+1)
            curlsd = modtable.col('angle_sd_%d' % grid)
            ax.violinplot([curlsd[rest], curlsd[spc]], widths=0.7, positions=(1,2),
                    showmedians=True)
            ax.set_xticks([1,2])
            ax.set_xticklabels(['Non-SPC', 'spc'])
            ax.set_ylabel('Angular S.D. (a.u.)')
            quicktitle(ax, 'Phase coding curl (%d-grids)' % grid)
