"""
Figure and description of top-level example of spatial phase coder (phaser).
"""

import os

import tables as tb
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator, MaxNLocator
from matplotlib.cm import get_cmap
import seaborn as sns
sns.reset_orig()
# sns.set(context='notebook', style='ticks', palette='pastel', color_codes=True,
        # rc={'figure.subplot.bottom': 0.15})

from roto.circstats import mean_angle as cmean
from roto.radians import RadianFormatter, cdiff
from roto.plots import condmap
import roto.axis as ra

from .. import store, RES_ROOT
from ..context import SPCAnalysis, step
from ..lib import data, phasers, cell, session, labels, motion, theta
from ..tools import maps, spikes, binned, adaptive


CHARACTERIZE_ROOT = os.path.join(RES_ROOT, '2016-08-15+cell-properties')


class IntroExample(SPCAnalysis):

    """
    Figure and description of an introductory example phaser cell.
    """

    def cleanup(self):
        store.close()

    def plot_all_phasers(self, sortby='C_rp_r', levels=18):
        """Iterate through all phaser cells to plot and save figures."""
        P = phasers.filtered_dataframe()
        sortix = np.argsort(P[sortby])

        for cell_id in P.index[sortix]:
            self.setup(cell_id)
            self.find_example_LFP_traces()
            self.plot_figure(levels=levels)
            self.savefig()
            self.closefig()

    @step
    def setup(self, example_cell_id, adaptive=True, **mapkwds):
        """Retrieve all the data about the example cell for the figure."""
        f = store.get()
        props = f.root.tables.properties
        recs = f.root.recordings

        # Cell ID and array index
        self.c.cell_id = c_id = int(example_cell_id)
        self.c.cell_ix = c_ix = int(np.argmax(props.col('id') == c_id))
        c_id_str = 'id%04d' % c_id

        # Cell information
        self.c.label = labels.session_cell_id(c_id)
        cell_rec = cell.record(c_id)
        self.c.ttc = ttc = int(cell_rec['tetrode']), int(cell_rec['cluster'])
        self.c.session_id = int(cell_rec['session_id'])

        # Save spike train
        st = cell.spikes(c_id)
        self.save_array('spikes', cell.spikes(c_id))

        # Get session data and save trajectory
        S = cell.session_for(c_id, **mapkwds)
        self.save_array('T', S.motion.t)
        self.save_array('X', S.motion.x)
        self.save_array('Y', S.motion.y)
        self.save_array('spikes_X', S.motion.F('x', st))
        self.save_array('spikes_Y', S.motion.F('y', st))

        # Compute adaptive ratemap
        R = S.ratemap(ttc, adaptive=adaptive)
        self.save_array('ratemap', R)

        # Compute adaptive phasemap
        P = S.phasemap(ttc, adaptive=adaptive)
        self.save_array('phasemap', P)

        # Get temporal autocorrelation
        self['maxlag'] = 0.25
        lags = f.root.theta.autocorr.lags.read()
        autocorr = f.get_node('/theta/autocorr', c_id_str).read()
        self.save_array('autocorr_lags', lags)
        self.save_array('autocorr', autocorr)

        # Get phase distribution
        phase = S.lfp.F('phase', st)
        centers = f.root.theta.phasedist.centers.read()
        pdsm = f.get_node('/theta/phasedist', c_id_str + '_smooth').read()
        self.save_array('phasedist_centers', centers)
        self.save_array('spk_phase', phase)
        self.save_array('phasedist_smooth', pdsm)

        # Get rate/phase pixels and regression results
        with tb.open_file(os.path.join(CHARACTERIZE_ROOT, 'data.h5')) as fc:
            Rpixels = fc.get_node('/pixels/rate', c_id_str).read()
            Ppixels = fc.get_node('/pixels/phase', c_id_str).read()
            Crp = fc.get_node('/rp_corr').read()[c_ix]
        self.save_array('rate_pixels', Rpixels)
        self.save_array('phase_pixels', Ppixels)
        self.save_array('rp_corr', Crp)

    @step
    def plot_phase_modulation_figure(self, win=(7147.07,7162.07),
        ibi_min=0.025, axes=None, interactive=True, spike_cmap='coolwarm',
        figsize=(9.32653061, 5.97278912)):
        """Plot phase-modulation time series example."""
        if hasattr(self.c, 'c_id'):
            self.disconnect_figure()

        R = self.read_array('ratemap')
        rmin, rmax = np.nanmin(R), np.nanmax(R)
        T = self.read_array('T')
        X = self.read_array('X') * adaptive.CM_SCALE
        Y = self.read_array('Y') * adaptive.CM_SCALE

        S = cell.session_for(self.c.cell_id)
        st = bt = self.read_array('spikes')
        if ibi_min:
            bt = spikes.burst_filter(st, isi_thresh=ibi_min)
        self.out('N spikes/bursts = {}/{}', st.size, bt.size)
        sx = S.motion.F('x', bt) * adaptive.CM_SCALE
        sy = S.motion.F('y', bt) * adaptive.CM_SCALE
        sphase = S.lfp.F('phase', bt)

        # Compute map-based average firing rates for all spikes
        xmin, xmax, ymin, ymax = motion.arena_extent(0)
        xedges = np.linspace(xmin, xmax, R.shape[0] + 1)
        yedges = np.linspace(ymin, ymax, R.shape[1] + 1)
        self.out('Finding map-based spike rates...')
        srate = np.zeros_like(sphase)
        for i, (x, y) in enumerate(zip(sx, sy)):
            xix = (xedges < x).nonzero()[0].max()
            yix = (yedges < y).nonzero()[0].max()
            srate[i] = R[xix, yix]

        fig = self.figure('rate-phase-modulation', clear=True, figsize=figsize,
                title='Rate-Phase Modulation: {}'.format(self.c.label))
        if axes is None:
            axm = fig.add_subplot(221)
            axp = fig.add_subplot(222)
        else:
            axm, axp = axes
        axs = fig.add_subplot(212)

        # Slice time-interval
        t0, t1 = win[0], win[1]
        ix = (T >= t0) & (T < t1)
        six = (bt >= t0) & (bt < t1)

        # Plot formatting
        radfmt = RadianFormatter(inline=False)
        maxfivebins = MaxNLocator(prune='both', nbins=5, integer=True)
        trajfmt = dict(c='k', lw=0.8, zorder=10, solid_capstyle='round', alpha=0.7)
        spkfmt = dict(s=12, linewidths=0, edgecolors='k', alpha=0.7, zorder=20)
        regfmt = dict(ls='-', c='k', lw=1.2, alpha=0.5, zorder=-1)

        # Ratemap plot with sliced spike-trajectory overlay
        im = maps.plot(R, ax=axm, cmap='CMRmap', cmin=0.0, cbar=False,
                levels=None)
        im.set_alpha(0.25)
        spkcmap = get_cmap(spike_cmap)
        spkcol = spkcmap(np.linspace(0, 1, six.sum()))
        [htraj] = axm.plot(X[ix], Y[ix], **trajfmt)
        hspk = axm.scatter(sx[six], sy[six], c=spkcol, **spkfmt)
        axm.set(xlabel='X', ylabel='Y')
        axm.axis('equal')
        axm.set_axis_off()

        # Rate-phase plot with burst-phase/rate regression lines
        # for i in six.nonzero()[0]:
            # self.out('Spike {}: t = {:.1f}s, r = {:.1f}, ({:.1f}, {:.1f})',
                    # i, bt[i], srate[i], sx[i], sy[i])
        hphase = axp.scatter(srate[six], sphase[six], c=spkcol, **spkfmt)
        hphase2 = axp.scatter(srate[six], 2*np.pi+sphase[six], c=spkcol, **spkfmt)
        axp.set(xlabel='Avg. firing rate (spikes/s)', ylabel='Burst theta phase')
        axp.xaxis.set_major_locator(maxfivebins)
        axp.yaxis.set_major_formatter(radfmt)
        axp.set_yticks([-np.pi, 0, np.pi, 2*np.pi, 3*np.pi])
        axp.set_xlim(np.nanmin(R), np.nanmax(R))
        axp.set_ylim(-np.pi, 3*np.pi)
        ra.despine(axp)

        # Calculate cicular-linear correlation line here for plotting
        sl,phi,r,p = theta.phaseregress(srate[six], sphase[six])
        rlim = np.array([srate[six].min(), srate[six].max()])
        reglines = [axp.plot(rlim, m * 2*np.pi + phi + sl * rlim, **regfmt)
                    for m in range(-3,4)]

        # Time-series of x/y data for interactive scrubbing
        axs.plot(T, np.c_[X, Y])
        axs.set_xlim(t0, t1)
        ra.despine(axs)

        if not interactive:
            return

        plt.ion()
        plt.show()

        # Set up callback for mouse release (after pan/zoom in scrub axes)
        class draw_plots(object):
            def __init__(self, **kwargs):
                for k in kwargs:
                    setattr(self, k, kwargs[k])

            def __call__(self, event):
                if event.inaxes is not self.axs: return
                t0, t1 = self.axs.get_xlim()
                ix = (self.T >= t0) & (self.T < t1)
                six = (self.bt >= t0) & (self.bt < t1)
                self.htraj.remove()
                self.hspk.remove()
                self.hphase.remove()
                self.hphase2.remove()
                [rl[0].remove() for rl in self.reglines]
                [self.htraj] = self.axm.plot(self.X[ix], self.Y[ix], **self.trajfmt)
                spkcol = self.spkcmap(np.linspace(0, 1, six.sum()))
                self.hspk = self.axm.scatter(self.sx[six], self.sy[six], c=spkcol, **self.spkfmt)
                self.hphase = self.axp.scatter(self.srate[six], self.sphase[six], c=spkcol, **self.spkfmt)
                self.hphase2 = self.axp.scatter(self.srate[six], 2*np.pi+self.sphase[six],
                        c=spkcol, **self.spkfmt)
                sl,phi,r,p = theta.phaseregress(self.srate[six], self.sphase[six])
                rlim = np.array([self.srate[six].min(), self.srate[six].max()])
                self.reglines = [
                        axp.plot(rlim, m * 2*np.pi + phi + sl * rlim, **regfmt)
                            for m in range(-3,4)]
                self.axs.set_ylim(0,80)

        out = self.out
        scope = {k:v for k,v in locals().items() if k != 'self'}
        self['c_id'] = fig.canvas.mpl_connect('button_release_event',
                draw_plots(**scope))

    def disconnect_figure(self):
        self.figure('rate-phase-modulation').canvas.mpl_disconnect(self.c.c_id)

    @step
    def print_cell_properties(self):
        """Print cell properties record for the example cell."""
        P = data.properties_dataframe()
        self.out(str(P.loc[self.c.cell_id]))

    @step
    def find_example_LFP_traces(self, dur=1.0, min_spikes=2):
        """Search for low/high firing example intervals of phase code."""
        self.get_datafile(False)

        c_id = self.c.cell_id
        C = data.properties_dataframe().loc[c_id]
        S = session.get(self.c.session_id)

        t = S.lfp.t
        phase = S.lfp.phase
        spk = self.read_array('spikes')

        edges = np.arange(S.tstart, S.tend, dur)
        bins = np.c_[edges[:-1], edges[1:]]
        N_bins = bins.shape[0]

        spk_count, _ = np.histogram(spk, bins=edges)
        spk_phase = self.read_array('spk_phase')

        rmap = self.read_array('rate_pixels')
        rmin, rmax = rmap.min(), rmap.max()

        plow = C.C_rp_phi + C.C_rp_sl * rmin
        phigh = C.C_rp_phi + C.C_rp_sl * rmax

        valid = np.nonzero(spk_count >= min_spikes)[0]
        win_rate = spk_count[valid] / dur
        win_phase = np.zeros_like(win_rate)

        for i in range(win_rate.size):
            ix = np.logical_and(spk>=bins[valid[i],0], spk<bins[valid[i],1])
            win_phase[i] = cmean(spk_phase[ix])

        rmin_diff = np.abs(win_rate - rmin) / (win_rate + rmin)
        rmax_diff = np.abs(win_rate - rmax) / (win_rate + rmax)

        plow_diff = np.abs(cdiff(win_phase, plow)) / np.pi
        phigh_diff = np.abs(cdiff(win_phase, phigh)) / np.pi

        min_diff = rmin_diff + plow_diff
        max_diff = rmax_diff + phigh_diff

        i_min = valid[np.argmin(min_diff)]
        i_max = valid[np.argmin(max_diff)]

        min_bin = bins[i_min]
        max_bin = bins[i_max]

        min_lfp_ix = np.logical_and(t>=min_bin[0], t<min_bin[1])
        min_spk_ix = np.logical_and(spk>=min_bin[0], spk<min_bin[1])

        max_lfp_ix = np.logical_and(t>=max_bin[0], t<max_bin[1])
        max_spk_ix = np.logical_and(spk>=max_bin[0], spk<max_bin[1])

        self.save_array('lfp/min/t', t[min_lfp_ix])
        self.save_array('lfp/min/phase', phase[min_lfp_ix])
        self.save_array('lfp/min/spk', spk[min_spk_ix])

        self.save_array('lfp/max/t', t[max_lfp_ix])
        self.save_array('lfp/max/phase', phase[max_lfp_ix])
        self.save_array('lfp/max/spk', spk[max_spk_ix])

    @step
    def plot_lfp_example_traces(self, figsize=(5.5,3.5)):
        """Test figure plot of LFP phase traces."""
        f = self.figure('LFP-traces-%d' % self.c.cell_id, clear=True,
                figsize=figsize, title='LFP Phase : ' + self.c.label)

        axlow = f.add_subplot(211)
        axhigh = f.add_subplot(212)

        self._plot_lfp_example_traces(axlow, axhigh)

    def _plot_lfp_example_traces(self, axlow, axhigh):
        """Private method for plotting LPF traces and spikes."""
        tkw = dict(ls='-', c='0.2')
        skw = dict(c='slateblue', solid_capstyle='butt')

        def plot_trace_spikes(ax, which, t, phase, spk):
            t0 = t[0]
            t -= t0
            spk -= t0
            trace = 0.65 * np.cos(phase) + 1.5

            ax.plot(t, trace, **tkw)
            sns.rugplot(spk, height=0.3, ax=ax, **skw)
            ax.set_xlim(t.min() - 0.02*t.ptp(), t.max() + 0.08*t.ptp())
            ax.set_ylim(0, 2.3)
            ax.set_axis_off()
            ra.ylabel(ax, '{} rate'.format(which))

        plot_trace_spikes(axlow, 'Low',
                self.read_array('lfp/min/t'),
                self.read_array('lfp/min/phase'),
                self.read_array('lfp/min/spk'))

        plot_trace_spikes(axhigh, 'High',
                self.read_array('lfp/max/t'),
                self.read_array('lfp/max/phase'),
                self.read_array('lfp/max/spk'))

    @step
    def plot_spike_trajectory(self, figsize=(4.5, 4.5)):
        """Create a separate figure of the entire spike-trajectory data."""
        f = self.figure('spike-trajectory-%d' % self.c.cell_id, clear=True,
                figsize=figsize, title='Spike Trajectory')

        # Axis and plot formatting
        trajkw = dict(ls='-', c='k', alpha=0.25, lw=0.25, zorder=-1)
        spkkw = dict(marker='.', s=2.5, c='#FF5555', alpha=0.4, linewidths=0)

        def plot_spike_trajectory(ax):
            plt.plot(self.read_array('X'), self.read_array('Y'), **trajkw)
            plt.scatter(self.read_array('spikes_X'),
                    self.read_array('spikes_Y'), **spkkw)
            ax.axis('scaled')
            ax.set_xlim(0, 100)
            ax.set_ylim(0, 100)
            ax.set_axis_off()
            # ra.quicktitle(ax, self.c.label)

        ax = f.add_subplot(111)
        plot_spike_trajectory(ax)
        ra.xlabel(ax, 'X')
        ra.ylabel(ax, 'Y')

    @step
    def plot_figure(self, levels=18, phasemod_return=False,
        figsize=(8.16326531, 7.19727891)):
        """Create multi-panel introductory figure using example cell data."""
        f = self.figure('intro-cell-%d' % self.c.cell_id, clear=True,
                figsize=figsize, title='Phase Data Intro : ' + self.c.label)

        # Axis and plot formatting
        gshape = (3,3)
        mapkw = dict(cmin=0.0, cmap='CMRmap', cbar=True, levels=levels,
                mask_color='0.9')
        cbargs = dict(pad=0.005, shrink=0.73, aspect=12)

        # Radian colorbar parameters
        radfmt = RadianFormatter(inline=True)
        radticks = FixedLocator([-np.pi,0,np.pi])
        radcbargs = cbargs.copy()
        radcbargs.update(ticks=radticks, format=radfmt)
        phasekw = dict(cmin=-np.pi, cmax=np.pi, cbar=True, levels=levels,
                mask_color='0.9', cbargs=radcbargs)

        maxfivebins = MaxNLocator(prune='both', nbins=5, integer=True)

        def plot_ratemap(ax):
            R = self.read_array('ratemap')
            Rmax = np.floor(R[np.isfinite(R)].max())
            ticks = FixedLocator([0,Rmax/2,Rmax])
            thiscbargs = cbargs.copy()
            thiscbargs.update(ticks=ticks)
            im, cb = maps.plot(R, cbargs=thiscbargs, **mapkw)
            cb.ax.tick_params(pad=0, length=0)
            ax.set_axis_off()
            # ra.xlabel(ax, 'X')

        plot_ratemap(plt.subplot2grid(gshape, (0,0)))

        def plot_phasemap(ax):
            P = self.read_array('phasemap')
            ticks = FixedLocator([-np.pi,0,np.pi])
            cbargs = dict(ticks=ticks, format=radfmt)
            im, cb = maps.plot(P[0], circular=True, **phasekw)
            cb.ax.tick_params(pad=0, length=0)
            ax.set_axis_off()
            # ra.xlabel(ax, 'X')

        plot_phasemap(plt.subplot2grid(gshape, (0,1)))

        def plot_coherencemap(ax):
            P = self.read_array('phasemap')
            im, cb = maps.plot(P, satnorm=True, **phasekw)
            cb.ax.tick_params(pad=0, length=0)
            ax.set_axis_off()
            # ra.xlabel(ax, 'X')

        plot_coherencemap(plt.subplot2grid(gshape, (0,2)))

        def temporal_autocorr(ax):
            lags = self.read_array('autocorr_lags')
            Ct = self.read_array('autocorr')
            Ct = np.ma.masked_where(lags == 0, Ct)
            spikes.corrplot((Ct, lags), ax=ax, colors='b', style='verts')
            ax.set(xlim=(-self.c.maxlag, self.c.maxlag), xlabel='Lag (s)',
                ylim=(0, 1.2 * np.max(Ct[:int(Ct.size/2)])))
            ax.set_yticks([])
            ra.despine(ax, left=False)

        temporal_autocorr(plt.subplot2grid(gshape, (1,0)))

        def phase_dist(ax, bins=binned.DEFAULT_PHASE_BINS):
            phase = self.read_array('spk_phase')
            edges = np.linspace(-np.pi, np.pi, bins + 1)
            centers = self.read_array('phasedist_centers')
            pdistsm = self.read_array('phasedist_smooth')
            ax.hist(phase, bins=edges, color='r', lw=0, alpha=0.9, zorder=0)
            ax.plot(centers, pdistsm, 'k-', alpha=0.7, zorder=1)
            ax.axvline(x=0, ls=':', c='k', alpha=0.8, zorder=2)
            ax.xaxis.set_major_formatter(radfmt)
            ax.set_xticks([-np.pi, -np.pi/2, 0, np.pi/2, np.pi])
            ax.set_yticks([])
            ax.set(xlim=(-np.pi, np.pi), xlabel='Spike phase',
                ylim=(0, 1.2 * np.max(pdistsm)))
            ra.despine(ax, left=False)

        phase_dist(plt.subplot2grid(gshape, (1,1)))

        self._plot_lfp_example_traces(
                plt.subplot2grid((6,3), (2,2)),
                plt.subplot2grid((6,3), (3,2)))

        def rate_phase_regression(ax):
            rate = self.read_array('rate_pixels')
            phase = self.read_array('phase_pixels')
            sl,phi,r,p = self.read_array('rp_corr')
            self.out('N pixels = {}', len(rate))
            self.out('r = {:.4f}, p = {:g}', r, p)

            rlim = np.array([rate.min(), rate.max()])
            rate = np.tile(rate, 2)
            phase = np.r_[phase, phase + 2*np.pi]
            condmap(rate, phase, ax=ax, cmap='gray_r')
            def _plot_regline(m):
                plt.plot(rlim, m * 2 *np.pi + phi + sl * rlim, '-',
                        c='#FF3333', lw=1.6, alpha=0.7)
            for cycle in range(-3,4):
                _plot_regline(cycle)
            ax.set(xlabel='Avg. firing rate (spikes/s)',
                   ylabel='Avg. theta phase')
            # ax.tick_params(pad=0)
            ax.set_yticks([-np.pi, 0, np.pi, 2*np.pi, 3*np.pi])
            ax.set_ylim(-np.pi, 3*np.pi)
            ax.yaxis.set_major_formatter(radfmt)
            ax.xaxis.set_major_locator(maxfivebins)
            ra.despine(ax)

        rate_phase_regression(plt.subplot2grid(gshape, (2,2)))

        if phasemod_return:
            return dict(axes=[plt.subplot2grid(gshape, (2,0)),
                              plt.subplot2grid(gshape, (2,1))])
