"""
Direction-dependent maps of firing rate and phase.
"""

import os

from matplotlib import ticker
from scipy.stats import pearsonr
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from pouty import log
from roto import reports, decorators, circstats, radians, arrays, axis

from .. import store, DATA_ROOT
from ..context import SPCAnalysis, step
from ..lib import cell, motion, labels, space, time, session, theta
from ..tools import maps, adaptive


DEFAULT_NDIRS = 8
DEFAULT_MIN_DUR = 0.25
DEFAULT_MIN_SPEED = 5.0

TEMPLATE_PATH = os.path.join(DATA_ROOT, 'figure-templates')


@decorators.memoize
def fast_directional_intervals(sdata, speed, dbin, dur):
    log('Generating intervals: Rat {}, Day {} ({}) for [{:.2f}, {:.2f}]',
            sdata.attrs['rat'], sdata.attrs['day'], sdata.attrs['comment'],
            dbin[0], dbin[1])
    t, sp, md = sdata.motion.t, sdata.motion.speed_cm, sdata.motion.md
    d0, d1 = dbin
    if d0 > d1: # circular discontinuity, thus OR the conditions
        conds = [np.logical_or(md>=d0, md<d1)]
    else:
        conds = [md>=d0, md<d1] # conditions are ANDed by default
    conds += [sp >= speed]
    return time.intervals_where(t, conds, min_dur=dur)


class MovementDirectionMaps(SPCAnalysis):

    """
    Generate spatial maps across directions of movement.
    """

    logcolor = 'green'

    @step
    def script(self, cell_id_range=None, **mapkwds):
        self.setup(cell_id_range=cell_id_range)
        self.create_directional_bins()
        self.direction_tuning()
        self.generate_movement_maps(**mapkwds)
        self.compute_correlation_matrixes()
        self.compute_allocentricity_indexes()
        self.render_report()

    @step
    def setup(self, cell_id_range=None, min_dur=DEFAULT_MIN_DUR,
        min_speed=DEFAULT_MIN_SPEED):
        proptable = store.get().root.tables.properties
        valid = cell_ids = proptable.col('id')

        if cell_id_range is not None:
            cell_ids = sorted(set(range(*cell_id_range)).intersection(valid))

        self.save_array('cell_ids', cell_ids, title='Cell IDs')
        self.out('Creating maps for {} cells', len(cell_ids))
        self['min_dur'] = min_dur
        self['min_speed'] = min_speed

    @step
    def create_directional_bins(self, ndirs=DEFAULT_NDIRS):
        """Set the limits for each of the directional bins."""
        df = self.get_datafile(False)
        smin, smax = motion.MDIR_RANGE
        self['ndirs'] = ndirs

        edges = np.linspace(smin, smax, ndirs + 1)
        edges -= np.pi/ndirs
        edges[edges<0] += 2*np.pi
        self.save_array('dirbins', edges, title='Direction Bin Edges')

        dirs = circstats.mean_vector(np.c_[edges[:-1], edges[1:]], axis=1)[0]
        dirs[dirs<0] += 2*np.pi
        dirs %= 2*np.pi

        #NOTE: Hack to fix some fp error in the 5/4pi value
        for i in range(ndirs):
            if abs(dirs[i] - (5/4) * np.pi) < 0.01:
                dirs[i] = (5/4) * np.pi

        self.save_array('dirs', dirs, title='Direction Bin Centers')

    @step
    def direction_tuning(self, blur=10.0):
        """Compute directional tuning curves of firing rate."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')
        dirmod = np.empty(cell_ids.shape)

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)

            dmap, angles = S.dirmap(C)
            dmap_blur = arrays.circblur(dmap, degrees=10.0)
            dirmod[i] = dmap_blur.ptp() / dmap_blur.max()

            self.save_array('tuning/id%04d' % c_id, dmap,
                    title='Direction Map: %s' % tag)
            self.save_array('tuning/id%04d_smooth' % c_id, dmap_blur,
                    title='Direction Map (Smoothed): %s' % tag)

        self.save_array('tuning/angles', angles,
                title='Direction Map Angles')
        self.save_array('tuning/modulation_index', dirmod,
                title='Direction Modulation Index')

    @step
    def generate_movement_maps(self, **mapkwds):
        """Create adaptive rate and phase maps in each direction."""
        df = self.get_datafile(False)
        self['map_pixels'] = mapres = mapkwds.get('res', adaptive.NPIXELS)
        cell_ids = self.read_array('cell_ids')
        edges = self.read_array('dirbins')

        for c_id in cell_ids:
            C = cell.record(c_id)
            S = cell.session_for(c_id, **mapkwds)
            R = np.empty((self.c.ndirs, mapres, mapres))
            P = np.empty((self.c.ndirs, 2, mapres, mapres))
            tag = labels.session_cell_id(c_id)
            self.out('Computing maps for {}', tag)
            for i in range(self.c.ndirs):
                dbin = (edges[i], edges[i+1])
                intervals = fast_directional_intervals(S, self.c.min_speed,
                        dbin, self.c.min_dur)
                R[i] = S.ratemap(C, intervals=intervals, speed=None)
                P[i] = S.phasemap(C, intervals=intervals, speed=None)

            self.save_array('ratemaps/id%04d' % c_id, R,
                    title='Directional Ratemaps: {}'.format(tag))
            self.save_array('phasemaps/id%04d' % c_id, P,
                    title='Directional Phasemaps: {}'.format(tag))

    @step
    def compute_correlation_matrixes(self):
        """Compute pair-wise spatial correlation matrixes for display."""
        df = self.get_datafile(False)

        for i, c_id in enumerate(self.get_node('cell_ids')):
            tag = labels.session_cell_id(c_id)
            R = self.get_node('ratemaps/id%04d' % c_id)
            P = self.get_node('phasemaps/id%04d' % c_id)[:,0,:]  # phase mean
            C_R = np.zeros((self.c.ndirs, self.c.ndirs))
            C_P = np.zeros_like(C_R)

            for j in range(self.c.ndirs):
                for k in range(j + 1, self.c.ndirs):
                    Rvalid = np.logical_and(np.isfinite(R[j]),
                            np.isfinite(R[k]))
                    if Rvalid.any():
                        C_R[j,k] = C_R[k,j] = pearsonr(
                                R[j][Rvalid], R[k][Rvalid])[0]

                    Pvalid = np.logical_and(np.isfinite(P[j]),
                            np.isfinite(P[k]))
                    if Pvalid.any():
                        C_P[j,k] = C_P[k,j] = theta.phaseregress(
                                P[j][Pvalid], P[k][Pvalid])[2]

            self.save_array('ratecorrs/id%04d' % c_id, C_R,
                    title='Rate Directional Correlation: {}'.format(tag))
            self.save_array('phasecorrs/id%04d' % c_id, C_P,
                    title='Phase Directional Correlation: {}'.format(tag))

    @step
    def compute_allocentricity_indexes(self):
        """Compute measures of rate and phase allocentricity."""
        df = self.get_datafile(False)
        rmapgrp = store.get().root.ratemaps
        pmapgrp = store.get().root.phasemaps
        cell_ids = self.read_array('cell_ids')
        rate_index = np.empty(cell_ids.size)
        phase_index = np.empty(cell_ids.size)

        for i, c_id in enumerate(cell_ids):
            tag = labels.session_cell_id(c_id)
            self.out('Computing {}', tag)
            Rfull = rmapgrp._f_get_child('id%04d' % c_id)
            Pfull = pmapgrp._f_get_child('id%04d' % c_id)[0]  # phase mean
            R = self.get_node('ratemaps/id%04d' % c_id)
            P = self.get_node('phasemaps/id%04d' % c_id)[:,0,:]  # phase mean

            rate_index[i] = space.ratemap_allocentricity(Rfull, R)
            phase_index[i] = space.phasemap_allocentricity(Pfull, P)

        self.save_array('properties/rate_allocentricity', rate_index,
                title='Spatial Ratemap Allocentricity Index')
        self.save_array('properties/phase_allocentricity', phase_index,
                title='Spatial Phasemap Allocentricity Index')

    @step
    def render_report(self, levels=12):
        """Make a report showing all directional maps and other data."""
        df = store.get()
        rmapgrp = df.root.ratemaps
        pmapgrp = df.root.phasemaps
        cell_ids = self.read_array('cell_ids')
        dirs = self.read_array('dirs')
        angles = self.read_array('tuning/angles')
        mod_index = self.read_array('tuning/modulation_index')
        rate_index = self.read_array('properties/rate_allocentricity')
        phase_index = self.read_array('properties/phase_allocentricity')

        mpl.rc('xtick', labelsize=7.0)
        mpl.rc('ytick', labelsize=7.0)
        mpl.rc('axes', labelsize='x-small')
        mpl.rc('axes', titlesize='x-small')

        radfmt = radians.RadianFormatter(inline=False)
        radinline = radians.RadianFormatter(inline=True)
        threeticks = ticker.LinearLocator(numticks=3)

        cbargs = {'pad':0.0, 'fraction':0.07, 'format':'%.1f',
                  'ticks':threeticks}
        pcbargs = cbargs.copy()
        pcbargs.update({'format':radfmt})
        mkw = dict(cbar=True, cbargs=cbargs)
        pmkw = dict(cbar=True, cbargs=pcbargs)

        templpath = os.path.join(TEMPLATE_PATH,
                'directional-report-%02d' % self.c.ndirs)
        desc = self.figinfo(templpath)
        width, height = desc['figsize']
        axrect = desc['axes']
        report = reports.PageReport(self.path('maps'), width=width,
                height=height, numloc='bottom right')

        for i, c_id in enumerate(cell_ids):
            fig = report.page
            tag = labels.session_cell_id(c_id)
            fig.suptitle(tag)
            self.out('Rendering {}', tag)

            ax = plt.axes(axrect['tuning'])
            D = self.read_array('tuning/id%04d' % c_id)
            Ds = self.read_array('tuning/id%04d_smooth' % c_id)
            ax.plot(angles, D, 'r-', lw=1, alpha=0.6)
            ax.plot(angles, Ds, 'b-', lw=2, alpha=0.6)
            ax.set_title('Direction Index = {:.2f}'.format(mod_index[i]),
                    va='top')
            axis.despine(ax)
            ax.set_ylim(0, 1.2 * max(1.0, np.nanmax(D)))
            ax.set_ylabel('Rate')
            ax.set_xlabel('Direction (rad)')
            ax.xaxis.set_major_formatter(radinline)
            ax.set_xticks([0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi])
            ax.set_xlim(0, 2 * np.pi)

            ax = plt.axes(axrect['Rcorr'])
            C_R = self.read_array('ratecorrs/id%04d' % c_id)
            plt.imshow(C_R, origin='lower', interpolation='nearest',
                    aspect='auto', cmap='bwr_r', vmin=-1, vmax=1)
            plt.colorbar(**cbargs)
            ax.set_title(
                'Ratemap Correlations\nAllocentricity = {:.3f}'.format(
                    rate_index[i]))
            ax.set(xticks=[], xlabel='Directions', yticks=[],
                ylabel='Directions')

            ax = plt.axes(axrect['Pcorr'])
            C_P = self.read_array('phasecorrs/id%04d' % c_id)
            plt.imshow(C_P, origin='lower', interpolation='nearest',
                    aspect='auto', cmap='bwr_r', vmin=-1, vmax=1)
            plt.colorbar(**cbargs)
            ax.set_title(
                'Phasemap Correlations\nAllocentricity = {:.3f}'.format(
                    phase_index[i]))
            ax.set(xticks=[], xlabel='Directions', yticks=[],
                ylabel='Directions')

            R = df.get_node(rmapgrp, 'id%04d' % c_id).read()
            P = df.get_node(pmapgrp, 'id%04d' % c_id).read()
            Rdir = self.get_node('ratemaps/id%04d' % c_id)
            Pdir = self.get_node('phasemaps/id%04d' % c_id)

            rmax = max(np.nanmax(R), np.nanmax(Rdir))

            for j in range(self.c.ndirs):
                ax = plt.axes(axrect['R%d' % j])
                maps.plot(Rdir[j], cmin=0.0, cmax=rmax, levels=levels)
                ax.set_title(r'Ratemap: %s' % radians.prettify(dirs[j],
                    inline=True))
                ax.set_axis_off()

            ax = plt.axes(axrect['R'])
            maps.plot(R, cmin=0.0, cmax=rmax, levels=levels, **mkw)
            ax.set_title('Ratemap')
            ax.set_axis_off()

            for j in range(self.c.ndirs):
                ax = plt.axes(axrect['P%d' % j])
                maps.plot(Pdir[j], satnorm=True, levels=levels)
                ax.set_title(r'Phasemap: %s' % radians.prettify(dirs[j],
                    inline=True))
                ax.set_axis_off()

            ax = plt.axes(axrect['P'])
            maps.plot(P, satnorm=True, levels=levels, **pmkw)
            ax.set_title('Phasemap')
            ax.set_axis_off()

            ax = plt.axes(axrect['phasedisc'])
            maps.phasedisc(res=256, levels=levels, inline=True, zoom=0.75,
                labelsize='x-small')

            report.save_page()

        report.close()
        mpl.rc_file_defaults()
