"""
Trajectory movie of head and movement direction.
"""

import matplotlib.pyplot as plt
import numpy as np

from roto import lines, movies

from ..context import SPCAnalysis, step
from ..lib import labels, motion, parsers


EXAMPLE_S_ID = 27


class MovementTrajectoryMovie(SPCAnalysis):

    """
    Movie of trajectory with arrows for HD/MD motion signals.
    """

    @step
    def make(self, s_id=EXAMPLE_S_ID, name=None, arrowlen=1.0, duration=5.0,
        speedup=2.0, t0=None, snapshot_only=False):
        """Make the trajectory data movie."""
        mdata = motion.MotionData(s_id)

        if name is None:
            name = 'traj-movie-s%03d' % s_id
        if not name.endswith('.mp4'):
            name += '.mp4'
        mpath = self.path(name)

        T, X, Y = [getattr(mdata, k) for k in ('t', 'x', 'y')]

        if t0 is None:
            t0 = T[0]

        movie = movies.FigureMovie(mpath, duration=duration)

        ax = plt.axes([0,0,1,1])  # full-figure axes
        ax.axis('equal')
        ax.set(xticks=[], yticks=[])
        for spine in ax.spines.values():
            spine.set_visible(False)

        akw = dict(width=0.15, lw=1, length_includes_head=True, zorder=4)
        arrows = []

        ax.plot(X, Y, 'k-', lw=0.5, alpha=0.4, zorder=-1)
        trace = lines.TimeTracePlot(T, X, Y, dt=15.0, tau=5.0, cmap='CMRmap_r', lw=3.5)
        h = ax.plot([0], [0], 'wo', mew=1.5, mfc='w', mec='k', ms=9, zorder=5)
        head = h[0]

        def update(t):
            t1 = speedup * t + t0
            trace.plot(t1, ax=ax)

            x_t = mdata.F('x', t1)
            y_t = mdata.F('y', t1)
            hd_t = np.deg2rad(mdata.F('hd_orig', t1))
            md_t = mdata.F('md', t1)
            # v_t = mdata.F('speed_cm', t1)

            head.set_xdata([x_t])
            head.set_ydata([y_t])

            if arrows:
                [arr.remove() for arr in arrows]
                while arrows:
                    arrows.pop()
            sgain = 15.0  # v_t / 1.5
            mda = ax.arrow(x_t, y_t, sgain * arrowlen * np.cos(md_t),
                sgain * arrowlen * np.sin(md_t), shape='left', ec='r',
                fc='r', **akw)
            hda = ax.arrow(x_t, y_t, sgain * arrowlen * np.cos(hd_t),
                sgain * arrowlen * np.sin(hd_t), shape='right', ec='b',
                fc='b', **akw)
            arrows.extend([hda, mda])

        def label(t):
            t1 = speedup * t + t0
            return 't = %.2fs' % t1

        movie.set_update_function(update)
        movie.set_label_function(label)
        movie.set_label_fmt(size='large')

        if snapshot_only:
            movie.snapshot([0,duration/2,duration])
            return

        movie.render()
        movie.view()
