"""
Descriptive analysis of criteria-selected phaser cells.
"""

from collections import namedtuple
from functools import reduce

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl
import scipy.stats as stat
import scipy.optimize as opt
import seaborn as sns

from roto.radians import RadianFormatter
from roto.circstats import mean_angle
from roto.stats import smooth_pdf, empirical_CI, t_welch
from roto.plots import grouped_bar_plot
from roto.misc import merge_two_dicts
import roto.axis as ra

from .. import store
from ..context import SPCAnalysis, step
from ..lib import phasers, data, labels, anatomy
from ..lib.data import dataframe_from_table
from ..lib.motion import arena_size
from ..lib.phasers import PHASER_IPHASE_IMIN, PHASER_RP_SMIN

IRATE_PMAX = 0.02
IDIR_PMAX = 0.02
ISPEED_PMAX = 0.02

POSCOLOR = 'sandybrown'
NEGCOLOR = 'slateblue'
NONCOLOR = 'dimgray'
PHCOLOR = (0.3, 0.73372549, 0.61568627)  # lightened colorblind 'g'
SIGCOLOR = (0.3, 0.61294117, 0.78862745)  # lightened colorblind 'b'
PI2 = 2*np.pi

MeanSD = namedtuple('MeanSD', ['mean', 'SD'])


class PhaserCellDescription(SPCAnalysis):

    """
    Compute and show basic properties of the phaser cell subpopulation.
    """

    def cleanup(self):
        store.close()

    def setup(self):
        """Cell properties dataframe with phaser subtype columns."""
        P = phasers.properties_dataframe()

        self.c.N_rats = len(set(P.loc[P.phaser, 'rat']))
        self.c.N_sessions = len(set(P.loc[P.phaser, 'session_id']))

        self.out('Found {} phasers across {} rats and {} sessions.',
                P.phaser.sum(), self.c.N_rats, self.c.N_sessions)

    @step
    def phaser_cell_stats(self, ddof=0):
        """Basic statistics for theta cell recordings."""
        props = phasers.properties_dataframe()
        phaser = props.loc[props.phaser]
        nonphaser = props.loc[props.phaser == False]

        phasers_by_rat = phaser.groupby('rat')
        nonphasers_by_rat = nonphaser.groupby('rat')

        stats = [
            ('Avg. Rate', 'avg_rate'),
            ('Max. Rate', 'ratemap_max'),
            ('Theta Index', 'theta_mod'),
            ('Rhythmicity', 'theta_rhythm'),
            ('Phase Coherence', 'phase_coh'),
            ('Freq. Estimate', 'f_hat'),
            ('Rate/phase r', 'C_rp_r'),
            ('Rate/phase slope', 'C_rp_sl'),
            ('Total Phase Shift', 'rp_shift')
        ]

        data = [(phasers_by_rat, 'Phaser'), (nonphasers_by_rat, 'Non-Phaser')]

        for name, field in stats:
            for props, label in data:
                values = props[field].mean()
                avg = MeanSD(mean=values.mean(), SD=values.std(ddof=ddof))
                self.out('{}: {} = {:.3f} +- {:.3f}', label, name, *avg)

    def plot_violin_strips(self, ax, ylim, phcolor, sigcolor, df, values, name,
        valname):
        """Violin plot of cell properties values."""
        if type(values) is str:
            values = [values]

        plot_data = pd.melt(df, id_vars=['id', 'phaser'], value_vars=values,
                var_name=name, value_name=valname)

        sns.violinplot(x=name, y=valname, data=plot_data, hue='phaser',
                cut=0.0, scale='count', split=True, ax=ax,
                palette={True:phcolor, False:sigcolor}, inner='quartiles')

        # Strip plot for phaser selection
        for i, col in enumerate(values):
            nondata = df.loc[df.phaser == False, col].values
            ax.scatter(i - 0.55 - 0.025 * np.random.randn(nondata.size),
                    nondata, s=20, c=sigcolor, linewidths=0, alpha=0.5)
            data = df.loc[df.phaser, col].values
            ax.scatter(i + 0.4 + 0.025 * np.random.randn(data.size),
                    data, s=20, c=phcolor, linewidths=0, alpha=0.5)

        ax.set_xlim(-0.7, len(values) - 0.3)
        ax.set_ylim(*ylim)
        for x0 in range(len(values)):
            ax.axvline(x=x0, color='0.5', linewidth=1)

        return ax

    @step
    def rate_rhythm_comparison(self, phcolor=PHCOLOR, sigcolor=SIGCOLOR,
        figsize=(6,5.5)):
        """Univarite distros for max. spatial rate and theta rhythmicity."""
        props = phasers.properties_dataframe()
        props = props.loc[props.I_phase_p <= phasers.PHASER_IPHASE_PMAX]

        f = self.figure('rate-rhythm', clear=True, figsize=figsize,
                title='Rate/Rhythm Comparison for Phasers/Non-Phasers')

        # Max. spatial firing rate distributions
        self.plot_violin_strips(f.add_subplot(121), (0.0, 30.0), phcolor,
                sigcolor, props, 'ratemap_max', 'Spatial peak firing rate',
                'Firing rate (spikes/s)')

        # Theta rhythmicity distributions
        self.plot_violin_strips(f.add_subplot(122), (0.0, 1.0), phcolor,
                sigcolor, props.loc[props.theta_rhythm != -1],
                'theta_rhythm', 'Theta rhythmicity index', 'Index')

        sns.despine()

    @step
    def phase_shift_accuracy(self, poscolor=POSCOLOR, negcolor=NEGCOLOR,
        shiftbins=36, model_pos_shift=1.2, model_neg_shift=-1.5,
        figsize=(6.6,5.15)):
        """Distributions of phase shift and implied place errors."""
        P = phasers.properties_dataframe()
        recs = data.dataframe_from_table('/recordings', index='id')
        P['cell_id'] = recs.cell_id

        f = self.figure('shift-error', clear=True, figsize=figsize,
                title='Phase Shift x Implied Place-Position Error')

        axe = f.add_subplot(121)
        axd = f.add_subplot(122)

        # Phase shifts for neg/pos phaser cells
        shiftneg = P.rp_shift[P.subtype == 'negative']
        shiftpos = P.rp_shift[P.subtype == 'positive']

        # Compute the implied errors
        radius = arena_size(0) / 2
        Ineg = P.I_phase[P.subtype == 'negative']
        Eneg = radius / np.sqrt(2**Ineg)
        Ipos = P.I_phase[P.subtype == 'positive']
        Epos = radius / np.sqrt(2**Ipos)
        P['error'] = radius / np.sqrt(2**P.I_phase[P.phaser])

        # Plot formatting
        errfmt = dict(s=24, linewidths=0.5, alpha=0.8)
        critfmt = dict(closed=True, hatch=r'\\\\\\', edgecolor=(1,0,0),
                alpha=0.7, lw=0.8, facecolor='none', zorder=-5)
        bins = np.linspace(-PI2, PI2, shiftbins+1)
        modelfmt = dict(lw=1.0, ls='--', zorder=2)
        smax = 6.48

        def finish_yaxis(ax):
            ax.set_ylim(-smax, smax)
            ax.set_yticks(np.linspace(-PI2, PI2, 5))
            ax.yaxis.set_major_formatter(RadianFormatter(inline=False))

        def plot_criteria(ax, pad=0.1):
            x0, x1 = ax.get_xlim()
            x1 += pad
            y0, y1 = -smax - pad, smax + pad
            Emax = radius / np.sqrt(2**PHASER_IPHASE_IMIN)
            smin = PHASER_RP_SMIN
            verts = np.array(
                [[x1, x1, Emax, Emax, x0,    x0,   Emax, Emax],
                 [y1, y0, y0,  -smin, -smin, smin, smin, y1  ]]).T
            p = mpl.patches.Polygon(verts, clip_box=ax.bbox, **critfmt)
            ax.add_artist(p)

        def plot_errors(ax):
            ax.scatter(Eneg, shiftneg, c=negcolor, label='Negative', **errfmt)
            ax.scatter(Epos, shiftpos, c=poscolor, label='Positive', **errfmt)
            ax.legend(loc=6)
            ax.set_xlabel('Implied position error (cm)')
            ax.set_ylabel('Total phase shift')
            plot_criteria(ax)
            finish_yaxis(ax)
        plot_errors(axe)

        # Output regression and summary statistics
        _, _, rneg, pneg, _ = stat.linregress(Eneg, shiftneg)
        self.out('Negative error regression: r = {:.4f}, p = {:g}',
                rneg, pneg)
        _, _, rpos, ppos, _ = stat.linregress(Epos, shiftpos)
        self.out('Positive error regression: r = {:.4f}, p = {:g}',
                rpos, ppos)
        self.out('Negative error = {:.4f}, {:.4f} s.e.m.', Eneg.mean(),
                Eneg.std()/np.sqrt(Eneg.size))
        self.out('Positive error = {:.4f}, {:.4f} s.e.m.', Epos.mean(),
                Epos.std()/np.sqrt(Epos.size))

        # Test for negative/positive differences
        ncells = P.loc[P.subtype == 'negative'].groupby('cell_id').error.mean()
        pcells = P.loc[P.subtype == 'positive'].groupby('cell_id').error.mean()
        Nncells = len(ncells)
        Npcells = len(pcells)
        t, p = t_welch(ncells, pcells, tails=2)
        self.out('Welch\'s t-test ({},{}): t = {:.6f}, p = {:.6f}', Nncells,
                Npcells, t, p)

        def dist_kde_plot(ax, shifts, color, label):
            sns.distplot(shifts.values, ax=ax, kde=False, hist=True,
                    bins=bins, hist_kws={'alpha':0.7}, vertical=True,
                    norm_hist=False, color=color)

            kx, ky = smooth_pdf(shifts, sd=np.pi/4)
            ky *= 0.4 * shifts.size
            ax.plot(ky, kx, c=color, alpha=0.85, zorder=1,
                    solid_capstyle='butt', label=label)

        def plot_distros(ax):
            dist_kde_plot(ax, shiftneg, negcolor, 'Negative')
            dist_kde_plot(ax, shiftpos, poscolor, 'Positive')
            ax.axhline(y=model_neg_shift, c=negcolor, **modelfmt)
            ax.axhline(y=model_pos_shift, c=poscolor, **modelfmt)
            ax.legend(loc='upper right', frameon=False)
            ax.set_xlabel('Phaser cell count')
            finish_yaxis(ax)
        plot_distros(axd)

        sns.despine()

    @step
    def spatial_coherence(self, CI_alpha=0.1, poscolor=POSCOLOR,
        negcolor=NEGCOLOR, noncolor=NONCOLOR, show_non=False,
        figsize=(6.6,5.15)):
        """Analysis of spatial coherence of the phaser code."""
        P = phasers.properties_dataframe()
        recs = data.dataframe_from_table('/recordings', index='id')
        P['cell_id'] = recs.cell_id
        P['coh'] = 0.0
        df = store.get()

        # Phase coherence values across space for neg/pos phaser cells
        non = []
        neg = []
        pos = []
        shiftnon = []
        shiftneg = []
        shiftpos = []

        self.out('Reading phasemaps...')
        for row in P.itertuples():
            coh_map = df.get_node('/phasemaps/id%04d' % row.Index).read()[1]
            coh = coh_map[np.isfinite(coh_map)]
            if coh.size == 0:
                continue
            mu = coh.mean()
            CI = empirical_CI(coh, alpha=CI_alpha)
            cdata = (mu, mu - CI.lower, CI.upper - mu, coh.max())
            if row.phaser == True:
                P.loc[row.Index, 'coh'] = mu
                if row.subtype == 'negative':
                    neg.append(cdata)
                    shiftneg.append(row.rp_shift)
                    self.box(filled=True, color='purple')
                else:
                    pos.append(cdata)
                    shiftpos.append(row.rp_shift)
                    self.box(filled=True, color='brown')
            else:
                non.append(cdata)
                shiftnon.append(row.rp_shift)
                self.box(filled=False, color='white')
        self.newline()

        non = np.atleast_1d(non).T
        neg = np.atleast_1d(neg).T
        pos = np.atleast_1d(pos).T

        f = self.figure('phase-coherence', clear=True, figsize=figsize,
                title='Spatial Phase-Coherence of Phaser Cells')

        ax = f.add_subplot(121)

        # Plot formatting
        critfmt = dict(closed=True, hatch=r'\\\\\\', edgecolor=(1,0,0),
                alpha=0.7, lw=0.8, facecolor='none', zorder=-5)
        errfmt = dict(s=24, linewidths=0.5, alpha=0.8)
        nonfmt = dict(s=1, linewidths=0.5, zorder=-10, c=noncolor, alpha=0.3)
        CIfmt = dict(fmt='none', elinewidth=0.5, capsize=0, zorder=-1)
        smax = 6.48

        def finish_yaxis(ax, pad=0.05):
            ax.set_xlim(-pad, 1 + pad)
            ax.set_ylim(-smax, smax)
            ax.set_yticks(np.linspace(-PI2, PI2, 5))
            ax.yaxis.set_major_formatter(RadianFormatter(inline=False))

        def plot_criteria(ax, pad=0.1):
            x0, x1 = -pad, 1 + pad
            smin = PHASER_RP_SMIN
            verts = np.array(
                [[x0,   x1,   x1,    x0],
                 [smin, smin, -smin, -smin]]).T
            p = mpl.patches.Polygon(verts, clip_box=ax.bbox, **critfmt)
            ax.add_artist(p)

        def plot_coherence(ax):
            if show_non:
                ax.scatter(non[0], shiftnon, label='Nonphasers', **nonfmt)
            ax.errorbar(neg[0], shiftneg, xerr=neg[1:3], ecolor=negcolor,
                    **CIfmt)
            ax.scatter(neg[0], shiftneg, c=negcolor, label='Negative',
                    **errfmt)
            ax.errorbar(pos[0], shiftpos, xerr=pos[1:3], ecolor=poscolor,
                    **CIfmt)
            ax.scatter(pos[0], shiftpos, c=poscolor, label='Positive',
                    **errfmt)

            # ax.legend(loc=2)
            ax.set_xlabel('Spatial phase coherence')
            ax.set_ylabel('Total phase shift')
            plot_criteria(ax)
            finish_yaxis(ax)
        plot_coherence(ax)

        # Report the range and median of the maximum coherence
        negmax = neg[3]
        posmax = pos[3]
        allmax = np.r_[negmax, posmax]
        self.out('Negative max. range = [{:.3f}, {:.3f}], median = {:.3f}',
                np.min(negmax), np.max(negmax), np.median(negmax))
        self.out('Positive max. range = [{:.3f}, {:.3f}], median = {:.3f}',
                np.min(posmax), np.max(posmax), np.median(posmax))
        self.out('All max. range = [{:.3f}, {:.3f}], median = {:.3f}',
                np.min(allmax), np.max(allmax), np.median(allmax))

        self.out('Negative coh. = {:.4f}, {:.4f} s.e.m.', neg[0].mean(),
                neg[0].std()/np.sqrt(neg[0].size))
        self.out('Positive coh. = {:.4f}, {:.4f} s.e.m.', pos[0].mean(),
                pos[0].std()/np.sqrt(pos[0].size))

        # Test for negative/positive differences
        ncells = P.loc[P.subtype == 'negative'].groupby('cell_id').coh.mean()
        pcells = P.loc[P.subtype == 'positive'].groupby('cell_id').coh.mean()
        Nncells = len(ncells)
        Npcells = len(pcells)
        t, p = t_welch(ncells, pcells, tails=2)
        self.out('Welch\'s t-test ({},{}): t = {:.6f}, p = {:.6f}', Nncells,
                Npcells, t, p)

        sns.despine()

    @step
    def phase_shift_distros(self, poscolor=POSCOLOR, negcolor=NEGCOLOR,
        phasemap_avg=True, phase_dist_bins=48, figsize=(10,5.5)):
        """Phase shift and distributions for phaser subtypes."""
        P = phasers.properties_dataframe()

        f = self.figure('phase-shift-distros', clear=True, figsize=figsize,
                title='Phaser Subtypes by Phase Shift/Preference')

        # Regressed spatial phase shifts
        H = P.loc[P.phaser]
        plow = H.C_rp_phi + H.C_rp_sl * H.ratemap_min
        phigh = H.C_rp_phi + H.C_rp_sl * H.ratemap_max
        clevel = np.ceil(3 * np.abs(H.C_rp_r)) / 2
        subtype = H.subtype == 'positive'

        plow = plow.values
        phigh = phigh.values
        clevel = clevel.values
        subtype = subtype.values

        # Plot low-to-high firing rate phase shifts
        axs = plt.subplot2grid((1,4), (0,0))
        for i in range(len(H)):
            c = {True:poscolor, False:negcolor}[subtype[i]]
            for k in range(-2,3):
                axs.plot([0,1], [2*np.pi*k + plow[i], 2*np.pi*k + phigh[i]],
                        c=c, lw=clevel[i], alpha=0.7, zorder=-plow[i]-10)

        # axlw = mpl.rcParams['axes.linewidth']
        # axs.axvline(x=0, c='k', lw=1.2, zorder=100)
        axs.axvline(x=1, c='k', lw=1.4, zorder=100)
        ylim = -2*np.pi, 2*np.pi
        axs.set_ylim(*ylim)
        axs.set_xlim(0, 1)
        axs.set_xticks([0,1])
        axs.set_xticklabels(['Min.', 'Peak'])
        axs.set_xlabel('Spatial firing rate')
        axs.set_yticks(np.linspace(ylim[0], ylim[1], 5))
        axs.set_ylabel('Rate-phase regression')
        axs.yaxis.set_major_formatter(RadianFormatter(inline=False))
        axs.yaxis.set_tick_params(which='major', pad=1)

        # Preferred phase distributions
        axp = plt.subplot2grid((1,4), (0,1), colspan=3)
        bins = np.linspace(-3*np.pi, 3*np.pi, phase_dist_bins+1)

        # Compute the preferred phase from the phase-average map pixels
        pref_phase = 'phase_avg'
        if phasemap_avg:
            df = store.get()
            self.out('Computing preferred phase from phasemaps...')
            pmap_pref = np.zeros(len(P))
            for i, cell_id in enumerate(P.index.values):
                pmap = df.get_node('/phasemaps/id%04d' % cell_id).read()
                valid = np.isfinite(pmap[0])
                pmap_pref[i] = mean_angle(pmap[0][valid])
            pmap_pref[np.isnan(pmap_pref)] = 0.0
            P['pmap_pref'] = pmap_pref
            pref_phase = 'pmap_pref'

        def dist_kde_plot(subtype, color):
            data = P.loc[P.subtype == subtype, pref_phase].values
            data = np.r_[data - 2*np.pi, data, data + 2*np.pi]

            sns.distplot(data, ax=axp, kde=False, hist=True, bins=bins,
                hist_kws={'alpha':0.7}, norm_hist=False, color=color)

            kx, ky = smooth_pdf(data, sd=np.pi/4)
            valid = np.logical_and(kx>=ylim[0], kx<=ylim[1])
            kx, ky = kx[valid], ky[valid]
            ky *= 2 * (P.subtype == subtype).sum()
            axp.plot(kx, ky, c=color, alpha=0.85, zorder=1,
                    solid_capstyle='butt', label=subtype.title())

        dist_kde_plot('negative', negcolor)
        dist_kde_plot('positive', poscolor)
        axp.legend(loc='upper center', frameon=False)

        axp.set_xlim(*ylim)
        axp.set_xticks(np.linspace(ylim[0], ylim[1], 9))
        axp.xaxis.set_major_formatter(RadianFormatter(inline=True))
        axp.set_xlabel('Preferred theta phase of spikes')
        axp.set_ylabel('Count')

        sns.despine()

    @step
    def phase_information_comparison(self, poscolor=POSCOLOR,
        negcolor=NEGCOLOR, ymax=1.8, figsize=(9.5, 4.5)):
        """Show distributions of phase information vs. other types."""
        P = phasers.properties_dataframe()

        # Add hippocampal flag to dataframe for differential display
        R = data.dataframe_from_table('/recordings', index='id')
        P['hipp'] = (R.area == anatomy.area_code('hippocampus'))

        def optslope(which):
            x = P.loc[P.phaser, 'I_phase'].values
            y = P.loc[P.phaser, 'I_' + which].values
            m0 = stat.linregress(x, y)[0]
            mopt = float(opt.leastsq(lambda m,x,y: y-m*x, m0, args=(x,y))[0])
            self.out('Optimal {}-phase slope = {:.6f}', which, mopt)
            return mopt

        f = self.figure('phase-info-comparison', clear=True, figsize=figsize,
                title='Comparison of Spatial Phase Information')

        axm = f.add_subplot(131)
        axd = f.add_subplot(132, sharex=axm)
        axs = f.add_subplot(133, sharex=axm)

        # Scatter parameters
        skw = dict(s=48, linewidths=0.5, edgecolor='w', marker='o', alpha=0.7)
        hippskw = merge_two_dicts(skw, dict(s=56, edgecolor='0.3', marker='*',
            alpha=0.7, zorder=5))
        sigkw = merge_two_dicts(skw, dict(s=36, linewidths=1.2, alpha=0.85,
            marker='x', zorder=10))
        psigkw = merge_two_dicts(sigkw, {'c':'r'})
        nsigkw = merge_two_dicts(sigkw, {'c':'m'})
        parkw = dict(ls=':', c='0.4', zorder=-10)
        regkw = dict(ls='-', c='0.7', zorder=-5)

        # Data indexes
        posix = (P.subtype == 'positive').values
        negix = (P.subtype == 'negative').values
        poshippix = np.logical_and(posix, P.hipp)    # positive, hippocampal
        posnohippix = np.logical_and(posix, ~P.hipp) # positive, not hipp.
        neghippix = np.logical_and(negix, P.hipp)    # negative, hippocamapl
        negnohippix = np.logical_and(negix, ~P.hipp) # negative, not hipp.
        posrateix = np.logical_and(posix, P.I_rate_p > IRATE_PMAX)
        negrateix = np.logical_and(negix, P.I_rate_p > IRATE_PMAX)
        posdirix = np.logical_and(posix, P.I_dir_p > IDIR_PMAX)
        negdirix = np.logical_and(negix, P.I_dir_p > IDIR_PMAX)
        posspeedix = np.logical_and(posix, P.I_speed_p > ISPEED_PMAX)
        negspeedix = np.logical_and(negix, P.I_speed_p > ISPEED_PMAX)

        # Legend labels
        PosHipp =   'Positive (hipp.)'
        NegHipp =   'Negative (hipp.)'
        PosNoHipp = 'Positive (not hipp.)'
        NegNoHipp = 'Negative (not hipp.)'
        PosNS =     'Positive (n.s.)'
        NegNS =     'Negative (n.s.)'

        # Scatter plot of I_phase vs. I_rate
        axm.scatter(P.loc[poshippix].I_phase, P.loc[poshippix].I_rate,
                facecolor=poscolor, **hippskw)
        axm.scatter(P.loc[neghippix].I_phase, P.loc[neghippix].I_rate,
                facecolor=negcolor, **hippskw)
        axm.scatter(P.loc[posnohippix].I_phase, P.loc[posnohippix].I_rate,
                facecolor=poscolor, **skw)
        axm.scatter(P.loc[negnohippix].I_phase, P.loc[negnohippix].I_rate,
                facecolor=negcolor, **skw)
        axm.scatter(P.loc[posrateix].I_phase, P.loc[posrateix].I_rate,
                **psigkw)
        axm.scatter(P.loc[negrateix].I_phase, P.loc[negrateix].I_rate,
                **nsigkw)

        axm.set(xlabel=r'$I_{\rm phase}$ (bits)',
                ylabel='Spatial rate information (bits/spike)')
        axm.set_xlim(left=0)

        # Scatter plot of I_phase vs. I_dir
        axd.scatter(P.loc[poshippix].I_phase, P.loc[poshippix].I_dir,
                label=PosHipp, facecolor=poscolor, **hippskw)
        axd.scatter(P.loc[neghippix].I_phase, P.loc[neghippix].I_dir,
                label=NegHipp, facecolor=negcolor, **hippskw)
        axd.scatter(P.loc[posnohippix].I_phase, P.loc[posnohippix].I_dir,
                label=PosNoHipp, facecolor=poscolor, **skw)
        axd.scatter(P.loc[negnohippix].I_phase, P.loc[negnohippix].I_dir,
                label=NegNoHipp, facecolor=negcolor, **skw)
        axd.scatter(P.loc[posdirix].I_phase, P.loc[posdirix].I_dir,
                label=PosNS, **psigkw)
        axd.scatter(P.loc[negdirix].I_phase, P.loc[negdirix].I_dir,
                label=NegNS, **nsigkw)

        axd.set(xlabel=r'$I_{\rm phase}$ (bits)',
                ylabel='Directional information (bits/spike)')
        axd.set_xlim(left=0)
        axd.legend(loc='lower left', frameon=False, bbox_to_anchor=(0.03,0.7))

        # Scatter plot of I_phase vs. I_speed
        axs.scatter(P.loc[poshippix].I_phase, P.loc[poshippix].I_speed,
                facecolor=poscolor, **hippskw)
        axs.scatter(P.loc[neghippix].I_phase, P.loc[neghippix].I_speed,
                facecolor=negcolor, **hippskw)
        axs.scatter(P.loc[posnohippix].I_phase, P.loc[posnohippix].I_speed,
                facecolor=poscolor, **skw)
        axs.scatter(P.loc[negnohippix].I_phase, P.loc[negnohippix].I_speed,
                facecolor=negcolor, **skw)
        axs.scatter(P.loc[posspeedix].I_phase, P.loc[posspeedix].I_speed,
                **psigkw)
        axs.scatter(P.loc[negspeedix].I_phase, P.loc[negspeedix].I_speed,
                **nsigkw)

        axs.set(xlabel=r'$I_{\rm phase}$ (bits)',
                ylabel='Speed information (bits/spike)')
        axs.set_xlim(left=0)

        # Equalize y-axis and plot parkw and regression lines
        xmax = axs.get_xlim()[1]
        axm.plot([0, xmax], [0, xmax], **parkw)
        axd.plot([0, xmax], [0, xmax], **parkw)
        axs.plot([0, xmax], [0, xmax], **parkw)
        axm.plot([0, xmax], [0, xmax*optslope('rate')], **regkw)
        axd.plot([0, xmax], [0, xmax*optslope('dir')], **regkw)
        axs.plot([0, xmax], [0, xmax*optslope('speed')], **regkw)
        axm.set_ylim(bottom=0.0, top=ymax)
        axd.set_ylim(bottom=0.0, top=ymax)
        axs.set_ylim(bottom=0.0, top=ymax)

        sns.despine()

    @step
    def speed_direction_modulation(self, poscolor=POSCOLOR, negcolor=NEGCOLOR,
        noncolor=NONCOLOR, bins=24, figsize=(10, 5.5)):
        """Show distributions of speed and direction modulation."""
        P = phasers.properties_dataframe()

        f = self.figure('speed-dir-modulation', clear=True, figsize=figsize,
                title='Speed and Direction Modulation of Phasers')

        axd = f.add_subplot(121)
        axs = f.add_subplot(122)

        xmin, xmax = 0, 1
        xlim = xmin, xmax
        bins = np.linspace(xmin, xmax, bins + 1)

        def modulation_dist_plot(ax, subtype, which, color):
            data = P.loc[P.subtype == subtype, which].values
            sns.distplot(data, ax=ax, kde=False, hist=True, bins=bins,
                hist_kws={'alpha':0.7}, norm_hist=False, color=color,
                label=subtype.title())

        def modulation_kde_plot(ax, ymax, subtype, which, color, **kwds):
            data = P.loc[P.subtype == subtype, which].values
            kx, ky = smooth_pdf(data, sd=0.05)
            valid = np.logical_and(kx>=xmin, kx<=xmax)
            kx, ky = kx[valid], ky[valid]
            ky *= ymax / ky.max()
            ax.plot(kx, ky, c=color, alpha=0.7, label='Nonphaser', **kwds)

        modulation_dist_plot(axd, 'negative', 'dir_mod', negcolor)
        modulation_dist_plot(axd, 'positive', 'dir_mod', poscolor)
        modulation_kde_plot(axd, 6.25, 'none', 'dir_mod', noncolor, zorder=-5)

        axd.set_xlim(*xlim)
        axd.set_ylim(top=8.5)
        axd.set(xlabel='Direction Modulation Index', ylabel='Count')

        modulation_dist_plot(axs, 'negative', 'speed_mod', negcolor)
        modulation_dist_plot(axs, 'positive', 'speed_mod', poscolor)
        modulation_kde_plot(axs, 7.25, 'none', 'speed_mod', noncolor, zorder=-5)

        axs.set_xlim(*xlim)
        axs.set_ylim(top=8.5)
        axs.set(xlabel='Speed Modulation Index', ylabel='Count')
        axs.legend(loc='best', frameon=False)

        sns.despine()

    @step
    def anatomical_distribution(self, poscolor=POSCOLOR, negcolor=NEGCOLOR,
        ymax=35, figsize=(6, 5.5)):
        """Show distribution of unique phaser cell counts by recording area."""
        P = phasers.properties_dataframe()

        # Add unique cell IDs and site/area information from /recordings
        recs = dataframe_from_table('/recordings', index='id')
        P['cell_id'] = recs.loc[P.index, 'cell_id']
        P['area'] = recs.loc[P.index, 'area']

        # Get the recording area and phaser subtype categories
        codes = sorted(anatomy.code_for_area.keys())
        area_names = ['Septum', 'Hipp.', 'Thal.', 'Midbrain', 'Other']
        N_groups = len(codes)
        subtypes = [('negative', negcolor), ('positive', poscolor)]
        N_values = len(subtypes)

        data = np.zeros((N_groups, N_values), 'i')
        for i in range(N_groups):
            code = codes[i]
            for j in range(N_values):
                subtype = subtypes[j][0]
                selection = (P.subtype == subtype) & (P.area == code)
                data[i,j] = len(P.loc[selection].cell_id.unique())

        # Create the figure and the grouped bar plot
        f = self.figure('anatomical-distro', clear=True, figsize=figsize,
                title='Anatomical Distribution')

        ax = f.add_subplot(111)
        h = grouped_bar_plot(data, area_names, subtypes, ax=ax, alpha=0.7)
        ax.legend(h[:2], ['Negative', 'Positive'], loc='best', frameon=False)
        ax.set(xlabel='Recording area', ylabel='Count')
        ax.set_ylim(top=ymax)

        sns.despine()

    @step
    def anatomical_distribution_table(self):
        """Print out a prevalence table including non-phasers."""
        P = phasers.properties_dataframe()

        # Add unique cell IDs and site/area information from /recordings
        recs = data.dataframe_from_table('/recordings', index='id')
        P['cell_id'] = recs.loc[P.index, 'cell_id']
        P['area'] = recs.loc[P.index, 'area']
        P['site'] = recs.loc[P.index, 'site']
        U = P.groupby('cell_id')

        # Get the area codes and names
        # Note: the prepended 1 below accounts for the division of septum
        # into LS and MS, thus needing 2 septum codes in the loop
        codes = [1] + sorted(anatomy.code_for_area.keys())
        area_names = ['LS', 'MS', 'Hippocampus', 'Thalamus', 'Midbrain',
                'Other']

        # Generate the table
        columns = ['negative', 'positive', 'mixed', 'none']
        self.out('Multimarkdown phaser cell anatomical prevalence table:')
        self.printf('[**Table X** Prevalence of phaser cells across'
                    ' recorded brain areas][Table X]\n')
        self.printf('| Area | Total | ' + ' | '.join(
            map(str.title, columns)) + ' |\n')
        self.printf('|:-|' + ':-:|' * len(columns) + '\n')

        # Create indexes for each subtype
        subtype_data = {col: np.zeros(len(U), '?') for col in columns}
        subtype_data['area'] = []
        subtype_data['site'] = []
        for i, (cell, recs) in enumerate(U):
            subtype_data['area'].append(recs.area.iloc[0])  # grab first area
            subtype_data['site'].append(recs.site.iloc[0])  # grab first site
            subtypes = set(recs.subtype)
            if len(subtypes) == 1:
                subtype_data[subtypes.pop()][i] = True
            elif len(subtypes) == 2:
                if 'none' in subtypes:
                    subtypes.remove('none')
                    subtype_data[subtypes.pop()][i] = True
                    continue
                subtype_data['mixed'][i] = True
            elif len(subtypes) == 3:
                subtype_data['mixed'][i] = True
            else:
                self.out('No subtypes found for {}', cell)
        C = pd.DataFrame(data=subtype_data)

        for j, code in enumerate(codes):
            row_name = area_names[j]
            if row_name == 'LS':
                match = (C.area == code) & (C.site != 'MS')
            elif row_name == 'MS':
                match = (C.site == 'MS')
            else:
                match = (C.area == code)
            total = np.sum(match)
            self.printf('| ' + row_name + ' | ' + str(total) + ' |')
            for col in columns:
                count = np.sum(C[col] & match)
                self.printf(' {} ({:.1f}%) |'.format(count, 100*count/total))
            self.printf('\n')

        # Grand total row
        total = len(C)
        self.printf('| Total | ' + str(total) + ' |')
        for col in columns:
            count = np.sum(C[col])
            self.printf(' {} ({:.1f}%) |'.format(count, 100*count/total))
        self.printf('\n')
        self.printf('---\n')

    @step
    def recording_site_distribution(self, area='septum', ignore_zeros=True,
        poscolor=POSCOLOR, negcolor=NEGCOLOR, figsize=(7, 5.5)):
        """Show distribution of phaser cells by recording site."""
        P = phasers.properties_dataframe()

        # Add unique cell IDs and site/area information from /recordings
        recs = dataframe_from_table('/recordings', index='id')
        P['cell_id'] = recs.loc[P.index, 'cell_id']
        P['area'] = recs.loc[P.index, 'area']
        P['site'] = recs.loc[P.index, 'site']

        # Get the categorical grouped counts
        target = anatomy.area_code(area)
        sites = sorted(set(P.site[P.area == target]))
        N_groups = len(sites)
        subtypes = [('negative', negcolor), ('positive', poscolor)]
        N_values = len(subtypes)

        data = np.zeros((N_groups, N_values), 'i')
        for i in range(N_groups):
            site = sites[i]
            for j in range(N_values):
                subtype = subtypes[j][0]
                selection = (P.subtype == subtype) & (P.site == site)
                data[i,j] = len(P.loc[selection].cell_id.unique())

        if ignore_zeros:
            valid = data.sum(axis=1) > 0
            data = data[valid]
            sites = [sites[i] for i in range(N_groups) if valid[i]]

        # Create the figure and the grouped bar plot
        f = self.figure('site-distro-%s' % area, clear=True, figsize=figsize,
                title='%s: Site Distribution' % area.title())

        ax = f.add_subplot(111)
        h = grouped_bar_plot(data, sites, subtypes, ax=ax, alpha=0.7)
        ax.legend(h[:2], ['Negative', 'Positive'], loc='best', frameon=False)
        ax.set(xlabel='Recording site within %s' % area, ylabel='Count')

        sns.despine()
