"""
Small repeats multi-panel figure of various example phaser cells.
"""

import os

import tables as tb
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator, MaxNLocator
import seaborn as sns
sns.set(context='notebook', style='ticks', palette='pastel', color_codes=True,
        rc={'figure.subplot.bottom': 0.15})

from roto.circstats import mean_angle as cmean
from roto.radians import RadianFormatter, cdiff
from roto.plots import condmap
import roto.axis as ra

from .. import store, RES_ROOT
from ..context import SPCAnalysis, step
from ..lib import data, phasers, cell, parsers, session, labels
from ..tools import maps, spikes, binned


class PhaserExamples(SPCAnalysis):

    """
    Multiples figure of example phaser cells.
    """

    def cleanup(self):
        store.close()

    def plot_all_phasers(self, sortby='C_rp_r', levels=18, figsize=(9,6.5)):
        """Iterate through all phaser cells to plot and save figures."""
        P = phasers.filtered_dataframe()
        sortix = np.argsort(P[sortby])
        cells = P.index[sortix]

        for i in range(0,cells.size,4):
            clist = list(cells[i:i+4])
            self.plot_examples(cell_list=clist, levels=levels, figsize=figsize)
            self.savefig()
            self.closefig()

    @step
    def plot_examples(self, cell_list=None, levels=18, mapres=64,
        figsize=(9,6.5)):
        """Plot figure with 4 example phaser cells."""
        props = data.properties_dataframe()
        df = store.get()

        N_cells = len(cell_list)
        assert N_cells < 6, "can't plot that many cells!"

        clist_str = [str(c) for c in cell_list]
        f = self.figure('examples-' + '-'.join(clist_str),
                clear=True, figsize=figsize,
                title='Example Phaser Cells: ' + ', '.join(clist_str))

        gshape = (3,N_cells)

        # Axis and plotting parameters
        mapkw = dict(cmin=0.0, cbar=True, cmap='CMRmap', levels=levels,
                mask_color='0.9')
        cbargs = dict(pad=0.005, shrink=0.97, aspect=15)

        # Radian colorbar parameters
        radfmt = RadianFormatter(inline=False)
        radticks = FixedLocator([-np.pi,0,np.pi])
        radcbargs = cbargs.copy()
        radcbargs.update(ticks=radticks, format=radfmt)
        phasekw = dict(cmin=-np.pi, cmax=np.pi, cbar=True, levels=levels,
                mask_color='0.9', cbargs=radcbargs)

        def plot_ratemap(ax, R):
            Rmax = np.floor(R[np.isfinite(R)].max())
            ticks = FixedLocator([0,Rmax/2,Rmax])
            thiscbargs = cbargs.copy()
            thiscbargs.update(ticks=ticks)
            im, cb = maps.plot(R, ax=ax, cbargs=thiscbargs, **mapkw)
            cb.ax.tick_params(pad=-1, length=0, labelsize='x-small')
            ax.set_axis_off()

        def plot_phasemap(ax, P):
            ticks = FixedLocator([-np.pi,0,np.pi])
            cbargs = dict(ticks=ticks, format=radfmt)
            im, cb = maps.plot(P, ax=ax, satnorm=True, **phasekw)
            cb.ax.tick_params(pad=-1, length=0, labelsize='x-small')
            ax.set_axis_off()

        def rate_phase_regression(ax, rate, phase, phi, slope):
            rlim = np.array([rate.min(), rate.max()])
            rate = np.tile(rate, 2)
            phase = np.r_[phase, phase + 2*np.pi]
            condmap(rate, phase, ax=ax, cmap='gray_r')
            def _plot_regline(m):
                ax.plot(rlim, m * 2 *np.pi + phi + slope * rlim, '-',
                        c='#FF3333', lw=2, alpha=0.6)
            for cycle in range(-3,4):
                _plot_regline(cycle)
            ax.set(xlabel='Firing rate (spikes/s)')
            ax.set_yticks([-np.pi, 0, np.pi, 2*np.pi, 3*np.pi])
            ax.set_ylim(-np.pi, 3*np.pi)
            # ax.tick_params(which='major', pad=4)
            ax.yaxis.set_major_formatter(radfmt)
            maxfivebins = MaxNLocator(prune='both', nbins=5, integer=True)
            ax.xaxis.set_major_locator(maxfivebins)
            ra.despine(ax)

        first = True
        for i, c_id in enumerate(cell_list):
            axr = plt.subplot2grid(gshape, (0,i))
            axp = plt.subplot2grid(gshape, (1,i))
            axc = plt.subplot2grid(gshape, (2,i))

            rec = parsers.parse_recording(c_id, raise_on_fail=True)
            ttc = (rec['tetrode'], rec['cluster'])
            S = cell.session_for(c_id, res=mapres)

            # Retrieve or compute the ratemap
            rpath = os.path.join('/ratemaps_%d/id%04d' % (mapres, c_id))
            if self.has_node(rpath):
                R = self.read_array(rpath)
            else:
                self.out('Computing ratemap ({}) for recording #{}...',
                        mapres, c_id)
                R = S.ratemap(ttc)
                self.save_array(rpath, R, title='{}-pixel Ratemap: Recording'
                        ' #{}'.format(mapres, c_id))

            # Retrieve or compute the phase map
            ppath = os.path.join('/phasemaps_%d/id%04d' % (mapres, c_id))
            if self.has_node(ppath):
                P = self.read_array(ppath)
            else:
                self.out('Computing phase map ({}) for recording #{}...',
                        mapres, c_id)
                P = S.phasemap(ttc)
                self.save_array(ppath, P, title='{}-pixel Phasemap: Recording'
                        ' #{}'.format(mapres, c_id))

            # R = df.get_node('/ratemaps', 'id%04d' % c_id)
            plot_ratemap(axr, R)
            title = labels.rat_cell_id(c_id)
            ra.quicktitle(axr, title, size='x-small', fontweight='light')

            # P = df.get_node('/phasemaps', 'id%04d' % c_id)
            plot_phasemap(axp, P)

            valid = np.logical_and(np.isfinite(R), np.isfinite(P[0]))
            rate = R[valid]
            phase = P[0][valid]
            phi = float(props.loc[c_id, 'C_rp_phi'])
            slope = float(props.loc[c_id, 'C_rp_sl'])
            rate_phase_regression(axc, rate, phase, phi, slope)
            if i == 0:
                ra.ylabel(axr, 'Ratemaps')
                ra.ylabel(axp, 'Phasemaps')
                axc.set_ylabel('Spike phase')
            else:
                axc.set_yticklabels([])
