"""
Model simulation of phaser cells correcting phase integrators.
"""

import os
from itertools import product

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator, FixedLocator

from brian2.units import *
import brian2 as br
import seaborn as sns

import roto.axis as ra
import roto.circstats as rcs
from roto.radians import RadianFormatter
from roto.plots import condmap
from roto.arrays import circblur
from roto.misc import merge_two_dicts

from .spatial_input import SpatialInputModel, Fplot
from .phaser_cells import NEGCOLOR, POSCOLOR
from ..context import SPCSimulation, step
from ..lib import sim, theta, labels, motion
from ..tools import adaptive, maps, binned
from .. import DATA_ROOT, store


DEFAULT_TRAJ_SESSION = 27
DEFAULT_SPATIAL_DIR = '/Users/joe/projects/spc/results/' \
                      '2017-04-20+spatial-input-model'

DEFAULT_GE = 21.0
DEFAULT_GTHETA = 5.0
DEFAULT_TAU = 7.0
DEFAULT_F_THETA = 7.5


class PhaserSimulation(SPCSimulation):

    """
    Phaser cell simulation with spatial phase synchronization.
    """

    def cleanup(self):
        store.close()

    @step
    def setup(self, ge=DEFAULT_GE, gtheta=DEFAULT_GTHETA, tau=DEFAULT_TAU,
        ftheta=DEFAULT_F_THETA, spatial_dir=DEFAULT_SPATIAL_DIR,
        traj_session=DEFAULT_TRAJ_SESSION):
        self['ge'] = ge
        self['gtheta'] = gtheta
        self['tau'] = tau
        self['ftheta'] = ftheta
        self['spatial_model_dir'] = spatial_dir
        self['traj_session_id'] = traj_session

    def search_test_model_parameters(self):
        """Perform a grid search of the test phaser simulation."""
        ge_grid = np.linspace(10, 30, 21)
        gtheta_grid = np.linspace(1, 21, 21)
        tau_grid = np.linspace(5, 15, 11)

        searchdir = self.path('search')
        if not os.path.isdir(searchdir):
            os.makedirs(searchdir)

        for ge, gtheta, tau in product(ge_grid, gtheta_grid, tau_grid):
            self.test_phaser_simulation(ge=ge, gtheta=gtheta, tau=tau)
            self.plot_test_phaser_simulation()
            self.savefig()
            self.closefig()
            newfn = os.path.join(searchdir,
                'ge{:02d}-gtheta{:02d}-tau{:02d}.png'.format(
                    int(ge), int(gtheta), int(tau)))
            os.rename(self.path('phaser-demo-00.png'), newfn)
            self.out('Moved image to: {}', newfn)

    @step
    def test_phaser_simulation(self, ge=None, gtheta=None, tau=None,
        ftheta=None, full_sample=False):
        """Run an artifical test simulation of a phaser pair for testing."""
        br.defaultclock.dt = dt = 1 * ms

        if full_sample:
            reportdt = 30.0 * second
            periods = np.linspace(10, 62, 100)
            duration = periods.sum() * second
            x_traj = None
            for period in periods:
                wave = sim.triangle_wave(
                        np.arange(0, period*second+dt, dt), period)
                if x_traj is None:
                    x_traj = wave
                else:
                    x_traj = np.r_[x_traj, wave]
        else:
            reportdt = 2.0 * second
            duration = 20.0 * second
            x_traj = sim.triangle_wave(np.arange(0, duration+dt, dt), 20)

        self.out('Trajectory duration: {}', duration)
        F_input = br.TimedArray(x_traj, dt=dt)

        a = 0.02
        b = 0.2
        c = -50.0
        d = 4.0
        f_theta = (ftheta or self.c.ftheta)*Hz
        ge = ge or self.c.ge
        gtheta = gtheta or self.c.gtheta
        gtheta_pos = 5.0*self.c.gtheta
        tau = (tau or self.c.tau)*ms
        v0 = -75.0  # initial voltage

        # Inhibitory weight and conductance parameters
        Wneg = 3.0      # IPSP strength
        Ei = -80.0      # reversal potential
        tau_i = 100*ms  # time constant

        negative_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ie + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Ramping triangle-wave input
            Ie = ge*F_input(t) : 1

            # Theta phase and inhibitory (antiphase) input
            Itheta = -0.5*gtheta*(cos(phase) + 1) : 1
            phase = 2*pi*f_theta*t : 1
        """
        Gneg = br.NeuronGroup(1, model=negative_eqns, threshold='v > 30',
                order=1, reset='v = c; u = u + d', name='NegativeTestPhaser',
                method='rk4')
        Gneg.v = v0

        positive_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ineg + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Slow inhibitory synaptic current from negative phasers
            Ineg = -gi*(v - Ei) : 1
            dgi/dt = -gi/tau_i : 1
            Ie : 1 (linked)

            # Theta phase and excitatory (inphase) input
            Itheta = 0.5*gtheta_pos*(cos(phase) + 1) : 1
            phase = 2*pi*f_theta*t : 1
        """
        Gpos = br.NeuronGroup(1, model=positive_eqns, threshold='v > 30',
                order=2, reset='v = c; u = u + d', name='PositiveTestPhaser',
                method='rk4')
        Gpos.v = v0
        Gpos.Ie = br.linked_var(Gneg.Ie)

        C = br.Synapses(Gneg, Gpos, on_pre='gi += Wneg')
        C.connect(i=0, j=0)

        NegSpkMon = br.SpikeMonitor(Gneg, ['Ie', 'phase'], record=[0],
                name='negspikes')
        PosSpkMon = br.SpikeMonitor(Gpos, ['Ie', 'phase'], record=[0],
                name='posspikes')
        NegMon = br.StateMonitor(Gneg, ['Ie', 'Itheta', 'v'], record=[0],
                name='negmon')
        PosMon = br.StateMonitor(Gpos, ['Itheta', 'v'], record=[0],
                name='posmon')

        @br.network_operation(dt=reportdt, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        Net = br.Network(Gneg, Gpos, C, NegSpkMon, PosSpkMon, NegMon, PosMon,
                progress, name='PhaserDemo')
        Net.run(duration)

        # Save the demo or sampling simulation
        simname = 'phaser_sample' if full_sample else 'phaser_demo'
        subtitle = 'Correlation Sample' if full_sample else 'Test Demo'
        simtitle = 'Negative/Positive Phaser Pair: {}'.format(subtitle)
        self.save_simulation(simname, Net, title=simtitle)

    @step
    def plot_test_phaser_simulation(self, interval_dur=0.7, negcolor=NEGCOLOR,
        poscolor=POSCOLOR, Iebins=512, corrbins=16, figsize=(7.2,4.5)):
        """Plot a figure of the phaser demonstration simulation."""
        demo = self.read_simulation('phaser_demo')
        fullsample = self.read_simulation('phaser_sample')

        # Read the continuous state for negative phaser
        t = demo.negmon.t
        Fs = 1 / np.median(np.diff(t))
        Vneg = demo.negmon.v
        Ie = demo.negmon.Ie
        Itheta = demo.negmon.Itheta

        # Read the continuous state for positive phaser
        Vpos = demo.posmon.v
        pos_Itheta = demo.posmon.Itheta

        # Read the spike data for the negative phaser
        neg_spk_t = demo.negspikes.t
        neg_spk_Ie = demo.negspikes.Ie
        sim.add_wrapped_phase_column(demo.negspikes)
        neg_spk_phase = demo.negspikes.phase0

        # Read the spike data for the positive phaser
        pos_spk_t = demo.posspikes.t
        pos_spk_Ie = demo.posspikes.Ie
        sim.add_wrapped_phase_column(demo.posspikes)
        pos_spk_phase = demo.posspikes.phase0

        # Load figure template and create figure
        templ = os.path.join(DATA_ROOT, 'figure-templates', 'phaser-demo')
        _, axmap = self.figure_from_template(templ, figsize=figsize,
                title='Phaser Simulation')

        # Plot formatting
        span_fmt = dict(ymin=0.085, ymax=0.92, zorder=-1, fc='gold', alpha=0.4,
                lw=0)
        corr_fmt = dict(s=12, marker='o', linewidths=0.8, facecolor='none', alpha=0.6, zorder=1)
        regline_fmt = dict(zorder=0)  # lw=2, alpha=0.6, c='#FF3333'
        radfmt = RadianFormatter(inline=False)

        def plot_Ie(ax):
            ax.plot(t, Ie + Itheta, lw=0.6, c='g')
            ra.ypadding(ax)
            ax.set_axis_off()
        plot_Ie(axmap['Ie'])

        # Find intervals for lowest/highest input spikes
        if len(neg_spk_t):
            imin = np.argmin(neg_spk_Ie)
            imax = np.argmax(neg_spk_Ie)
            tmin = neg_spk_t[imin]
            tmax = neg_spk_t[imax]

            # Get the trough spike from the second half of simulation
            late = neg_spk_t > tmax
            neg_spk_t_late = neg_spk_t[late]
            neg_spk_Ie_late = neg_spk_Ie[late]
            imin2 = np.argmin(neg_spk_Ie_late)
            tmin2 = neg_spk_t_late[imin2]

            dt = interval_dur / 2
            trough = (tmin - dt, tmin + dt)
            peak = (tmax - dt, tmax + dt)
            trough2 = (tmin2 - dt, tmin2 + dt)

        def plot_V(ax, V, col, label):
            ax.plot(t, V, c=col)
            if len(neg_spk_t):
                ax.axvspan(*trough, **span_fmt)
                ax.axvspan(*peak, **span_fmt)
                ax.axvspan(*trough2, **span_fmt)
            ax.set_xlim(t[0], t[t.size-1])
            ax.set_axis_off()
        plot_V(axmap['Vneg'], Vneg, negcolor, '$V$ (neg)')
        plot_V(axmap['Vpos'], Vpos, poscolor, '$V$ (pos)')

        if not len(neg_spk_t):
            [axmap[k].set_axis_off() for k in ('negspikes', 'posspikes',
                'trough', 'peak', 'trough2', 'inputspikes')]
            return

        # Compute rate/phase circular-linear correlation
        negslope, negphi, negrpcorr, negpval = theta.phaseregress(
                neg_spk_Ie / self.c.ge, neg_spk_phase)
        posslope, posphi, posrpcorr, pospval = theta.phaseregress(
                pos_spk_Ie / self.c.ge, pos_spk_phase)

        def plot_spikes(ax, spk_t, label, c):
            sns.rugplot(spk_t, 1, ax=ax, c=c, lw=0.5)
            ax.set_xlim(t[0], t[t.size-1])
            ax.set_axis_off()
        plot_spikes(axmap['negspikes'], neg_spk_t, 'neg', negcolor)
        plot_spikes(axmap['posspikes'], pos_spk_t, 'pos', poscolor)

        def plot_V_Itheta_interval(ax, interval, label):
            ix = (t >= interval[0]) & (t <= interval[1])
            ax.plot(t[ix], pos_Itheta[ix], zorder=-1, alpha=0.8, c='m')
            ra.ypadding(ax, 0.4)
            axv = ax.twinx()
            axv.plot(t[ix], Vpos[ix], c=poscolor)
            axv.plot(t[ix], Vneg[ix], c=negcolor)
            ax.set_xlim(interval)
            ax.set_axis_off()
            axv.set_axis_off()
        plot_V_Itheta_interval(axmap['trough'], trough, 'early low')
        plot_V_Itheta_interval(axmap['peak'], peak, 'high')
        plot_V_Itheta_interval(axmap['trough2'], trough2, 'late low')

        def plot_input_spikes(ax, spk_Ie, spk_phase, phi, slope, col):
            spk_Ie_F = spk_Ie / self.c.ge
            ilim = np.array([spk_Ie_F.min(), spk_Ie_F.max()])
            ax.scatter(spk_Ie_F, spk_phase, edgecolors=col, **corr_fmt)
            def _plot_regline(m):
                ax.plot(ilim, m * 2 * np.pi + phi + slope * ilim, c=col,
                        **regline_fmt)
            for cycle in range(-2,3):
                _plot_regline(cycle)
        plot_input_spikes(axmap['inputspikes'], pos_spk_Ie, pos_spk_phase,
                posphi, posslope, poscolor)
        plot_input_spikes(axmap['inputspikes'], neg_spk_Ie, neg_spk_phase,
                negphi, negslope, negcolor)

        def finish_input_spikes(ax):
            ax.set(xlabel=r'External input, $I_{\rm ramp}$', xlim=(0,1),
                    xticks=[0,0.5,1], xticklabels=['0',0.5,'1'])
            ax.set(ylabel='Spike theta phase', ylim=(-np.pi, np.pi))
            ax.set_yticks([-np.pi, -np.pi/2, 0, np.pi/2, np.pi])
            ax.yaxis.set_major_formatter(radfmt)
            ra.xpadding(ax, 0.05)
            ra.despine(ax)
        finish_input_spikes(axmap['inputspikes'])

        def compute_binned_rate_phase(spk_Ie, spk_phase, Ie):
            edges = np.linspace(0, Ie.max(), Iebins + 1)
            bins = (edges[:-1] + edges[1:]) / 2

            rate = np.empty(Iebins)
            phase = np.empty_like(rate)
            phase[:] = np.nan
            for i in range(Iebins):
                ix = (spk_Ie >= edges[i]) & (spk_Ie < edges[i+1])

                nspk = np.sum(ix)
                nocc = np.sum((Ie >= edges[i]) & (Ie < edges[i+1]))
                rate[i] = Fs * nspk / nocc

                if ix.any():
                    phi = spk_phase[ix].values
                    phase[i] = rcs.mean_angle(phi)

            return rate, phase

        sim.add_wrapped_phase_column(fullsample.negspikes)
        sim.add_wrapped_phase_column(fullsample.posspikes)

        Rneg, Pneg = compute_binned_rate_phase(
                fullsample.negspikes.Ie, fullsample.negspikes.phase0,
                    fullsample.negmon.Ie)
        Rpos, Ppos = compute_binned_rate_phase(
                fullsample.posspikes.Ie, fullsample.posspikes.phase0,
                    fullsample.negmon.Ie)

        def plot_rate_phase_correlation(ax, rate, phase, label):
            valid = np.isfinite(phase)
            rate = rate[valid]
            phase = phase[valid]
            rlim = np.array([rate.min(), rate.max()])

            sl,phi,r,p = theta.phaseregress(rate, phase)
            self.out('Phase shift ({}) = {:.6f} radians', label, sl*rlim.ptp())
            self.out('N pixels = {}', len(rate))
            self.out('r = {:.4f}, p = {:g}', r, p)

            # rate = np.tile(rate, 2)
            # phase = np.r_[phase, phase + 2*np.pi]
            bins = [np.linspace(rlim[0], rlim[1], corrbins+1),
                    np.linspace(-np.pi, np.pi, corrbins+1)]
            condmap(rate, phase, ax=ax, bins=bins, cmap='gray_r')
            def _plot_regline(m):
                ax.plot(rlim, m * 2 * np.pi + phi + sl * rlim, '-',
                        c='#FF3333', lw=2, alpha=0.6)
            for cycle in range(-2,3):
                _plot_regline(cycle)
            ax.set_yticks([-np.pi, -np.pi/2, 0, np.pi/2, np.pi])
            ax.set_ylim(-np.pi, np.pi)
            ax.yaxis.set_major_formatter(radfmt)
            # ax.xaxis.set_major_locator(maxfivebins)
            ra.despine(ax)
        plot_rate_phase_correlation(axmap['negcorr'], Rneg, Pneg, 'negative')
        plot_rate_phase_correlation(axmap['poscorr'], Rpos, Ppos, 'positive')
        axmap['negcorr'].set(xlabel='Avg. rate (spikes/s)',
               ylabel='Avg. theta phase')
        axmap['poscorr'].set(xlabel='Avg. rate (spikes/s)')

    @step
    def spatial_phaser_simulation(self, N_or_id, example=False, duration=60.0):
        """Run a simulation of the phaser layer (or single example).

        Arguments:
        N_or_id -- number of phasers in layer, OR cell_id if example is True
        example -- whether to simulate an example cell or randomized layer
        duration -- length of simulation in seconds, OR 'session' to simulate
            the entire duration of the trajectory recording session
        """
        self.get_datafile(False)
        br.defaultclock.dt = dt = 1*ms

        trajmodel = sim.TrajectoryModel(self.c.traj_session_id)
        Trajectory = trajmodel.new_group()

        datamodel = SpatialInputModel.load(self.c.spatial_model_dir)

        if example:
            SpatialLayer = datamodel.generate_spatial_layer(1,
                    cell_ids=[N_or_id], trajgrp=Trajectory, fuzz=0.0,
                    rotate=False, plot=False)
        else:
            SpatialLayer = datamodel.generate_spatial_layer(N_or_id,
                    trajgrp=Trajectory, plot=False)

        # Get full session duration if specified
        if duration == 'session':
            mdata = motion.MotionData(self.c.traj_session_id)
            tstart, tend = mdata.t[0], mdata.t[-1]
            duration = tend - tstart
            self.out('Simulating full session ({:.1f}s)'.format(duration))

        a = 0.02
        b = 0.2
        c = -50.0
        d = 4.0
        f_theta = self.c.ftheta*Hz
        gx = self.c.ge
        gtheta = self.c.gtheta
        gtheta_pos = 5.0*self.c.gtheta
        Wneg = 3.0
        tau = self.c.tau*ms
        X = Trajectory.namespace['X']  # needed in namespace to record spike
        Y = Trajectory.namespace['Y']  # positions from trajectory

        negative_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ispatial + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Spatial layer input
            Ispatial = gx*Fx : 1
            Fx : 1 (linked)
            x = X(t) : 1 (shared)
            y = Y(t) : 1 (shared)

            # Inhibitory (antiphase) theta input
            Itheta = -0.5*gtheta*(cos(phase) + 1) : 1 (shared)
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        Gneg = br.NeuronGroup(len(SpatialLayer), model=negative_eqns,
                threshold='v > 30', reset='v = c; u = u + d',
                name='NegativePhasers', order=1, method='rk4')

        Gneg.v[:] = -80.0
        Gneg.Fx = br.linked_var(SpatialLayer.f)

        positive_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ineg + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Slow inhibitory synaptic current from negative phasers
            Ineg = -gi*(v + 80) : 1
            dgi/dt = -gi/(100*ms) : 1

            # Position input for spike recording
            x = X(t) : 1 (shared)
            y = Y(t) : 1 (shared)

            # Excitatory (inphase) theta input
            Itheta = 0.5*gtheta_pos*(cos(phase) + 1) : 1 (shared)
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        Gpos = br.NeuronGroup(len(SpatialLayer), model=positive_eqns,
                threshold='v > 30', reset='v = c; u = u + d',
                name='PositivePhasers', order=2, method='rk4')

        Gpos.v[:] = -80.0

        C = br.Synapses(Gneg, Gpos, on_pre='gi += Wneg')
        C.connect(j='i')

        NegSpkMon = br.SpikeMonitor(Gneg, ['Ispatial', 'phase', 'x', 'y'],
                record=True, name='neg_spikes')
        NegExMon = br.StateMonitor(Gneg, ['Ispatial', 'Itheta', 'v'],
                record=[0], name='neg_example')
        PosSpkMon = br.SpikeMonitor(Gpos, ['phase', 'x', 'y'],
                record=True, name='pos_spikes')
        PosExMon = br.StateMonitor(Gpos, ['Itheta', 'v'],
                record=[0], name='pos_example')
        TrajMon = br.StateMonitor(SpatialLayer, ['x','y'], record=[0],
                name='traj')

        @br.network_operation(dt=10*second, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        Net = br.Network(Gneg, Gpos, C, SpatialLayer, NegSpkMon, NegExMon,
                PosSpkMon, PosExMon, TrajMon, progress,
                name='SpatialPhaserNetwork')
        Net.run(duration*second)

        # Save data traces from the simulation
        if example:
            grp = 'spatial_example_%04d' % N_or_id
            title = 'Spatial Phaser Simulation: Example Cell #%d' % N_or_id
        else:
            grp = 'spatial_sim_N%d' % N_or_id
            title = 'Spatial Phaser Simulation: N=%d' % N_or_id

        self.save_simulation(grp, Net, title=title)

    @step
    def plot_example_cell_simulation(self, cell_id, subtype='negative',
        levels=18, Fbins=256, mapres=128, figsize=(9.5,12)):
        """Plot figure showing results from example phaser simulation."""
        S = self.read_simulation('spatial_example_%04d' % cell_id)
        spikes = {'negative':S.neg_spikes, 'positive':S.pos_spikes}[subtype]
        sim.add_wrapped_phase_column(spikes)

        tag = labels.session_cell_id(cell_id)
        f = self.figure('example-%s-phaser-%04d' % (subtype, cell_id),
                clear=True, figsize=figsize,
                title='Example {} Phaser Simulation: {}'.format(
                    subtype.title(), tag))

        # Compute adpative rate/phase maps
        mdata = motion.MotionData(self.c.traj_session_id)
        amapkw = dict(res=mapres, alim=(8.0, 24.0), scale='cm')
        ratemap = adaptive.AdaptiveRatemap(mdata, **amapkw)
        phasemap = adaptive.AdaptivePhasemap(mdata, **amapkw)

        # Axis and plot formatting
        gshape = (3,2)
        trajkw = dict(ls='-', c='k', alpha=0.4, lw=1, zorder=-1)
        spkkw = dict(marker='o', s=16, c='#FF5555', alpha=0.6, linewidths=0)
        cbtickkw = dict(axis='y', labelsize='small', pad=-3, length=3)
        cbargs = dict(pad=0.01, shrink=0.7, aspect=12)
        mapkw = dict(cmin=0.0, cbar=True, mask_color='0.8', levels=levels)
        Fmapkw = merge_two_dicts(mapkw, dict(cbticks=cbtickkw, cbargs=cbargs,
            levels=None))

        # Radian colorbar parameters
        radfmt = RadianFormatter(inline=False)
        radticks = FixedLocator([-np.pi,0,np.pi])
        radcbargs = merge_two_dicts(cbargs, dict(ticks=radticks, format=radfmt))
        phasekw = dict(cmin=-np.pi, cmax=np.pi, cbar=True, levels=levels,
                mask_color='0.9', cbargs=radcbargs)
        maxfivebins = MaxNLocator(prune='both', nbins=5, integer=True)

        # Generate spatial model function and pixel evaluation grid
        datamodel = SpatialInputModel.load(self.c.spatial_model_dir)
        B = datamodel.read_dataframe('spatial_data/all').loc[cell_id:cell_id+1]
        S_ = datamodel.sample_spatial_model(B, fuzz=0, rotate=False)
        M = datamodel.normalize_spatial_model(S_)
        Fargs = tuple(M.iloc[0]) + (datamodel.c.wall_transition,
                datamodel.c.wall_k)
        pts = np.linspace(0, motion.CIRCLE_DIAMETER, Fbins)
        xgrid, ygrid = np.meshgrid(pts, pts, indexing='ij')

        def plot_spatial_function(ax):
            Fplot(ax, xgrid, ygrid, Fargs, clip=True, **Fmapkw)
        plot_spatial_function(plt.subplot2grid(gshape, (0,0)))

        def plot_spike_trajectory(ax):
            plt.plot(S.traj.x, S.traj.y, **trajkw)
            plt.scatter(spikes.x, spikes.y, **spkkw)
            ax.axis('scaled')
            ax.set_xlim(0, 80)
            ax.set_ylim(0, 80)
            ax.set_axis_off()
        plot_spike_trajectory(plt.subplot2grid(gshape, (0,1)))

        R = ratemap(spikes.x, spikes.y, S.traj.x, S.traj.y)
        def plot_ratemap(ax):
            Rmax = np.floor(R[np.isfinite(R)].max())
            ticks = FixedLocator([0,Rmax/2,Rmax])
            thiscbargs = cbargs.copy()
            thiscbargs.update(ticks=ticks)
            _, cb = maps.plot(R, cbargs=thiscbargs, **mapkw)
            cb.ax.tick_params(**cbtickkw)

            ax.set_axis_off()
        plot_ratemap(plt.subplot2grid(gshape, (1,0)))

        P = phasemap(spikes.x, spikes.y, spikes.phase)
        def plot_phasemap(ax):
            _, cb = maps.plot(P, satnorm=True, **phasekw)
            cb.ax.tick_params(**cbtickkw)
            ax.set_axis_off()
        plot_phasemap(plt.subplot2grid(gshape, (1,1)))

        def plot_phase_dist(ax, bins=binned.DEFAULT_PHASE_BINS):
            phase = spikes.phase0
            edges = np.linspace(-np.pi, np.pi, bins + 1)
            centers = (edges[:-1] + edges[1:]) / 2
            H, _, _ = ax.hist(phase, bins=edges, color='r', lw=0, alpha=0.9,
                    zorder=0)
            pdistsm = circblur(H, radians=2*np.pi/36)
            ax.plot(centers, pdistsm, 'k-', lw=1.5, alpha=0.7, zorder=1)
            ax.axvline(x=0, ls=':', c='k', alpha=0.8, zorder=2)
            ax.xaxis.set_major_formatter(radfmt)
            ax.set_xticks([-np.pi, -np.pi/2, 0, np.pi/2, np.pi])
            ax.set_yticks([])
            ax.set(xlim=(-np.pi, np.pi), xlabel='Theta Phase (rad)',
                ylim=(0, 1.2 * np.max(pdistsm)))
            ra.despine(ax, left=False)
        plot_phase_dist(plt.subplot2grid(gshape, (2,0)))

        def plot_rate_phase_regression(ax):
            valid = np.isfinite(R) & np.isfinite(P[0])
            rate = R[valid]
            phase = P[0][valid]
            sl,phi,r,p = theta.phaseregress(rate, phase)
            rlim = np.array([rate.min(), rate.max()])
            rate = np.tile(rate, 2)
            phase = np.r_[phase, phase + 2*np.pi]
            condmap(rate, phase, ax=ax, cmap='gray_r')
            def _plot_regline(m):
                plt.plot(rlim, m * 2 *np.pi + phi + sl * rlim, '-',
                        c='#FF3333', lw=2, alpha=0.6)
            for cycle in range(-3,4):
                _plot_regline(cycle)
            ax.set(xlabel='Firing Rate (spk/s)', ylabel='Theta Phase')
            ax.tick_params(pad=4)
            ax.set_yticks([0, np.pi, 2*np.pi])
            ax.set_ylim(-np.pi, 3*np.pi)
            ax.yaxis.set_major_formatter(radfmt)
            ax.xaxis.set_major_locator(maxfivebins)
            ra.despine(ax)
        plot_rate_phase_regression(plt.subplot2grid(gshape, (2,1)))

    @step
    def plot_phaser_layer_simulation(self, N, subtype='negative', **figkw):
        """Create small multiples figure of simulation rate/phasemaps.i

        Arguments:
        N -- integer size of simulation layer to be plotted
        subtype -- phaser subtype to plot

        Remaining keywords are passed to `spc.lib.sim.plot_phaser_maps_figure`:
        figname, figtitle, maxN, levels, mapres, figsize.
        """
        S = self.read_simulation('spatial_sim_N%d' % N)
        spikes = {'negative':S.neg_spikes, 'positive':S.pos_spikes}[subtype]
        figname = '%s-layer-N%d' % (subtype, N)
        figtitle = '{} Phaser Simulation ({})'.format(N, subtype)

        sim.plot_phaser_maps_figure(self, S.traj, spikes, **figkw)
