"""
Vizualize pair-wise correlations between cell properties.
"""

from functools import reduce
import matplotlib.pyplot as plt
import numpy as np
import operator as op
import pandas as pd
import seaborn as sns
sns.set_style('darkgrid')

from .. import store
from ..context import SPCAnalysis, step
from ..lib.data import properties_dataframe


allprops = [ 'id', 'session_id', 'rat', 'avg_rate', 'phase_avg',
        'phase_coh', 'ratemap_avg', 'ratemap_max', 'Ct_rp_sl', 'Ct_rp_phi',
        'Ct_rp_r', 'Ct_rp_p', 'C_rp_sl', 'C_rp_phi', 'C_rp_r', 'C_rp_p',
        'C_rc_sl', 'C_rc_int', 'C_rc_r', 'C_rc_p', 'I_rate', 'I_rate_p',
        'I_phase', 'I_phase_p', 'curl_angle', 'curlfree', 'dir_avg', 'dir_coh',
        'I_dir', 'I_dir_p', 'dir_mod', 'rate_alloc', 'phase_alloc', 'speed_mod',
        'C_sr_sl', 'C_sr_int', 'C_sr_r', 'C_sr_p', 'f_hat', 'theta_mod',
        'theta_rhythm' ]

plim = {
    'Ct_rp_sl': (-1, 1),
    'C_rp_sl': (-15, 15),
    'C_rc_sl': (-5, 2),
    'f_hat': (4, 14),
    'theta_rhythm': (0, 1),
    'curlfree': (0, 1)
}

categories = {
    'regslopes': [ 'Ct_rp_sl', 'C_rp_sl', 'C_rc_sl', 'C_sr_sl' ],
    'regcorrs':  [ 'Ct_rp_r', 'C_rp_r', 'C_rc_r', 'C_sr_r' ],
    'regpvals':  [ 'Ct_rp_p', 'C_rp_p', 'C_rc_p', 'C_sr_p' ],
    'ratevars':  [ 'avg_rate', 'ratemap_avg', 'ratemap_max', 'I_rate',
                   'rate_alloc' ],
    'phasevars': [ 'phase_avg', 'phase_coh', 'I_phase', 'phase_alloc' ],
    'dirvars':   [ 'dir_avg', 'dir_coh', 'I_dir', 'dir_mod', 'curlfree' ],
    'allocvars': [ 'curl_angle', 'curlfree', 'rate_alloc', 'phase_alloc' ],
    'thetavars': [ 'f_hat', 'theta_mod', 'theta_rhythm', 'speed_mod' ],
    'inforates': [ 'I_rate', 'I_phase', 'I_dir' ],
    #'infopvals': [ 'I_rate_p', 'I_phase_p', 'I_dir_p' ]
    'infologp': [ 'I_rate_logp', 'I_phase_logp', 'I_dir_logp' ]
}

sorted_keys = sorted(categories.keys(), key=lambda a: len(categories[a]),
                reverse=True)

crosses = [(x,y) for i, x in enumerate(sorted_keys) for y in sorted_keys[i+1:]]


def clean_index(data, fields):
    """Index into data array where given fields are within finite limits."""
    AND = op.and_
    ix = reduce(AND, (np.isfinite(data[f]) for f in fields))
    ix = reduce(AND, (data[f] >= plim[f][0] for f in fields if f in plim), ix)
    ix = reduce(AND, (data[f] <= plim[f][1] for f in fields if f in plim), ix)
    return ix.values

def significance_selection(data, field_level):
    """Add a selection field for highlighting significant data points.

    Arguments:
    data -- cell properties dataframe
    field_level -- None or (field, critical_pval) tuple

    Returns:
    None or string, column name for the selection
    """
    if field_level is None:
        return None
    field, level = field_level
    field_p = field + '_p'
    if field_p not in data:
        raise KeyError('missing p-values for {}'.format(field))
    selection = '%s_p%.2f' % field_level
    data[selection] = data[field_p] <= level
    return selection

def conditional_selection(data, conditions):
    """Add a field named 'selection' for a conditional selection of points.

    Arguments:
    data -- cell properties dataframe
    conditions -- list of (field, opstr, value) tuples

    Returns:
    int, number of selected points
    """
    AND = op.and_
    F = {'==':op.eq, '!=':op.ne, '<':op.lt, '<=':op.le, '>':op.gt, '>=':op.ge}
    data['selection'] = True
    data['selection'] = reduce(AND,
        (F[opstr](data[field], value) for field, opstr, value in
            conditions), data['selection'])
    return data['selection'].sum()

def conditional_string(conditions):
    """String description for a set of conditions."""
    if not conditions:
        return 'No selection'
    return ', '.join(('{} {} {:g}'.format(field, opstr, value) for
        field, opstr, value in conditions))


class VisualizeCellProperties(SPCAnalysis):

    """
    Create plots and visualizations of cell properties.
    """

    @step
    def script(self, corr_pval=0.05, info_pval=0.01):
        """Generate many paired-data scatter plots and densities."""
        self.plot_within_category_pairs()
        self.plot_within_category_pairs(select_sig=('C_rp', corr_pval))
        self.plot_within_category_pairs(select_sig=('I_rate', info_pval))
        self.plot_within_category_pairs(select_sig=('I_phase', info_pval))
        self.plot_within_category_pairs(select_sig=('I_dir', info_pval))
        self.plot_cross_category_pairs()
        self.plot_cross_category_pairs(select_sig=('C_rp', corr_pval))
        self.plot_cross_category_pairs(select_sig=('I_rate', info_pval))
        self.plot_cross_category_pairs(select_sig=('I_phase', info_pval))
        self.plot_cross_category_pairs(select_sig=('I_dir', info_pval))

    @step
    def plot_within_category_pairs(self, select_sig=None):
        """Plot paired-data figures for within-category variables.

        Keyword arguments:
        select_sig -- (colname, siglevel) tuple to highlight significant points
        """
        df = properties_dataframe(add_logp_cols=True)
        selection = significance_selection(df, select_sig)

        ion = plt.isinteractive()

        ppkwds = dict(diag_kind='kde')
        if selection is None:
            cname = 'dirty orange'
            col = sns.xkcd_rgb[cname]
            sns.set_palette(sns.xkcd_palette([cname]))
            ppkwds.update(plot_kws={'facecolor':col}, diag_kws={'lw':2})
        else:
            ppkwds.update(hue=selection, hue_order=[False, True],
                    palette=sns.diverging_palette(240, 10, n=2)[::-1])

        for cat, plotvars in categories.items():
            if selection is None:
                label = 'within-figures/%s' % cat
            else:
                label = 'within-figures-%s/%s-%s' % (selection, cat, selection)
            self.out('Plotting {}', label)
            ix = clean_index(df, plotvars)
            if not ix.any():
                continue
            sns.pairplot(df.iloc[ix], vars=plotvars, **ppkwds)
            self.figure(label, handle=plt.gcf())
            self.savefig()
            if not ion:
                self.closefig()

        if ion:
            plt.ion()
            plt.show()

    @step
    def plot_cross_category_pairs(self, select_sig=None):
        """Plot paired-data figures for cross-category variables.

        Keyword arguments:
        select_sig -- (colname, siglevel) tuple to highlight significant points
        """
        df = properties_dataframe(add_logp_cols=True)
        selection = significance_selection(df, select_sig)

        ion = plt.isinteractive()

        ppkwds = dict(diag_kind='kde')
        if selection is None:
            col = sns.xkcd_rgb['dirty orange']
            ppkwds.update(plot_kws={'facecolor':col})
        else:
            ppkwds.update(hue=selection, hue_order=[False, True],
                    palette=sns.diverging_palette(240, 10, n=2)[::-1])

        for xcat, ycat in crosses:
            if selection is None:
                label = 'cross-figures/%s-%s' % (xcat, ycat)
            else:
                label = 'cross-figures-%s/%s-%s-%s' % (
                        selection, xcat, ycat, selection)
            self.out('Plotting {}', label)
            xvars = categories[xcat]
            yvars = categories[ycat]
            ix = clean_index(df, xvars + yvars)
            sns.pairplot(df.iloc[ix], x_vars=xvars, y_vars=yvars, **ppkwds)
            self.figure(label, handle=plt.gcf())
            self.savefig()
            if not ion:
                self.closefig()

        if ion:
            plt.ion()
            plt.show()

    @step
    def plot_custom_panel(self, which, *conds):
        """Plot custom panels of paired-data figures.

        Arguments:
        which -- string or (xcat, ycat) tuple, variable category names
        *conds -- tuples representing selection conditionals, e.g.
            ('I_rate', '>', 0.4), ('avg_rate', '>=', 5.0), etc.
        """
        if type(which) is str:
            xcat = ycat = which
        else:
            xcat, ycat = which
        df = properties_dataframe(add_logp_cols=True)
        npts = conditional_selection(df, conds)
        selstr = conditional_string(conds)
        self.out('Selected {} points: {}', npts, selstr)

        ion = plt.isinteractive()

        ppkwds = dict(diag_kind='kde')
        if not conds:
            col = sns.xkcd_rgb['dirty orange']
            ppkwds.update(plot_kws={'facecolor':col})
        else:
            ppkwds.update(hue='selection', hue_order=[False, True],
                    palette=sns.diverging_palette(240, 10, n=2)[::-1])

        label = 'panels/%s-%s' % (xcat, ycat)
        self.out('Plotting {}', label)
        xvars = categories[xcat]
        yvars = categories[ycat]
        ix = clean_index(df, xvars + yvars)
        sns.pairplot(df.iloc[ix], x_vars=xvars, y_vars=yvars, **ppkwds)
        fig = plt.gcf()
        fig.suptitle(selstr, y=.995)
        self.figure(label, handle=plt.gcf())
        self.savefig()
        if not ion:
            self.closefig()

        if ion:
            plt.ion()
            plt.show()
