"""
Spatial input layer based on GLM decoding model for driving phaser simulations.
"""

import os

import pandas as pd
import numpy as np
import scipy.optimize as opt
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator, MaxNLocator

from brian2.units import *
from brian2 import implementation
import brian2 as br

from roto.axis import quicktitle
from roto.images import tiling_dims
from roto.radians import rotoffset
from roto.reports import AxesReport
from roto.misc import unique
from roto.paths import uniquify

from .trajectory_independence import (TrajectoryIndependence,
        add_variables_to_context, CIRCLE_RADIUS, DEFAULT_WALL_TRANSITION,
        DEFAULT_WALL_K, wall_sigmoid)

from .. import store, RES_ROOT
from ..context import SPCSimulation, step
from ..lib import phasers, data, sim, labels
from ..tools import maps


DEFAULT_FUZZ = 0.1
DEFAULT_TRAJ_SESSION = 27

CENTER_X = CENTER_Y = CIRCLE_RADIUS
TRAJECTORY_ANALYSIS = os.path.join(RES_ROOT,
        '2017-04-15+trajectory-independence+pure-spatial-model')


def F(x, y, Lx, Ly, Lxx, Lyy, Lxy, Lw, b0, phi, w50, k):
    x, y = rotoffset(x, y, CENTER_X, CENTER_Y, -phi)
    r = np.hypot(x - CENTER_X, y - CENTER_Y)
    return (Lx*x + Ly*y + Lxx*x**2 + Lyy*y**2 + Lxy*x*y +
            Lw*wall_sigmoid(r, w50=w50, k=k) + b0)

def clip_sigmoid(x):
    """Sigmoid to smoothly clip spatial inputs to {0,1}."""
    return 1 / (1 + np.exp(-10 * (x - 0.5)))

def Fclip(x, y, Lx, Ly, Lxx, Lyy, Lxy, Lw, b0, phi, w50, k):
    x, y = rotoffset(x, y, CENTER_X, CENTER_Y, -phi)
    r = np.hypot(x - CENTER_X, y - CENTER_Y)
    return clip_sigmoid(
            Lx*x + Ly*y + Lxx*x**2 + Lyy*y**2 + Lxy*x*y +
            Lw*wall_sigmoid(r, w50=w50, k=k) + b0)

@implementation('numpy', discard_units=True)
@check_units(x0=1, y0=1, r=1, Lx=1, Ly=1, Lxx=1, Lyy=1, Lxy=1,
             Lw=1, b0=1, phi=1, w50=1, k=1, result=1)
def Fsim(x0, y0, r, Lx, Ly, Lxx, Lyy, Lxy, Lw, b0, phi, w50, k):
    cs = np.cos(-phi)   # offset-rotation of position by angle phi
    sn = np.sin(-phi)   # ...
    dx = x0 - 40.0      # ...
    dy = y0 - 40.0      # ...
    xr = cs*dx - sn*dy  # ...
    yr = sn*dx + cs*dy  # ...
    x = xr + 40.0       # ...
    y = yr + 40.0       # x/y position is now rotated
    w = 1 / (1 + np.exp(-k * (r - w50)))
    h = Lx*x + Ly*y + Lxx*x**2 + Lyy*y**2 + Lxy*x*y + Lw*w + b0
    f = 1 / (1 + np.exp(-10 * (h - 0.5)))
    return f

def Fopt(X, Lx, Ly, Lxx, Lyy, Lxy, Lw, b0, phi, w50, k):
    x, y = rotoffset(X[0], X[1], CENTER_X, CENTER_Y, -phi)
    r = np.hypot(x - CENTER_X, y - CENTER_Y)
    return (Lx*x + Ly*y + Lxx*x**2 + Lyy*y**2 + Lxy*x*y +
            Lw*wall_sigmoid(r, w50=w50, k=k))

def Finv(X, Lx, Ly, Lxx, Lyy, Lxy, Lw, b0, phi, w50, k):
    x, y = rotoffset(X[0], X[1], CENTER_X, CENTER_Y, -phi)
    r = np.hypot(x - CENTER_X, y - CENTER_Y)
    return -(Lx*x + Ly*y + Lxx*x**2 + Lyy*y**2 + Lxy*x*y +
             Lw*wall_sigmoid(r, w50=w50, k=k))

def Fplot(ax, xgrid, ygrid, args, clip=True, wall=CIRCLE_RADIUS, cbticks=None,
    **mapkw):
    """Plot the spatial response function to the given axis."""
    if clip:
        resp = Fclip(xgrid, ygrid, *args)
    else:
        resp = F(xgrid, ygrid, *args)
    rad2 = np.power(xgrid - CENTER_X, 2) + np.power(ygrid - CENTER_Y, 2)
    resp[rad2 > wall**2] = np.nan

    rmin = round(resp[np.isfinite(resp)].min(), 2)
    rmax = round(resp[np.isfinite(resp)].max(), 2)
    ticks = FixedLocator([rmin, (rmin + rmax) / 2, rmax])
    cbargs = mapkw.pop('cbargs', {})
    thiscbargs = cbargs.copy()
    thiscbargs.update(ticks=ticks)
    thismapkw = dict(cmin=0.0, cmax=1.0, cbar=True, cbargs=thiscbargs)
    thismapkw.update(mapkw)

    _, cb = maps.plot(resp, ax=ax, **thismapkw)

    if cbticks is None:
        cbticks = dict(pad=0, length=0)
    cb.ax.tick_params(**cbticks)
    # cb.ax.yaxis.tick_right()
    ax.set_axis_off()
    return ax


class SpatialInputModel(SPCSimulation):

    """
    Simulate and study a path-integating network with phaser input.
    """

    def cleanup(self):
        store.close()

    @step
    def setup(self, wall_transition=DEFAULT_WALL_TRANSITION,
        wall_k=DEFAULT_WALL_K):
        self['wall_transition'] = wall_transition
        self['wall_k'] = wall_k
        add_variables_to_context(self.c, only_spatial=True)

    @step
    def save_spatial_data(self, trajdir=TRAJECTORY_ANALYSIS):
        """Load the GLM spatial weights and rescale for the spatial model."""
        rdf = data.dataframe_from_table('/recordings')
        dataf = self.get_datafile(False)

        if not self.has_node('/GLM/beta_1'):
            traj = TrajectoryIndependence.load(trajdir)
            data.clone_node(traj.get_node('beta_1'), dataf, parent='/GLM')
            data.clone_node(traj.get_node('intercept_1'), dataf, parent='/GLM')
            traj.close_datafile()

        beta = np.squeeze(self.read_array('/GLM/beta_1'))
        intercept = np.squeeze(self.read_array('/GLM/intercept_1'))

        # Create a combined dataframe for the spatial model
        P = phasers.properties_dataframe()
        B = pd.DataFrame(data=beta, columns=self.c.predictors, index=P.index)
        B['B0'] = intercept

        self.save_dataframe('spatial_data/all', B,
                title='Spatial Model Matrix: All Cells')
        for subtype in ('negative', 'positive', 'none'):
            self.save_dataframe('spatial_data/%s' % subtype,
                B.loc[P.subtype == subtype],
                title='Spatial Model Matrix: %s' % subtype.title())

    def sample_spatial_model(self, B, n=None, fuzz=0.0, rotate=False):
        """Randomly sample phaser data for randomized spatial inputs."""
        if n is None:
            M = B.copy()
        else:
            M = B.sample(n=n, replace=True)

        M *= 1 + fuzz * np.random.randn(*M.shape)

        # Add new column for rotation angle parameter
        M['phi'] = 0.0
        if rotate:
            M['phi'] = 2 * np.pi * np.random.rand(len(M))

        return M

    def normalize_spatial_model(self, S):
        """Find minimum/maximum of spatial model for normalization."""
        def wallcon(X):
            return CIRCLE_RADIUS - np.hypot(X[0] - CENTER_X, X[1] - CENTER_Y)

        M = S.copy()
        M['B0'] = np.nan
        center_xy = CENTER_X, CENTER_Y
        for i in range(len(M)):
            args = tuple(S.iloc[i]) + (self.c.wall_transition, self.c.wall_k)

            # COBYLA search for min
            xmin = opt.fmin_cobyla(Fopt, center_xy, wallcon, args=args,
                    consargs=(), disp=0)
            fmin = Fopt(xmin, *args)

            # COBYLA search for max
            xmax = opt.fmin_cobyla(Finv, center_xy, wallcon, args=args,
                    consargs=(), disp=0)
            fmax = Fopt(xmax, *args)

            # Divisive normalization of model coefficients with intercept.
            # 1) Only normalize the weight columns before rotation angle
            # 2) Last two columns are intercept (B0) and angle (phi)
            M.iloc[i,:-2] /= fmax - fmin
            M.iloc[i,-2] = -fmin / (fmax - fmin)

            self.box(color='green')
        self.newline()

        return M

    @step
    def spatial_model_validation_report(self, subtype='negative', n_samples=6,
        bins=64, fuzz=DEFAULT_FUZZ, rotate=False, trajdir=TRAJECTORY_ANALYSIS):
        """Create report of observed ratemaps juxtaposed with model samples."""
        traj = TrajectoryIndependence.load(trajdir)
        pdfpath = uniquify(self.path('validation-report-%s' % subtype),
            fmt='%s-%02d', ext='pdf')
        report = AxesReport(pdfpath)

        # Load datasets for the report
        dataf = store.get()
        rdf = data.dataframe_from_table('/recordings')
        P = phasers.filtered_dataframe()
        P = P.loc[P.subtype == subtype]
        P['cell_id'] = rdf.loc[P.index, 'cell_id']

        B = self.read_dataframe('spatial_data', subtype)
        S = self.sample_spatial_model(B)
        M = self.normalize_spatial_model(S)

        wallargs = (self.c.wall_transition, self.c.wall_k)
        pts = np.linspace(0, 2*CIRCLE_RADIUS, bins)
        xgrid, ygrid = np.meshgrid(pts, pts, indexing='ij')

        cbtickkw = dict(axis='y', labelsize='x-small', pad=-3, length=3)
        cbargs = dict(pad=0.01, shrink=0.7, aspect=10)
        mapkw = dict(cbar=True, cmin=0, mask_color='0.8')
        bmapkw = mapkw.copy()
        bmapkw.update(cbticks=cbtickkw, cmin=None, cmax=None, cbargs=cbargs,
                mask_color='paleturquoise')
        nmapkw = mapkw.copy()
        nmapkw.update(cbticks=cbtickkw, cbargs=cbargs, mask_color='oldlace')
        smapkw = mapkw.copy()
        smapkw.update(cbticks=cbtickkw, cbargs=cbargs, mask_color='salmon')

        def plot_ratemap(ax, R, title):
            ticks = MaxNLocator(nbins=3)
            thiscbargs = cbargs.copy()
            thiscbargs.update(ticks=ticks)
            _, cb = maps.plot(R, ax=ax, cbargs=thiscbargs, **mapkw)
            cb.ax.tick_params(**cbtickkw)
            quicktitle(ax, title, size='small')
            ax.set_axis_off()

        for cell_id in unique(P.cell_id):
            for rec_id in P.loc[P.cell_id == cell_id].index:
                rmap = dataf.get_node('/ratemaps', 'id%04d' % rec_id).read()
                plot_ratemap(report.next_axes(), rmap, '{} [#{}]'.format(
                    cell_id, rec_id))

                rpred = traj.read_array('/ratemaps_1', 'id%04d' % rec_id)
                plot_ratemap(report.next_axes(), rpred, 'recon.')

                args = tuple(S.loc[rec_id]) + wallargs
                ax = Fplot(report.next_axes(), xgrid, ygrid, args, clip=False,
                        **bmapkw)
                quicktitle(ax, 'predictor', size='small')

                args = tuple(M.loc[rec_id]) + wallargs
                ax = Fplot(report.next_axes(), xgrid, ygrid, args, **nmapkw)
                quicktitle(ax, 'normalized', size='small')

                # Create and plot randomized models of the given recording
                S_rand = B.iloc[:n_samples].copy()
                for col in B.columns:
                    S_rand[col] = B.loc[rec_id, col]

                M_rand = self.normalize_spatial_model(
                               self.sample_spatial_model(S_rand, fuzz=fuzz,
                                   rotate=rotate))

                for i in range(n_samples):
                    args = tuple(M_rand.iloc[i]) + wallargs
                    ax = Fplot(report.next_axes(), xgrid, ygrid, args, **smapkw)
                    quicktitle(ax, 'sample %d' % (i+1), size='small')

                report.breakline()
            report.breakline()
        report.close()

    @step
    def plot_spatial_model_validation(self, cell_list=[97,444], bins=64,
        fuzz=DEFAULT_FUZZ, rotate=True, trajdir=TRAJECTORY_ANALYSIS,
        figsize=(7.2, 3.6)):
        """Create figure of example cells showing sample validation."""
        traj = TrajectoryIndependence.load(trajdir)
        n_samples = 1
        subtype = 'negative'

        f = self.figure('example-validation', clear=True, figsize=figsize,
                title='Spatial Input Validation: ' +
                      ', '.join(map(str, cell_list)))
        N_cells = len(cell_list)
        gshape = (N_cells+1, 5)

        # Load datasets for the report
        dataf = store.get()
        P = phasers.filtered_dataframe()
        P = P.loc[P.subtype == subtype]

        B = self.read_dataframe('spatial_data', subtype)
        S = self.sample_spatial_model(B)
        M = self.normalize_spatial_model(S)

        wallargs = (self.c.wall_transition, self.c.wall_k)
        pts = np.linspace(0, 2*CIRCLE_RADIUS, bins)
        xgrid, ygrid = np.meshgrid(pts, pts, indexing='ij')

        cbtickkw = dict(pad=-1, length=0, labelsize='x-small')
        cbargs = dict(pad=0.005, shrink=0.98, aspect=15)
        mapkw = dict(cbar=True, cmin=0, mask_color='0.9')
        bmapkw = mapkw.copy()
        bmapkw.update(cbticks=cbtickkw, cmin=None, cmax=None, cbargs=cbargs,
                mask_color='paleturquoise')
        nmapkw = mapkw.copy()
        nmapkw.update(cbticks=cbtickkw, cbargs=cbargs,
                mask_color='paleturquoise')
        smapkw = mapkw.copy()
        smapkw.update(cbticks=cbtickkw, cbargs=cbargs,
                mask_color='paleturquoise')

        def plot_ratemap(ax, R, title):
            Rmax = np.floor(R[np.isfinite(R)].max())
            ticks = FixedLocator([0,Rmax/2,Rmax])
            # ticks = MaxNLocator(nbins=3)
            thiscbargs = cbargs.copy()
            thiscbargs.update(ticks=ticks)
            _, cb = maps.plot(R, ax=ax, cbargs=thiscbargs, **mapkw)
            cb.ax.tick_params(**cbtickkw)
            quicktitle(ax, title, size='x-small', fontweight='light')
            ax.set_axis_off()

        for i,rec_id in enumerate(cell_list):
            rmap = dataf.get_node('/ratemaps', 'id%04d' % rec_id).read()
            plot_ratemap(plt.subplot2grid(gshape, (i,0)), rmap,
                labels.rat_cell_id(rec_id))

            rpred = traj.read_array('/ratemaps_1', 'id%04d' % rec_id)
            plot_ratemap(plt.subplot2grid(gshape, (i,1)), rpred,
                    r'1$\times$1 LQW')

            args = tuple(S.loc[rec_id]) + wallargs
            ax = Fplot(plt.subplot2grid(gshape, (i,2)), xgrid, ygrid, args,
                    clip=False, **bmapkw)
            quicktitle(ax, 'Predictor', size='x-small', fontweight='light')

            args = tuple(M.loc[rec_id]) + wallargs
            ax = Fplot(plt.subplot2grid(gshape, (i,3)), xgrid, ygrid, args,
                    **nmapkw)
            quicktitle(ax, 'Normalized', size='x-small', fontweight='light')

            # Create and plot randomized models of the given recording
            S_rand = B.iloc[:n_samples].copy()
            for col in B.columns:
                S_rand[col] = B.loc[rec_id, col]

            M_rand = self.normalize_spatial_model(
                           self.sample_spatial_model(S_rand, fuzz=fuzz,
                               rotate=rotate))

            for j in range(n_samples):
                args = tuple(M_rand.iloc[j]) + wallargs
                ax = Fplot(plt.subplot2grid(gshape, (i,4+j)), xgrid, ygrid,
                        args, **smapkw)
                quicktitle(ax, 'Sample', size='x-small', fontweight='light')

    def plot_normalized_model_samples(self, M, maxN=36, wall=CIRCLE_RADIUS,
        bins=128, figsize=(12,12)):
        """Plot figure of normalized spatial profile samples."""
        pts = np.linspace(0, 2*CIRCLE_RADIUS, bins)
        xgrid, ygrid = np.meshgrid(pts, pts, indexing='ij')

        f = self.figure('spatial-samples', clear=True, figsize=figsize,
                title='Examples of Randomly Sampled Spatial Inputs')

        # Colorbar formatting parameters
        cbargs = dict(pad=0.01, shrink=0.7, aspect=10)

        N = min(len(M), maxN)
        r, c = tiling_dims(N)
        for i in range(N):
            args = tuple(M.iloc[i]) + (self.c.wall_transition, self.c.wall_k)
            ax = f.add_subplot(r, c, i+1)
            Fplot(ax, xgrid, ygrid, args, mask_color='0.8', cbargs=cbargs)
            quicktitle(ax, M.index[i])

        plt.ion()
        plt.show()

    def create_spatial_group(self, params, trajgrp):
        """Create a group for the spatial inputs for the simulations."""
        eqns = br.Equations("""
            f = F(x,y,r,Lx,Ly,Lxx,Lyy,Lxy,Lw,b0,phi,w50,k) : 1 (constant over dt)
            x : 1 (linked)
            y : 1 (linked)
            r : 1 (linked)
            Lx : 1
            Ly : 1
            Lxx : 1
            Lyy : 1
            Lxy : 1
            Lw : 1
            b0 : 1
            phi : 1
        """)
        G = br.NeuronGroup(len(params), model=eqns, name='SpatialLayer',
                clock=trajgrp.clock, order=0)

        G.namespace['F'] = Fsim
        G.namespace['w50'] = self.c.wall_transition
        G.namespace['k'] = self.c.wall_k
        G.namespace.update(trajgrp.namespace)

        G.x = br.linked_var(trajgrp.x)
        G.y = br.linked_var(trajgrp.y)
        G.r = br.linked_var(trajgrp.r)

        P = params.to_records()
        G.Lx = P.X
        G.Ly = P.Y
        G.Lxx = P.XX
        G.Lyy = P.YY
        G.Lxy = P.XY
        G.Lw = P.W
        G.b0 = P.B0
        G.phi = P.phi

        return G

    def generate_spatial_layer(self, N, trajgrp=None, cell_ids=None,
        subtype='negative', fuzz=DEFAULT_FUZZ, rotate=True, plot=False,
        return_model=False):
        """Generate a simulation spatial layer group from a new data sample.

        Arguments:
        N -- number of spatial inputs in the layer

        Keyword arguments:
        trajgrp -- trajectory simulation group (see sim.TrajectoryModel)
        cell_ids -- optional, list of cell ids to restrict sampling
        subtype -- if cell_ids not specified, phaser subtype for sampling
        fuzz -- randomization of spatial weights in s.d. units
        rotate -- whether spatial inputs should be reandomly rotated
        plot -- whether to plot the evaluated spatial functions
        return_model -- whether to return (G, M) tuple including model

        Returns:
        G -- simulation group of size N representing spatial inputs
        G, M -- if return_model is True, also return normalized model matrix
        """
        if trajgrp is None:
            trajgrp = sim.TrajectoryModel(DEFAULT_TRAJ_SESSION).new_group()

        if cell_ids is None:
            B = self.read_dataframe('spatial_data', subtype)
        else:
            B = self.read_dataframe('spatial_data/all').loc[cell_ids]

        S = self.sample_spatial_model(B, n=N, fuzz=fuzz, rotate=rotate)
        M = self.normalize_spatial_model(S)
        G = self.create_spatial_group(M, trajgrp)

        if plot:
            self.plot_normalized_model_samples(M)

        if return_model:
            return G, M
        return G

    @step
    def test_spatial_simulation(self, N, duration=300*second,
        subtype='negative', fuzz=DEFAULT_FUZZ, rotate=True):
        """Run a test simulation of just the spatial input layer."""
        traj = sim.TrajectoryModel(DEFAULT_TRAJ_SESSION)
        Trajectory = traj.new_group()

        SpatialLayer = self.generate_spatial_layer(N, trajgrp=Trajectory,
                plot=True, subtype=subtype, fuzz=fuzz, rotate=rotate)

        Mon = br.StateMonitor(SpatialLayer, 'f', record=True)
        TrajMon = br.StateMonitor(Trajectory, ['x','y'], record=[0])

        TestNetwork = br.Network(name='TestNetwork')
        TestNetwork.add(SpatialLayer)
        TestNetwork.add(Trajectory)
        TestNetwork.add(Mon)
        TestNetwork.add(TrajMon)

        self.out('Running simulation...')
        TestNetwork.run(duration)

        self.save_array('test_sim/T', TrajMon.t, title='Test Simulation: T')
        self.save_array('test_sim/X', TrajMon.x, title='Test Simulation: X')
        self.save_array('test_sim/Y', TrajMon.y, title='Test Simulation: Y')
        self.save_array('test_sim/Fx', Mon.f, title='Test Simulation: Fx')

    @step
    def plot_test_simulation(self, maxN=36, figsize=(12,12)):
        """Plot the output from the test spatial simulation."""
        self.out('Plotting simulation output...')
        f = self.figure('spatial-simulation', clear=True, figsize=figsize,
                title='Test of Spatial Input Layer Simulation')

        # Load simulation data
        X = np.squeeze(self.read_array('test_sim/X'))
        Y = np.squeeze(self.read_array('test_sim/Y'))
        Fx = self.read_array('test_sim/Fx')
        N = len(Fx)
        self.out('Loaded data for {} spatial inputs.'.format(N))

        # Colorbar formatting parameters
        cbticks = dict(pad=0, length=0)
        cbargs = dict(pad=0.01, shrink=0.7, aspect=10)
        trkw = dict(linewidths=3, cbar=True)

        N_plots = min(N, maxN)
        r, c = tiling_dims(N_plots)
        for i in range(N_plots):
            ax = f.add_subplot(r, c, i+1)
            fmin = round(Fx[i].min(), 2)
            fmax = round(Fx[i].max(), 2)
            ticks = FixedLocator([fmin, (fmin + fmax) / 2, fmax])
            cbargs.update(ticks=ticks)
            _, cb = maps.trace(X, Y, Fx[i], ax=ax, cbargs=cbargs, **trkw)
            cb.ax.tick_params(**cbticks)
            cb.ax.yaxis.tick_right()
            ax.set_axis_off()
            quicktitle(ax, 'input {}'.format(i))
