"""
Spatial stability analysis of criteria-selected phaser cells.
"""

from itertools import product

import numpy as np
import pandas as pd
import scipy.stats as st
import matplotlib.pyplot as plt
import matplotlib.ticker as tk
import matplotlib as mpl

from roto.radians import RadianFormatter, cdiff
from roto.stats import freedman_diaconis_bins as fdbins, empirical_CI
from roto.misc import unique_pairs
import roto.axis as ra

from .. import store
from ..context import SPCAnalysis, step
from ..lib import phasers, session, data, labels
from ..lib.theta import phaseregress
from ..tools import maps


POSCOLOR = 'sandybrown'
NEGCOLOR = 'slateblue'
NONCOLOR = 'dimgray'
PHCOLOR = (0.3, 0.73372549, 0.61568627)  # lightened colorblind 'g'
SIGCOLOR = (0.3, 0.61294117, 0.78862745)  # lightened colorblind 'b'


class PhaserStabilityAnalysis(SPCAnalysis):

    """
    Conduct spatial correlation analyses to assess phaser stability.
    """

    def cleanup(self):
        store.close()

    def get_dataframe(self):
        """Phaser properties dataframe with cell-id column."""
        P = phasers.filtered_dataframe()
        rdf = data.dataframe_from_table('/recordings', index='id')
        P['cell_id'] = rdf.loc[P.index, 'cell_id']
        P['ttc'] = rdf.loc[P.index, 'ttc']
        P['num_recs'] = -1  # new column for number of per-cell recordings
        for c_id, group in P.groupby('cell_id'):
            P.loc[P.cell_id == c_id, 'num_recs'] = len(group)
        return P

    @step
    def compute_maps_shifts(self, maxwin=3600.0, adaptive=True, mapres=64):
        """Compute and store firing-rate maps including early/late splits."""
        P = self.get_dataframe()
        N = len(P)

        def calc_phase_shift(rmap, pmap):
            valid = np.isfinite(rmap) & np.isfinite(pmap)
            R, P = rmap[valid], pmap[valid]
            sl, _, _, _ = phaseregress(rmap[valid], pmap[valid])
            return R.ptp() * sl

        ratemaps = []
        early = []
        late = []
        phasemaps = []
        pearly = []
        plate = []
        shift_early = np.empty(N)
        shift_late = np.empty(N)
        for i, (rec_id, rec) in enumerate(P.iterrows()):
            S = session.get(rec['session_id'], res=mapres)
            self.out('Computing ratemaps for recording {}', rec_id)

            # Split early/late intervals as first/last maxwin of session
            # except meeting at the midpoint if the session is shorter than
            # twice the maxwin length
            early_interval = [S.tstart, S.tstart+maxwin]
            late_interval = [S.tend-maxwin, S.tend]
            if early_interval[1] > late_interval[0]:
                midpoint = S.tstart + S.duration / 2
                early_interval[1] = late_interval[0] = midpoint
                self.out('Warning: session {} duration ({:.1f}s) is less than '
                        '2*{:.1f}s.', S.attrs['id'], S.duration, maxwin,
                        error=True)

            # Full, early, and late ratemaps
            full_map = S.ratemap(rec['ttc'], adaptive=adaptive)
            early_map = S.ratemap(rec['ttc'], intervals=[early_interval],
                    adaptive=adaptive)
            late_map = S.ratemap(rec['ttc'], intervals=[late_interval],
                    adaptive=adaptive)

            ratemaps.append(full_map)
            early.append(early_map)
            late.append(late_map)

            # Full, early, and late phase maps
            full_pmap = S.phasemap(rec['ttc'], adaptive=adaptive)
            early_pmap = S.phasemap(rec['ttc'], intervals=[early_interval],
                    adaptive=adaptive)
            late_pmap = S.phasemap(rec['ttc'], intervals=[late_interval],
                    adaptive=adaptive)

            shift_early[i] = calc_phase_shift(early_map, early_pmap[0])
            shift_late[i] = calc_phase_shift(late_map, late_pmap[0])

            phasemaps.append(full_pmap)
            pearly.append(early_pmap)
            plate.append(late_pmap)

        # Save ratemaps
        self.save_array('ratemaps/full', ratemaps)
        self.save_array('ratemaps/early', early)
        self.save_array('ratemaps/late', late)

        # Save phase maps
        self.save_array('phasemaps/full', phasemaps)
        self.save_array('phasemaps/early', pearly)
        self.save_array('phasemaps/late', plate)
        self.save_array('shifts/early', shift_early)
        self.save_array('shifts/late', shift_late)

    @step
    def example_phase_coherence(self, cohmax=1.0, example_id=97, CI_alpha=0.1,
        nbins=32, minpixels=5, figsize=(7.5,3.4)):
        """Example relationship between early/late avg. phase and coherence."""
        P = self.get_dataframe()
        F = self.read_array('phasemaps/full')
        E = self.read_array('phasemaps/early')
        L = self.read_array('phasemaps/late')

        # Get example phase and coherence maps
        ex_id = (P.index == example_id).nonzero()[0][0]
        ex_coh = F[ex_id,1]
        ex_phase_early = E[ex_id,0]
        ex_phase_late = L[ex_id,0]

        # Process coherence/phase maps into valid-pixel arrays
        valid = np.isfinite(ex_coh) & np.isfinite(ex_phase_early) & \
                    np.isfinite(ex_phase_late)
        ex_coh = ex_coh[valid]
        ex_phase_early = ex_phase_early[valid]
        ex_phase_late = ex_phase_late[valid]
        ex_phase_diff = np.abs(cdiff(ex_phase_early, ex_phase_late))

        # Binned and masked data for mean+CI summary curve
        cbins = np.linspace(0, 1, nbins+1)
        centers = (cbins[:-1] + cbins[1:]) / 2
        coh_bin_ix = np.digitize(ex_coh, cbins)
        diff_mu = np.zeros(nbins)
        diff_CI = np.zeros((2,nbins))
        mask = np.zeros(nbins, '?')
        for i in range(nbins):
            ix = (coh_bin_ix == i)
            if ix.sum() < minpixels:
                mask[i] = True
                continue
            diff_mu[i] = mu = ex_phase_diff[ix].mean()
            l, u = empirical_CI(ex_phase_diff[ix], alpha=CI_alpha)
            diff_CI[:,i] = mu - l, u - mu
        diff_mu = np.ma.array(diff_mu, mask=mask)
        diff_CI = np.ma.array(diff_CI, mask=np.c_[mask,mask].T)

        f = self.figure('example-coherence-cell-%04d' % example_id,
                clear=True, figsize=figsize, title='Spatial Phase Stability: '
                      '{}'.format(labels.session_cell_id(example_id)))

        axm = f.add_subplot(121)
        axi = f.add_subplot(122)

        # Plot formatting
        pixelfmt = dict(marker=',', s=4, c='k', alpha=0.2, linewidths=0)
        errfmt = dict(fmt='o-', c='k', ecolor='k', capsize=0, ms=4, mew=0,
                lw=1, elinewidth=1)
        diffmax = np.pi

        def plot_pixels(ax):
            ax.scatter(ex_coh, ex_phase_diff, **pixelfmt)
            ax.set_xlabel('Mean resultant vector length')
            ax.set_ylabel(r'Early-late mean phase $|\Delta|$')
            ax.set_xlim(left=0)
            ax.set_ylim(0, diffmax)
            ax.set_yticks(np.linspace(0,1,5)*np.pi)
            ax.yaxis.set_major_formatter(RadianFormatter(inline=False))
            ra.despine(ax)
        plot_pixels(axm)
        ymax = axm.get_ylim()[1]

        def plot_summary(ax):
            ax.errorbar(centers, diff_mu, yerr=diff_CI, **errfmt)
            ax.set_xlabel('Mean resultant vector length')
            ax.set_xlim(0, cohmax)
            ax.set_ylim(bottom=0, top=diffmax)
            ax.set_yticks(np.linspace(0,1,5)*np.pi)
            ax.yaxis.set_major_formatter(RadianFormatter(inline=False))
            ra.despine(ax)
        plot_summary(axi)

        # Save images of early/late phase-coherence maps
        savekw = dict(cmin=-np.pi, cmax=np.pi, levels=12, mask_color='w',
                circular=True)
        maps.save(E[ex_id,0], self.path('phasemap-early.png'), **savekw)
        self.out('Saved: {}'.format(self.path('phasemap-early.png')))
        maps.save(L[ex_id,0], self.path('phasemap-late.png'), **savekw)
        self.out('Saved: {}'.format(self.path('phasemap-late.png')))
        maps.save(F[ex_id,1], self.path('coherence-full.png'), cmin=0,
                mask_color='w', cmap='gray')
        self.out('Saved: {}'.format(self.path('coherence-full.png')))

    @step
    def analyze_phase_coherence(self, cohmax=0.6, cohbins=64, diffbins=64,
        cmap='cubehelix_r', cbar=False, report_thresh=np.pi/4,
        figsize=(7.5,3.4)):
        """Compute phase-coherence curves for all phaser cells."""
        P = self.get_dataframe()
        F = self.read_array('phasemaps/full')
        E = self.read_array('phasemaps/early')
        L = self.read_array('phasemaps/late')

        neg = np.zeros((cohbins, diffbins))
        pos = np.zeros_like(neg)
        Nneg = 0
        Npos = 0
        negreport = []
        posreport = []

        cmin, cmax = 0, cohmax
        cbins = np.linspace(cmin, cmax, cohbins+1)

        dmin, dmax = 0, np.pi
        dbins = np.linspace(dmin, dmax, diffbins+1)
        bins = (cbins, dbins)
        extent = [cmin, cmax, dmin, dmax]

        for c_id, rec in P.iterrows():

            # Get phase and coherence maps
            data_id = (P.index == c_id).nonzero()[0][0]
            coh = F[data_id,1]
            phase_early = E[data_id,0]
            phase_late = L[data_id,0]

            # Process coherence/phase maps into valid-pixel arrays
            valid = np.isfinite(coh) & np.isfinite(phase_early) & \
                        np.isfinite(phase_late)
            coh = coh[valid]
            phase_early = phase_early[valid]
            phase_late = phase_late[valid]
            phase_diff = np.abs(cdiff(phase_early, phase_late))

            # Binned heatmap of coherence x phase diff. across pixels
            H, _, _ = np.histogram2d(coh, phase_diff, normed=False, bins=bins)
            Hdiff = H.T / coh.size

            # Concatenate curve to neg/pos categories
            report = (phase_diff < report_thresh).sum() / phase_diff.size
            if rec.subtype == 'negative':
                self.box(filled=True, color='purple')
                negreport.append(report)
                neg += Hdiff
                Nneg += 1
            elif rec.subtype == 'positive':
                self.box(filled=True, color='brown')
                posreport.append(report)
                pos += Hdiff
                Npos += 1

        neg /= Nneg
        pos /= Npos

        self.newline()
        self.out('Percent negative <{:.3f} diff = {:.3f} +/- {:.3f} s.e.m.',
                report_thresh, 100*np.mean(negreport),
                100*np.std(negreport)/np.sqrt(len(negreport)))
        self.out('Percent positive <{:.3f} diff = {:.3f} +/- {:.3f} s.e.m.',
                report_thresh, 100*np.mean(posreport),
                100*np.std(posreport)/np.sqrt(len(posreport)))

        f = self.figure('phase-stability', clear=True, figsize=figsize,
                title='Phase-Code Stability: Negative/Positive Phaser Cells')

        axn = f.add_subplot(121)
        axp = f.add_subplot(122)

        # Plot formatting
        imfmt = dict(interpolation='nearest', origin='lower', extent=extent,
                vmin=0, cmap=cmap)
        errfmt = dict(fmt='s-', capsize=0, ms=3, mew=0, lw=1, elinewidth=1)
        cbarfmt = dict(pad=0, shrink=0.6, fraction=0.07)

        def plot_data(ax, data, label, N):
            im = ax.imshow(data, **imfmt)
            ra.quicktitle(ax, '{} (n={} recordings)'.format(label, N),
                    size='small', weight='light')
            return im
        imn = plot_data(axn, neg, 'Negative', Nneg)
        imp = plot_data(axp, pos, 'Positive', Npos)

        def finish_plot(ax, im, ylabel=False):
            ax.set_xlabel('Mean resultant vector length')
            if ylabel:
                ax.set_ylabel(r'Early-late mean phase $|\Delta|$')
            ax.set_xlim(cmin, cmax)
            ax.set_ylim(dmin, dmax)
            ax.set_yticks(np.linspace(0,1,5)*np.pi)
            ax.yaxis.set_major_formatter(RadianFormatter(inline=False))
            if cbar:
                cb = plt.colorbar(ax=ax, mappable=im, **cbarfmt)
                cb.set_ticks([])
            ra.despine(ax)
        finish_plot(axn, imn, ylabel=True)
        finish_plot(axp, imp)

    @step
    def analyze_maps_shift(self):
        """Analyze intertmeporal spatial correlations and phase shifts."""
        self.get_datafile(False)
        P = self.get_dataframe()
        N = len(P)
        R = self.read_array('ratemaps/full')
        E = self.read_array('ratemaps/early')
        L = self.read_array('ratemaps/late')
        shifts_early = self.read_array('shifts/early')
        shifts_late = self.read_array('shifts/late')

        # Create NxN cell-identity mask matrix
        same = np.zeros((N,N), '?')
        for i, (rec_id, rec) in enumerate(P.iterrows()):
            this_cell = rec['cell_id']
            same[i] = (P.cell_id == this_cell)

        # Intertemporal phase shifts
        def shift_distance(a, b):
            return np.abs(a - b)
        adist = shift_distance(shifts_late, shifts_early)
        xdist = np.array([])
        for i in range(N):
            this_xdist = shift_distance(shifts_late[~same[i]], shifts_early[i])
            xdist = np.r_[xdist, this_xdist]

        # Intertemporal spatial correlations
        def correlation(a, b):
            valid = np.isfinite(a) & np.isfinite(b)
            a = a[valid]
            b = b[valid]
            A = a - a.mean()
            B = b - b.mean()
            Anorm = np.sqrt(np.dot(A, A))
            Bnorm = np.sqrt(np.dot(B, B))
            return np.dot(A, B) / (Anorm * Bnorm)
        acorr = [correlation(E[i], L[i]) for i in range(N)]
        xcorr = np.array([])
        for i in range(N):
            for other_late in L[~same[i]]:
                this_xcorr = correlation(E[i], other_late)
                xcorr = np.r_[xcorr, this_xcorr]

        # Multiple recordings phase shifts and shift distances
        mult_adist = np.array([])
        mult_xdist = np.array([])
        mult_shifts = dict(cell_id=[], shift=[], I_phase=[])
        for c_id in P.loc[P.num_recs > 1, 'cell_id'].unique():
            info = P.loc[P.cell_id == c_id, 'I_phase'].values
            shifts = P.loc[P.cell_id == c_id, 'rp_shift'].values
            mult_shifts['I_phase'].extend(info)
            mult_shifts['shift'].extend(shifts)
            mult_shifts['cell_id'].extend([c_id] * len(shifts))

            adists = [shift_distance(s, t) for s, t in unique_pairs(shifts)]
            mult_adist = np.r_[mult_adist, adists]

            other_shifts = P.loc[P.cell_id != c_id, 'rp_shift'].values
            xdists = np.ravel([shift_distance(s, other_shifts) for s in shifts])
            mult_xdist = np.r_[mult_xdist, xdists]

        # Multiple recordings spatial correlations
        mult_acorr = np.array([])
        mult_xcorr = np.array([])
        for c_id in P.loc[P.num_recs > 1, 'cell_id'].unique():
            same_cell = (P.cell_id == c_id).nonzero()[0]
            other_cells = (P.cell_id != c_id).nonzero()[0]

            acorrs = [correlation(R[a], R[b]) for a, b in
                        unique_pairs(same_cell)]
            mult_acorr = np.r_[mult_acorr, acorrs]

            xcorrs = [correlation(R[a], R[b]) for a, b in
                        product(same_cell, other_cells)]
            mult_xcorr = np.r_[mult_xcorr, xcorrs]

        # Save all the intertemporal phase shift and correlation data
        self.save_array('analysis/cell_identity', same)
        self.save_array('analysis/adist', adist)
        self.save_array('analysis/xdist', xdist)
        self.save_array('analysis/acorr', acorr)
        self.save_array('analysis/xcorr', xcorr)
        mult_shifts_df = pd.DataFrame(data=mult_shifts)
        self.save_dataframe('analysis/mult_shifts', mult_shifts_df)
        self.save_array('analysis/mult_adist', mult_adist)
        self.save_array('analysis/mult_xdist', mult_xdist)
        self.save_array('analysis/mult_acorr', mult_acorr)
        self.save_array('analysis/mult_xcorr', mult_xcorr)

    @step
    def plot_stability_analysis(self, noncolor=NONCOLOR, phcolor=PHCOLOR,
        figsize=(7.05442177, 4.11564626)):
        """Create figure for intertemporal phase shift and correlation data."""
        adist = self.read_array('analysis/adist')
        xdist = self.read_array('analysis/xdist')
        acorr = self.read_array('analysis/acorr')
        xcorr = self.read_array('analysis/xcorr')
        mult_shifts = self.read_dataframe('analysis/mult_shifts')
        mult_adist = self.read_array('analysis/mult_adist')
        mult_xdist = self.read_array('analysis/mult_xdist')
        mult_acorr = self.read_array('analysis/mult_acorr')
        mult_xcorr = self.read_array('analysis/mult_xcorr')

        fig = self.figure('stability', clear=True, figsize=figsize,
                title='Stability of Phase Shifts and Spatial Correlations')
        gshape = (2,3)

        # Plot formatting
        phaselabel = r'Total phase shift $|\Delta|$'
        xhistfmt = dict(lw=0.9, ec=noncolor, normed=True, histtype='step')
        ahistfmt = dict(lw=0, fc=phcolor, normed=True, histtype='stepfilled')
        shiftfmt = dict(lw=1.2, marker='+', mew=1.2, ms=6,
                solid_capstyle='round', alpha=0.7)

        def plot_spatial_corrs(ax, adata, xdata, ylabel):
            ax.hist(adata, bins=fdbins(adata), **ahistfmt)
            ax.hist(xdata, bins=fdbins(xdata), **xhistfmt)
            ax.yaxis.set_major_locator(tk.MaxNLocator(nbins=5))
            ax.set(xlabel='Spatial correlation', ylabel=ylabel)
            ax.set_xlim(-1, 1)
            ra.despine(ax)
            self.out('Corr. {}, within = {}', ylabel, adata.size)
            self.out('Corr. {}, across = {}', ylabel, xdata.size)
        plot_spatial_corrs(plt.subplot2grid(gshape, (0,0)), acorr, xcorr,
                'Early-late pair density')

        def plot_phase_dists(ax, adata, xdata, ylabel):
            ax.hist(adata, bins=fdbins(adata), **ahistfmt)
            ax.hist(xdata, bins=fdbins(xdata), **xhistfmt)
            ax.yaxis.set_major_locator(tk.MaxNLocator(nbins=5))
            ax.set_xticks(np.pi*np.linspace(0, 4, 5))
            ax.xaxis.set_major_formatter(RadianFormatter(inline=True))
            ax.set(xlabel=phaselabel, ylabel=ylabel)
            ax.set_xlim(0, 4*np.pi)
            ra.despine(ax)
            self.out('Phase dist. {}, within = {}', ylabel, adata.size)
            self.out('Phase dist. {}, across = {}', ylabel, xdata.size)
        plot_phase_dists(plt.subplot2grid(gshape, (0,1)), adist, xdist,
                'Early-late pair density')

        def plot_mult_phase_curves(ax, df, column, ylabel):
            maxpts = 0
            unique_cells = df.cell_id.unique()
            n = len(unique_cells)
            cols = mpl.cm.Paired(np.linspace(0,1,n))
            for i, c_id in enumerate(unique_cells):
                shift = df.loc[df.cell_id == c_id, column]
                npts = len(shift)
                maxpts = max(maxpts, npts)
                ax.plot(np.arange(1, npts+1), shift, c=cols[i], label=c_id,
                        **shiftfmt)
            ax.set_xlabel('Day number')
            ax.set_ylabel(ylabel)
            ax.set_xlim(0.5, maxpts+0.5)
            ra.despine(ax)
            return ax

        ax = plot_mult_phase_curves(plt.subplot2grid(gshape, (0,2)),
                mult_shifts, 'I_phase', ylabel=r'$I_{\rm phase}$ (bits)')
        ax.set_ylim(0, ax.axis()[3]+0.1)

        ax = plot_mult_phase_curves(plt.subplot2grid(gshape, (1,2)),
                mult_shifts, 'shift', ylabel='Total phase shift')
        ax.set_ylim(-2*np.pi, 2*np.pi)
        ax.set_yticks(np.pi*np.linspace(-2,2,5))
        ax.yaxis.set_major_formatter(RadianFormatter(inline=False))

        plot_phase_dists(plt.subplot2grid(gshape, (1,1)), mult_adist,
                mult_xdist, 'Day pair density')

        plot_spatial_corrs(plt.subplot2grid(gshape, (1,0)), mult_acorr,
                mult_xcorr, 'Day pair density')

    @step
    def compute_stability_stats(self):
        """Statistics for within-cell vs. cross-cell differences."""
        adist = self.read_array('analysis/adist')
        xdist = self.read_array('analysis/xdist')
        acorr = self.read_array('analysis/acorr')
        xcorr = self.read_array('analysis/xcorr')
        mult_adist = self.read_array('analysis/mult_adist')
        mult_xdist = self.read_array('analysis/mult_xdist')
        mult_acorr = self.read_array('analysis/mult_acorr')
        mult_xcorr = self.read_array('analysis/mult_xcorr')

        def report_stat(label, within, cross):
            self.out('{}: median = {:.6f}', label, np.median(within))

            D, p = st.ks_2samp(within, cross)
            self.out('{}: D = {:.3f}, p = {:g}', label, D, p)

            U, p = st.mannwhitneyu(within, cross)
            self.out('{}: U = {:.3f}, p = {:g}', label, U, p)
            self.newline()

        report_stat('Early-late phase distance', adist, xdist)
        report_stat('Early-late spatial corr.', acorr, xcorr)
        report_stat('Cross-recording phase distance', mult_adist, mult_xdist)
        report_stat('Cross-recording spatial corr.', mult_acorr, mult_xcorr)

    @step
    def split_correlation_matrix(self, cmap='CMRmap', figsize=(6,4.5)):
        """Calculate the correlation (vector similarity) matrix for splits."""
        P = self.get_dataframe()
        rec_ids = P.loc[P.phaser].index
        N = len(rec_ids)

        self.out('Load early/late split ratemaps...')
        early = self.read_array('ratemaps/early')
        late = self.read_array('ratemaps/late')
        allmaps = np.r_[early, late]

        self.out('Calculating correlation matrix...')
        C = np.empty((2*N, 2*N), 'd')
        for i, j in product(range(2*N), repeat=2):
            U = allmaps[i]
            V = allmaps[j]
            valid = np.isfinite(U) & np.isfinite(V)
            U = U[valid]
            V = V[valid]
            U -= np.mean(U)  # adjusted cosine similarity
            V -= np.mean(V)  #
            Unorm = np.sqrt(np.dot(U, U))
            Vnorm = np.sqrt(np.dot(V, V))
            C[i,j] = np.dot(U, V) / (Unorm * Vnorm)
            # C[i,j], _ = st.pearsonr(U, V)

        self.save_array('corr_split', C, title='Early/Late Split Ratemaps: '
            'Correlation Matrix')

        fig = self.figure('split-corrmat', clear=True, figsize=figsize,
                title='Early/Late Correlation Matrix for Phaser Recordings')
        axm = fig.add_subplot(121)
        axc = fig.add_subplot(122)

        def plot_matrix(ax):
            ax.imshow(C, origin='lower', interpolation='nearest', cmap=cmap)
            ax.axis('equal')
            ax.axis('tight')

        plot_matrix(axm)
