"""
Model simulation of spatial phase-stripe (VCO-like) synchronization to phasers.
"""

from itertools import product

import numpy as np
import pandas as pd
import scipy.signal as sig
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator

from brian2.units import *
import brian2 as br

import roto.axis as ra
from roto.misc import merge_two_dicts
from roto.arrays import maxima
from roto.radians import RadianFormatter, cdiff
from roto.axis import quicktitle, despine
from roto.dicts import Reify
from roto.circstats import mean_resultant_vector, mean_angle, mean_length
from roto.movies import FigureMovie
from roto.images import tiling_dims, array_to_image
from roto.colormaps import huslmap
from roto.stats import confidence_interval, empirical_CI, Interval

import spc.ana.spatial_input as spi
import spc.ana.phaser_model as phm
import spc.ana.sync_model as sym

from ..context import SPCSimulation, step
from ..lib import sim, motion
from ..tools import spikes as spk, binned, maps, adaptive, decoder
from .. import store


DEFAULT_BURSTSIM_DIR = '/Users/joe/projects/spc/results/' \
                       '2017-07-29+burst-sync-simulation+beta'

# VCO-like parameters
DEFAULT_X0 = 40.0           # cm, spatial reference point (x)
DEFAULT_Y0 = 40.0           # cm, spatial reference point (y)
DEFAULT_THETA0 = -1.2763    # radians, phase at spatial reference point
DEFAULT_THETA1 = 1.9635     # radians, ibid. (alternate value)
DEFAULT_L = 80.0            # cm, VCO spatial scale
DEFAULT_PHI = np.pi/4       # radians, VCO preferred direction

# Learning parameters
DEFAULT_EMAX = 0.035        # fraction of all-or-none winners

# Spatial feedback paramters
VCO_NEG_FDBK_GAIN = 10.0
VCO_POS_FDBK_GAIN = 5.0

# Burst parameters
IBI_MIN = 0.025             # seconds, inter-burst interval threshold
DEFAULT_WIN_THETA = 2.0     # number of theta cycles to decode
DEFAULT_WINC_FRAC = 0.5     # fraction of decoding time window to slide
DEFAULT_PRIOR = 15.0        # Gaussian kernel width (cm) for continuity prior


# Decoding performance plot colors
# sns.xkcd_palette(['reddish orange', 'teal blue', 'teal green', 'deep lilac'])
PHASECOL, DIR1COL, DIR2COL, ALLCOL = [
 (0.9725490196078431, 0.2823529411764706, 0.10980392156862745),
 (0.00392156862745098, 0.5333333333333333, 0.6235294117647059),
 (0.1450980392156863, 0.6392156862745098, 0.43529411764705883),
 (0.5882352941176471, 0.43137254901960786, 0.7411764705882353)]


class StripeSyncSimulation(SPCSimulation):

    """
    Simulations synchronizing phase-strip burster cell with spatial phasers.
    """

    def cleanup(self):
        store.close()

    @step
    def setup(self, x0=DEFAULT_X0, y0=DEFAULT_Y0, theta0=DEFAULT_THETA0,
        theta1=DEFAULT_THETA1, L=DEFAULT_L, phi=DEFAULT_PHI, Emax=DEFAULT_EMAX,
        traj_session_id=None, burstsim_dir=DEFAULT_BURSTSIM_DIR):
        """Load parameters for phaser and burster models."""
        self['x0'] = x0
        self['y0'] = y0
        self['theta0'] = theta0
        self['theta1'] = theta1
        self['L'] = L
        self['phi_pref'] = phi
        self['Emax'] = Emax
        self['vco_neg_fdbk'] = VCO_NEG_FDBK_GAIN
        self['vco_pos_fdbk'] = VCO_POS_FDBK_GAIN

        # Load burster model parameters
        bsim = sym.BurstSyncSimulation.load(burstsim_dir)
        self['Ie'] = bsim.c.Ie
        self['tau'] = bsim.c.tau
        self['phaser_model_dir'] = bsim.c.phaser_model_dir
        self['sigma'] = sym.DEFAULT_SIGMA
        self['neg_fdbk'] = sym.DEFAULT_NEG_FDBK_GAIN
        self['pos_fdbk'] = sym.DEFAULT_POS_FDBK_GAIN

        # Load phaser model parameters
        psim = phm.PhaserSimulation.load(self.c.phaser_model_dir)
        self['ftheta'] = psim.c.ftheta
        self['tau_phaser'] = psim.c.tau
        self['ge_phaser'] = psim.c.ge
        self['gtheta_phaser'] = psim.c.gtheta
        self['spatial_model_dir'] = psim.c.spatial_model_dir
        self['traj_session_id'] = traj_session_id or psim.c.traj_session_id

        # Full session duration
        mdata = motion.MotionData(self.c.traj_session_id)
        tstart, tend = mdata.t[0], mdata.t[-1]
        self['session_duration'] = tend - tstart

    @step
    def spatial_phaser_training_simulation(self, N, dur=1000.0):
        """Run a simulation of spatial phasers for training."""
        br.defaultclock.dt = 1*ms
        duration = dur*second

        self['N_phaser_training'] = N

        # Generate spatial model for training and (later) testing
        trajmodel = sim.TrajectoryModel(self.c.traj_session_id)
        Trajectory = trajmodel.new_group()
        datamodel = spi.SpatialInputModel.load(self.c.spatial_model_dir)
        SpatialLayer, SpatialModel = datamodel.generate_spatial_layer(
                N, trajgrp=Trajectory, plot=False, return_model=True)

        # Save spatial model matrix to be loaded for test simulations
        self.save_dataframe('training/spatial_model', SpatialModel,
            title='Spatial Model Matrix: Phaser Training')

        # Bring trajectory data into namespace for recording spike positions
        X = Trajectory.namespace['X']
        Y = Trajectory.namespace['Y']

        a = 0.02
        b = 0.2
        c = -50.0
        d = 4.0
        f_theta = self.c.ftheta*Hz
        ge = self.c.ge_phaser
        gtheta = self.c.gtheta_phaser
        gtheta_pos = 5.0*gtheta
        tau = self.c.tau_phaser*ms
        v0 = -75.0
        Wneg = 3.0
        Ei = -80.0
        tau_i = 100*ms

        negative_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ispatial + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Spatial drive from generative model of spatial input layer
            Ispatial = ge*Fx : 1
            Fx : 1 (linked)

            # Trajectory variables to record spike positions
            x = X(t) : 1 (shared, constant over dt)
            y = Y(t) : 1 (shared, constant over dt)

            # Theta phase and inhibitory (antiphase) input
            Itheta = -0.5*gtheta*(cos(phase) + 1) : 1 (shared)
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        Gneg = br.NeuronGroup(N, model=negative_eqns, threshold='v > 30',
                order=1, reset='v = c; u = u + d', name='NegativePhasers',
                method='rk4')
        Gneg.v = v0
        Gneg.Fx = br.linked_var(SpatialLayer.f)

        positive_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ineg + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Slow inhibitory synaptic current from negative phasers
            Ineg = -gi*(v - Ei) : 1
            dgi/dt = -gi/tau_i : 1

            # Theta phase and excitatory (inphase) input
            Itheta = 0.5*gtheta_pos*(cos(phase) + 1) : 1 (shared)

            # Trajectory variables to record spike positions
            x : 1 (linked)
            y : 1 (linked)

            # Theta phase
            phase : 1 (linked)
        """
        Gpos = br.NeuronGroup(N, model=positive_eqns, threshold='v > 30',
                order=2, reset='v = c; u = u + d', name='PositivePhasers',
                method='rk4')
        Gpos.v = v0
        Gpos.phase = br.linked_var(Gneg.phase)
        Gpos.x = br.linked_var(Gneg.x)
        Gpos.y = br.linked_var(Gneg.y)

        Cnp = br.Synapses(Gneg, Gpos, on_pre='gi += Wneg')
        Cnp.connect(j='i')

        NegMon = br.SpikeMonitor(Gneg, ['phase', 'x', 'y'], record=True,
                name='nspikes')
        PosMon = br.SpikeMonitor(Gpos, ['phase', 'x', 'y'], record=True,
                name='pspikes')
        TrajMon = br.StateMonitor(Trajectory, ['x','y'], record=[0],
                name='traj')

        @br.network_operation(dt=10*second, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        Net = br.Network(Gneg, Gpos, Cnp, NegMon, PosMon, TrajMon, Trajectory,
                SpatialLayer, progress, name='SpatialPhaserTrainingNetwork')
        Net.run(duration)

        self.save_simulation('training', Net,
                title='Spatial Phaser Training Simulation')

    @step
    def plot_training_simulation(self, **kwds):
        """Plot figure of rate/phasemap pairs for training phasers."""
        S = self.read_simulation('training')
        title = '2D Spatial Phaser Training'

        figkw = dict(maxN=50, levels=18, figsize=(11,11))
        figkw = merge_two_dicts(figkw, kwds)

        sim.plot_phaser_maps_figure(self, S.traj, S.nspikes,
                figname='training-negative-layer',
                figtitle='{}: Negative Layer'.format(title), **figkw)

        sim.plot_phaser_maps_figure(self, S.traj, S.pspikes,
                figname='training-positive-layer',
                figtitle='{}: Positive Layer'.format(title), **figkw)

    @step
    def compute_space_phase_distros(self, ibi_min=IBI_MIN):
        """Compute space-phase conditional probability distributions."""
        S = self.read_simulation('training')
        N = self.c.N_phaser_training
        Nx = binned.INFO_PHASE_POS_BINS
        Np = binned.INFO_PHASE_BINS

        sim.add_wrapped_phase_column(S.nspikes)
        sim.add_wrapped_phase_column(S.pspikes)

        S.nspikes['burst'] = spk.burst_onset(S.nspikes.t, ibi_min)
        S.pspikes['burst'] = spk.burst_onset(S.pspikes.t, ibi_min)

        sh = N, Nx, Nx, Np
        Pneg = np.empty(sh, 'd')
        Ppos = np.empty_like(Pneg)

        data = [(Pneg, S.nspikes, 'neg'), (Ppos, S.pspikes, 'pos')]
        for P, spikes, label in data:
            self.out('Computing {} phaser distros:', label)
            for n in range(N):
                xb = spikes.loc[spikes.neuron == n, 'x']
                yb = spikes.loc[spikes.neuron == n, 'y']
                bphase = spikes.loc[spikes.neuron == n, 'phase0']
                P[n] = binned.phasedist2d(xb, yb, bphase)
                self.box()
            self.newline()

        self.save_array('phasedists/neg', Pneg,
                title='Spatial Negative Phasers: Phase-Space Distribution')
        self.save_array('phasedists/pos', Ppos,
                title='Spatial Positive Phasers: Phase-Space Distribution')

    def create_vco_target_map(self, tag=None, Fbins=256, x0=None, y0=None,
        theta0=None, L=None, phi_pref=None, plot=True, mask=True,
        figsize=(4,4)):
        """Create VCO-like phase-stripe target map for learning weights."""
        diam = motion.CIRCLE_DIAMETER
        radius = diam / 2

        # Persistently set VCO function parameters in this context
        x0 = self['x0'] = self.c.x0 if x0 is None else x0
        y0 = self['y0'] = self.c.y0 if y0 is None else y0
        phi0 = self['theta0'] = self.c.theta0 if theta0 is None else theta0
        L = self['L'] = self.c.L if L is None else L
        phi_pref = self['phi_pref'] = \
                self.c.phi_pref if phi_pref is None else phi_pref
        args = (phi_pref, 1/L, phi0, x0, y0)

        # Evaluate the VCO spatial function on high-res pixel grid
        pts = np.linspace(0, diam, Fbins)
        xgrid, ygrid = np.meshgrid(pts, pts, indexing='ij')
        P = sim.Fvco(xgrid, ygrid, *args)

        # Save phasemap to tagged path/title
        if mask:
            rad2 = np.power(xgrid - radius, 2) + np.power(ygrid - radius, 2)
            P[rad2 > radius**2] = np.nan  # mask pixels outside arena
            savepath = 'vco/phasemap'
            title = 'VCO-like Phase Map Target'
            if tag:
                savepath += '_{}'.format(tag)
                title += ' ({})'.format(tag)
        else:
            savepath = 'decode/maps/{}'.format(tag)
            title = 'VCO-like Unmasked Decoding Target ({})'.format(tag)
        self.save_array(savepath, P, title=title)
        self.save_array('vco/xgrid', xgrid, title='Phase Map Pixel X-Values')
        self.save_array('vco/ygrid', ygrid, title='Phase Map Pixel Y-Values')

        if plot:
            figname = 'vco-target'
            if tag:
                figname += '-{}'.format(tag)
            f = self.figure(figname, clear=True, figsize=figsize,
                    title='VCO-like Phase Target Map')
            self._plot_vco_phasemap(P, f.add_subplot(111), circular=True)
            f.tight_layout(pad=1.05)

    @staticmethod
    def _plot_vco_phasemap(P, ax, **mapkw):
        """Plot VCO-like spatial phasemap to the given axis."""
        cbargs = dict(pad=0.03, shrink=0.6, aspect=15,
                ticks=FixedLocator([-np.pi,0,np.pi]),
                format=RadianFormatter(inline=False))
        cbtickkw = dict(axis='y', labelsize='large', pad=0, length=0)
        mapkw = merge_two_dicts(
                    dict(cmin=-np.pi, cmax=np.pi, cbar=True, levels=None,
                        mask_color='0.9', cbargs=cbargs),
                    mapkw)

        _, cb = maps.plot(P, ax=ax, **mapkw)

        cb.ax.tick_params(**cbtickkw)
        # cb.ax.yaxis.tick_right()
        ax.set_axis_off()

    @step
    def compute_weight_vector(self, Emax=None, tag=None, plot_distro=False):
        """Perform competitive Emax%-winners-take-all to create weights."""
        self.get_datafile(False)
        N = self.c.N_phaser_training

        # Load space-phase distributions for neg/pos phasers
        Pneg = self.read_array('phasedists/neg')
        Ppos = self.read_array('phasedists/pos')

        # Load VCO phasemap and pixel evaluation grid data
        path = 'vco/phasemap'
        if tag:
            path += '_{}'.format(tag)
        phasemap = self.read_array(path)
        xgrid = self.read_array('vco/xgrid')
        ygrid = self.read_array('vco/ygrid')

        # Get valid arena pixels and compute VCO-like phase distribution
        valid = np.isfinite(phasemap)
        xp = xgrid[valid]
        yp = ygrid[valid]
        phase = phasemap[valid]
        Pvco = binned.phasedist2d(xp, yp, phase)

        if plot_distro:
            figname = 'vco-phase-distro'
            if tag:
                figname += '-{}'.format(tag)
            f = self.figure(figname, clear=True, figsize=(9,9),
                    title='VCO-like Phase Distribution')
            for i in range(36):
                ax = f.add_subplot(6,6,i+1)
                maps.plot(Pvco[...,i], ax=ax, cmap='bone')
                ax.set_axis_off()
                quicktitle(ax, 'phase {}'.format(i))

        # Compute the 3D cosine-vector similarity between phaser distros
        # and the VCO target distro
        def cosine_sim(P):
            Pnorm = np.sqrt((P * P).sum(axis=(1,2,3)))
            vco_norm = np.sqrt((Pvco * Pvco).sum())
            norm = Pnorm * vco_norm
            return (P * Pvco).sum(axis=(1,2,3)) / norm

        # Phaser-subtype competitive (Emax %)-winners-take-all
        Emax = Emax or self.c.Emax
        Wneg = sim.Emax_winners(cosine_sim(Pneg), Emax)
        Wpos = sim.Emax_winners(cosine_sim(Ppos), Emax)

        self.out('Emax = {:.3f}', Emax)
        self.out('Active neg. synapses: {} / {}', np.sum(Wneg>0), Wneg.size)
        self.out('Active pos. synapses: {} / {}', np.sum(Wpos>0), Wpos.size)

        negpath = 'feedback/Wneg'
        pospath = 'feedback/Wpos'
        if tag:
            negpath += '_{}'.format(tag)
            pospath += '_{}'.format(tag)
        self.save_array(negpath, Wneg, title='Negative Phaser-VCO Weights')
        self.save_array(pospath, Wpos, title='Positive Phaser-VCO Weights')

    @step
    def compute_weighted_inputs(self, tag=None):
        """Compute total weighted inputs from phaser space-phase distros."""
        N = self.c.N_phaser_training
        Nx = binned.INFO_PHASE_POS_BINS
        Np = binned.INFO_PHASE_BINS

        # Load the phaser space-phase distributions
        Pneg = self.read_array('phasedists/neg')
        Ppos = self.read_array('phasedists/pos')

        # Load the negative/positive phaser-burst weight vectors
        Wneg_path = 'feedback/Wneg'
        Wpos_path = 'feedback/Wpos'
        if tag:
            Wneg_path += '_{}'.format(tag)
            Wpos_path += '_{}'.format(tag)
        Wneg = self.read_array(Wneg_path)
        Wpos = self.read_array(Wpos_path)

        # Compute weighted-average space-phase distributions
        Ineg = (Wneg[...,np.newaxis,np.newaxis] * Pneg).sum(axis=0)
        Ipos = (Wpos[...,np.newaxis,np.newaxis] * Ppos).sum(axis=0)
        TotalInput = Ineg + Ipos

        # Get center values and binsize for phase bins
        pedges = np.linspace(-np.pi, np.pi, Np + 1)
        pcenters = (pedges[:-1] + pedges[1:]) / 2
        pbinsize = 2*np.pi/Np

        # Compute bin-corrected weighted mean resultant vectors across space
        Pinp = np.zeros((2, Nx, Nx))
        for i, j in product(range(Nx), range(Nx)):
            Pinp[:,i,j] = mean_resultant_vector(pcenters,
                    weights=TotalInput[i,j], binsize=pbinsize)

        # Save the total input coherence map
        savepath = 'vco/inputmap'
        savetitle = 'Coherence Map of Total Phaser Input'
        if tag:
            savepath += '_{}'.format(tag)
            savetitle += ' ({})'.format(tag)
        self.save_array(savepath, Pinp, title=savetitle)
        self.out('Max. coherence of total weighted input = {:.3f}',
                np.nanmax(Pinp[1]))

    @step
    def plot_weighted_inputs(self, tag=None, figsize=(6,3.5)):
        """Plot figure of VCO target and resulting total input maps."""
        vcopath = 'vco/phasemap'
        inputpath = 'vco/inputmap'
        if tag:
            vcopath += '_{}'.format(tag)
            inputpath += '_{}'.format(tag)
        Pvco = self.read_array(vcopath)
        Pinp = self.read_array(inputpath)

        figname = 'weighted-inputs'
        figtitle = 'VCO-Weighted Phaser Space-Phase Inputs'
        if tag:
            figname += '-{}'.format(tag)
            figtitle += ' ({})'.format(tag)
        f = self.figure(figname, clear=True, figsize=figsize, title=figtitle)
        axv = f.add_subplot(121)
        axi = f.add_subplot(122)

        self._plot_vco_phasemap(Pvco, axv, circular=True)
        quicktitle(axv, 'VCO phase-code target')

        self._plot_vco_phasemap(Pinp, axi, satnorm=True)
        quicktitle(axi, 'Total phaser input')

        f.tight_layout(pad=1.02)

    def _sync_simulation_path(self, sigma, fdbk, tag=None):
        noise_str = 'sigma_{:.2f}'.format(sigma).replace('.', '_')
        fdbk = fdbk or (self.c.vco_neg_fdbk, self.c.vco_pos_fdbk)
        fdbk_str = 'neg_{:.2f}_pos_{:.2f}'.format(*fdbk).replace('.', '_')
        param_str = '{}_{}'.format(noise_str, fdbk_str)
        if tag:
            param_str += '_{}'.format(tag)
        path = 'vcosim/{}'.format(param_str)
        return path

    @step
    def spatial_phaser_burster_simulation(self, tag=None, noise=False,
        neg_fdbk=None, pos_fdbk=None, dur=60.0):
        """Data-driven spatial phasers synchronizing a VCO-like burster.

        Arguments:
        sigma -- burster voltage noise level in s.d.
        dur -- simulation duration in seconds, or 'session'
        """
        sigma = self.c.sigma if noise else 0.0

        if dur == 'session':
            dur = self.c.session_duration

        self.out('Simulating {:.1f} seconds'.format(dur))

        br.defaultclock.dt = 1*ms
        duration = dur*second
        N = self.c.N_phaser_training

        # Create the trajectory group
        trajmodel = sim.TrajectoryModel(self.c.traj_session_id)
        Trajectory = trajmodel.new_group()

        # Load spatial model matrix to recreate the training spatial layer
        datamodel = spi.SpatialInputModel.load(self.c.spatial_model_dir)
        SpatialModel = self.read_dataframe('training/spatial_model')
        SpatialLayer = datamodel.create_spatial_group(SpatialModel, Trajectory)

        # Bring trajectory data into namespace for recordings spike positions
        X = Trajectory.namespace['X']
        Y = Trajectory.namespace['Y']

        a = 0.02
        b = 0.2
        c = -50.0
        d = 4.0
        d_b = 5.0
        f_theta = self.c.ftheta*Hz
        ge = self.c.ge_phaser
        gtheta = self.c.gtheta_phaser
        gtheta_pos = 5.0*gtheta
        tau = self.c.tau_phaser*ms
        tau_b = self.c.tau*ms
        v0 = -75.0
        v0_b = -80.0
        Ie_b = self.c.Ie
        Wneg = 3.0
        Ei = -80.0
        tau_i = 100*ms

        # Load the negative/positive phaser-burst weight vectors
        Wneg_path = 'feedback/Wneg'
        Wpos_path = 'feedback/Wpos'
        if tag:
            Wneg_path += '_{}'.format(tag)
            Wpos_path += '_{}'.format(tag)
        Wnb = self.read_array(Wneg_path)
        Wpb = self.read_array(Wpos_path)

        negative_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ispatial + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Spatial drive from generative model of spatial input layer
            Ispatial = ge*Fx : 1
            Fx : 1 (linked)

            # Trajectory variables to record spike positions
            x = X(t) : 1 (shared, constant over dt)
            y = Y(t) : 1 (shared, constant over dt)

            # Theta phase and inhibitory (antiphase) input
            Itheta = -0.5*gtheta*(cos(phase) + 1) : 1 (shared)
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        Gneg = br.NeuronGroup(N, model=negative_eqns, threshold='v > 30',
                order=1, reset='v = c; u = u + d', name='NegativePhaser',
                method='rk4')
        Gneg.v = v0
        Gneg.Fx = br.linked_var(SpatialLayer.f)

        positive_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ineg + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Slow inhibitory synaptic current from negative phasers
            Ineg = -gi*(v - Ei) : 1
            dgi/dt = -gi/tau_i : 1

            # Theta phase and excitatory (inphase) input
            Itheta = 0.5*gtheta_pos*(cos(phase) + 1) : 1 (shared)

            # Trajectory variables to record spike positions
            x : 1 (linked)
            y : 1 (linked)

            # Theta phase
            phase : 1 (linked)
        """
        Gpos = br.NeuronGroup(N, model=positive_eqns, threshold='v > 30',
                order=2, reset='v = c; u = u + d', name='PositivePhaser',
                method='rk4')
        Gpos.v = v0
        Gpos.phase = br.linked_var(Gneg.phase)
        Gpos.x = br.linked_var(Gneg.x)
        Gpos.y = br.linked_var(Gneg.y)

        Cnp = br.Synapses(Gneg, Gpos, on_pre='gi += Wneg')
        Cnp.connect(j='i')

        burst_eqns = br.Equations("""
            dv/dt = (0.04*v**2 + 5*v + 140 + Ie - u)/tau + sigma/sqrt(dt)*xi : 1
            du/dt = a*(b*v - u) / tau : 1

            # Trajectory variables to record spike positions
            x : 1 (linked)
            y : 1 (linked)

            # Theta phase
            phase : 1 (linked)
        """, Ie=Ie_b, tau=tau_b)
        Gb = br.NeuronGroup(1, model=burst_eqns, threshold='v > 30', order=3,
                reset='v=c; u=u+d_b', name='SyncBurster', method='euler')
        Gb.v = v0_b
        Gb.phase = br.linked_var(Gneg.phase)
        Gb.x = br.linked_var(Gneg.x)
        Gb.y = br.linked_var(Gneg.y)

        # Phaser-burster synapse model
        gnb = self.c.vco_neg_fdbk if neg_fdbk is None else neg_fdbk
        gpb = self.c.vco_pos_fdbk if pos_fdbk is None else pos_fdbk
        Cnb = br.Synapses(Gneg, Gb, model='W : 1 (constant)',
                on_pre='v += gnb*W')
        Cpb = br.Synapses(Gpos, Gb, model='W : 1 (constant)',
                on_pre='v += gpb*W')
        Cnb.connect(i=np.arange(N), j=0)
        Cpb.connect(i=np.arange(N), j=0)

        Cnb.W = Wnb[:,0]  # negative-burster weights
        Cpb.W = Wpb[:,0]  # positive-burster weights

        VCOMon = br.SpikeMonitor(Gb, ['phase', 'x', 'y'],
                record=True, name='vco')
        TrajMon = br.StateMonitor(Trajectory, ['x','y'],
                record=[0], name='traj')

        @br.network_operation(dt=10*second, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        Net = br.Network(Gneg, Gpos, Cnp, Gb, Cnb, Cpb, VCOMon, TrajMon,
                Trajectory, SpatialLayer, progress, name='PhaserVCONetwork')
        Net.run(duration)

        fdbk = (gnb, gpb)
        simpath = self._sync_simulation_path(sigma, fdbk, tag=tag)
        self.save_simulation(simpath, Net,
                title='Spatial VCO-like Burst Synchronization by Phasers')

    def run_example_simulations(self, tag1='theta0', tag2='theta1', dur=1000.0):
        """Run example VCO simulations with noise."""
        for tag, theta in [(tag1, self.c.theta0), (tag2, self.c.theta1)]:
            self.create_vco_target_map(tag=tag, theta0=theta, plot=False)
            self.compute_weight_vector(tag=tag, plot_distro=False)
            self.spatial_phaser_burster_simulation(tag=tag, noise=True,
                    dur=dur)

    @step
    def plot_vco_simulation(self, tag=None, noise=False, neg_fdbk=None,
        pos_fdbk=None, mapres=128, levels=18, ibi_min=IBI_MIN,
        figsize=(6, 3.5)):
        """Plot figure of the VCO-like phasemap for the burster."""
        sigma = self.c.sigma if noise else 0.0
        neg_fdbk = self.c.vco_neg_fdbk if neg_fdbk is None else neg_fdbk
        pos_fdbk = self.c.vco_pos_fdbk if pos_fdbk is None else pos_fdbk
        fdbk = (neg_fdbk, pos_fdbk)

        simpath = self._sync_simulation_path(sigma, fdbk, tag=tag)
        S = self.read_simulation(simpath)
        sim.add_wrapped_phase_column(S.vco)
        S.vco['bursts'] = spk.burst_onset(S.vco.t, isi_thresh=ibi_min)

        # Load the adaptive phasemap
        mdata = motion.MotionData(self.c.traj_session_id)
        phasemap = adaptive.AdaptivePhasemap(mdata, scale='cm', res=mapres)

        # Compute the phasemap
        self.out('Computing VCO phasemap...')
        bx = S.vco.loc[S.vco.bursts, 'x']
        by = S.vco.loc[S.vco.bursts, 'y']
        bphase = S.vco.loc[S.vco.bursts, 'phase']
        P = phasemap(bx, by, bphase)

        figname = 'vco-sim-noise' if noise else 'vco-sim-nonoise'
        figtitle = 'Spatial Phaser-VCO Simulation: {:.1f} sigma'.format(sigma)
        if tag:
            figname += '-{}'.format(tag)
            figtitle += ' ({})'.format(tag)
        f = self.figure(figname, clear=True, figsize=figsize, title=figtitle)
        axp = f.add_subplot(121)
        axd = f.add_subplot(122)

        maps.plot(P, ax=axp, satnorm=True, levels=levels, mask_color='0.9')
        axp.set_axis_off()

        maps.phasedisc(ax=axd, res=256, levels=levels, labelsize='x-large',
                labelpad=0.07, inline=True, zoom=1)

        f.tight_layout()

    @step
    def plot_vco_example_simulations(self, tags=None, neg_fdbk=None,
        pos_fdbk=None, mapres=128, levels=18, ibi_min=IBI_MIN,
        figsize=(7.2, 3.6)):
        """Plot figure of example VCO simulations."""
        sigma = self.c.sigma
        tags = ['theta0', 'theta1'] if tags is None else tags
        neg_fdbk = self.c.vco_neg_fdbk if neg_fdbk is None else neg_fdbk
        pos_fdbk = self.c.vco_pos_fdbk if pos_fdbk is None else pos_fdbk
        fdbk = (neg_fdbk, pos_fdbk)

        # Load the adaptive phasemap and ratemap
        mdata = motion.MotionData(self.c.traj_session_id)
        phasemap = adaptive.AdaptivePhasemap(mdata, scale='cm', res=mapres)
        ratemap = adaptive.AdaptiveRatemap(mdata, scale='cm', res=mapres)

        # Compute the no-feedback simulation spatial maps
        self.out('Computing no-feedback phasemap...')
        B = self.read_simulation(
                self._sync_simulation_path(sigma, (0.0, 0.0), tag='nofdbk'))
        sim.add_wrapped_phase_column(B.vco)
        B.vco['bursts'] = spk.burst_onset(B.vco.t, isi_thresh=ibi_min)
        bx = B.vco.loc[B.vco.bursts, 'x']
        by = B.vco.loc[B.vco.bursts, 'y']
        bphase = B.vco.loc[B.vco.bursts, 'phase']
        Pb = phasemap(bx, by, bphase)

        # Plotting parameters
        mapkw = dict(levels=levels, mask_color='0.9')
        cbargs = dict(pad=0.005, shrink=0.97, aspect=15)
        tkw = dict(size='small', fontweight='light')

        figname = 'vco-simulations'
        figtitle = 'Example VCO Synchronization Simulations'
        f = self.figure(figname, clear=True, title=figtitle, figsize=figsize)
        gshape = (2, 4)

        def plot_target_weights(row, tag, title):
            vcopath = 'vco/phasemap_{}'.format(tag)
            inputpath = 'vco/inputmap_{}'.format(tag)
            Pvco = self.read_array(vcopath)
            Pinp = self.read_array(inputpath)
            self.out('Input max. coherence [{}] = {:.3f}', tag,
                    np.nanmax(Pinp[1]))

            axv = plt.subplot2grid(gshape, (row, 1))
            axi = plt.subplot2grid(gshape, (row, 2))
            maps.plot(Pvco, ax=axv, circular=True)
            maps.plot(Pinp, ax=axi, mask_color=mapkw['mask_color'],
                    satnorm=True)
            axv.set_axis_off()
            axi.set_axis_off()

            quicktitle(axv, title, **tkw)
            quicktitle(axi, 'Learned phaser input', **tkw)

        def plot_phasemap(ax, tag, title, pmap=None):
            if pmap is not None:
                P = pmap
            else:
                self.out('Computing {} phasemap...', tag)
                S = self.read_simulation(
                        self._sync_simulation_path(sigma, fdbk, tag=tag))
                sim.add_wrapped_phase_column(S.vco)
                S.vco['bursts'] = spk.burst_onset(S.vco.t, isi_thresh=ibi_min)
                bx = S.vco.loc[S.vco.bursts, 'x']
                by = S.vco.loc[S.vco.bursts, 'y']
                bphase = S.vco.loc[S.vco.bursts, 'phase']
                P = phasemap(bx, by, bphase)
            self.out('Max. coherence [{}] = {:.3f}', tag, np.nanmax(P[1]))
            maps.plot(P, ax=ax, satnorm=True, **mapkw)
            ax.set_axis_off()
            quicktitle(ax, title, **tkw)

        plot_phasemap(plt.subplot2grid(gshape, (0, 0)), 'nofdbk',
                'Target burster\nWithout phaser input', pmap=Pb)
        plot_target_weights(0, 'theta0', 'Phase code 1')
        plot_phasemap(plt.subplot2grid(gshape, (0, 3)), tags[0],
                'Target burster')
        plot_target_weights(1, 'theta1', 'Phase code 2')
        plot_phasemap(plt.subplot2grid(gshape, (1, 3)), tags[1],
                'Target burster')

        maps.phasedisc(ax=plt.subplot2grid(gshape, (1, 0)), res=384,
                levels=levels, labelsize='medium', labelpad=0.07, inline=False,
                zoom=0.7, orient='natural')

    def run_phase_movie_simulations(self, mapres=128, dur=600.0, nframes=64):
        """Run frame simulations for an animated movie of initial phase."""
        sigma = self.c.sigma
        fdbk = (self.c.vco_neg_fdbk, self.c.vco_pos_fdbk)
        self['phase_movie_nframes'] = nframes
        phase_list = np.linspace(2*np.pi, 0.0, nframes+1)[:-1]
        self.save_array('phasemovie/values', phase_list,
                title='Initial Phase Values for Frames')

        # Load the adaptive phasemap
        mdata = motion.MotionData(self.c.traj_session_id)
        phasemap = adaptive.AdaptivePhasemap(mdata, scale='cm', res=mapres)

        def run_frame(i, phase0):
            tag = 'phase%03d' % i

            # Saves VCO target phasemap at: /vco/phasemap_phaseXXX
            self.create_vco_target_map(tag=tag, Fbins=mapres, theta0=phase0,
                    plot=False)

            # Saves total weighted input phasemap at: /vco/inputmap_phaseXXX
            self.compute_weight_vector(tag=tag, plot_distro=False)
            self.compute_weighted_inputs(tag=tag)

            # Runs simulation based on this frame
            self.spatial_phaser_burster_simulation(tag=tag, noise=True, dur=dur)

            # Compute simulated phasemap and save at:
            # /phasemovie/burstmap_phaseXXX
            S = self.read_simulation(
                    self._sync_simulation_path(sigma, fdbk, tag=tag))
            sim.add_wrapped_phase_column(S.vco)
            S.vco['bursts'] = spk.burst_onset(S.vco.t, isi_thresh=IBI_MIN)
            bx = S.vco.loc[S.vco.bursts, 'x']
            by = S.vco.loc[S.vco.bursts, 'y']
            bphase = S.vco.loc[S.vco.bursts, 'phase']
            Pout = phasemap(bx, by, bphase)
            self.save_array('phasemovie/burstmap_{}'.format(tag), Pout,
                    title='Phase Movie Output Phasemap: Frame %03d' % i)

        for i, phase0 in enumerate(phase_list):
            run_frame(i, phase0)

        self.close_datafile()

    def run_dir_movie_simulations(self, mapres=128, dur=600.0, ex=0,
        nframes=64):
        """Run frame simulations for an animated movie of initial phase."""
        sigma = self.c.sigma
        fdbk = (self.c.vco_neg_fdbk, self.c.vco_pos_fdbk)
        self['dir_movie_nframes'] = nframes
        phi_list = np.linspace(2*np.pi, 0.0, nframes+1)[:-1]
        self.save_array('dirmovie/values', phi_list,
                title='Preferred Direction Values for Frames')

        # Load the adaptive phasemap
        mdata = motion.MotionData(self.c.traj_session_id)
        phasemap = adaptive.AdaptivePhasemap(mdata, scale='cm', res=mapres)

        def run_frame(i, phi_pref):
            if ex == 0:
                tag = 'dir%03d' % i
            elif ex == 1:
                tag = 'altdir%03d' % i

            # Saves VCO target phasemap at: /vco/phasemap_dirXXX
            self.create_vco_target_map(tag=tag, Fbins=mapres,
                    phi_pref=phi_pref, plot=False)

            # Saves total weighted input phasemap at: /vco/inputmap_dirXXX
            self.compute_weight_vector(tag=tag, plot_distro=False)
            self.compute_weighted_inputs(tag=tag)

            # Runs simulation based on this frame
            self.spatial_phaser_burster_simulation(tag=tag, noise=True, dur=dur)

            # Compute simulated phasemap and save at:
            # /dirmovie/burstmap_dirXXX
            S = self.read_simulation(
                    self._sync_simulation_path(sigma, fdbk, tag=tag))
            sim.add_wrapped_phase_column(S.vco)
            S.vco['bursts'] = spk.burst_onset(S.vco.t, isi_thresh=IBI_MIN)
            bx = S.vco.loc[S.vco.bursts, 'x']
            by = S.vco.loc[S.vco.bursts, 'y']
            bphase = S.vco.loc[S.vco.bursts, 'phase']
            Pout = phasemap(bx, by, bphase)
            self.save_array('dirmovie/burstmap_{}'.format(tag), Pout,
                    title='Directional Movie Output Phasemap: Frame %03d' % i)

        for i, phi_pref in enumerate(phi_list):
            run_frame(i, phi_pref)

        self.close_datafile()

    def run_scale_movie_simulations(self, mapres=128, dur=600.0, nframes=64):
        """Run frame simulations for an animated movie of spatial scale."""
        sigma = self.c.sigma
        fdbk = (self.c.vco_neg_fdbk, self.c.vco_pos_fdbk)
        self['scale_movie_nframes'] = nframes
        L_list = np.linspace(35, 99, nframes+1)[:-1]
        self.save_array('scalemovie/values', L_list,
                title='Spatial Scale Values for Frames')

        # Load the adaptive phasemap
        mdata = motion.MotionData(self.c.traj_session_id)
        phasemap = adaptive.AdaptivePhasemap(mdata, scale='cm', res=mapres)

        def run_frame(i, L):
            tag = 'scale%03d' % i

            # Saves VCO target phasemap at: /vco/phasemap_dirXXX
            self.create_vco_target_map(tag=tag, Fbins=mapres, L=L, plot=False)

            # Saves total weighted input phasemap at: /vco/inputmap_dirXXX
            self.compute_weight_vector(tag=tag, plot_distro=False)
            self.compute_weighted_inputs(tag=tag)

            # Runs simulation based on this frame
            self.spatial_phaser_burster_simulation(tag=tag, noise=True, dur=dur)

            # Compute simulated phasemap and save at:
            # /phasemovie/burstmap_dirXXX
            S = self.read_simulation(
                    self._sync_simulation_path(sigma, fdbk, tag=tag))
            sim.add_wrapped_phase_column(S.vco)
            S.vco['bursts'] = spk.burst_onset(S.vco.t, isi_thresh=IBI_MIN)
            bx = S.vco.loc[S.vco.bursts, 'x']
            by = S.vco.loc[S.vco.bursts, 'y']
            bphase = S.vco.loc[S.vco.bursts, 'phase']
            Pout = phasemap(bx, by, bphase)
            self.save_array('scalemovie/burstmap_{}'.format(tag), Pout,
                    title='Directional Movie Output Phasemap: Frame %03d' % i)

        for i, L in enumerate(L_list):
            run_frame(i, L)

        self.close_datafile()

    @step
    def generate_movie(self, tagstem, varname, title, ex=0, levels=18,
        figsize=(8,6)):
        """Generate the movie figure and render the simulation movies.

        Arguments:
        tagstem -- frame tag stem used like 'stem%03d' to create data tags
        varname -- name used in figure label to annotate frame value
        title -- title displayed at top of movie figure
        levels -- phasemap contour levels
        figsize -- movie figure size
        """
        nframes = self['{}_movie_nframes'.format(tagstem)]
        values_list = self.read_array('{}movie/values'.format(tagstem))

        if ex == 0:
            moviename = '{}-movie.mp4'.format(tagstem)
        elif ex == 1:
            moviename = 'alt{}-movie.mp4'.format(tagstem)
        movie = FigureMovie(self.path(moviename),
                fps=10.0, res=1080, duration=(nframes-0.7)/10.0,
                width=figsize[0], height=figsize[1])

        # Initial setup of static elements of the figure
        f = movie.get_frame()
        f.suptitle(title)
        axv = f.add_subplot(221)
        axi = f.add_subplot(222)
        axb = f.add_subplot(223)
        axp = f.add_subplot(224)
        f.tight_layout(rect=(0,0,1,0.94))
        maps.phasedisc(ax=axp, res=384, levels=levels, labelsize='large',
                labelpad=0.08, orient='natural', inline=False, zoom=0.7)

        def update_frame(t):
            i = round(10*t)
            if ex == 0:
                tag = '%s%03d' % (tagstem, i)
            elif ex == 1:
                tag = 'alt%s%03d' % (tagstem, i)

            [ax.cla() for ax in [axv, axi, axb]]

            Pvco = self.read_array('vco/phasemap_{}'.format(tag))
            Pinp = self.read_array('vco/inputmap_{}'.format(tag))
            Pout = self.read_array('{}movie/burstmap_{}'.format(tagstem, tag))

            maps.plot(Pvco, ax=axv, circular=True, mask_color='w')
            quicktitle(axv, 'Desired phase code')
            axv.set_axis_off()

            maps.plot(Pinp, ax=axi, satnorm=True, mask_color='w')
            quicktitle(axi, 'Learned phaser input')
            axi.set_axis_off()

            maps.plot(Pout, ax=axb, satnorm=True, levels=levels, mask_color='w')
            quicktitle(axb, 'Target burster')
            axb.set_axis_off()

        def label_frame(t):
            i = round(10*t)
            value = values_list[i]
            return '{} = {:.3f}'.format(varname, value)

        movie.set_update_function(update_frame)
        movie.set_label_function(label_frame)
        movie.set_label_location('bottom center')
        movie.render()

    def save_vco_array_image(self, P, tag=None):
        """Save a flattened image of VCO phasemaps going into the decoder."""
        N = P.shape[0]
        s = P.shape[1]
        r, c  = tiling_dims(N)
        res = s * max(r, c)
        Pimg = np.zeros((res, res))

        for i in reversed(range(N)):
            x0 = s * (i % c)
            x1 = x0 + s
            y0 = s * (r - 1 - int(i / r))
            y1 = y0 + s
            Pimg[x0:x1,y0:y1] = P[i]
        Pimg[0,0] = np.pi/2

        fn = 'vco-network'
        if tag is not None:
            fn += '-%s' % tag
        fn += '.png'
        savepath = self.path(fn)
        cmap = huslmap()
        array_to_image(Pimg.T, savepath, cmap=cmap, cmin=-np.pi, cmax=np.pi,
                mask=np.isnan(Pimg), mask_color='w')
        self.out('Saved VCO array: {}', savepath)

    def create_unmasked_dir_vco_maps(self, mapres=256, ex=0):
        """Create unmasked VCO target maps from the directional movies."""
        phi_list = self.read_array('dirmovie/values')

        def run_frame(i, phi_pref):
            if ex == 0:
                tag = 'dir%03d' % i
            elif ex == 1:
                tag = 'altdir%03d' % i

            # Saves VCO decoder phasemap at: /vco/decodemap_dirXXX
            self.create_vco_target_map(tag=tag, mask=False, Fbins=mapres,
                    phi_pref=phi_pref, theta0=ex*np.pi, plot=False)

        for i, phi_pref in enumerate(phi_list):
            run_frame(i, phi_pref)

        self.close_datafile()

    def create_unmasked_phase_vco_maps(self, mapres=256):
        """Create unmasked VCO target maps from the phase-offset movie."""
        theta_list = self.read_array('phasemovie/values')

        def run_frame(i, theta0):
            tag = 'phase%03d' % i

            # Saves VCO decoder phasemap at: /vco/decodemap_dirXXX
            self.create_vco_target_map(tag=tag, mask=False, Fbins=mapres,
                    phi_pref=np.pi/4, theta0=theta0, plot=False)

        for i, theta0 in enumerate(theta_list):
            run_frame(i, theta0)

        self.close_datafile()

    def load_decoding_data(self):
        """Load 3 populations of VCO phase to set up Bayesian decoding."""
        Nvco = min(self.c.phase_movie_nframes, self.c.dir_movie_nframes)

        # Save unmasked phase-code maps of all VCOs for decoding
        #
        Pphase = np.array([self.read_array('/decode/maps/phase%03d' % i)
                            for i in range(Nvco)])
        self.save_vco_array_image(Pphase, tag='phase')
        self.save_array('decode/Pphase', Pphase,
                title='Phase Decode Maps for the Phase VCO Network')

        Pdir1 = np.array([self.read_array('/decode/maps/dir%03d' % i)
                            for i in range(Nvco)])
        self.save_vco_array_image(Pdir1, tag='dir1')
        self.save_array('decode/Pdir1', Pdir1,
                title='Phase Decode Maps for the Dir-1 VCO Network')

        Pdir2 = np.array([self.read_array('/decode/maps/altdir%03d' % i)
                            for i in range(Nvco)])
        self.save_vco_array_image(Pdir2, tag='dir2')
        self.save_array('decode/Pdir2', Pdir2,
                title='Phase Decode Maps for the Dir-2 VCO Network')

        Pvco = np.concatenate((Pphase, Pdir1, Pdir2), axis=0)
        self.save_array('decode/Pvco', Pvco,
                title='Phase Decode Maps for 3 VCO Networks')

        # Save trajectory data
        sigma = self.c.sigma
        fdbk = (self.c.vco_neg_fdbk, self.c.vco_pos_fdbk)
        traj = self.read_simulation(
                self._sync_simulation_path(sigma, fdbk, tag='phase000')).traj
        self.save_dataframe('decode/traj', traj)

        # Load individual VCO simulations for burst-phase data
        data = dict(network=[], label=[], neuron=[], t=[], phase=[])
        k = 0
        for tag in ('phase', 'dir', 'altdir'):
            for i in range(Nvco):
                simtag = '%s%03d' % (tag, i)
                simpath = self._sync_simulation_path(sigma, fdbk, tag=simtag)
                sim = self.read_simulation(simpath).vco
                bursts = sim
                sim['bursts'] = spk.burst_onset(sim.t, isi_thresh=IBI_MIN)
                bursts = sim.loc[sim.bursts]
                nbursts = len(bursts)

                if tag == 'phase':
                    name = tag
                if tag == 'dir':
                    name = 'dir1'
                elif tag == 'altdir':
                    name = 'dir2'

                data['network'].extend([name]*nbursts)
                data['label'].extend([i]*nbursts)
                data['neuron'].extend([k]*nbursts)
                data['t'].extend(list(bursts.t))
                data['phase'].extend(list(bursts.phase))

                k += 1

        self.save_dataframe('decode/bursts', pd.DataFrame(data=data,
                columns=['network', 'label', 'neuron', 't', 'phase']),
                title='Burst-Phase Data for All 3 VCO Networks')

    def phase_decoding_analysis(self, interval=[0,1], choose=None,
        network=None, win_theta=DEFAULT_WIN_THETA, winc_frac=DEFAULT_WINC_FRAC,
        prior=DEFAULT_PRIOR, return_posterior=False):
        """Perform Bayesian decoding of phaser spiking into position."""
        traj = self.read_dataframe('decode/traj')
        bursts = self.read_dataframe('decode/bursts')
        dur = traj.t.max()
        if interval[0] < 0 or interval[1] > dur:
            raise ValueError('interval out of bounds: {}'.format(interval))

        # Filter data to a specific network
        if network is not None:
            if network in ('phase', 'dir1', 'dir2'):
                bursts = bursts.query('network == "%s"' % network)
            else:
                raise ValueError('bad network name: {}'.format(network))
        neurons = list(np.sort(bursts.neuron.unique()))

        # Randomly subsample the network
        if choose is not None:
            if choose <= 0 or choose > len(neurons):
                raise ValueError('bad choose value ({})'.format(choose))
            neurons = list(np.unique(np.array(neurons)[
                np.random.randint(len(neurons), size=choose)]))
            bursts = bursts.query('neuron in @neurons')

        Nvco = len(neurons)
        Pvco = self.read_array('decode/Pvco')[neurons]
        dec = decoder.BayesPhaseDecoder(Pvco)
        self.out('Created decoder for {} VCOs.', Nvco)

        # Set up sliding window accumulation
        window = win_theta / self.c.ftheta
        winc = winc_frac * window
        phase = pd.Series(data=np.zeros(Nvco), index=neurons)
        Px = np.zeros((dec.ngrid,)*2)
        t0, t1 = interval[0], interval[0] + window
        path = []
        actual = []
        t = []

        def phase_avg(grp):
            phi = grp.values
            n = len(phi)
            if n == 0:
                return -99
            if n == 1:
                return float(phi)
            return mean_angle(phi)

        while t1 <= interval[1]:
            avgxy = traj.query('(t >= @t0) & (t < @t1)')[['x', 'y']].mean()
            actual.append(list(avgxy))
            t.append((t0 + t1) / 2)

            diff = bursts.query('(t >= @t0) & (t < @t1)'
                       ).groupby('neuron', sort=False).phase.apply(phase_avg)
            phase.update(diff[diff != -99])

            Px_win = dec.decode(phase, window=window, continuity=prior)
            if return_posterior:
                Px += Px_win

            path.append(dec.argmax)

            t0 += winc
            t1 += winc

            self.box(color='green', filled=False)
        self.newline()

        actx, acty = np.array(actual).T
        decx, decy = np.array(path).T
        t = np.array(t)

        errors = np.sqrt((actx - decx)**2 + (acty - decy)**2)
        MSE = np.mean(errors**2)

        packet = dict(t=t, actx=actx, acty=acty, decx=decx, decy=decy,
                        errors=errors, MSE=MSE)

        if return_posterior:
            packet['Px'] = Px

        return Reify(packet)

    def plot_decoder_packet(self, p, figsize=(9.88435374, 4.14285714)):
        """Plot a figure for a given decoder output packet."""
        plt.ioff()

        fig = self.figure('decoder-output', clear=True, figsize=figsize,
                title='Decoder Output')
        axp = fig.add_subplot(121)
        axe = fig.add_subplot(122)

        actfmt = dict(ls='-', lw=1.2, zorder=1)
        decfmt = dict(ls=':', marker='.', lw=1.2, ms=4, mew=0.2, zorder=2)

        def plot_paths(ax):
            if hasattr(p, 'Px'):
                maps.plot(p.Px, ax=ax, cmap='afmhot', cmin=0.0)
            ax.plot(p.actx, p.acty, c='b', label='Actual', **actfmt)
            ax.plot(p.decx, p.decy, c='m', label='Decoded', **decfmt)
            ax.set(xlabel='X (cm)', ylabel='Y (cm)')
            quicktitle(ax, 'P(x|phase)')
            despine(ax)
            # ax.legend()
        plot_paths(axp)

        def plot_errors(ax):
            ax.plot(p.t, p.errors, c='k', ls='-', lw=1.2)
            ax.set(xlabel='Time (s)', ylabel='Error')
            quicktitle(ax, 'Decoding Error: MSE = {:.1f}'.format(p.MSE))
            despine(ax)
        plot_errors(axe)

        plt.ion()
        plt.show()

    @step
    def plot_network_parameters(self, figsize=(3.7,3.7)):
        """Plot preferred direction and phase offset for the 3 networks."""
        pmin = -np.pi/8
        pmax = 17*np.pi/8
        phase_values = self.read_array('phasemovie/values')
        phase_dir = np.zeros_like(phase_values) + np.pi/4
        dir_values = self.read_array('dirmovie/values')
        dir_phase0 = np.zeros_like(phase_values) + 0.0
        dir_phase1 = np.zeros_like(dir_phase0) + np.pi

        f = self.figure('VCO-params', clear=True, figsize=figsize,
                title='VCO Netowrk Parameters')
        ax = f.add_subplot(111)

        fmt = dict(marker='o', c='k', lw=1, ls='-', mew=1, mfc='none', ms=12,
                alpha=0.7)

        ax.plot(phase_dir, phase_values, mec=PHASECOL, label='Phase network', **fmt)
        ax.plot(dir_values, dir_phase0, mec=DIR1COL, label='Dir1 network', **fmt)
        ax.plot(dir_values, dir_phase1, mec=DIR2COL, label='Dir2 network', **fmt)

        ax.legend(frameon=False)
        ax.set_xlim(pmin, pmax)
        ax.set_ylim(pmin, pmax)
        ax.set_xticks([0, np.pi, 2*np.pi])
        ax.set_yticks([0, np.pi, 2*np.pi])
        ax.xaxis.set_major_formatter(RadianFormatter(inline=True))
        ax.yaxis.set_major_formatter(RadianFormatter(inline=False))
        ax.set_xlabel('Preferred direction')
        ax.set_ylabel('Reference phase offset')
        despine(ax)

    @step
    def compute_decoding_traces(self, interval=[108,114],
        win_theta=DEFAULT_WIN_THETA, winc_frac=DEFAULT_WINC_FRAC,
        prior=DEFAULT_PRIOR):
        """Decode a sample interval for each network condition for plotting."""
        decparms = dict(win_theta=win_theta, winc_frac=winc_frac, prior=prior,
                return_posterior=True)

        decode_phase = self.phase_decoding_analysis(interval=interval,
                network='phase', **decparms)
        decode_dir1 = self.phase_decoding_analysis(interval=interval,
                network='dir1', **decparms)
        decode_dir2 = self.phase_decoding_analysis(interval=interval,
                network='dir2', **decparms)
        decode_all = self.phase_decoding_analysis(interval=interval,
                network=None, **decparms)

        return (interval, decode_phase, decode_dir1, decode_dir2, decode_all)

    @step
    def plot_decoding_traces(self, traces, figsize=(8.5, 2.5)):
        """Plot figure of example trace decoding across network conditions."""
        diam = motion.CIRCLE_DIAMETER
        radius = diam / 2
        xgrid = self.read_array('vco/xgrid')
        ygrid = self.read_array('vco/ygrid')

        interval, decode_phase, decode_dir1, decode_dir2, decode_all = traces

        f = self.figure('decoding-traces', clear=True, figsize=figsize,
                title='Example Decoding for Interval: {}'.format(interval))
        axa = f.add_subplot(141)
        axb = f.add_subplot(142)
        axc = f.add_subplot(143)
        axd = f.add_subplot(144)

        actfmt = dict(ls='-', lw=1.2, solid_capstyle='round', alpha=1.0,
                zorder=1)
        decfmt = dict(ls='-', lw=0.6, marker='o', mfc='m', mew=0.0, ms=2,
                alpha=1.0, zorder=2)

        def plot_trace(ax, dec, label):
            rad2 = np.power(xgrid - radius, 2) + np.power(ygrid - radius, 2)
            P = dec.Px.copy()
            P[rad2 > radius**2] = np.nan  # mask pixels outside arena
            maps.plot(P, ax=ax, cmap='afmhot', cmin=0.0)
            ax.plot(dec.actx, dec.acty, mec='b', **actfmt)
            ax.plot(dec.decx, dec.decy, c='m', **decfmt)
            quicktitle(ax, label)
            ax.set_axis_off()

        plot_trace(axa, decode_phase, 'Phase')
        plot_trace(axb, decode_dir1, 'Dir1')
        plot_trace(axc, decode_dir2, 'Dir2')
        plot_trace(axd, decode_all, 'Phase + Dir1 + Dir2')

    def _perf_savepath(self, name, size, bootstrapped):
        suffix = '_bs' if bootstrapped else ''
        return 'decode/performance/{}_{:03d}{}'.format(name, size, suffix)

    @step
    def collect_decoding_data(self, bootstraps=10, interval=[108,114],
        win_theta=DEFAULT_WIN_THETA, winc_frac=DEFAULT_WINC_FRAC,
        prior=DEFAULT_PRIOR):
        """Run all network conditions with bootstraps for performance."""
        Nvco = min(self.c.phase_movie_nframes, self.c.dir_movie_nframes)
        decparms = dict(interval=interval, win_theta=win_theta,
                winc_frac=winc_frac, prior=prior, return_posterior=False)

        # Run a test decode to find out number of windows
        df = self.get_datafile(False)  # open the data for writing
        test = self.phase_decoding_analysis(network='dir1', choose=5,
                            **decparms)
        self.save_array('decode/performance/t', test.t, title='Time Index')
        nwins = test.t.size

        # Set up network sizes to sample
        net_max = Nvco
        net_sizes = (4, 8, 16, 32, 48, 64)
        all_max = 3*Nvco
        all_sizes = (4, 8, 16, 32, 48, 64, 96, 128, 192)

        for which in ('phase', 'dir1', 'dir2', None):
            if which is None:
                name = 'all'
                nmax = all_max
                sizes = all_sizes
            else:
                name = which
                nmax = net_max
                sizes = net_sizes

            for size in sizes:
                if size == nmax:
                    decode = self.phase_decoding_analysis(network=which,
                                    choose=None, **decparms)
                    self.save_array(self._perf_savepath(name, size, False),
                            decode.errors, title='Decoding Error for Network: '
                                '{}'.format(name.title()))

                errors = np.empty((bootstraps, nwins))
                for i in range(bootstraps):
                    decode = self.phase_decoding_analysis(network=which,
                                    choose=size, **decparms)
                    errors[i] = decode.errors
                self.save_array(self._perf_savepath(name, size, True), errors,
                        title='Bootstrapped Decoding Errors: {}, N={}'.format(
                            name.title(), size))

    @step
    def plot_size_performance(self, CIalpha=None, maxshift=0.0,
        figsize=(6, 3.5)):
        """Plot a figure comparing decoding performance across network size."""
        Nvco = min(self.c.phase_movie_nframes, self.c.dir_movie_nframes)
        net_max = Nvco
        net_sizes = (4, 8, 16, 32, 48, 64)
        nnet = len(net_sizes)
        all_max = 3*Nvco
        all_sizes = (4, 8, 16, 32, 48, 64, 96, 128, 192)
        nall = len(all_sizes)

        fig = self.figure('decoding-by-size', clear=True, figsize=figsize,
                title='Decoding Performance Across Network Size')
        ax = fig.add_subplot(111)

        xnet = np.arange(nnet)
        xall = np.arange(nall)

        efmt = dict(ls='-', elinewidth=1.1, capsize=0, mew=0, ms=5)

        namecols = [('phase', PHASECOL), ('dir1', DIR1COL), ('dir2', DIR2COL),
                    ('all', ALLCOL)]

        for i, (name, col) in enumerate(namecols):
            n = nall if name == 'all' else nnet
            bsCI = np.empty((2,n))
            errmean = np.empty(n)
            sizes = all_sizes if name == 'all' else net_sizes
            for j, size in enumerate(sizes):
                errors = self.read_array(self._perf_savepath(name, size, True))
                nbs, npts = errors.shape
                MSE = np.mean(np.power(errors, 2), axis=1)
                mu = MSE.mean()
                if CIalpha is None:
                    bsCI[:,j] = MSE.std()/np.sqrt(nbs)
                else:
                    CI = empirical_CI(MSE, alpha=CIalpha)
                    bsCI[:,j] = mu - CI.lower, CI.upper - mu
                errmean[j] = mu

            fullerr = self.read_array(self._perf_savepath(name, size, False))
            fullMSE = np.mean(np.power(fullerr, 2))

            x = xall if name == 'all' else xnet
            shift = maxshift * ((i/3) - 0.5)
            mkr = 'o^vs'[{'phase':0,'dir1':1,'dir2':2,'all':3}[name]]
            ax.errorbar(x + shift, errmean, yerr=bsCI, c=col, marker=mkr,
                    mfc=col, label=name.title(), zorder=0, **efmt)
            ax.plot(x[n-1] + shift, fullMSE, c='k', marker='+', ms=10, mew=1.5, zorder=5+i)
            ax.plot(x[n-1] + shift, fullMSE, c=col, marker='+', ms=9, mew=1.0, zorder=5+i)
            self.out('{} network full-network MSE = {:.4f}', name, fullMSE)

        ax.set_xlim(-0.5, nall-0.5)
        ax.set_xticks(xall)
        ax.set_xticklabels(all_sizes)
        ax.set_xlabel('Network size')
        ax.set_ylim(bottom=0.0)
        ax.set_ylabel('MSE')
        ax.legend(loc=1, frameon=False)
        despine(ax)

    @step
    def plot_correction_timescale(self, maxlag=6.0, corr_CIalpha=None,
        summary_CIalpha=0.05, figsize=(7, 4.5)):
        """Plot time-lag curves of bootstrapped error correction."""
        Nvco = min(self.c.phase_movie_nframes, self.c.dir_movie_nframes)
        Nall = 3*Nvco

        t = self.read_array('decode/performance/t')
        errors = {name: self.read_array(self._perf_savepath(name, sz, True))
                    for name, sz in [('phase', Nvco), ('dir1', Nvco),
                        ('dir2', Nvco), ('all', Nall)]}
        nobs = {name: self.read_array(self._perf_savepath(name, sz, False))
                    for name, sz in [('phase', Nvco), ('dir1', Nvco),
                        ('dir2', Nvco), ('all', Nall)]}


        fig = self.figure('correction-time', clear=True, figsize=figsize,
                title='Time-Scale of Burst-Synchronized Error Correction')
        axt = plt.subplot2grid((1,4), (0,0), colspan=3)
        axs = fig.add_subplot(144)

        efmt = dict(ls='-', elinewidth=1.1, capsize=0, mew=0, ms=5)

        def plot_time_data(ax, ax2, err, nobs, order, name, col):
            SE = np.power(err, 2)
            nbs, npts = SE.shape
            C = np.empty_like(SE)
            for i in range(nbs):
                C[i] = sig.correlate(SE[i], SE[i], mode='same')
            peakix = np.argmax(C.sum(axis=0))
            tpeak = t[peakix]
            lagix = (t >= tpeak - maxlag) & (t <= tpeak + maxlag)
            tlag = t[lagix] - tpeak
            Clag = C[:,lagix]
            Cnorm = (lambda m: (Clag - m.min()) / m.ptp())(Clag.mean(axis=0))
            Cmean = Cnorm.mean(axis=0)

            bsCI = np.empty((2,Cnorm.shape[1]))
            for j in range(Cnorm.shape[1]):
                if corr_CIalpha is None:
                    bsCI[:,j] = Cnorm[:,j].std()/np.sqrt(nbs)
                else:
                    CI = empirical_CI(Cnorm[:,j], alpha=corr_CIalpha)
                    bsCI[:,j] = Cmean[j] - CI.lower, CI.upper - Cmean[j]

            mkr = 'o^vs'[{'phase':0,'dir1':1,'dir2':2,'all':3}[name]]
            ax.errorbar(tlag, Cmean, yerr=bsCI, c=col, marker=mkr,
                    mfc=col, label=name.title(), **efmt)

            # Compute HWHM as summary statistic of time scale
            Cbs = (Clag - Clag.min(axis=1).reshape(-1,1)) / \
                        Clag.ptp(axis=1).reshape(-1,1)
            thalfix = tlag >= 0
            thalf = tlag[thalfix]
            Clow = (Cbs[:, thalfix] < 0.5).astype('d')
            tlow = Clow * thalf
            tlow[tlow == 0.0] = np.inf  # exclude lags before half-cross
            hwhm = tlow.min(axis=1)
            hwhm_mu = hwhm.mean()
            hwhm_CI = empirical_CI(hwhm, alpha=summary_CIalpha)
            hwhm_err = hwhm_mu - hwhm_CI.lower, hwhm_CI.upper - hwhm_mu
            hwhm_err = np.reshape(hwhm_err, (-1,1))
            ax2.errorbar(order, hwhm.mean(), yerr=hwhm_err, c=col, marker=mkr,
                    mfc=col, **efmt)

            # Compute HWHM of full (non-bootstrapped) decoding
            SEnobs = np.power(nobs, 2)
            Cnobs = sig.correlate(SEnobs, SEnobs, mode='same')[lagix]
            Clownobs = (Cnobs[thalfix] < 0.5).astype('d')
            tlownobs = Clownobs * thalf
            tlownobs[tlownobs == 0.0] = np.inf
            hwhmnobs = tlow.min()
            ax2.plot(order, hwhmnobs, c='k', marker='+', ms=10, mew=1.5, zorder=5)
            ax2.plot(order, hwhmnobs, c=col, marker='+', ms=9, mew=1.0, zorder=10)
            self.out('HWHM (no-bootstrap) for {} network = {:.3f}s'.format(name,
                hwhmnobs))

        plot_time_data(axt, axs, errors['phase'], nobs['phase'], 0, 'phase', PHASECOL)
        plot_time_data(axt, axs, errors['dir1'], nobs['dir1'], 1, 'dir1', DIR1COL)
        plot_time_data(axt, axs, errors['dir2'], nobs['dir2'], 2, 'dir2', DIR2COL)
        plot_time_data(axt, axs, errors['all'], nobs['all'], 3, 'all', ALLCOL)

        def finish_time_curves(ax, tpad=0.1):
            ax.axvline(x=0, c='k', ls=':', zorder=-10)
            ax.set_xlim(-maxlag-tpad, maxlag+tpad)
            ax.set_ylim(bottom=0.0)
            ax.set_xlabel('Time lag (s)')
            ax.set_ylabel('Normalized MSE')
            ax.legend(loc=1, frameon=False)
            despine(ax)
        finish_time_curves(axt)

        def finish_summary_plot(ax):
            ax.set_xlim(-0.5, 3.5)
            ax.set_xticks([0,1,2,3])
            ax.set_xticklabels(['Phase', 'Dir1', 'Dir2', 'All'])
            ax.set_ylim(bottom=0.0)
            ax.set_xlabel('Decoding Network')
            ax.set_ylabel('Error-correction HWHM (s)')
            despine(ax)
        finish_summary_plot(axs)
