"""
Model simulation of intrinsic bursting theta cell learning from phasers.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import ColorConverter

from brian2.units import *
import brian2 as br

import roto.axis as ra
from roto.radians import RadianFormatter
from roto.images import tiling_dims
from roto.misc import merge_two_dicts

from .phaser_model import PhaserSimulation
from .phaser_cells import NEGCOLOR, POSCOLOR
from ..context import SPCSimulation, step
from ..lib import sim
from ..tools import spikes as spk, binned, maps
from .. import store


DEFAULT_PHASER_DIR = '/Users/joe/projects/spc/results/' \
                      '2017-05-22+phaser-simulation+beta'

DEFAULT_IE = 12.65  # tuned external input to match theta frequency
DEFAULT_TAU = 3.0   # ms, membrane time-constant for intrinsic burster
DEFAULT_EMAX = 0.2  # percent winners-take-all for phaser-burster synapses

# Default noise and feedback strength for simulations
DEFAULT_SIGMA = 0.3
DEFAULT_NEG_FDBK_GAIN = 1.0
DEFAULT_POS_FDBK_GAIN = 2.0

IBI_MIN = 0.025     # seconds, inter-burst interval threshold


class BurstSyncSimulation(SPCSimulation):

    """
    Intrinsic bursting theta cell simulation with phaser synchronization.
    """

    def cleanup(self):
        store.close()

    @step
    def setup(self, Ie=DEFAULT_IE, tau=DEFAULT_TAU,
        Emax=DEFAULT_EMAX, phaser_dir=DEFAULT_PHASER_DIR):
        self['Ie'] = Ie
        self['tau'] = tau
        self['Emax'] = Emax
        self['phaser_model_dir'] = phaser_dir

        # Load phaser simulation parameters
        psim = PhaserSimulation.load(self.c.phaser_model_dir)
        self['ftheta'] = psim.c.ftheta
        self['tau_phaser'] = psim.c.tau
        self['ge_phaser'] = psim.c.ge
        self['gtheta_phaser'] = psim.c.gtheta
        self['spatial_model_dir'] = psim.c.spatial_model_dir
        self['traj_session_id'] = psim.c.traj_session_id

    def intrinsic_burster_simulation(self, duration=30*second, dvsync=20.0,
        sigma=0.0, ftheta=None, tau=None, name='burster_demo'):
        """Test simulation for tuning intrinsic burster model parameters."""
        br.defaultclock.dt = 1*ms

        a = 0.02
        b = 0.2
        c = -50.0
        d = 5.0
        f_theta = (ftheta or self.c.ftheta)*Hz
        tau = self.c.tau*ms
        v0 = -80.0
        Ie = self.c.Ie

        eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 + Ie - u)/tau + sigma/sqrt(dt)*xi : 1
            du/dt = a*(b*v - u) / tau : 1
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        G = br.NeuronGroup(1, model=eqns, threshold='v > 30',
                reset='v=c; u=u+d', name='ThetaBurster', method='euler')
        G.v = v0

        SpkMon = br.SpikeMonitor(G, 'phase', record=[0], name='spikes')
        Mon = br.StateMonitor(G, 'v', record=[0], name='mon')

        @br.network_operation(dt=2*second, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        @br.network_operation(dt=2/f_theta, when='start')
        def sync(t):
            G.v += dvsync  # approximating input synchronous burst

        Net = br.Network(G, SpkMon, Mon, progress, sync, name='BursterTest')
        Net.run(duration)

        self.save_simulation(name, Net,
                title='Intrinsic Bursting Theta Cell Test')

    def analyze_test_simulation(self, ibi_min=IBI_MIN, name=None):
        """Compute burst statistics and tuning for test burster simulation."""
        name = name or 'burster_demo'
        S = self.read_simulation(name)
        sim.add_wrapped_phase_column(S.spikes)
        S.spikes['bursts'] = spk.burst_onset(S.spikes.t, isi_thresh=ibi_min)
        bt = S.spikes.t[S.spikes.bursts & (S.spikes.t>10)]
        db = np.diff(bt)
        bfreq = 1 / db.mean()

        self.out('N bursts = {}', len(bt))
        self.out('ISI bursts = {:.4f}, {:.5f} s.d.', db.mean(), db.std())
        self.out('Freq = {:.3f} Hz', bfreq)

        f = self.figure('burst-test', clear=True, figsize=(17, 9.5),
                title='Burst Test Simulation: f = {:.3f} Hz'.format(bfreq))
        ax = f.add_subplot(111)

        ax.plot(S.mon.t, S.mon.v)
        ymin, ymax = ax.get_ylim()
        ax.vlines(bt, ymin, ymax, linestyles=':')

        f = self.figure('burst-test-pdist', clear=True, figsize=(8,9),
                title='Burst Test Simulation: Phase Distribution')
        ax = f.add_subplot(111)
        ax.hist(S.spikes.phase0[S.spikes.bursts], 128)
        ax.set_xlim(-np.pi, np.pi)
        ax.set_xlabel('theta phase')
        ax.set_ylabel('count')

    @step
    def demo_burst_synchronization(self, dur=30.0, ntrials=30, dvsync=15.0,
        sigma=DEFAULT_SIGMA):
        """Run burster synchronization simulations for intro figure."""
        dur = dur*second

        self.intrinsic_burster_simulation(duration=dur, dvsync=0, sigma=0,
                name='intro/nonoise')
        self.intrinsic_burster_simulation(duration=dur, dvsync=dvsync, sigma=0,
                name='intro/nonoise_sync')
        self.intrinsic_burster_simulation(duration=dur, dvsync=dvsync,
                sigma=sigma, name='intro/noise_sync')

        for i in range(ntrials):
            self.intrinsic_burster_simulation(duration=dur, dvsync=0,
                    sigma=sigma, name='intro/noise_%02d'%i)

        df = self.get_datafile(False)
        intro = df.get_node('/intro')
        intro._v_attrs['dur'] = dur
        intro._v_attrs['ntrials'] = ntrials
        intro._v_attrs['dvsync'] = dvsync
        intro._v_attrs['sigma'] = sigma

    @step
    def plot_demo_burst_synchronization(self, interval=(0,1), ibi_min=IBI_MIN,
        nonoise_lw=0.8, figsize=(8,5)):
        """Create intro figure of burster synchronization with test pulses."""
        Snn = self.read_simulation('intro/nonoise')
        Snnsync = self.read_simulation('intro/nonoise_sync')
        Ssync = self.read_simulation('intro/noise_sync')
        Sn0 = self.read_simulation('intro/noise_00')

        # Retrieve study parameters
        params = self.get_node('intro')._v_attrs
        ntrials = params['ntrials']
        dur = params['dur']
        sigma = params['sigma']

        f = self.figure('burst-sync-intro', clear=True, figsize=figsize,
                title='Burst Synchronization with Test Pulses')

        # Plot formatting
        V_fmt = dict(c='b')
        radfmt = RadianFormatter(inline=False)

        def plot_v(ax, S, label):
            ix = (S.mon.t >= interval[0]) & (S.mon.t <= interval[1])
            ax.plot(S.mon.t[ix], S.mon.v[ix], **V_fmt)
            ax.set_xlim(*interval)
            ax.set_axis_off()
            ra.ylabel(ax, label)
        plot_v(f.add_subplot(221), Snn, "$V$, no noise")
        plot_v(f.add_subplot(223), Sn0, "$V$, $%.1f\sigma$"%sigma)

        sim.add_wrapped_phase_column(Snn.spikes)
        sim.add_wrapped_phase_column(Snnsync.spikes)
        sim.add_wrapped_phase_column(Ssync.spikes)

        Snn.spikes['bursts'] = spk.burst_onset(Snn.spikes.t,
                isi_thresh=ibi_min)
        Snnsync.spikes['bursts'] = spk.burst_onset(Snnsync.spikes.t,
                isi_thresh=ibi_min)
        Ssync.spikes['bursts'] = spk.burst_onset(Ssync.spikes.t,
                isi_thresh=ibi_min)

        def _plot_phase_finish(ax):
            ax.set_ylim(-np.pi, np.pi)
            ax.set_yticks([-np.pi, 0, np.pi])
            ax.yaxis.set_major_formatter(radfmt)
            ax.set_xlim(0, dur)
            ra.despine(ax)
            ax.set_ylabel(r'Burst $\theta$-phase')

        rclw = mpl.rcParams['lines.linewidth']
        def plot_nn_phase(ax, spk, spksync):
            for s, c, lw in [(spk, '0.5', nonoise_lw), (spksync, 'r', rclw)]:
                ax.plot(s.t[s.bursts], s.phase0[s.bursts], ls='-', lw=lw, c=c)
            _plot_phase_finish(ax)
        plot_nn_phase(f.add_subplot(222), Snn.spikes, Snnsync.spikes)

        # Collect burst-phase curves for all trials of the noise simulations
        db_noise = []
        t_noise = []
        phase_noise = []
        for i in range(ntrials):
            Snx = self.read_simulation('intro/noise_%02d'%i)
            sim.add_wrapped_phase_column(Snx.spikes)
            Snx.spikes['bursts'] = spk.burst_onset(Snx.spikes.t,
                    isi_thresh=ibi_min)
            bt = Snx.spikes.t[Snx.spikes.bursts]
            db_noise.append(np.diff(bt))
            t_noise.append(bt)
            phase_noise.append(Snx.spikes.phase0[Snx.spikes.bursts])

        # Compute some burst statistics
        db = np.diff(Snn.spikes.t[Snn.spikes.bursts])
        self.out('No-noise pd. = {:.5f} mean ({:.5f} Hz), {:.5f} var',
                db.mean(), 1/db.mean(), db.var())
        mean_db = np.mean([d.mean() for d in db_noise])
        var_db = np.mean([d.var() for d in db_noise])
        self.out('{:.1f}sigma = {:.5f} avg. mean ({:.5} Hz), {:.6f} avg. var',
                sigma, mean_db, 1/mean_db, var_db)

        def plot_noise_phase(ax, y):
            for i in range(ntrials):
                ax.plot(t_noise[i], phase_noise[i], marker='.', ls='', ms=3,
                        alpha=0.3, zorder=-1, c='0.5')
            ax.plot(y.t[y.bursts], y.phase0[y.bursts], ls='-', zorder=0, c='r')
            _plot_phase_finish(ax)
            ax.set_xlabel('Simulation time (s)')
        plot_noise_phase(f.add_subplot(224), Ssync.spikes)

    @step
    def phaser_training_simulation(self, N=128, dur=1000.0, lap_period=60.0):
        """Run a simulation of 1D spatially tuned phasers for training."""
        br.defaultclock.dt = dt = 1*ms
        duration = dur*second

        F_input = br.TimedArray(
                sim.triangle_wave(np.arange(0, duration+dt, dt), lap_period),
                dt=dt)

        self['N_phaser_training'] = N
        self['N_subset'] = Nsub = int(N/2)
        self['width_phaser_tuning'] = phaser_width = 1/Nsub

        a = 0.02
        b = 0.2
        c = -50.0
        d = 4.0
        f_theta = self.c.ftheta*Hz
        ge = self.c.ge_phaser
        gtheta = self.c.gtheta_phaser
        gtheta_pos = 5.0*gtheta
        tau = self.c.tau_phaser*ms
        v0 = -75.0
        Wneg = 3.0
        Ei = -80.0
        tau_i = 100*ms

        negative_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ie + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # 1D spatially-tuned gaussian input with inverse inputs
            Ie = ge*(isinv*h_inv + (1-isinv)*h) : 1
            h_inv = 1 - h : 1
            h = exp(-((x-xpref)**2)/(2*width)) : 1
            x = F_input(t) : 1 (shared)
            isinv : 1 (constant)
            xpref : 1 (constant)
            width : 1 (shared)

            # Theta phase and inhibitory (antiphase) input
            Itheta = -0.5*gtheta*(cos(phase) + 1) : 1 (shared)
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        Gneg = br.NeuronGroup(N, model=negative_eqns, threshold='v > 30',
                order=1, reset='v = c; u = u + d', name='Negative1DPhaser',
                method='rk4')
        Gneg.v = v0
        Gneg.isinv[:Nsub] = 0
        Gneg.isinv[Nsub:] = 1
        Gneg.xpref[:Nsub] = np.linspace(0, 1, Nsub)
        Gneg.xpref[Nsub:] = np.linspace(0, 1, Nsub)
        Gneg.width = phaser_width

        positive_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ineg + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Slow inhibitory synaptic current from negative phasers
            Ineg = -gi*(v - Ei) : 1
            dgi/dt = -gi/tau_i : 1

            # Theta phase and excitatory (inphase) input
            Itheta = 0.5*gtheta_pos*(cos(phase) + 1) : 1 (shared)
            phase : 1 (linked)

            x = F_input(t) : 1 (shared)
        """
        Gpos = br.NeuronGroup(N, model=positive_eqns, threshold='v > 30',
                order=2, reset='v = c; u = u + d', name='Positive1DPhaser',
                method='rk4')
        Gpos.v = v0
        Gpos.phase = br.linked_var(Gneg.phase)

        C = br.Synapses(Gneg, Gpos, on_pre='gi += Wneg')
        C.connect(j='i')

        NegSpkMon = br.SpikeMonitor(Gneg, ['x', 'phase'], record=True,
                name='nspikes')
        PosSpkMon = br.SpikeMonitor(Gpos, ['x', 'phase'], record=True,
                name='pspikes')

        @br.network_operation(dt=10*second, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        Net = br.Network(Gneg, Gpos, C, NegSpkMon, PosSpkMon, progress,
                name='Phaser1DNetwork')
        Net.run(duration)

        self.save_simulation('phaser1d', Net,
                title='1D Training Simulation of Negative/Positive Phasers')

    @step
    def compute_space_phase_distros(self, ibi_min=IBI_MIN):
        """Compute conditional probability of phase in space for 1D phasers."""
        S = self.read_simulation('phaser1d')
        N = self.c.N_phaser_training
        Nsub = self.c.N_subset

        sim.add_wrapped_phase_column(S.nspikes)
        sim.add_wrapped_phase_column(S.pspikes)

        S.nspikes['burst'] = spk.burst_onset(S.nspikes.t, ibi_min)
        S.pspikes['burst'] = spk.burst_onset(S.pspikes.t, ibi_min)

        sh = Nsub, binned.INFO_PHASE_POS_BINS, binned.INFO_PHASE_BINS
        Pneg = np.empty(sh, 'd')
        Ppos = np.empty_like(Pneg)
        Pninv = np.empty_like(Pneg)
        Ppinv = np.empty_like(Pneg)

        data = [(Pneg, S.nspikes, list(range(Nsub))),
                (Pninv, S.nspikes, list(range(Nsub,N))),
                (Ppos, S.pspikes, list(range(Nsub))),
                (Ppinv, S.pspikes, list(range(Nsub,N)))]

        for P, spikes, neurons in data:
            for i,n in enumerate(neurons):
                xb = spikes.loc[spikes.neuron == n, 'x']
                bphase = spikes.loc[spikes.neuron == n, 'phase0']
                P[i] = binned.phasedist1d(xb, bphase, xmin=0, xmax=1)

        self.save_array('phaser1d/phasedists/neg', Pneg,
                title='1D Negative Phasers: Phase-Space Distribution')
        self.save_array('phaser1d/phasedists/neg_inverse', Pninv,
                title='1D Negative-Inverse Phasers: Phase-Space Distribution')
        self.save_array('phaser1d/phasedists/pos', Ppos,
                title='1D Positive Phasers: Phase-Space Distribution')
        self.save_array('phaser1d/phasedists/pos_inverse', Ppinv,
                title='1D Positive-Inverse Phasers: Phase-Space Distribution')

    @step
    def plot_space_phase_distros(self, cmap='viridis', example_x=0.5,
        figsize=(5,5), full_figsize=(8,8)):
        """Create figure of negative/positive space-phase distributions."""
        N = self.c.N_phaser_training
        Nsub = self.c.N_subset

        fneg = self.figure('neg-phase-distros', clear=True, figsize=full_figsize,
                title='Negative 1D Phasers: Phase-Space Distributions')
        fninv = self.figure('neg-inv-phase-distros', clear=True, figsize=full_figsize,
                title='Negative-Inverse 1D Phasers: Phase-Space Distributions')
        fpos = self.figure('pos-phase-distros', clear=True, figsize=full_figsize,
                title='Positive 1D Phasers: Phase-Space Distributions')
        fpinv = self.figure('pos-inv-phase-distros', clear=True, figsize=full_figsize,
                title='Positive-Inverse 1D Phasers: Phase-Space Distributions')

        fex = self.figure('example-phase-distros', clear=True, figsize=figsize,
                title='1D Phasers: Phase-Space Distributions')

        # Read phase distribution data
        Pneg = self.read_array('phaser1d/phasedists/neg')
        Pninv = self.read_array('phaser1d/phasedists/neg_inverse')
        Ppos = self.read_array('phaser1d/phasedists/pos')
        Ppinv = self.read_array('phaser1d/phasedists/pos_inverse')

        # Get index of example cell based on preferred location
        xpref = np.linspace(0, 1, Nsub)
        ex_ix = np.argmin(np.abs(example_x - xpref))

        # Plot an example phase from each subset for a main figure
        mapextent = (0, 1, binned.PHASE_MIN, binned.PHASE_MAX)
        mapkw = dict(extent=mapextent, cmap=cmap)
        radfmt = RadianFormatter(inline=False)
        axdata = [(Pneg, 'Negative'), (Ppos, 'Positive'),
                  (Pninv, 'Negative/Inverse'), (Ppinv, 'Positive/Inverse')]
        for i, (P, title) in enumerate(axdata):
            ax = fex.add_subplot(221 + i)
            maps.plot(P[ex_ix], ax=ax, **mapkw)
            ax.axis('tight')
            ax.set(xticks=[0,0.5,1], xticklabels=['0',0.5,'1'],
                    yticks=[-np.pi,0,np.pi])
            ax.yaxis.set_major_formatter(radfmt)
            ra.quicktitle(ax, title, weight='light', size='x-small')
            if i % 2 == 0:
                ax.set_ylabel('Burst phase')
            if i >= 2:
                ax.set_xlabel('Position')
            if i == 0:
                ra.despine(ax, bottom=False)
            if i == 1:
                ax.set_axis_off()
            if i == 2:
                ra.despine(ax)
            if i == 3:
                ra.despine(ax, left=False)

        # Plot each figure as small-multiples grid of space-phase plots
        figdata = [(fneg, Pneg), (fninv, Pninv), (fpos, Ppos), (fpinv, Ppinv)]
        r, c = tiling_dims(Nsub)
        for f, P in figdata:
            for i in range(Nsub):
                ax = f.add_subplot(r, c, i+1)
                maps.plot(P[i], ax=ax, **mapkw)
                ax.axis('tight')
                if i % c == 0:
                    ax.set_yticks([-np.pi,0,np.pi])
                    if i == 0:
                        ax.set_ylabel('Burst phase')
                    if i == (r-1)*c:  # last row
                        ax.set_xticks([0,1])
                        ax.set_xlabel('Position')
                        ra.despine(ax)
                    else:
                        ax.set_xticks([])
                        ra.despine(ax, bottom=False)
                    ax.yaxis.set_major_formatter(radfmt)
                else:
                    ax.set_axis_off()
            f.tight_layout(h_pad=0, w_pad=0, rect=(0,0,1,0.97))

    @step
    def compute_weight_vector(self, Emax=None):
        """Perform competitive Emax%-winners-take-all to create weights."""
        self.get_datafile(False)
        N = self.c.N_phaser_training

        # Load space-phase distributions for neg/pos phasers
        Pneg = np.r_[
                self.read_array('phaser1d/phasedists/neg'),
                self.read_array('phaser1d/phasedists/neg_inverse')]
        Ppos = np.r_[
                self.read_array('phaser1d/phasedists/pos'),
                self.read_array('phaser1d/phasedists/pos_inverse')]

        # Define the three-part pixel-based phase code target pattern
        target = np.zeros_like(Pneg[0])
        target[0:4,:3] = 1
        target[0:4,33:36] = 1
        target[11:15,16:22] = 1
        self.save_array('phaser1d/target2d', target,
                title='Phase-Code Target: Joint Image')

        # Compute cosine similarity as basis of phaser-burster weights
        def cosine_sim(P):
            Pnorm = np.sqrt((P * P).sum(axis=(1,2)))
            target_norm = np.sqrt((target * target).sum())
            norm = Pnorm * target_norm
            return (P * target).sum(axis=(1,2)) / norm

        # Phaser-subtype competitive (Emax %)-winners-take-all
        Emax = Emax or self.c.Emax
        Wneg = sim.Emax_winners(cosine_sim(Pneg), Emax)
        Wpos = sim.Emax_winners(cosine_sim(Ppos), Emax)

        self.out('Emax = {:.3f}', Emax)
        self.out('Active neg. synapses: {} / {}', np.sum(Wneg>0), Wneg.size)
        self.out('Active pos. synapses: {} / {}', np.sum(Wpos>0), Wpos.size)

        self.save_array('phaser1d/Wneg', Wneg,
            title='Negative Phaser-Burster Weights')
        self.save_array('phaser1d/Wpos', Wpos,
            title='Positive Phaser-Burster Weights')

    @step
    def plot_target_weights_inputs(self, poscolor=POSCOLOR, negcolor=NEGCOLOR,
        figsize=(6.1,7.9)):
        """Plot figures of the phase-code target and weighted inputs."""
        N = self.c.N_phaser_training
        Nsub = self.c.N_subset
        Wneg = self.read_array('phaser1d/Wneg')
        Wpos = self.read_array('phaser1d/Wpos')
        target = self.read_array('phaser1d/target2d')
        Pneg = self.read_array('phaser1d/phasedists/neg')
        Pninv = self.read_array('phaser1d/phasedists/neg_inverse')
        Ppos = self.read_array('phaser1d/phasedists/pos')
        Ppinv = self.read_array('phaser1d/phasedists/pos_inverse')

        f = self.figure('phase-code-weights', clear=True, figsize=figsize,
                title='Phase Code Target: {} x {}'.format(*target.shape))
        axt = f.add_subplot(311)
        axw = f.add_subplot(312)
        axi = f.add_subplot(313)

        # Weight curves plot data and formatting
        W_fmt = dict(zorder=1)
        conv = ColorConverter()
        ninvcolor = tuple(map(lambda v: 0.7*v, conv.to_rgb(negcolor)))
        pinvcolor = tuple(map(lambda v: 0.7*v, conv.to_rgb(poscolor)))
        W_curve_fmt = [dict(c=negcolor, ls='-', label='Negative'),
                       dict(c=ninvcolor, ls='-', label='Negative/Inverse'),
                       dict(c=poscolor, ls='-', label='Positive'),
                       dict(c=pinvcolor, ls='-', label='Positive/Inverse')]
        W2 = np.c_[Wneg[:Nsub], Wneg[Nsub:], Wpos[:Nsub], Wpos[Nsub:]].T
        x_sub = np.linspace(0, 1, Nsub)

        def plot_distro(ax, M, cmap, label):
            radfmt = RadianFormatter(inline=False)
            maps.plot(M, ax=ax, cmap=cmap, extent=(0,1,-np.pi,np.pi))
            ax.set(xticks=[0,0.5,1], xticklabels=['0',0.5,'1'],
                    yticks=[-np.pi,0,np.pi], ylabel=label)
            # maps.plot(M, ax=ax, cmap=cmap, extent=(0,1,np.pi,3*np.pi))
            # ax.set(xticks=[0,0.5,1], yticks=[-np.pi,0,np.pi,2*np.pi,3*np.pi],
                    # ylabel=label)
            ax.tick_params(axis='x', pad=1)
            ax.yaxis.set_major_formatter(radfmt)
            ax.axis('tight')
            ra.despine(ax)
        plot_distro(axt, target, 'gray_r', 'Phase target')

        def plot_weight_curves(ax):
            for weights, fmt in zip(W2, W_curve_fmt):
                ax.plot(x_sub, weights, **merge_two_dicts(W_fmt, fmt))
            ax.set(ylim=(0,1), yticks=[0,0.5,1], yticklabels=['0',0.5,'1'],
                    xticks=[0,0.5,1], xticklabels=['0',0.5,'1'])
            ax.set_ylabel('Weights')
            ax.legend(loc='upper center')
            ax.tick_params(axis='x', pad=1)
            ra.despine(ax)
            ra.ypadding(ax, pad=0.2)
            ax.set_ylim(bottom=0.0)
        plot_weight_curves(axw)

        # Compute and plot the weighted-average phaser distros
        Pnall = np.concatenate((Pneg, Pninv), axis=0)
        Ppall = np.concatenate((Ppos, Ppinv), axis=0)
        Pw = (Wneg[...,np.newaxis]*Pnall).sum(axis=0)
        Pw += (Wpos[...,np.newaxis]*Ppall).sum(axis=0)
        plot_distro(axi, Pw, 'viridis', 'Weighted input')
        axi.set_xlabel('Position')

    @staticmethod
    def _sync_simulation_path(sigma, fdbk):
        noise_str = 'sigma_{:.1f}'.format(sigma).replace('.', '_')
        fdbk = fdbk or (DEFAULT_NEG_FDBK_GAIN, DEFAULT_POS_FDBK_GAIN)
        fdbk_str = 'neg_{:.2f}_pos_{:.2f}'.format(*fdbk).replace('.', '_')
        param_str = '{}_{}'.format(noise_str, fdbk_str)
        path = 'syncsim/{}'.format(param_str)
        return path

    @step
    def phaser_burster_sync_simulation(self, dur=60.0, lap_period=60.0,
        sigma=DEFAULT_SIGMA,
        neg_fdbk=DEFAULT_NEG_FDBK_GAIN,
        pos_fdbk=DEFAULT_POS_FDBK_GAIN):
        """Run simulation of phasers and intrinsic burster with weights."""
        br.defaultclock.dt = step = 1*ms
        duration = dur*second
        N = self.c.N_phaser_training
        Nsub = self.c.N_subset

        F_input = br.TimedArray(
                sim.triangle_wave(np.arange(0, duration+step, step),
                    lap_period), dt=step)

        a = 0.02
        b = 0.2
        c = -50.0
        d = 4.0
        d_b = 5.0
        f_theta = self.c.ftheta*Hz
        ge = self.c.ge_phaser
        gtheta = self.c.gtheta_phaser
        gtheta_pos = 5.0*gtheta
        tau = self.c.tau_phaser*ms
        tau_b = self.c.tau*ms
        v0 = -75.0
        v0_b = -80.0
        Ie_b = self.c.Ie
        Wneg = 3.0
        Ei = -80.0
        tau_i = 100*ms
        Wnb = self.read_array('phaser1d/Wneg')
        Wpb = self.read_array('phaser1d/Wpos')

        negative_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ie + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # 1D spatially-tuned gaussian input with inverse inputs
            Ie = ge*(isinv*h_inv + (1-isinv)*h) : 1
            h_inv = 1 - h : 1
            h = exp(-((x-xpref)**2)/(2*width)) : 1
            x = F_input(t) : 1 (shared)
            isinv : 1 (constant)
            xpref : 1 (constant)
            width : 1 (shared)

            # Theta phase and inhibitory (antiphase) input
            Itheta = -0.5*gtheta*(cos(phase) + 1) : 1 (shared)
            phase = 2*pi*f_theta*t : 1 (shared)
        """
        Gneg = br.NeuronGroup(N, model=negative_eqns, threshold='v > 30',
                order=1, reset='v = c; u = u + d', name='Negative1DPhaser',
                method='rk4')
        Gneg.v = v0
        Gneg.isinv[:Nsub] = 0
        Gneg.isinv[Nsub:] = 1
        Gneg.xpref[:Nsub] = np.linspace(0, 1, Nsub)
        Gneg.xpref[Nsub:] = np.linspace(0, 1, Nsub)
        Gneg.width = self.c.width_phaser_tuning

        positive_eqns = """
            dv/dt = (0.04*v**2 + 5*v + 140 - u + Ineg + Itheta) / tau : 1
            du/dt = a*(b*v - u) / tau : 1

            # Slow inhibitory synaptic current from negative phasers
            Ineg = -gi*(v - Ei) : 1
            dgi/dt = -gi/tau_i : 1

            # Theta phase and excitatory (inphase) input
            Itheta = 0.5*gtheta_pos*(cos(phase) + 1) : 1 (shared)
            phase : 1 (linked)
        """
        Gpos = br.NeuronGroup(N, model=positive_eqns, threshold='v > 30',
                order=2, reset='v = c; u = u + d', name='Positive1DPhaser',
                method='rk4')
        Gpos.v = v0
        Gpos.phase = br.linked_var(Gneg.phase)

        Cnp = br.Synapses(Gneg, Gpos, on_pre='gi += Wneg')
        Cnp.connect(j='i')

        burst_eqns = br.Equations("""
            dv/dt = (0.04*v**2 + 5*v + 140 + Ie - u)/tau + sigma/sqrt(dt)*xi : 1
            du/dt = a*(b*v - u) / tau : 1
            phase : 1 (linked)
            x : 1 (linked)
        """, Ie=Ie_b, tau=tau_b)
        Gb = br.NeuronGroup(1, model=burst_eqns, threshold='v > 30', order=3,
                reset='v=c; u=u+d_b', name='SyncBurster', method='euler')
        Gb.v = v0_b
        Gb.phase = br.linked_var(Gneg.phase)
        Gb.x = br.linked_var(Gneg.x)

        # Phaser-burster synapse model
        gnb = neg_fdbk
        gpb = pos_fdbk
        Cnb = br.Synapses(Gneg, Gb, model='W : 1 (constant)',
                on_pre='v += gnb*W')
        Cpb = br.Synapses(Gpos, Gb, model='W : 1 (constant)',
                on_pre='v += gpb*W')
        Cnb.connect(i=np.arange(N), j=0)
        Cpb.connect(i=np.arange(N), j=0)
        Cnb.W = Wnb[:,0]  # negative-burster weights
        Cpb.W = Wpb[:,0]  # positive-burster weights

        BurstSpkMon = br.SpikeMonitor(Gb, ['x', 'phase'], record=True,
                name='spikes')
        BurstMon = br.StateMonitor(Gb, 'v', record=[0], name='mon')

        @br.network_operation(dt=10*second, when='end')
        def progress(t):
            self.out('T = {:.1f} seconds', t/second)

        Net = br.Network(Gneg, Gpos, Cnp, Gb, Cnb, Cpb, BurstSpkMon, BurstMon,
                progress, name='Sync1DNetwork')
        Net.run(duration)

        simpath = self._sync_simulation_path(sigma, (neg_fdbk, pos_fdbk))
        self.save_simulation(simpath, Net,
                title='1D Feedback Simulation of Phasers and a Burster')

    @step
    def plot_phaser_burster_sync_simulation(self, sigma=(0.0,DEFAULT_SIGMA),
        ibi_min=IBI_MIN, fdbk=(DEFAULT_NEG_FDBK_GAIN, DEFAULT_POS_FDBK_GAIN),
        figsize=(6.5,10.5)):
        """Plot figure for phaser-burster feedback simulation."""
        P = self.read_simulation(self._sync_simulation_path(sigma[0], fdbk))
        S = self.read_simulation(self._sync_simulation_path(sigma[1], fdbk))
        sim.add_wrapped_phase_column(P.spikes)
        sim.add_wrapped_phase_column(S.spikes)
        P.spikes['bursts'] = spk.burst_onset(P.spikes.t, isi_thresh=ibi_min)
        S.spikes['bursts'] = spk.burst_onset(S.spikes.t, isi_thresh=ibi_min)
        dur = S.mon.t.max()
        target = self.read_array('phaser1d/target2d')

        f = self.figure('feedback-sync', clear=True, figsize=figsize,
                title='1D Phaser Sync: {:.1f}s, {:.1f} and '
                      '{:.1f} s.d.'.format(dur, sigma[0], sigma[1]))

        # Plot formatting parameters
        burst_fmt = dict(c='#000080', marker='.', ls='', ms=6, alpha=0.1,
                zorder=-1)
        radfmt = RadianFormatter(inline=False)

        def plot_bursts(ax, spk):
            maps.plot(target, ax=ax, cmap='gray_r', cmin=0, cmax=1,
                    extent=(0,1,-np.pi,np.pi)).set_alpha(0.3)
            maps.plot(target, ax=ax, cmap='gray_r', cmin=0, cmax=1,
                    extent=(0,1,np.pi,3*np.pi)).set_alpha(0.3)
            ax.plot(spk.x[spk.bursts], spk.phase0[spk.bursts], **burst_fmt)
            ax.plot(spk.x[spk.bursts], spk.phase0[spk.bursts] + 2*np.pi,
                    **burst_fmt)
            ax.axis('tight')
            ax.set(ylim=(-np.pi,3*np.pi),
                    yticks=[-np.pi,0,np.pi,2*np.pi,3*np.pi])
            ax.yaxis.set_major_formatter(radfmt)
            ax.set(xticks=(0,0.5,1), xticklabels=['0',0.5,'1'], xlim=(0,1))
            ax.set_ylabel('Target burster phase')
            ra.despine(ax)
        plot_bursts(f.add_subplot(211), P.spikes)
        plot_bursts(f.add_subplot(212), S.spikes)
        plt.gca().set_xlabel(xlabel='Position')

    @step
    def plot_supplemental_sync_simulations(self, sigma_list=None,
        fdbk_list=None, fdbk_sigma=None, ibi_min=IBI_MIN, figsize=(11,8)):
        """Plot figure for phaser-burster feedback simulation.

        Set `sigma_list` to a list of noise values to generate a figure
        of space-phase burst-timing plots
        Set `fdbk_list` to a list of `(neg_fdbk, pos_fdbk)` tuples to load
        simulation data for those parameters. This will create a second
        figure for varying feedback gains.
        """
        target = self.read_array('phaser1d/target2d')

        # Plot formatting parameters
        burst_fmt = dict(c='#000080', marker='.', ls='', ms=6, alpha=0.1,
                zorder=-1)
        radfmt = RadianFormatter(inline=False)

        def plot_bursts(ax, spk, title):
            maps.plot(target, ax=ax, cmap='gray_r', cmin=0, cmax=1,
                    extent=(0,1,-np.pi,np.pi)).set_alpha(0.3)
            maps.plot(target, ax=ax, cmap='gray_r', cmin=0, cmax=1,
                    extent=(0,1,np.pi,3*np.pi)).set_alpha(0.3)
            ax.plot(spk.x[spk.bursts], spk.phase0[spk.bursts], **burst_fmt)
            ax.plot(spk.x[spk.bursts], spk.phase0[spk.bursts] + 2*np.pi,
                    **burst_fmt)
            ax.axis('tight')
            ax.set(ylim=(-np.pi,3*np.pi),
                    yticks=[-np.pi,0,np.pi,2*np.pi,3*np.pi])
            ax.yaxis.set_major_formatter(radfmt)
            ax.set(xticks=(0,0.5,1), xticklabels=['0',0.5,'1'], xlim=(0,1))
            ra.quicktitle(ax, title)
            ra.despine(ax)

        def load_and_plot_simulation(f, r, c, i, title, sigma, fdbk):
            simpath = self._sync_simulation_path(sigma, fdbk)
            self.out('Loading simulation: {}', simpath)

            S = self.read_simulation(simpath)
            sim.add_wrapped_phase_column(S.spikes)
            S.spikes['bursts'] = spk.burst_onset(S.spikes.t, isi_thresh=ibi_min)

            ax = f.add_subplot(r,c,i+1)
            plot_bursts(ax, S.spikes, title)
            if i >= (r-1)*c:
                ax.set_xlabel(xlabel='Position')
            if i % c == 0:
                ax.set_ylabel('Target burster phase')

        if sigma_list:
            parstr = ', '.join(map(lambda s: '{:.1f}'.format(s), sigma_list))
            f = self.figure('feedback-sync-noise', clear=True, figsize=figsize,
                    title='1D Synchonization Noise Levels: '
                          '{}'.format(parstr))
            r,c = tiling_dims(len(sigma_list))

            for i in range(len(sigma_list)):
                sigma = sigma_list[i]
                title = r'{:.1f}$\sigma$'.format(sigma)
                self.out('Plotting sigma = {:.1f}', sigma)
                load_and_plot_simulation(f, r, c, i, title, sigma, None)

        if fdbk_list:
            parstr = ', '.join(map(str, fdbk_list))
            f = self.figure('feedback-sync-gain', clear=True, figsize=figsize,
                    title='1D Synchonization Feedback Gains: '
                          '{}'.format(parstr))
            r,c = tiling_dims(len(fdbk_list))
            sigma = fdbk_sigma or DEFAULT_SIGMA

            for i in range(len(fdbk_list)):
                fdbk = fdbk_list[i]
                title = '{:.1f} Negative gain, {:.1f} Positive gain'.format(*fdbk)
                self.out('Plotting feedback = {}', str(fdbk))
                load_and_plot_simulation(f, r, c, i, title, sigma, fdbk)
