"""
Compute phase estimates for Theta 11 based on the TT9 theta cell.
"""

import os

import matplotlib as mpl
import numpy as np
import tables as tb

from roto import arrays, radians, reports, plots

from .. import store
from ..context import SPCAnalysis, step
from ..lib import labels, cell, data, session
from ..tools import spikes, maps, binned


T9LABEL = '110900'


class Theta11PhaseEstimates(SPCAnalysis):

    """
    Add phase estimates to all sessions with theta cell 110900.
    """

    @step
    def find_cells(self):
        """Find recordings of theta cell 110900 and simultaneous cells."""
        df = store.get()
        cell_ids = []
        session_ids = []

        for rec in df.root.recordings.where('cell_id=="%s"' % T9LABEL):
            s_id = rec['session_id']
            if df.root.sessions[s_id]['arena'] != data.Arenas.circle:
                continue
            cell_ids.append(rec['id'])
            session_ids.append(rec['session_id'])

        reclock_cell_ids = []
        for s_id in session_ids:
            for C in df.root.recordings.where('session_id==%d' % s_id):
                reclock_cell_ids.append(C['id'])

        self['N'] = len(cell_ids)
        self['N_reclock'] = len(reclock_cell_ids)
        self.out('Found {} tt9 sessions: {} to reclock.', self['N'],
                self['N_reclock'])

        self.save_array('cell_ids', np.asarray(cell_ids))
        self.save_array('session_ids', np.asarray(session_ids))
        self.save_array('reclock_ids', np.asarray(reclock_cell_ids))

    @step
    def compute_phase_estimates(self, parallel=False):
        """Compute phase estimates for all reference spike trains."""
        self.get_datafile(False)
        cell_ids = self.read_array('cell_ids')

        for i in range(self.c.N):
            c_id = cell_ids[i]
            tag = labels.session_cell_id(c_id)
            S = cell.session_for(c_id)
            C = cell.record(c_id)
            lfp_grp = os.path.join(S.attrs['group'], 'LFP')
            spike_train = S.get_cluster(C).spikes
            self.out('Computing {}: {} spikes', tag, spike_train.size)

            estimator = spikes.FiringPhaseEstimate(spike_train,
                    tlim=(S.tstart, S.tend), parallel=parallel)
            ts, phase = estimator.compute()

            self.save_array(os.path.join(lfp_grp, 'TS_tt9'), ts,
                    title='Timestamps for 110900-based theta estimate')
            self.save_array(os.path.join(lfp_grp, 'thetaphase_tt9'), phase,
                    title='Theta phase estimated from 110900 spike train')

    @step
    def export_to_datastore(self):
        """Copy the phase estimate data arrays to the main data store."""
        df = self.get_datafile(True)
        for node in df.walk_nodes('/data', classname='Array'):
            path, name = os.path.split(node._v_pathname)
            self.out('Copying {}', node._v_pathname)
            data.new_array(path, name, node.read(), title=node.title)

    @step
    def compute_updated_phasemaps(self):
        """Compute and store phase maps from the updated phase data."""
        self.get_datafile(False)
        reclock_ids = self.get_node('reclock_ids')

        for c_id in reclock_ids:
            tag = labels.session_cell_id(c_id)
            C = cell.record(c_id)
            S = cell.session_for(c_id)
            P = S.phasemap(C, adaptive=True)
            self.save_array('phasemaps/id%04d' % c_id, P,
                    title='Phasemap: %s' % tag, cell_id=c_id)

    @step
    def compute_updated_distros(self, bins=binned.DEFAULT_PHASE_BINS):
        """Compute and store updated phase distributions."""
        self.get_datafile(False)
        edges = np.linspace(-np.pi, np.pi, bins + 1)
        centers = (edges[:-1] + edges[1:]) / 2
        self.save_array('phasedist/centers', centers, title='Phase Bins')
        reclock_ids = self.get_node('reclock_ids')

        for c_id in reclock_ids:
            tag, st = cell.tag_and_spikes(c_id)
            S = cell.session_for(c_id)
            sphase = S.lfp.F('phase', st)
            H = np.histogram(sphase, bins=edges)[0]
            Hg = arrays.circblur(H, radians=2*np.pi/36)
            self.save_array('phasedist/id%04d' % c_id, H,
                title='Phase Distribution: %s' % tag, cell_id=c_id)
            self.save_array('phasedist/id%04d_smooth' % c_id, Hg,
                title='Phase Distribution: %s' % tag, cell_id=c_id)

    @step
    def generate_comparison_report(self):
        """Make a report comparing the old to the new phase data."""
        df = store.get()
        phigrp = df.get_node('/theta/phasedist')
        mapgrp = df.get_node('/phasemaps')
        centers = df.get_node(phigrp, 'centers').read()
        reclock_ids = self.read_array('reclock_ids')

        mpl.rc('xtick', labelsize=7.0, direction='out')
        mpl.rc('ytick', labelsize=7.0, direction='out')
        mpl.rc('axes', titlesize='xx-small', labelsize='xx-small')
        radfmt = radians.RadianFormatter(inline=True)

        report = reports.AxesReport(self.path('update-comparison'),
                naxes=4*self.c.N_reclock, rows=5, cols=4,
                xnorm=False, ynorm=False)
        report.set_label_fmt(size='x-small')

        for i, c_id in enumerate(reclock_ids):
            tag = labels.session_cell_id(c_id)
            self.out('Plotting {}', tag)

            Pnew = self.read_array('phasemaps/id%04d' % c_id)
            Pold = df.get_node(mapgrp, 'id%04d' % c_id).read()

            phiold = df.get_node(phigrp, 'id%04d' % c_id).read()
            phioldsm = df.get_node(phigrp, 'id%04d_smooth' % c_id).read()
            modold = phioldsm.ptp() / phioldsm.max()

            phinew = self.read_array('phasedist/id%04d' % c_id)
            phinewsm = self.read_array('phasedist/id%04d_smooth' % c_id)
            modnew = phinewsm.ptp() / phinewsm.max()

            ax = report.next_axes()
            maps.plot(Pold, ax=ax, levels=24, mask_color='0.8', satnorm=True)
            ax.set_axis_off()
            if report.firstrow:
                ax.set_title('{}\nOriginal Phasemap'.format(tag))
            else:
                ax.set_title(tag)
            if report.firstcol:
                plots.ylabel(ax, 'Y', size='x-small')
            if report.lastrow:
                plots.xlabel(ax, 'X', size='x-small')

            ax = report.next_axes()
            maps.plot(Pnew, ax=ax, levels=24, mask_color='0.8', satnorm=True)
            ax.set_axis_off()
            if report.firstrow:
                ax.set_title('Updated Phasemap')
            if report.lastrow:
                plots.xlabel(ax, 'X', size='x-small')

            def _plot_phase_dist(name, phi, phis, mod):
                ax = report.next_axes()
                ax.plot(centers, phi, 'r-', lw=1, alpha=0.6)
                ax.plot(centers, phis, 'b-', lw=2, alpha=0.6)
                if report.firstrow:
                    ax.set_title('{}\nIndex = {:.3f}'.format(name, mod), va='top')
                else:
                    ax.set_title('\nIndex = {:.3f}'.format(mod), va='top')
                ax.set_ylim(0.0, 1.2*np.max([phi,phis]))
                ax.set_xlim(-np.pi, np.pi)
                ax.set_xticks([-np.pi, -np.pi/2, 0, np.pi/2, np.pi])
                ax.xaxis.set_major_formatter(radfmt)
                ax.set_yticks([])
                ax.set_yticklabels([])
                plots.despine(ax)
                if report.lastrow:
                    ax.set_xlabel('Phase (rad)')

            _plot_phase_dist('Original', phiold, phioldsm, modold)
            _plot_phase_dist('Updated', phinew, phinewsm, modnew)

        report.close()
        mpl.rc_file_defaults()

    @step
    def export_phasemaps_and_distros(self):
        """Replace old phase maps and distros with updated versions."""
        df = store.get(False)
        thisdf = self.get_datafile()
        data.clone_subtree(thisdf, '/phasemaps', df)
        data.clone_subtree(thisdf, '/phasedist', df, root='/theta')

    @step
    def store_tt9phase_flags(self, backup='recordings_orig'):
        """Add flag columns to /recordings indicating updated phase values."""
        df = store.get(False)
        try:
            orig = df.root.recordings
        except tb.NoSuchNodeError:
            pass
        else:
            orig._f_rename(backup)
        orig = df.get_node('/', backup)

        newdesc = data.recordings_desc
        newdesc['tt9ref'] = tb.Boospcl(pos=len(newdesc)+1)
        newdesc['tt9clocked'] = tb.Boospcl(pos=len(newdesc)+1)
        tbl = data.new_table('/', 'recordings', newdesc, h5file=df,
                title='Cell-Session Recordings')

        row = tbl.row
        fields = orig.colnames
        reclock_ids = set(self.read_array('reclock_ids'))
        for record in orig.iterrows():
            for col in fields:
                row[col] = record[col]
            row['tt9ref'] = row['cell_id'] == '110900'
            row['tt9clocked'] = row['id'] in reclock_ids
            row.append()

        self.out('Done!')
