"""
Descriptive analysis of theta cell recording dataset
"""

from collections import namedtuple

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from roto.radians import RadianFormatter

from .. import store
from ..context import SPCAnalysis, step
from ..lib import data, labels
from ..lib.phasers import PHASER_IPHASE_IMIN, PHASER_RP_SMIN


NONSIGCOLOR = (0.88470588, 0.55803921, 0.3)  # lightened colorblind 'r'
SIGCOLOR = (0.3, 0.61294117, 0.78862745)  # lightened colorblind 'b'
INFO_SIG_THRESH = 0.02
PI2 = 2*np.pi
MeanSD = namedtuple('MeanSD', ['mean', 'SD'])


class ThetaCellDescription(SPCAnalysis):

    """
    Compute basic statistics of theta cell properties
    """

    def cleanup(self):
        store.close()

    @step
    def setup(self):
        """Cell properties with addtional fields for this analysis."""
        P = data.properties_dataframe()
        self.c.N_rats = len(set(P.rat))
        self.c.N_sessions = len(set(P.session_id))

    @step
    def theta_cell_stats(self, ddof=0):
        """Basic statistics for theta cell recordings."""
        P = data.properties_dataframe()
        P['sig'] = P['I_phase_p'] <= INFO_SIG_THRESH
        props_by_rat = P.groupby('rat')

        stats = [
            ('Avg. Rate', 'avg_rate'),
            ('Max. Rate', 'ratemap_max'),
            ('Theta Index', 'theta_mod'),
            ('Rhythmicity', 'theta_rhythm'),
            ('Phase Coherence', 'phase_coh'),
            ('Freq. Estimate', 'f_hat'),
            ('Rate/phase r', 'C_rp_r'),
            ('Rate/phase slope', 'C_rp_sl'),
            ('Total Phase Shift', 'rp_shift')
        ]

        for name, field in stats:
            values = props_by_rat[field].mean()
            avg = MeanSD(mean=values.mean(), SD=values.std(ddof=ddof))
            self.out('{} = {:.3f} +- {:.3f}', name, *avg)

    def plot_violin_strips(self, ax, ylim, nonsigcolor, sigcolor, df, values,
        name, valname):
        """Violin plot of cell properties values."""
        if type(values) is str:
            values = [values]

        plot_data = pd.melt(df, id_vars=['sig'], value_vars=values,
                var_name=name, value_name=valname)

        sns.violinplot(x=name, y=valname, data=plot_data, hue='sig', cut=0.0,
                scale='count', split=True, inner='quartiles', ax=ax,
                palette={True:sigcolor, False:nonsigcolor})

        for i, col in enumerate(values):
            data = df.loc[df.sig == True, col].values
            ax.scatter(i + 0.25 + 0.02 * np.random.randn(data.size), data,
                    s=16, c='b', linewidths=0, alpha=0.5)

        ax.set_ylim(*ylim)
        for x0 in range(len(values)):
            ax.axvline(x=x0, color='0.5')

        return ax

    @step
    def maxrate_thetamod_distros(self, nonsigcolor=NONSIGCOLOR,
        sigcolor=SIGCOLOR, figsize=(10,5.5)):
        """Univarite max. rate and theta index distros by phase info."""
        P = data.properties_dataframe()
        P['sig'] = P['I_phase_p'] <= INFO_SIG_THRESH

        f = self.figure('rate-theta', clear=True, figsize=figsize,
                title='Firing Rates and Theta Modulation')

        # Firing rate distributions
        ax = self.plot_violin_strips(f.add_subplot(141), (0.0, 90.0),
                nonsigcolor, sigcolor, P,
                ['ratemap_max'], 'Firing rate', 'Rate (spikes/s)')

        # Theta frequency distributions
        ax = self.plot_violin_strips(f.add_subplot(142), (4.0, 12.0),
                nonsigcolor, sigcolor, P.loc[P.f_hat != -1],
                ['f_hat'], 'Theta frequency', 'Frequency (Hz)')

        # Theta modulation distributions
        ax = self.plot_violin_strips(f.add_subplot(122), (0.0, 1.0),
                nonsigcolor, sigcolor, P.loc[P.theta_rhythm != -1],
                ['theta_mod', 'theta_rhythm'], 'Theta', 'Index')

        sns.despine()

    @step
    def phase_coding_distros(self, nonsigcolor=NONSIGCOLOR, sigcolor=SIGCOLOR,
        figsize=(16,5.5)):
        """Univarite distros for phase information and correlations."""
        P = data.properties_dataframe()
        P['sig'] = P['I_phase_p'] <= INFO_SIG_THRESH

        f = self.figure('rate-phase', clear=True, figsize=figsize,
                title='Spatial Phase-Coding Properties')

        # Spatial phase information distributions
        self.plot_violin_strips(f.add_subplot(141), (0.0, 4.0), nonsigcolor,
                sigcolor, P, 'I_phase', 'Spatial phase',
                'Information (bits/spike)')

        # Correlation strength distributions
        self.plot_violin_strips(f.add_subplot(142), (-1, 1), nonsigcolor,
                sigcolor, P, 'C_rp_r', 'Rate-phase correlation', 'r')

        # Correlation slope distributions
        self.plot_violin_strips(f.add_subplot(143), (-5, 5), nonsigcolor,
                sigcolor, P, 'C_rp_sl', 'Rate-phase slope', 'Slope')

        # Total phase shift distributions
        self.plot_violin_strips(f.add_subplot(144), (-PI2, PI2), nonsigcolor,
                sigcolor, P, 'rp_shift', 'Phase shift', 'Shift (rad)')

        sns.despine()

    @step
    def phase_coding_scatter_plot(self, nonsigcolor=NONSIGCOLOR,
        sigcolor=SIGCOLOR, xlegend=3.5, ylegend=[0,-np.pi/4],
        figsize=(6.6,5.15)):
        """Scatter plot of phase information against rate-phase shift."""
        P = data.properties_dataframe()
        P['sig'] = P['I_phase_p'] <= INFO_SIG_THRESH

        f = self.figure('info-shift-scatter', clear=True, figsize=figsize,
                title='Phase Information x Total Shift')

        ax = f.add_subplot(111)

        # Plot formatting
        critfmt = dict(closed=True, hatch=r'\\\\\\', edgecolor=(1,0,0),
                alpha=0.7, lw=0.8, facecolor='none', zorder=5)
        smax = 6.48

        cmap = sns.light_palette(nonsigcolor, n_colors=10, as_cmap=True)
        valid = np.logical_and(np.logical_not(P.sig), np.isfinite(P.rp_shift))
        sns.kdeplot(P.I_phase[valid], P.rp_shift[valid],
                cmap=cmap, shade=True, shade_lowest=False, cut=5, ax=ax,
                alpha=0.9, zorder=-10)

        rsquare = np.power(P.C_rp_r[P.sig], 2)
        ax.scatter(P.I_phase[P.sig], P.rp_shift[P.sig],
                s=10 + 60 * rsquare,
                c=sigcolor, linewidths=0.5, alpha=0.8)

        # Plot min and max points as a legend
        rlegend = np.array([rsquare.min(), rsquare.max()])
        ax.scatter([xlegend]*2, ylegend,
                s=10 + 60 * rlegend,
                c=sigcolor, linewidths=0.5, alpha=0.8)

        self.out('Marker size = R^2 range ~ {:.4f}-{:.4f}', rsquare.min(),
                rsquare.max())

        ax.set(xlim=(0,4), xticks=[0,0.5,1,1.5,2,2.5,3,3.5,4],
               xticklabels=['0',0.5,'1',1.5,'2',2.5,'3',3.5,'4'],
               xlabel='Spatial phase information (bits/spike)',
               ylim=(-smax, smax), yticks=np.linspace(-PI2, PI2, 5),
               ylabel='Total phase shift')
        ax.yaxis.set_major_formatter(RadianFormatter(inline=False))

        def plot_criteria(ax, pad=0.05):
            x1, x0 = ax.get_xlim()
            x0 += pad
            y0, y1 = -smax - pad, smax + pad
            Imin = PHASER_IPHASE_IMIN
            smin = PHASER_RP_SMIN
            verts = np.array(
                [[x1, x1, Imin, Imin, x0,    x0,   Imin, Imin],
                 [y1, y0, y0,  -smin, -smin, smin, smin, y1  ]]).T
            p = mpl.patches.Polygon(verts, clip_box=ax.bbox, **critfmt)
            ax.add_artist(p)
        plot_criteria(ax)

        sns.despine()
