"""
Collection of all analysis and modeling data plots for paper figures.
"""

import os
from os.path import join as opj
import subprocess

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

import roto.axis as ra
from roto.misc import merge_two_dicts
from roto.images import stack_to_image
from roto.colormaps import huslmap

from .ana.sync_model import BurstSyncSimulation
from .ana.phaser_model import PhaserSimulation
from .ana.data_intro_example import IntroExample
from .ana.theta_cells import ThetaCellDescription
from .ana.phaser_cells import PhaserCellDescription
from .ana.stability import PhaserStabilityAnalysis
from .ana.phaser_examples import PhaserExamples
from .ana.trajectory_independence import TrajectoryIndependence
from .ana.spatial_input import SpatialInputModel
from .ana.stripe_model import StripeSyncSimulation

from . import RES_ROOT as ResultsPath, store
from .context import SPCAnalysis


# Paths to all analyses used for plotting figures
BurstSyncPath = opj(ResultsPath, '2017-08-03+burst-sync-simulation+fixed')
PhaserDemoPath = opj(ResultsPath, '2017-10-18+phaser-simulation+sampling')
DataExamplePath = opj(ResultsPath, '2017-08-29+intro-example')
ThetaCellPath = opj(ResultsPath, '2017-08-29+theta-cell-description')
PhaserCellPath = opj(ResultsPath, '2017-08-30+phaser-cell-description')
StabilityPath = opj(ResultsPath, '2018-05-25+phaser-stability-analysis+dev')
CoherencePath = opj(ResultsPath,
        '2018-10-31+phaser-stability-analysis+revision+coherence-update')
PhaserExPath = opj(ResultsPath, '2017-08-31+phaser-examples')
TrajGridPath = opj(ResultsPath,
        '2017-03-18+trajectory-independence+ridge-alpha-1.2496')
TrajModelPath = opj(ResultsPath,
        '2017-04-15+trajectory-independence+pure-spatial-model')
SpatialInputPath = opj(ResultsPath, '2017-04-20+spatial-input-model')
StripeSimPath = opj(ResultsPath, '2017-09-05+stripe-sync-simulation')

# Matplotlib settings for publication figures
axes_lw = 0.8
rc_paper = {
        'savefig.format': 'pdf',
        'savefig.frameon': False,
        'savefig.transparent': True,
        'font.size': 9.0,
        'font.weight': 'normal',
        'font.sans-serif': ['Helvetica Neue LT Std', 'Lucida Grande', 'Geneva',
                            'Verdana', 'Bitstream Vera Sans', 'sans-serif'],
        'pdf.fonttype': 'truetype',
        'mathtext.default': 'rm',
        'figure.titlesize': 'large',
        'figure.titleweight': 'light',
        'figure.subplot.wspace': 0.23,
        'figure.subplot.hspace': 0.23,
        'axes.titlesize': 8.0,
        'axes.titleweight': 'light',
        'axes.labelsize': 8.0,
        'axes.linewidth': axes_lw,
        'axes.labelpad': 1.6,
        'lines.solid_capstyle': 'round',
        'lines.linewidth': 1.2,
        'lines.markeredgewidth': 0.0,
        'lines.markersize': 5.6,
        'patch.linewidth': 0.24,
        'xtick.major.width': axes_lw,
        'ytick.major.width': axes_lw,
        'xtick.minor.width': axes_lw/2,
        'ytick.minor.width': axes_lw/2,
        'xtick.major.pad': 1.3,
        'ytick.major.pad': 1.3,
        'xtick.labelsize': 7.2,
        'ytick.labelsize': 7.2,
        'legend.fontsize': 8.0,
        'legend.labelspacing': 0.25
    }
rc_png = {
        'savefig.format': 'png',
        'savefig.frameon': True,
        'savefig.transparent': False,
        'savefig.dpi': 600
    }

mpl.rc_file_defaults()
mpl.rcParams.update(rc_paper)

# Seaborn settings
sns.set_palette('colorblind', color_codes=True)

# Color constants
nonsigcolor = (0.88470588, 0.55803921, 0.3)  # lightened colorblind 'r'
sigcolor = (0.3, 0.61294117, 0.78862745)  # lightened colorblind 'b'
phcolor = (0.3, 0.73372549, 0.61568627)  # lightened colorblind 'g'
poscolor = 'sandybrown'
negcolor = 'slateblue'
noncolor = 'dimgray'

# Savefig parameters
savekw = dict(tight_padding=0.01, closeafter=True)

# Standard reference figure dimensions
col_width = 3.4  # inches
full_width = 7.2
col_height = 8.2
half_height = 0.5*col_height
figsize_hh = (col_width, half_height)
figsize_hf = (col_width, col_height)
figsize_fh = (full_width, half_height)
figsize_ff = (full_width, col_height)

# Common analysis parameters
IBI_MIN = 0.025     # seconds, inter-burst interval threshold


class PaperFigures(SPCAnalysis):

    def cleanup(self):
        store.close()

    def run_all_figures(self):
        """Script method to run all figure group methods."""
        mpl.rcParams.update(rc_paper)

        # Initial simulations
        self.phaser_demo()
        self.phaser1d_distros()
        self.phaser1d_training()
        self.burster_model()
        self.phaser1d_simulations()

        # Theta cell data analysis
        self.intro_example()
        self.theta_cells()
        self.phaser_cells()
        self.phaser_stability()
        self.phaser_examples()

        # GLM statistical model
        self.glm_optimization()
        self.glm_metrics()
        self.glm_analysis()
        self.glm_reconstructions()

        # Spatial phaser simulations
        self.spatial_input()
        self.spatial_layers()
        self.spatial_simulations()
        if hasattr(self, 'tracedata'):
            self.tracedata = self.decoding_analysis(tracedata=self.tracedata)
        else:
            self.tracedata = self.decoding_analysis()

    #
    # Initial models and simulations
    #

    def phaser_demo(self):
        """Phaser bursting model demonstration."""
        plt.ioff()
        path = self.mkdir('00-phaser-demo')
        phaser = PhaserSimulation.load(PhaserDemoPath)
        phaser.set_figfmt('mpl')

        # Phaser demonstration figure
        savekw2 = merge_two_dicts(savekw, {'tight_padding': 0.0})
        mpl.rcParams['lines.linewidth'] = 0.7
        mpl.rcParams['axes.labelpad'] = 0.8
        mpl.rcParams['mathtext.default'] = 'it'
        phaser.plot_test_phaser_simulation(interval_dur=0.7, negcolor=negcolor,
                poscolor=poscolor, figsize=(full_width, 4.55))
        phaser.savefig(basepath=opj(path, '00-phaser-demo'), **savekw2)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def phaser1d_distros(self):
        """1D phaser model examples and supplemental figures."""
        plt.ioff()
        path = self.mkdir('01-phaser1d-distros')
        suppl = opj(path, 'supplemental')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Space-phase 1D phaser distros: example and supplemental figures
        mpl.rcParams['figure.subplot.hspace'] = 0.12
        mpl.rcParams['figure.subplot.wspace'] = 0.12
        burst.plot_space_phase_distros(example_x=0.5,
                figsize=(col_width, col_width),
                full_figsize=(full_width, full_width))
        burst.savefig('example-phase-distros',
            basepath=opj(path, '00-example-distros'), **savekw)
        burst.savefig('neg-phase-distros',
            basepath=opj(suppl, '00-neg-distros'), **savekw)
        burst.savefig('neg-inv-phase-distros',
            basepath=opj(suppl, '00-neg-inv-distros'), **savekw)
        burst.savefig('pos-phase-distros',
            basepath=opj(suppl, '00-pos-distros'), **savekw)
        burst.savefig('pos-inv-phase-distros',
            basepath=opj(suppl, '00-pos-inv-distros'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def phaser1d_training(self):
        """1D phaser model weights and phase-code targets."""
        plt.ioff()
        path = self.mkdir('02-phaser1d-weights')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Phaser 1D simulation phase-code target, weights, and weighted input
        mpl.rcParams['legend.fontsize'] = 7.2
        mpl.rcParams['legend.frameon'] = False
        burst.plot_target_weights_inputs(poscolor=poscolor, negcolor=negcolor,
                figsize=figsize_hh)
        burst.savefig(basepath=opj(path, '00-weighted-input'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def burster_model(self):
        """Output burster model figures: voltage trace and pulse sync."""
        plt.ioff()
        path = self.mkdir('03-burster-model')
        suppl = opj(path, 'supplemental')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Intro burst synchronization figure
        mpl.rcParams['figure.subplot.hspace'] = 0.18
        burst.plot_demo_burst_synchronization(interval=(0,1.0), ibi_min=IBI_MIN,
                figsize=figsize_fh)
        burst.savefig(basepath=opj(path, '00-burst-sync'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def phaser1d_simulations(self):
        """1D phaser model simulations: burst timing across space."""
        plt.ioff()
        path = self.mkdir('04-phaser1d-simulations')
        suppl = opj(path, 'supplemental')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Phaser 1D simulation output figures: burst timing across space
        mpl.rcParams.update(rc_png)
        mpl.rcParams['figure.subplot.hspace'] = 0.19
        mpl.rcParams['figure.subplot.wspace'] = 0.12
        burst.plot_phaser_burster_sync_simulation(ibi_min=IBI_MIN,
                sigma=(0.0, 0.3), fdbk=(1.0, 2.0),
                figsize=figsize_hh)
        burst.savefig(basepath=opj(path, '00-phaser-sync-sim'), **savekw)

        # Supplemental figures: Varying noise level or feedback level
        burst.plot_supplemental_sync_simulations(ibi_min=IBI_MIN,
                sigma_list=[0.0, 0.1, 0.3, 0.5],
                fdbk_list=[(0.0,0.0), #(0.06,0.12), (0.12,0.25), (0.25,0.5),
                           (0.5,1.0), (1.0,2.0), (2.0,4.0)],
                fdbk_sigma=0.3,
                figsize=figsize_fh)
        burst.savefig('feedback-sync-noise',
                basepath=opj(suppl, '00-feedback-noise'), **savekw)
        burst.savefig('feedback-sync-gain',
                basepath=opj(suppl, '00-feedback-gain'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    #
    # Theta cell data and analysis
    #

    def intro_example(self, plot_spike_trajectory=False, run_setup=False,
        mapres=256):
        """Intro example of spatial theta cell from dataset."""
        plt.ioff()
        path = self.mkdir('05-data-example')
        intro = IntroExample.load(DataExamplePath)
        intro.set_figfmt('mpl')

        # Set up data for example cell
        if run_setup:
            example_cell_id = 97
            intro.setup(example_cell_id, res=mapres)
            intro.find_example_LFP_traces()

        # Plot the whole trajectory with overlaid spikes
        if plot_spike_trajectory:
            mpl.rcParams.update(rc_png)
            intro.plot_spike_trajectory(figsize=(1.8, 2.2))
            intro.savefig(basepath=opj(path, '00-traj-%04d' % intro.c.cell_id),
                    **savekw)

        # Plot/save the main figure
        mpl.rcParams.update(rc_paper)
        mpl.rcParams['figure.subplot.hspace'] = 0.3
        mpl.rcParams['axes.labelpad'] = 1.2
        pmodkw = intro.plot_figure(levels=12, phasemod_return=True,
                figsize=(5.5, 5.5))
        figname = intro._lastfig
        intro.plot_phase_modulation_figure(ibi_min=IBI_MIN, interactive=False,
            spike_cmap='brg', figsize=(5.5, 5.5), **pmodkw)
        intro.savefig(basepath=opj(path, '01-intro-%04d' % intro.c.cell_id),
                label=figname, **savekw)
        intro.close_figures()

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def theta_cells(self, runsuppl=True):
        """Theta cell statistics description and distribution plots."""
        plt.ioff()
        path = self.mkdir('06-theta-cells')
        suppl = opj(path, 'supplemental')
        theta = ThetaCellDescription.load(ThetaCellPath)
        theta.set_figfmt('pdf')

        # Rate/phase regression and spatial phase information figure
        mpl.rcParams['axes.labelpad'] = 0.0
        theta.phase_coding_scatter_plot(nonsigcolor=nonsigcolor,
                sigcolor=sigcolor, figsize=(3.0, 3.2))
        theta.savefig(basepath=opj(path, '00-phase-corr-info'), **savekw)
        mpl.rcParams['axes.labelpad'] = rc_paper['axes.labelpad']

        if runsuppl:
            # Max. firing rate and theta frequency, modulation, rhythmicity
            theta.maxrate_thetamod_distros(nonsigcolor=nonsigcolor,
                sigcolor=sigcolor, figsize=figsize_fh)
            theta.savefig(basepath=opj(suppl, '00-rate-theta-distros'),
                    **savekw)

            # Spatial phase-coding distribution figures
            theta.phase_coding_distros(nonsigcolor=nonsigcolor,
                sigcolor=sigcolor, figsize=figsize_fh)
            theta.savefig(basepath=opj(suppl, '01-spatial-phase-distros'),
                    **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def phaser_stability(self):
        """Phaser cell stability analysis plots."""
        plt.ioff()
        path = self.mkdir('07-phaser-cells')
        ana = PhaserStabilityAnalysis.load(StabilityPath)
        ana.set_figfmt('pdf')

        # Rate/phase regression and spatial phase information figure
        # mpl.rcParams['axes.labelpad'] = 1.3
        mpl.rcParams['figure.subplot.hspace'] = 0.3
        mpl.rcParams['figure.subplot.wspace'] = 0.4
        ana.plot_stability_analysis(phcolor=phcolor, noncolor=noncolor,
                figsize=(6.5, 3.7))
        ana.savefig(basepath=opj(path, '02-stability-analysis'), **savekw)
        mpl.rcParams.update(rc_paper)

    def phaser_coherence(self):
        """Phaser cell spatial coherence analysis plots."""
        plt.ioff()
        path = self.mkdir('07-phaser-cells')
        ana = PhaserStabilityAnalysis.load(CoherencePath)
        ana.set_figfmt('pdf')

        # Plot parameters
        nbins = 16
        minpixels = 5
        cohmax = 0.75
        figsize = (5.0, 2.4)

        # Example cell figure
        mpl.rcParams['figure.subplot.wspace'] = 0.12
        ana.example_phase_coherence(cohmax=cohmax, example_id=97, CI_alpha=0.1,
                nbins=nbins, minpixels=minpixels, figsize=figsize)
        ana.savefig(basepath=opj(path, '03-coherence-example'), **savekw)
        subprocess.call(['cp', ana.path('phasemap-early.png'),
            opj(path, '03-phasemap-early.png')])
        subprocess.call(['cp', ana.path('phasemap-late.png'),
            opj(path, '03-phasemap-late.png')])
        subprocess.call(['cp', ana.path('coherence-full.png'),
            opj(path, '03-coherence-full.png')])
        mpl.rcParams.update(rc_paper)

        # Phaser cell population figure
        mpl.rcParams['figure.subplot.wspace'] = 0.12
        params = dict(cohmax=cohmax, cohbins=64, diffbins=64,
                cmap='cubehelix_r', report_thresh=np.pi/4, figsize=figsize)
        ana.analyze_phase_coherence(cbar=False, **params)
        ana.savefig(basepath=opj(path, '04-coherence-summary'), **savekw)
        mpl.rcParams.update(rc_paper)

        mpl.rcParams['figure.subplot.wspace'] = 0.12
        ana.analyze_phase_coherence(cbar=True, **params)
        ana.savefig(basepath=opj(path, '05-coherence-summary-cbar'), **savekw)

        mpl.rcParams.update(rc_paper)

    def phaser_cells(self, runsuppl=True):
        """Phaser cell statistics description and distribution plots."""
        plt.ioff()
        path = self.mkdir('07-phaser-cells')
        suppl = opj(path, 'supplemental')
        phaser = PhaserCellDescription.load(PhaserCellPath)
        phaser.set_figfmt('pdf')

        mpl.rcParams['axes.labelpad'] = 0.0
        mpl.rcParams['figure.subplot.wspace'] = 0.55

        # Rate/phase regression and spatial phase information figure
        phaser.phase_shift_distros(poscolor=poscolor, negcolor=negcolor,
                phasemap_avg=True, phase_dist_bins=36, figsize=(4.2, 3.2))
        phaser.savefig(basepath=opj(path, '00-shift-phase-distros'), **savekw)

        # Phase shift x subtype x place-position error figure
        mpl.rcParams['figure.subplot.wspace'] = 0.2
        phaser.phase_shift_accuracy(poscolor=poscolor, negcolor=negcolor,
            shiftbins=32, model_pos_shift=0.654152, model_neg_shift=-2.441130,
            figsize=(4.2, 3.2))
        phaser.savefig(basepath=opj(path, '01-phase-error'), **savekw)

        # Phase shift x subtype x spatial phase coherence figure
        mpl.rcParams['figure.subplot.wspace'] = 0.2
        phaser.spatial_coherence(CI_alpha=0.1, poscolor=poscolor,
            negcolor=negcolor, noncolor=noncolor, show_non=False,
            figsize=(4.2, 3.2))
        phaser.savefig(basepath=opj(path, '02-spatial-coherence'), **savekw)
        mpl.rcParams.update(rc_paper)

        if runsuppl:
            # Firing rate and rhythmicity distros for (non)phaser cells
            phaser.rate_rhythm_comparison(phcolor=phcolor, sigcolor=sigcolor,
                    figsize=(4.5, 3.4))
            phaser.savefig(basepath=opj(suppl, '00-rate-rhythm-distros'),
                    **savekw)

            # Scatter plots comparing phase information with rate, speed, dir.
            mpl.rcParams['mathtext.default'] = 'it'
            mpl.rcParams['figure.subplot.wspace'] = 0.26
            phaser.phase_information_comparison(ymax=1.6, poscolor=poscolor,
                    negcolor=negcolor, figsize=(full_width, 3.7))
            phaser.savefig(basepath=opj(suppl, '01-info-comparisons'),
                    **savekw)
            mpl.rcParams.update(rc_paper)

            # Speed and direction modulation histograms
            mpl.rcParams['figure.subplot.wspace'] = 0.20
            phaser.speed_direction_modulation(bins=24, poscolor=poscolor,
                    negcolor=negcolor, figsize=(5.8, 3.0))
            phaser.savefig(basepath=opj(suppl, '02-speed-dir-modulation'),
                    **savekw)
            mpl.rcParams.update(rc_paper)

            # Anatomical and recording site distribution bar plots
            phaser.anatomical_distribution(poscolor=poscolor,
                    negcolor=negcolor, figsize=(2.8, 3.0))
            phaser.recording_site_distribution(area='septum',
                    poscolor=poscolor, negcolor=negcolor, figsize=(4.4, 3.0))
            phaser.savefig(label='anatomical-distro',
                    basepath=opj(suppl, '03-anatomical-distro'), **savekw)
            phaser.savefig(label='site-distro-septum',
                    basepath=opj(suppl, '04-septum-distro'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def phaser_examples(self, res=256):
        """Phaser cell examples for main figures."""
        plt.ioff()
        path = self.mkdir('08-phaser-examples')
        examples = PhaserExamples.load(PhaserExPath)
        examples.set_figfmt('pdf')

        # Temporay rc settings
        mpl.rcParams.update(rc_paper)
        mpl.rcParams['figure.subplot.hspace'] = 0.08
        mpl.rcParams['figure.subplot.wspace'] = 0.1
        fsz = (full_width, 3.6)

        # Negative phaser examples
        # Ids: 97, 137, 288, 289, 444, 449, 530, 581, 626, 768, 799
        examples.plot_examples(
                cell_list=[137,478,444,768,626],
                levels=12, mapres=res, figsize=fsz)
        examples.savefig(basepath=opj(path, '00-negative-examples'), **savekw)

        # Positive phaser examples
        # Ids: 41, 52, 221, 708, 793, 857, 858, 906
        mpl.rcParams.update(rc_paper)
        mpl.rcParams['figure.subplot.hspace'] = 0.08
        mpl.rcParams['figure.subplot.wspace'] = 0.1
        examples.plot_examples(
                cell_list=[41,52,708,793,906],
                levels=12, mapres=res, figsize=fsz)
        examples.savefig(basepath=opj(path, '01-positive-examples'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    #
    # GLM spatial model and analysis
    #

    def glm_optimization(self):
        """Supplemental figures for fitting the GLM spatial model."""
        plt.ioff()
        path = self.mkdir('09-glm-fit')
        suppl = opj(path, 'supplemental')
        model = TrajectoryIndependence.load(TrajModelPath)
        grid = TrajectoryIndependence.load(TrajGridPath)
        model.set_figfmt('mpl')
        grid.set_figfmt('mpl')

        # Likelihood and shrinkage curves: nongrid pure spatial model
        kw = dict(alpha_logscale=True, figsize=(2.4, col_height))
        model.plot_likelihood_and_shrinkage_curves(grid=1, **kw)

        # Likelihood and shrinkage curves: grid models
        grid.plot_likelihood_and_shrinkage_curves(grid=2, **kw)
        grid.plot_likelihood_and_shrinkage_curves(grid=3, **kw)

        # Save the figures
        grid.savefig(label='optimizer-2',
                basepath=opj(suppl, '00-grid2-shrinkage'), **savekw)
        grid.savefig(label='optimizer-3',
                basepath=opj(suppl, '01-grid3-shrinkage'), **savekw)
        model.savefig(label='optimizer-1',
                basepath=opj(suppl, '02-model-shrinkage'), **savekw)

    def glm_metrics(self):
        """Supplemental figure for scatter plots of independence metrics."""
        plt.ioff()
        path = self.mkdir('10-glm-metrics')
        suppl = opj(path, 'supplemental')
        traj = TrajectoryIndependence.load(TrajGridPath)
        traj.set_figfmt('mpl')

        mpl.rcParams['figure.subplot.wspace'] = 0.14
        traj.plot_independence_metrics(grid=3, poscolor=poscolor,
                negcolor=negcolor, noncolor=noncolor, figsize=(6.5, 4.0))
        traj.savefig(basepath=opj(suppl, '00-directional-metrics'), **savekw)
        mpl.rcParams.update(rc_paper)

    def glm_analysis(self, runsuppl=True):
        """Main and supplemental figures for GLM directional analysis."""
        plt.ioff()
        path = self.mkdir('11-glm-analysis')
        suppl = opj(path, 'supplemental')
        traj = TrajectoryIndependence.load(TrajGridPath)
        traj.set_figfmt('pdf')
        mpl.rcParams.update(rc_paper)

        # Figure for model coefficients and contributions statistics
        fsz = (col_width, 2.6)
        traj.plot_model_statistics(grid=3, phcolor=phcolor, noncolor=noncolor,
                figsize=fsz)
        traj.savefig(basepath=opj(path, '00-model-stats'), **savekw)
        mpl.rcParams.update(rc_paper)

        traj.plot_contribution_statistics(grid=3, phcolor=phcolor,
                noncolor=noncolor, figsize=fsz)
        traj.savefig(basepath=opj(path, '00-contrib-stats'), **savekw)
        mpl.rcParams.update(rc_paper)

        if runsuppl:
            # Supplemental figure for raw and grid-average model matrixes
            figsz = (col_width, 2.6)
            traj.plot_model_grid_matrix(grid=3, subtype='phaser',
                    figsize=figsz, figsize_combined=figsz)
            mpl.rcParams.update(rc_paper)
            traj.plot_contribution_grid_matrix(grid=3, subtype='phaser',
                    figsize=figsz, figsize_combined=figsz)
            mpl.rcParams.update(rc_paper)

            # Plot negaetive and positive subtype contributions separately
            traj.plot_contribution_grid_matrix(grid=3, subtype='negative',
                    figsize=figsz, figsize_combined=figsz)
            mpl.rcParams.update(rc_paper)
            traj.plot_contribution_grid_matrix(grid=3, subtype='positive',
                    figsize=figsz, figsize_combined=figsz)
            mpl.rcParams.update(rc_paper)

            traj.savefig(label='model-phaser-3',
                    basepath=opj(suppl, '00-model'), **savekw)
            traj.savefig(label='model-phaser-3-combined',
                    basepath=opj(suppl, '00-model-avg'), **savekw)

            traj.savefig(label='contribution-phaser-3',
                    basepath=opj(suppl, '01-contrib'), **savekw)
            traj.savefig(label='contribution-phaser-3-combined',
                    basepath=opj(suppl, '01-contrib-avg'), **savekw)

            traj.savefig(label='contribution-negative-3',
                    basepath=opj(suppl, '02-negative'), **savekw)
            traj.savefig(label='contribution-negative-3-combined',
                    basepath=opj(suppl, '02-negative-avg'), **savekw)
            traj.savefig(label='contribution-positive-3',
                    basepath=opj(suppl, '03-positive'), **savekw)
            traj.savefig(label='contribution-positive-3-combined',
                    basepath=opj(suppl, '03-positive-avg'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def glm_reconstructions(self):
        """Main and supplemental figures for GLM predicted ratemaps."""
        plt.ioff()
        path = self.mkdir('12-glm-reconstructions')
        suppl = opj(path, 'supplemental')
        traj = TrajectoryIndependence.load(TrajGridPath)
        traj.set_figfmt('mpl')

        # Temporay rc settings
        mpl.rcParams['figure.subplot.hspace'] = 0.30
        mpl.rcParams['figure.subplot.wspace'] = 0.09
        fsz = (6.7, 3.6)

        # Main phaser examples and their 3x3 grid reconstructions
        traj.plot_reconstructed_examples(cell_list=[137,478,444,768,626],
                grid=3, levels=12, figsize=fsz)
        traj.savefig(basepath=opj(path, '00-recon-examples'), **savekw)

        # Supplemental figure examples
        traj.plot_reconstructed_examples(cell_list=[434,489,384,629],
                grid=3, levels=12, figsize=(5.3, 3.6))
        traj.savefig(basepath=opj(suppl, '00-coherent-examples'), **savekw)
        traj.plot_reconstructed_examples(cell_list=[83,114,342],
                grid=3, levels=12, figsize=(4.2, 3.6))
        traj.savefig(basepath=opj(suppl, '00-noncoherent-examples'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    #
    # Spatial 2D phaser model
    #

    def spatial_input(self):
        """Spatial input from GLM model normalized for simulation."""
        plt.ioff()
        path = self.mkdir('13-spatial-input')
        inp = SpatialInputModel.load(SpatialInputPath)
        inp.set_figfmt('mpl')

        # Temporay rc settings
        mpl.rcParams['figure.subplot.hspace'] = 0.30
        mpl.rcParams['figure.subplot.wspace'] = 0.09
        fsz = (6.7, 3.6)

        inp.plot_spatial_model_validation(cell_list=[444, 626], bins=128,
                rotate=True, figsize=fsz)
        inp.savefig(basepath=opj(path, '00-example-validation'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def spatial_layers(self):
        """Spatial input layers for the simulation: ratemaps and phasemaps."""
        plt.ioff()
        path = self.mkdir('13-spatial-input')
        suppl = opj(path, 'supplemental')
        stripe = StripeSyncSimulation.load(StripeSimPath)
        stripe.set_figfmt('mpl')

        # Temporary rc settings
        mpl.rcParams['figure.subplot.wspace'] = 0.1
        mpl.rcParams['figure.subplot.hspace'] = 0.1

        fsz = (6.5, 6.5)
        stripe.plot_training_simulation(maxN=50, mapres=112, levels=12,
                figsize=fsz)
        stripe.savefig(label='training-negative-layer',
                basepath=opj(suppl, '00-negative-layer'), **savekw)
        stripe.savefig(label='training-positive-layer',
                basepath=opj(suppl, '01-positive-layer'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def spatial_simulations(self):
        """Target burster synchronization simulations with spatial phasers."""
        plt.ioff()
        path = self.mkdir('14-spatial-simulations')
        stripe = StripeSyncSimulation.load(StripeSimPath)
        stripe.set_figfmt('mpl')

        # Temporary rc settings
        mpl.rcParams['figure.subplot.wspace'] = 0.14
        mpl.rcParams['figure.subplot.hspace'] = 0.12

        fsz = (full_width, 3.6)
        stripe.plot_vco_example_simulations(tags=('theta0', 'theta1'),
                mapres=256, levels=12, ibi_min=IBI_MIN, figsize=fsz)
        stripe.savefig(basepath=opj(path, '00-spatial-sync'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

    def decoding_analysis(self, tracedata=None):
        """Bayesian phase decoding of burst timing from VCO simulations."""
        plt.ioff()
        path = self.mkdir('15-decoding-analysis')
        stripe = StripeSyncSimulation.load(StripeSimPath)
        stripe.set_figfmt('pdf')
        panelh = 2.6
        Nvco = min(stripe.c.phase_movie_nframes, stripe.c.dir_movie_nframes)

        # Network diagram
        stripe.plot_network_parameters(figsize=(panelh,panelh))
        stripe.savefig(basepath=opj(path, '00-network-params'), **savekw)

        # Load arena-masked VCO targets and save a grid image
        cmap = huslmap()
        imgfmt = dict(cmap=cmap, cmin=-np.pi, cmax=np.pi, mask_color='w')
        Pphase = np.array([stripe.read_array('vco/phasemap_phase%03d' % i)
                            for i in range(Nvco)])
        Pdir1 = np.array([stripe.read_array('vco/phasemap_dir%03d' % i)
                            for i in range(Nvco)])
        Pdir2 = np.array([stripe.read_array('vco/phasemap_altdir%03d' % i)
                            for i in range(Nvco)])
        self.out('Saving phase network maps...')
        stack_to_image(Pphase, opj(path, '01-phase-network.png'), **imgfmt)
        self.out('Saving dir1 network maps...')
        stack_to_image(Pdir1, opj(path, '01-dir1-network.png'), **imgfmt)
        self.out('Saving dir2 network maps...')
        stack_to_image(Pdir2, opj(path, '01-dir2-network.png'), **imgfmt)

        # Show sample trace decoding plots
        if tracedata is None:
            tracedata = stripe.compute_decoding_traces(interval=[108,114])
        stripe.plot_decoding_traces(tracedata, figsize=(full_width, panelh))
        stripe.savefig(basepath=opj(path, '02-decoding-examples'),
                dpi=600, format='png', **savekw)

        # Plot decoding performance by network size
        stripe.plot_size_performance(CIalpha=None, maxshift=0.0,
                figsize=(col_width, panelh))
        stripe.savefig(basepath=opj(path, '03-decoding-by-size'), **savekw)

        # Plot autocorrelation analysis of error-correction timescale
        stripe.plot_correction_timescale(maxlag=6.0, corr_CIalpha=None,
            summary_CIalpha=0.05, figsize=(3.3*panelh, panelh))
        stripe.savefig(basepath=opj(path, '04-error-correction'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_paper)

        return tracedata
