"""
Collection of all analysis and modeling data plots for poster figures.
"""

import os
from os.path import join as opj

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

import roto.axis as ra
from roto.misc import merge_two_dicts

from .ana.sync_model import BurstSyncSimulation
from .ana.phaser_model import PhaserSimulation
from .ana.data_intro_example import IntroExample
from .ana.theta_cells import ThetaCellDescription
from .ana.phaser_cells import PhaserCellDescription
from .ana.phaser_examples import PhaserExamples
from .ana.trajectory_independence import TrajectoryIndependence
from .ana.spatial_input import SpatialInputModel
from .ana.stripe_model import StripeSyncSimulation

from . import RES_ROOT as ResultsPath, store
from .context import SPCAnalysis


# Paths to all analyses used for plotting figures
BurstSyncPath = opj(ResultsPath, '2017-08-03+burst-sync-simulation+fixed')
PhaserDemoPath = opj(ResultsPath, '2017-10-18+phaser-simulation+sampling')
DataExamplePath = opj(ResultsPath, '2017-08-29+intro-example')
ThetaCellPath = opj(ResultsPath, '2017-08-29+theta-cell-description')
PhaserCellPath = opj(ResultsPath, '2017-08-30+phaser-cell-description')
PhaserExPath = opj(ResultsPath, '2017-08-31+phaser-examples')
TrajGridPath = opj(ResultsPath,
        '2017-03-18+trajectory-independence+ridge-alpha-1.2496')
TrajModelPath = opj(ResultsPath,
        '2017-04-15+trajectory-independence+pure-spatial-model')
SpatialInputPath = opj(ResultsPath, '2017-04-20+spatial-input-model')
StripeSimPath = opj(ResultsPath, '2017-09-05+stripe-sync-simulation')

# Matplotlib settings for publication figures
axes_lw = 2.2  # 0.8
rc_paper = {
        'savefig.format': 'pdf',
        'savefig.frameon': False,
        'savefig.transparent': True,
        'font.size': 9.0,
        'font.weight': 'normal',
        'font.sans-serif': ['Helvetica Neue LT Std', 'Lucida Grande', 'Geneva',
                            'Verdana', 'Bitstream Vera Sans', 'sans-serif'],
        'pdf.fonttype': 'truetype',
        'mathtext.default': 'rm',
        'figure.titlesize': 'large',
        'figure.titleweight': 'light',
        'figure.subplot.wspace': 0.23,
        'figure.subplot.hspace': 0.23,
        'axes.titlesize': 8.0,
        'axes.titleweight': 'light',
        'axes.labelsize': 8.0,
        'axes.linewidth': axes_lw,
        'axes.labelpad': 1.6,
        'lines.solid_capstyle': 'round',
        'lines.linewidth': 1.2,
        'lines.markeredgewidth': 0.0,
        'lines.markersize': 5.6,
        'patch.linewidth': 0.24,
        'xtick.major.width': axes_lw,
        'ytick.major.width': axes_lw,
        'xtick.minor.width': axes_lw/2,
        'ytick.minor.width': axes_lw/2,
        'xtick.major.pad': 1.3,
        'ytick.major.pad': 1.3,
        'xtick.labelsize': 7.2,
        'ytick.labelsize': 7.2,
        'legend.fontsize': 8.0,
        'legend.labelspacing': 0.25
    }
rc_png = {
        'savefig.format': 'png',
        'savefig.frameon': True,
        'savefig.transparent': False,
        'savefig.dpi': 600
    }
rc_poster_context = {
        'axes.labelsize': 17.6,
        'axes.titlesize': 19.2,
        'axes.labelpad': 2.2,  # 1.6,
        'font.size': 19.2,
        'grid.linewidth': 1.6,
        'legend.fontsize': 16.0,
        'legend.labelspacing': 0.4,  # 0.25
        'lines.linewidth': 2.8,
        'lines.markeredgewidth': 0.0,
        'lines.markersize': 14.6,
        'patch.linewidth': 0.8,
        'xtick.labelsize': 16.0,
        'xtick.major.size': 7.2,
        'xtick.major.pad': 2.4,
        'xtick.major.width': 1.6,
        'xtick.minor.width': 0.8,
        'ytick.labelsize': 16.0,
        'ytick.major.size': 7.2,
        'ytick.major.pad': 2.4,
        'ytick.major.width': 1.6,
        'ytick.minor.width': 0.8
    }

# Merge poster parameters into rc settings
rc_poster = merge_two_dicts(rc_paper, rc_poster_context)
mpl.rc_file_defaults()
mpl.rcParams.update(rc_poster)

# Figure size factor increase (Poster X-factor)
PX = 2.0

# Seaborn settings
sns.set_palette('colorblind', color_codes=True)

# Color constants
nonsigcolor = (0.88470588, 0.55803921, 0.3)  # lightened colorblind 'r'
sigcolor = (0.3, 0.61294117, 0.78862745)  # lightened colorblind 'b'
phcolor = (0.3, 0.73372549, 0.61568627)  # lightened colorblind 'g'
poscolor = 'sandybrown'
negcolor = 'slateblue'
noncolor = 'dimgray'

# Savefig parameters
savekw = dict(tight_padding=0.01, closeafter=True)

# Standard reference figure dimensions
col_width = 3.4*PX  # inches
full_width = 7.2*PX
col_height = 8.2*PX
half_height = 0.5*col_height
figsize_hh = (col_width, half_height)
figsize_hf = (col_width, col_height)
figsize_fh = (full_width, half_height)
figsize_ff = (full_width, col_height)

# Common analysis parameters
IBI_MIN = 0.025     # seconds, inter-burst interval threshold


class PosterFigures(SPCAnalysis):

    def cleanup(self):
        store.close()

    def run_all_figures(self):
        """Script method to run all figure group methods."""
        # Initial simulations
        self.phaser_demo()
        self.phaser1d_distros()
        self.phaser1d_training()
        self.burster_model()
        self.phaser1d_simulations()

        # Theta cell data analysis
        self.intro_example()
        self.theta_cells()
        self.phaser_cells()
        self.phaser_examples()

        # GLM statistical model
        self.glm_optimization()
        self.glm_metrics()
        self.glm_analysis()
        self.glm_reconstructions()

        # Spatial phaser simulations
        self.spatial_input()
        self.spatial_layers()
        self.spatial_simulations()

    #
    # Initial models and simulations
    #

    def phaser_demo(self):
        """Phaser bursting model demonstration."""
        plt.ioff()
        path = self.mkdir('00-phaser-demo')
        phaser = PhaserSimulation.load(PhaserDemoPath)
        phaser.set_figfmt('mpl')

        # Phaser demonstration figure
        savekw2 = merge_two_dicts(savekw, {'tight_padding': 0.0})
        mpl.rcParams['lines.linewidth'] = 2.2  # 0.7
        mpl.rcParams['axes.labelpad'] = 0.8
        mpl.rcParams['mathtext.default'] = 'it'
        phaser.plot_test_phaser_simulation(interval_dur=0.7, negcolor=negcolor,
                poscolor=poscolor, figsize=(full_width, 4.55*PX))
        phaser.savefig(basepath=opj(path, '00-phaser-demo'), **savekw2)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def phaser1d_distros(self):
        """1D phaser model examples and supplemental figures."""
        plt.ioff()
        path = self.mkdir('01-phaser1d-distros')
        suppl = opj(path, 'supplemental')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Space-phase 1D phaser distros: example and supplemental figures
        mpl.rcParams['figure.subplot.hspace'] = 0.12
        mpl.rcParams['figure.subplot.wspace'] = 0.12
        burst.plot_space_phase_distros(example_x=0.5,
                figsize=(col_width, col_width),
                full_figsize=(full_width, full_width))
        burst.savefig('example-phase-distros',
            basepath=opj(path, '00-example-distros'), **savekw)
        burst.savefig('neg-phase-distros',
            basepath=opj(suppl, '00-neg-distros'), **savekw)
        burst.savefig('neg-inv-phase-distros',
            basepath=opj(suppl, '00-neg-inv-distros'), **savekw)
        burst.savefig('pos-phase-distros',
            basepath=opj(suppl, '00-pos-distros'), **savekw)
        burst.savefig('pos-inv-phase-distros',
            basepath=opj(suppl, '00-pos-inv-distros'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def phaser1d_training(self):
        """1D phaser model weights and phase-code targets."""
        plt.ioff()
        path = self.mkdir('02-phaser1d-weights')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Phaser 1D simulation phase-code target, weights, and weighted input
        mpl.rcParams['legend.fontsize'] = 14.0  # 7.2
        mpl.rcParams['legend.frameon'] = False
        burst.plot_target_weights_inputs(poscolor=poscolor, negcolor=negcolor,
                figsize=figsize_hh)
        burst.savefig(basepath=opj(path, '00-weighted-input'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def burster_model(self):
        """Output burster model figures: voltage trace and pulse sync."""
        plt.ioff()
        path = self.mkdir('03-burster-model')
        suppl = opj(path, 'supplemental')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Intro burst synchronization figure
        mpl.rcParams['figure.subplot.hspace'] = 0.18
        mpl.rcParams['lines.linewidth'] = 2.2
        burst.plot_demo_burst_synchronization(interval=(0,1.0), ibi_min=IBI_MIN,
                figsize=figsize_fh)
        burst.savefig(basepath=opj(path, '00-burst-sync'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def phaser1d_simulations(self):
        """1D phaser model simulations: burst timing across space."""
        plt.ioff()
        path = self.mkdir('04-phaser1d-simulations')
        suppl = opj(path, 'supplemental')
        burst = BurstSyncSimulation.load(BurstSyncPath)
        burst.set_figfmt('mpl')

        # Phaser 1D simulation output figures: burst timing across space
        mpl.rcParams.update(rc_png)
        mpl.rcParams['figure.subplot.hspace'] = 0.19
        mpl.rcParams['figure.subplot.wspace'] = 0.12
        burst.plot_phaser_burster_sync_simulation(ibi_min=IBI_MIN,
                sigma=(0.0, 0.3), fdbk=(1.0, 2.0),
                figsize=figsize_hh)
        burst.savefig(basepath=opj(path, '00-phaser-sync-sim'), **savekw)

        # Supplemental figures: Varying noise level or feedback level
        burst.plot_supplemental_sync_simulations(ibi_min=IBI_MIN,
                sigma_list=[0.0, 0.1, 0.3, 0.5],
                fdbk_list=[(0.0,0.0), #(0.06,0.12), (0.12,0.25), (0.25,0.5),
                           (0.5,1.0), (1.0,2.0), (2.0,4.0)],
                fdbk_sigma=0.3,
                figsize=figsize_fh)
        burst.savefig('feedback-sync-noise',
                basepath=opj(suppl, '00-feedback-noise'), **savekw)
        burst.savefig('feedback-sync-gain',
                basepath=opj(suppl, '00-feedback-gain'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    #
    # Theta cell data and analysis
    #

    def intro_example(self, run_setup=False):
        """Intro example of spatial theta cell from dataset."""
        plt.ioff()
        path = self.mkdir('05-data-example')
        intro = IntroExample.load(DataExamplePath)
        intro.set_figfmt('mpl')

        # Set up data for example cell
        if run_setup:
            example_cell_id = 97
            intro.setup(example_cell_id)
            intro.find_example_LFP_traces()

        # Plot/save the figure
        mpl.rcParams.update(rc_png)
        mpl.rcParams['figure.subplot.hspace'] = 0.1
        mpl.rcParams['axes.labelpad'] = 1.4  # 1.2
        intro.plot_figure(levels=12, figsize=(full_width, 3.7*PX))
        intro.savefig(basepath=opj(path, '00-cell-%04d' % intro.c.cell_id),
                **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def theta_cells(self, runsuppl=True):
        """Theta cell statistics description and distribution plots."""
        plt.ioff()
        path = self.mkdir('06-theta-cells')
        suppl = opj(path, 'supplemental')
        theta = ThetaCellDescription.load(ThetaCellPath)
        theta.set_figfmt('mpl')

        # Rate/phase regression and spatial phase information figure
        mpl.rcParams['axes.labelpad'] = 0.0
        theta.phase_coding_scatter_plot(nonsigcolor=nonsigcolor,
                sigcolor=sigcolor, figsize=(3.0*PX, 3.2*PX))
        theta.savefig(basepath=opj(path, '00-phase-corr-info'), **savekw)
        mpl.rcParams['axes.labelpad'] = rc_poster['axes.labelpad']

        if runsuppl:
            # Max. firing rate and theta frequency, modulation, rhythmicity
            theta.maxrate_thetamod_distros(nonsigcolor=nonsigcolor,
                sigcolor=sigcolor, figsize=figsize_fh)
            theta.savefig(basepath=opj(suppl, '00-rate-theta-distros'),
                    **savekw)

            # Spatial phase-coding distribution figures
            theta.phase_coding_distros(nonsigcolor=nonsigcolor,
                sigcolor=sigcolor, figsize=figsize_fh)
            theta.savefig(basepath=opj(suppl, '01-spatial-phase-distros'),
                    **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def phaser_cells(self, runsuppl=True):
        """Phaser cell statistics description and distribution plots."""
        plt.ioff()
        path = self.mkdir('07-phaser-cells')
        suppl = opj(path, 'supplemental')
        phaser = PhaserCellDescription.load(PhaserCellPath)
        phaser.set_figfmt('mpl')

        # Rate/phase regression and spatial phase information figure
        mpl.rcParams['axes.labelpad'] = 0.0
        mpl.rcParams['figure.subplot.wspace'] = 0.55
        phaser.phase_shift_distros(poscolor=poscolor, negcolor=negcolor,
                figsize=(4.2*PX, 3.2*PX))
        phaser.savefig(basepath=opj(path, '00-shift-phase-distros'), **savekw)
        mpl.rcParams.update(rc_poster)

        if runsuppl:
            # Firing rate and rhythmicity distros for (non)phaser cells
            phaser.rate_rhythm_comparison(phcolor=phcolor, sigcolor=sigcolor,
                    figsize=(4.5*PX, 3.4*PX))
            phaser.savefig(basepath=opj(suppl, '00-rate-rhythm-distros'),
                    **savekw)

            # Scatter plots comparing phase information with rate, speed, dir.
            mpl.rcParams['mathtext.default'] = 'it'
            mpl.rcParams['figure.subplot.wspace'] = 0.26
            phaser.phase_information_comparison(ymax=1.6, poscolor=poscolor,
                    negcolor=negcolor, figsize=(full_width, 3.7*PX))
            phaser.savefig(basepath=opj(suppl, '01-info-comparisons'),
                    **savekw)
            mpl.rcParams.update(rc_poster)

            # Speed and direction modulation histograms
            mpl.rcParams['figure.subplot.wspace'] = 0.20
            phaser.speed_direction_modulation(bins=24, poscolor=poscolor,
                    negcolor=negcolor, figsize=(5.8*PX, 3.0*PX))
            phaser.savefig(basepath=opj(suppl, '02-speed-dir-modulation'),
                    **savekw)
            mpl.rcParams.update(rc_poster)

            # Anatomical and recording site distribution bar plots
            phaser.anatomical_distribution(poscolor=poscolor,
                    negcolor=negcolor, figsize=(2.8*PX, 3.0*PX))
            phaser.recording_site_distribution(area='septum',
                    poscolor=poscolor, negcolor=negcolor, figsize=(4.4*PX, 3.0*PX))
            phaser.savefig(label='anatomical-distro',
                    basepath=opj(suppl, '03-anatomical-distro'), **savekw)
            phaser.savefig(label='site-distro-septum',
                    basepath=opj(suppl, '04-septum-distro'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def phaser_examples(self):
        """Phaser cell examples for main figures."""
        plt.ioff()
        path = self.mkdir('08-phaser-examples')
        examples = PhaserExamples.load(PhaserExPath)
        examples.set_figfmt('mpl')

        # Temporay rc settings
        mpl.rcParams['figure.subplot.hspace'] = 0.08
        mpl.rcParams['figure.subplot.wspace'] = 0.1
        fsz = (full_width, 3.6*PX)

        # Negative phaser examples
        # Ids: 97, 137, 288, 289, 444, 449, 530, 581, 626, 768, 799
        examples.plot_examples(
                cell_list=[137,478,444,768,626],
                levels=12, figsize=fsz)
        examples.savefig(basepath=opj(path, '00-negative-examples'), **savekw)

        # Positive phaser examples
        # Ids: 41, 52, 221, 708, 793, 857, 858, 906
        examples.plot_examples(
                cell_list=[41,52,708,793,906],
                levels=12, figsize=fsz)
        examples.savefig(basepath=opj(path, '01-positive-examples'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    #
    # GLM spatial model and analysis
    #

    def glm_optimization(self):
        """Supplemental figures for fitting the GLM spatial model."""
        plt.ioff()
        path = self.mkdir('09-glm-fit')
        suppl = opj(path, 'supplemental')
        model = TrajectoryIndependence.load(TrajModelPath)
        grid = TrajectoryIndependence.load(TrajGridPath)
        model.set_figfmt('mpl')
        grid.set_figfmt('mpl')

        # Likelihood and shrinkage curves: nongrid pure spatial model
        kw = dict(alpha_logscale=True, figsize=(2.4*PX, col_height))
        model.plot_likelihood_and_shrinkage_curves(grid=1, **kw)

        # Likelihood and shrinkage curves: grid models
        grid.plot_likelihood_and_shrinkage_curves(grid=2, **kw)
        grid.plot_likelihood_and_shrinkage_curves(grid=3, **kw)

        # Save the figures
        grid.savefig(label='optimizer-2',
                basepath=opj(suppl, '00-grid2-shrinkage'), **savekw)
        grid.savefig(label='optimizer-3',
                basepath=opj(suppl, '01-grid3-shrinkage'), **savekw)
        model.savefig(label='optimizer-1',
                basepath=opj(suppl, '02-model-shrinkage'), **savekw)

    def glm_metrics(self):
        """Supplemental figure for scatter plots of independence metrics."""
        plt.ioff()
        path = self.mkdir('10-glm-metrics')
        suppl = opj(path, 'supplemental')
        traj = TrajectoryIndependence.load(TrajGridPath)
        traj.set_figfmt('mpl')

        mpl.rcParams['figure.subplot.wspace'] = 0.14
        traj.plot_independence_metrics(grid=3, poscolor=poscolor,
                negcolor=negcolor, noncolor=noncolor, figsize=(6.5*PX, 4.0*PX))
        traj.savefig(basepath=opj(suppl, '00-directional-metrics'), **savekw)
        mpl.rcParams.update(rc_poster)

    def glm_analysis(self, runsuppl=True):
        """Main and supplemental figures for GLM directional analysis."""
        plt.ioff()
        path = self.mkdir('11-glm-analysis')
        suppl = opj(path, 'supplemental')
        traj = TrajectoryIndependence.load(TrajGridPath)
        traj.set_figfmt('mpl')

        # Figure for model coefficients and contributions statistics
        fsz = (col_width, 2.6*PX)
        traj.plot_model_statistics(grid=3, phcolor=phcolor, noncolor=noncolor,
                figsize=fsz)
        traj.savefig(basepath=opj(path, '00-model-stats'), **savekw)
        traj.plot_contribution_statistics(grid=3, phcolor=phcolor,
                noncolor=noncolor, figsize=fsz)
        traj.savefig(basepath=opj(path, '00-contrib-stats'), **savekw)

        if runsuppl:
            # Supplemental figure for raw and grid-average model matrixes
            figsz = (col_width, 2.6*PX)
            traj.plot_model_grid_matrix(grid=3, subtype='phaser',
                    figsize=figsz, figsize_combined=figsz)
            traj.plot_contribution_grid_matrix(grid=3, subtype='phaser',
                    figsize=figsz, figsize_combined=figsz)

            traj.savefig(label='model-phaser-3',
                    basepath=opj(suppl, '00-model'), **savekw)
            traj.savefig(label='model-phaser-3-combined',
                    basepath=opj(suppl, '00-model-avg'), **savekw)
            traj.savefig(label='contribution-phaser-3',
                    basepath=opj(suppl, '00-contrib'), **savekw)
            traj.savefig(label='contribution-phaser-3-combined',
                    basepath=opj(suppl, '00-contrib-avg'), **savekw)

    def glm_reconstructions(self):
        """Main and supplemental figures for GLM predicted ratemaps."""
        plt.ioff()
        path = self.mkdir('12-glm-reconstructions')
        suppl = opj(path, 'supplemental')
        traj = TrajectoryIndependence.load(TrajGridPath)
        traj.set_figfmt('mpl')

        # Temporay rc settings
        mpl.rcParams['figure.subplot.hspace'] = 0.30
        mpl.rcParams['figure.subplot.wspace'] = 0.09
        fsz = (6.7*PX, 3.6*PX)

        # Main phaser examples and their 3x3 grid reconstructions
        traj.plot_reconstructed_examples(cell_list=[137,478,444,768,626],
                grid=3, levels=12, figsize=fsz)
        traj.savefig(basepath=opj(path, '00-recon-examples'), **savekw)

        # Supplemental figure examples
        traj.plot_reconstructed_examples(cell_list=[434,489,384,629],
                grid=3, levels=12, figsize=(5.3*PX, 3.6*PX))
        traj.savefig(basepath=opj(suppl, '00-coherent-examples'), **savekw)
        traj.plot_reconstructed_examples(cell_list=[83,114,342],
                grid=3, levels=12, figsize=(4.2*PX, 3.6*PX))
        traj.savefig(basepath=opj(suppl, '00-noncoherent-examples'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    #
    # Spatial 2D phaser model
    #

    def spatial_input(self):
        """Spatial input from GLM model normalized for simulation."""
        plt.ioff()
        path = self.mkdir('13-spatial-input')
        inp = SpatialInputModel.load(SpatialInputPath)
        inp.set_figfmt('mpl')

        # Temporay rc settings
        mpl.rcParams['figure.subplot.hspace'] = 0.30
        mpl.rcParams['figure.subplot.wspace'] = 0.09
        fsz = (6.7*PX, 3.6*PX)

        inp.plot_spatial_model_validation(cell_list=[444, 626], bins=128,
                rotate=True, figsize=fsz)
        inp.savefig(basepath=opj(path, '00-example-validation'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def spatial_layers(self):
        """Spatial input layers for the simulation: ratemaps and phasemaps."""
        plt.ioff()
        path = self.mkdir('13-spatial-input')
        suppl = opj(path, 'supplemental')
        stripe = StripeSyncSimulation.load(StripeSimPath)
        stripe.set_figfmt('mpl')

        # Temporary rc settings
        mpl.rcParams['figure.subplot.wspace'] = 0.1
        mpl.rcParams['figure.subplot.hspace'] = 0.1

        fsz = (6.5*PX, 6.5*PX)
        stripe.plot_training_simulation(maxN=50, mapres=150, levels=12,
                figsize=fsz)
        stripe.savefig(label='training-negative-layer',
                basepath=opj(suppl, '00-negative-layer'), **savekw)
        stripe.savefig(label='training-positive-layer',
                basepath=opj(suppl, '01-positive-layer'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)

    def spatial_simulations(self):
        """Target burster synchronization simulations with spatial phasers."""
        plt.ioff()
        path = self.mkdir('14-spatial-simulations')
        stripe = StripeSyncSimulation.load(StripeSimPath)
        stripe.set_figfmt('mpl')

        # Temporary rc settings
        mpl.rcParams['figure.subplot.wspace'] = 0.14
        mpl.rcParams['figure.subplot.hspace'] = 0.12

        fsz = (full_width, 3.6*PX)
        stripe.plot_vco_example_simulations(tags=('theta0', 'theta1'),
                mapres=444, levels=12, ibi_min=IBI_MIN, figsize=fsz)
        stripe.savefig(basepath=opj(path, '00-spatial-sync'), **savekw)

        # Restore rc settings
        mpl.rcParams.update(rc_poster)
