"""
Managed context for data analysis runs.
"""

import re
import os
import pdb
import sys
import time
import json
import inspect
import datetime
import subprocess
from decorator import decorator
from collections import OrderedDict, namedtuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import tables as tb
import pandas as pd
import brian2 as br

from pouty import anybar, ConsolePrinter
from roto import data
from roto.figures import get_svg_figinfo
from roto.paths import uniquify
from roto.strings import snake2title, camel2snake, parse_value
from roto.misc import merge_two_dicts

from . import parallel
from .repo import git_revision
from .store import DataStore


CALL_LOG = 'calls.log'
INIT_FILENAME = 'init.json'
ENV_FILENAME = 'env.json'
LOGDIR = 'logs'


@decorator
def step(f, *args, **kwargs):
    """Declare a subclass method as an analysis step in this context."""
    self = args[0]
    status = { 'OK': False }
    self._step_enter(f, args, kwargs)
    res = self._step_execute(f, args, kwargs, status)
    self._step_exit(f, args, kwargs, status)
    return res


class AttrDict(object):

    def __init__(self, odict=None):
        if odict is None:
            odict = OrderedDict()
        self.__odict__ = odict

    def __getattr__(self, attr):
        if attr not in self.__odict__:
            raise AttributeError("'%s' attribute is not set" % attr)
        return self.__odict__[attr]

    def __setattr__(self, attr, value):
        if attr == '__odict__':
            object.__setattr__(self, attr, value)
            return
        self.__odict__[attr] = value


class AnalysisContext(object):

    """
    Smart context for open and reproducible data analysis.

    Subclass overrides:
    cleanup -- clean-up to perform after each analysis step

    Class methods:
    load -- create a new context object for a previous run

    Instance variables:
    c -- attribute access to the context namespace

    Instance methods:
    mkdir -- make a subdirectory
    reveal -- open the run directory in Finder (OS X)
    out -- send messages to console, logs, notifications, and anybar
    open_logfile -- open a new log file
    use_timestamps -- toggle timestamping for log file messages
    close_logfile -- close the most recent log file
    set_parallel_profile -- set ipython profile for the parallel client
    set_datafile -- change the datafile path
    get_datafile -- open the datafile, creating it if necessary
    flush_datafile -- flush the datafile
    close_datafile -- close the datafile
    create_group -- create a group in the datafile
    create_table -- create a table in the datafile
    save_array -- save a numpy array or matrix to the datafile
    save_simulation -- save Brian simulation output to the datafile
    read_simulation -- load simulation output into DataFrames
    set_static_figures -- set/toggle whether figure windows are reused
    figure -- open a new figure (wrapper for `plt.figure(...)`)
    savefig -- save a figure (default: most recent)
    closefig -- close a figure (default: most recent)
    save_figures -- save all open figure windows
    close_figures -- close all open figure windows
    set_figfmt -- change the image format for saving figures
    """

    rootdir = os.path.join(os.getenv('HOME'), 'tenko')
    repodir = None
    resdir = None
    logcolor = 'purple'

    def __init__(self, desc=None, rootdir=None, rundir=None, resdir=None,
        h5file=None, h5root='/', figfmt='mpl', staticfigs=True,
        logcolor=None, profile=None):
        """Set up the analysis context.

        Keyword arguments:
        desc -- short phrase describing the run
        rundir -- path to folder that will contain run output
        rootdir -- parent folder for all analysis output
        resdir -- parent folder for results
        h5file -- specify an alternate path for the datafile
        h5root -- path to root group for data in the datafile
        figfmt -- default figure format for saving images (mpl is equivalent
                  to using the rcParams['savefig.format'] setting)
        staticfigs -- reuse figure windows for labeled figures
        logcolor -- prefix color for shell log messages (default: purple)
        profile -- ipython profile to use for parallel client
        """
        self._desc = str(desc).strip().lower() if desc is not None else None

        self._set_rootdir_rundir(rootdir, rundir)
        self._set_results_dir(resdir)
        self._logdir = self.mkdir(LOGDIR)

        self._logcolor = logcolor is None and self.logcolor or logcolor
        self._out = ConsolePrinter(prefix=self.__class__.__name__,
            prefix_color=self._logcolor, message_color='lightgray')

        self._load_env()  # inits or loads _env namespace
        self.c = AttrDict(self._env)  # attribute access to namespace

        self._datafile = None
        self._set_datafile(h5file, h5root)

        if profile:
            self.set_parallel_profile(profile)

        self._figures = {}
        self._savefig = {}
        self._lastfig = None
        self._savefig_patha = None
        self._holdfigs = False
        self._staticfigs = staticfigs
        self._figfmt = figfmt

        self._lastcall = None

        self._save()
        self.out('{} initialized:\n{}', self.__class__.__name__, str(self))

    def __str__(self):
        col_w = 13
        s = ['Module:'.ljust(col_w) + self.__class__.__module__]
        s += ['Subclass:'.ljust(col_w) + self.__class__.__name__]
        if self._desc:
            s += ['Description:'.ljust(col_w) + "'{}'".format(self._desc)]
        s += ['Directory:'.ljust(col_w) + str(self._rundir)]
        s += ['Results:'.ljust(col_w) + str(self._resdir)]
        s += ['Datafile:'.ljust(col_w) + str(self._datafile)]
        s += ['Root group:'.ljust(col_w) + str(self._h5root)]
        return '\n'.join(s)

    def _set_rootdir_rundir(self, rootdir, rundir):
        """Create path for run outputs based on description."""
        root = self._rootdir = \
                os.path.abspath(rootdir is None and self.rootdir or rootdir)

        if rundir is not None:
            self._rundir = os.path.abspath(rundir)
            return

        path = camel2snake(self.__class__.__name__, sep='-')
        if self._desc:
            path = os.path.join(path, self._desc.replace(' ', '-'))

        self._rundir = uniquify(os.path.join(root, path), fmt='%s/%02d')

    def _set_results_dir(self, resdir):
        """Set path for storing an analysis context as project results."""
        if resdir is None:
            if self.resdir is None:
                self._resdir = os.path.join(self._rootdir, 'results'),
            else:
                self._resdir = os.path.abspath(self.resdir)
        else:
            self._resdir = os.path.abspath(resdir)

    # Mapping methods

    def __setitem__(self, key, item):
        if isinstance(key, str):
            if not key.isidentifier():
                raise ValueError('key is not a valid name: {}'.format(key))
        else:
            raise ValueError('key is not a string: {}'.format(key))
        self._env[key] = item

    def __getitem__(self, key):
        return self._env[key]

    def __delitem__(self, key):
        del self._env[key]

    def __contains__(self, key):
        return key in self._env

    def __len__(self):
        return len(self._env)

    def __iter__(self):
        return iter(self._env)

    def _save_env(self):
        with open(self.path(ENV_FILENAME), 'w') as f:
            json.dump(self._env, f, skipkeys=True, indent=2, sort_keys=False)

    def _load_env(self):
        sfn = self.path(ENV_FILENAME)
        if not os.path.isfile(sfn):
            self._env = OrderedDict()
            return

        with open(sfn, 'r') as f:
            env = json.load(f)

        self._env = OrderedDict(**env)  # TODO this loses sort

    # Load/save methods

    @classmethod
    def load(cls, instance_or_rundir):
        """Return a new context for a previous run directory."""
        if hasattr(instance_or_rundir, '_rundir'):
            rundir = instance_or_rundir._rundir
        else:
            rundir = str(instance_or_rundir)

        initpath = os.path.join(rundir, INIT_FILENAME)
        out = ConsolePrinter(prefix="{}Loader".format(cls.__name__),
                prefix_color='red')
        if not os.path.isfile(initpath):
            out('Bad path or missing file: {}', initpath, error=True)
            return

        try:
            with open(initpath, 'r') as f:
                initargs = json.load(f)
        except ValueError:
            out('Invalid JSON file: {}', initpath, error=True)
            return

        out('Loaded data context: {}', rundir)
        return cls(**initargs)

    def _save(self):
        """Save the constructor parameters for this object."""
        with open(self.path(INIT_FILENAME), 'w') as f:
            json.dump({
                'desc': self._desc,
                'rootdir': self._rootdir,
                'rundir': self._rundir,
                'resdir': self._resdir,
                'h5file': self._datafile.path(),
                'h5root': self._h5root,
                'figfmt': self._figfmt,
                'staticfigs': self._staticfigs,
                'logcolor': self._logcolor
            }, f, indent=2, sort_keys=True)

    def register_as_results(self, tag=None):
        """Transfer the analysis run from the root to the results directory."""
        self.close_logfile()
        self.close_datafile()

        dt = datetime.datetime
        rdate = dt.strftime(dt.now(), '%Y-%m-%d')

        resdesc = camel2snake(self.__class__.__name__, sep='-')
        if self._desc:
            resdesc += '+{}'.format(self._desc.replace(' ', '-'))
        if tag:
            tag = tag.lower().strip().replace(' ', '-')
            resdesc += '+{}'.format(tag)

        resdest = os.path.join(self._resdir, '{}+{}'.format(rdate, resdesc))
        if os.path.isdir(resdest):
            resdest = uniquify(resdest, fmt='%s+%02d')

        if not os.path.isdir(self._resdir):
            os.makedirs(self._resdir)

        cmd = ['mv', self._rundir, resdest]
        if subprocess.call(cmd) != 0:
            self.out('Error moving analysis to {}', resdest, error=True)
            return

        if self._datafile:
            parent, name = os.path.split(self._datafile.path())
            if os.path.abspath(parent) == self._rundir:
                self._set_datafile(os.path.join(resdest, name), self._h5root)

        self._rootdir = self._resdir
        self._rundir = resdest
        self._logdir = self.path(LOGDIR)
        self._save()

        self.out('This context has been registered to:\n{}', resdest)

    # Convenience methods

    def path(self, *relpath):
        """Get an absolute path within the run directory."""
        return os.path.join(self._rundir, *relpath)

    def mkdir(self, *subdir):
        """Create a subdirectory within the run directory."""
        dpath = self.path(*subdir)
        if os.path.isdir(dpath):
            return dpath
        os.makedirs(dpath)
        return dpath

    def subfolder(self, label, prefix=False):
        """Make a unique subfolder under the analysis directory."""
        stem = self.path(label)
        if prefix:
            path = uniquify(stem, fmt='%02d-%s', reverse_fmt=True)
        else:
            path = uniquify(stem, fmt='%s-%02d')
        _, subf = os.path.split(path)
        return self.mkdir(subf)

    def reveal(self):
        """Reveal the run directory in Finder (OS X only)."""
        if not os.path.isdir(self._rundir):
            self.out('Directory does not exist: {}', self._rundir, error=True)
            return
        if sys.platform != 'darwin':
            self.out('Finder open only available on OS X', error=True)
            return
        if subprocess.call(['open', self._rundir]) != 0:
            self.out('Unable to open directory in Finder', error=True)

    # Console output methods

    def out(self, *args, **kwargs):
        """Display and log an output message.

        Arguments (except `anybar`) are passed to `ConsolePrinter`.

        Keyword arguments:
        anybar -- optional, color name to set the AnyBar widget color
        """
        color = kwargs.pop('anybar', None)
        if color is not None:
            anybar.set_color(color)
        self._out(*args, **kwargs)

    def printf(self, *args, **kwargs):
        """Send characters to stdout."""
        self._out.printf(*args, **kwargs)

    def box(self, filled=True, color=None):
        self._out.box(filled=filled, color=color)

    def newline(self):
        self._out.newline()

    # Logging methods

    def open_logfile(self, stem, timestamps=True):
        """Start a new log file with optional timestamping."""
        fn = os.path.join(self._logdir, '%s.log' % stem)
        self.use_timestamps(timestamps)
        self._out.set_outputfile(fn)

    def use_timestamps(self, active):
        """Timestamp messages recorded in log files."""
        self._out.set_timestamps(active)

    def close_logfile(self):
        """Close the current log file."""
        self._out.closefile()

    # Step wrapping methods

    def _step_enter(self, method, args, kwargs):
        spec = inspect.getargspec(method)

        argnames = spec.args[1:]
        argvalues = args[1:]
        params = list(zip(argnames, argvalues))

        self._lastcall = info = {
            'time': time.localtime(),
            'revision': git_revision(self.repodir),
            'subclass': self.__class__.__name__,
            'step': method.__name__,
            'params': params,
            'defaults': spec.defaults,
            'kwname': spec.keywords,
            'kwvalues': kwargs
        }

        self.open_logfile(info['step'].replace('_', '-'))
        self.use_timestamps(False)
        self._out.hline()
        self.out('Running step: {}', info['step'], popup=True, anybar='green')
        if params:
            self.out('Call parameters:\n{}',
                     self._format_params(params, spec.defaults))
        if kwargs:
            self.out('Keywords (\'{}\'):\n{}',
                     info['kwname'], self._format_keywords(kwargs))
        self._out.hline()

    def _format_params(self, params, defaults, bullet='-'):
        plist = []
        for i, (name, value) in enumerate(reversed(params)):
            if type(value) is pd.DataFrame:
                repr_value = 'DataFrame{}'.format(value.shape)
            else:
                repr_value = repr(value)
            plist.append('{} {} = {}'.format(bullet, name, repr_value))
            if defaults is not None and i < len(defaults):
                dflt = defaults[-1-i]
                if value != dflt:
                    plist[-1] += ' [default: {}]'.format(repr(dflt))
        return '\n'.join(reversed(plist))

    def _format_keywords(self, kwds, bullet='-'):
        return '\n'.join(['{} {} = {}'.format(bullet, k, repr(v))
                          for k,v in kwds.items()])

    def _step_execute(self, method, args, kwargs, status):
        status['OK'] = True

        # Save figure and interactive state to go non-interactive during call
        prevfigset = frozenset(self._figures.keys())
        was_interactive = plt.isinteractive()
        if was_interactive:
            plt.ioff()

        try:
            result = method(*args, **kwargs)
        except Exception as e:
            self.out('Exception in {}:\n{}: {}',
                     method.__name__, e.__class__.__name__, str(e),
                     error=True, popup=True, anybar='exclamation')
            pdb.post_mortem(sys.exc_info()[2])
            status['OK'] = False
            result = None
        finally:
            self.close_datafile()
            self.cleanup()  # optional subclass-defined clean up method

            # Show any new figures and restore interactive state
            curfigset = frozenset(self._figures.keys())
            if was_interactive:
                plt.ion()
                if prevfigset.symmetric_difference(curfigset):
                    plt.show()
                if len(plt.get_figlabels()):
                    plt.draw()

            # Restore matplotlib configuration if it was changed locally
            mpl.rc_file_defaults()

        return result

    def cleanup(self):
        """Analysis subclasses should override this with any tear-down or
        clean-up functionality (close files, etc.) after error or successful
        completion. This will be called after *every* step method in the class.

        Note that closing the analysis data file and showing new figures is
        already handled automatically.
        """
        # store.close()
        # parallel.close()
        pass

    def  _step_exit(self, method, args, kwargs, status):
        step = method.__name__
        if status['OK']:
            self._save_env()
            self._save_call_log()
            dt = time.time() - time.mktime(self._lastcall['time'])
            self.out('Finished step: {}', step, popup=dt>60, anybar='white')
        self._out.hline()
        self.close_logfile()
        if not status['OK']:
            raise RuntimeError('Stopping due to exception in {}'.format(step))

    def _save_call_log(self):
        logfn = self.path(CALL_LOG)
        history = []
        if os.path.isfile(logfn):
            with open(logfn, 'r') as logfd:
                history = logfd.readlines()
        call = self._lastcall
        with open(logfn, 'w') as logfd:
            logfd.write('Time: {}\n'.format(time.strftime('%c', call['time'])))
            signature = ', '.join(['%s=%s' % (k,repr(v))
                for k,v in call['params']])
            if call['kwvalues']:
                if call['params']:
                    signature += ', '
                signature += ', '.join(['%s=%s' % (k,repr(v))
                    for k,v in call['kwvalues'].items()])
            logfd.write('Call: {}.{}({})\n'.format(call['subclass'],
                call['step'], signature))
            logfd.write('Revision: {}\n'.format(call['revision']))
            logfd.write('-' * 40 + '\n')
            logfd.writelines(history)

    # Parallel methods

    def set_parallel_profile(self, profile):
        """Set the ipython profile to use for the parallel client."""
        parallel.set_default_profile(profile)

    def get_parallel_client(self, profile=None):
        return parallel.client(profile)

    def close_parallel_client(self):
        parallel.close()

    # Datafile methods

    def _set_datafile(self, newpath, root):
        """Set a new datafile path and root data group."""
        if self._datafile:
            self._datafile.close()

        if newpath is None:
            self._datafile = DataStore(where=self._rundir, logfunc=self._out)
        else:
            where, name = os.path.split(os.path.splitext(newpath)[0])
            self._datafile = DataStore(name=name, where=where,
                    logfunc=self._out)

        if root is None:
            if not hasattr(self, '_h5root'):
                self._h5root = '/'
        else:
            self._h5root = os.path.join('/', root)

    def set_datafile(self, newpath, root=None):
        """Set the analysis context to a new HDF file."""
        update = self._datafile is not None

        self._set_datafile(newpath, root)
        self._save()

        if update:
            self.out('Switched to datafile: {} ({})', self._datafile,
                    self._h5root)

    def get_datafile(self, readonly=None):
        """Get a handle to the HDF data file for this analysis."""
        return self._datafile.get(readonly=readonly)

    def flush_datafile(self):
        """Flush the data store file to disk."""
        self._datafile.flush()

    def close_datafile(self):
        """Close the data store file."""
        self._datafile.close()

    def backup_datafile(self, tag=None):
        """Backup the data file and create a clean active copy."""
        self._datafile.backup()

    def read_node(self, where, name=None):
        """Read the given node."""
        node = self.get_node(where, name)
        if not hasattr(node, 'read'):
            self.out('Node is not readable: {}', node._v_pathname, error=True)
            raise IOError('Data node does not have a read method')
        return node.read()

    def read_array(self, where, name=None):
        """Read array data from the given node."""
        node = self.get_node(where, name)
        if not isinstance(node, tb.Array):
            self.out('Not an array: {}', node._v_pathname, error=True)
            raise TypeError('Can only read array data from an array node')
        return node.read()

    def read_dataframe(self, where, name=None):
        """Read pandas dataframe from the given node."""
        if isinstance(where, tb.Node):
            key = where._v_pathname
        else:
            key = os.path.join(self._h5root, where)
        if name is not None:
            key = os.path.join(key, name)
        self.close_datafile()
        pdf = pd.HDFStore(self._datafile.path())
        try:
            dframe = pdf[key]
        except KeyError:
            raise IOError('No dataframe stored at %s' % key)
        finally:
            pdf.close()
        return dframe

    def read_simulation(self, where, name=None):
        """Read Brian simulation output from the data path."""
        grp = self.get_node(where, name)
        if grp._v_attrs['tenko_type'] != 'brian':
            self.out('Not a Brian simulation: %s' % grp._v_pathname,
                    error=True)
            raise TypeError('Can only read stored Brian simulation data')

        # While HDF5 file is open, filter parent group for monitor nodes
        network_name = grp._v_attrs['name']
        parent = grp._v_pathname
        dfnames = []
        for mon_node in grp._f_iter_nodes(classname='Group'):
            if 'monitor_type' not in mon_node._v_attrs:
                self.out('skipping {}', mon_node._v_pathname, error=True)
                continue
            name = mon_node._v_name
            dfnames.append(name)

        # Create namedtuple with dataframes loaded from HDF5 file
        dfs = {}
        for name in dfnames:
            dfs[name] = self.read_dataframe(parent, name)

        simdata = namedtuple('%sData' % network_name, dfs.keys())
        return simdata(**dfs)

    def get_node(self, where, name=None):
        """Get a handle to a data node if it exists."""
        if isinstance(where, tb.Node):
            if name is None:
                return where
            where = where._v_pathname
        else:
            where = os.path.join(self._h5root, where)
        node = None
        try:
            node = self.get_datafile().get_node(where, name)
        except tb.NoSuchNodeError as e:
            if name is None:
                self.out('Node does not exist: {}', where, error=True)
            else:
                self.out('Node \'{}\' does not exist ({})', name, where,
                    error=True)
            raise(e)
        return node

    def has_node(self, where, name=None):
        """Whether a data node exists."""
        where = os.path.join(self._h5root, where)
        try:
            self.get_datafile().get_node(where, name)
        except tb.NoSuchNodeError as e:
            return False
        return True

    def create_group(self, path):
        """Create a new group in the datafile."""
        p = os.path.join(self._h5root, path)
        where, name = os.path.split(p)
        dfile = self.get_datafile(False)
        try:
            grp = dfile.get_node(where, name)
        except tb.NoSuchNodeError:
            grp = data.new_group(dfile, where, name)
        else:
            self.out('warning: group already exists: {}', grp._v_pathname,
                    error=True)
        return grp

    def create_table(self, path, descr, **attrs):
        """Create a new table in the datafile."""
        self._new_node('table', data.new_table, path, descr, attrs, force=True)
        return self.get_node(path)

    def save_array(self, path, X, **attrs):
        """Save a data array to the datafile."""
        self._new_node('array', data.new_array, path, np.asarray(X), attrs)

    def save_dataframe(self, path, X, **attrs):
        """Save a pandas Series/DataFrame/Panel to the datafile."""
        self._new_node('dataframe', data.new_dataframe, path, X, attrs,
                use_pandas=True)

    def save_simulation(self, path, network, **attrs):
        """Save Brian simulation output in group/DataFrame structure."""
        monitors = [obj for obj in network.objects if type(obj) in
                        (br.StateMonitor, br.SpikeMonitor)]

        for mon in monitors:
            monpath = os.path.join(path, mon.name)
            mon_attrs = {'name': mon.name}
            record_variables = [v for v in mon.record_variables
                                    if v not in ('t', 'i')]
            columns = {}

            if type(mon) is br.StateMonitor:
                mon_attrs['monitor_type'] = 'state'
                ts = np.tile(mon.t / mon.t.unit, len(mon.record))  # timestamps
                neuron_ix = mon.record.repeat(mon.t.shape[0])  # int unit index

            elif type(mon) is br.SpikeMonitor:
                mon_attrs['monitor_type'] = 'spike'
                ts = mon.t / mon.t.unit  # spike times in seconds
                neuron_ix = (mon.i / mon.i.unit).astype('i')  # int unit index

            columns.update(t=ts, neuron=neuron_ix)

            for varname in record_variables:
                trace = getattr(mon, varname)

                if type(trace) is np.ndarray:
                    unit = 'scalar'
                elif type(trace) is br.core.variables.VariableView:
                    unit = repr(trace.unit)
                    trace = trace / trace.unit
                elif type(trace) is br.units.fundamentalunits.Quantity:
                    if trace.is_dimensionless:
                        unit = 'scalar'
                    else:
                        unit = repr(trace.dimensions)
                    trace = np.asarray(trace)
                else:
                    self.out('skipping {} \'{}\': unknown type ({})'.format(
                        mon.name, varname, type(trace)), error=True)
                    continue

                if type(mon) is br.StateMonitor:
                    trace = trace.reshape(-1)
                mon_attrs['%s_unit' % varname] = unit
                columns[varname] = trace

            self.save_dataframe(monpath,
                    pd.DataFrame(data=columns,
                        columns=['t', 'neuron'] + record_variables),
                    title='Network: {}, Monitor: {}'.format(
                        network.name, mon.name), **mon_attrs)

        # Write context attributes to parent group of simulation data
        self.get_datafile(False)
        grp = self.get_node(path)
        grp_attrs = dict(tenko_type='brian', name=network.name)
        grp_attrs = merge_two_dicts(grp_attrs, attrs)
        self._write_v_attrs(grp, grp_attrs)

    def _new_node(self, ntype, new_node, path, X, attrs, use_pandas=False,
        **kwds):
        p = os.path.join(self._h5root, path)
        where, name = os.path.split(p)
        title = attrs.pop('title', snake2title(name))
        kwds.update(createparents=True, title=title)
        if use_pandas:
            self.close_datafile()
            dfile = pd.HDFStore(self._datafile.path())
        else:
            dfile = self.get_datafile(False)
        node = new_node(dfile, where, name, X, **kwds)
        self._write_v_attrs(node, attrs)
        self.out('Created {}: {} ("{}")', ntype, node._v_pathname, title)
        if use_pandas:
            dfile.close()

    def _write_v_attrs(self, node, attrs):
        call = self._lastcall
        na = node._v_attrs
        if attrs is not None:
            for k in list(attrs.keys()):
                na[k] = attrs[k]
        if call is None:
            na['time'] = time.localtime()
            return node
        na['time'] = time.mktime(call['time'])
        for key in ('revision', 'subclass', 'step'):
            na[key] = call[key]
        for name, value in call['params']:
            na[name] = value
        for kwd, value in call['kwvalues'].items():
            na[kwd] = value

    # Figure methods

    def figure_from_template(self, svgpath, label=None, figsize=None, **figkw):
        """Load figure template to create figure and axes.

        Arguments:
        svgpath -- path to SVG file (tiny profile) where rectangles define axes
        label -- optional figure label, defaults to template basename
        figsize -- optionally override the figure size specified in the svg

        Remaining keywords are passed to the `figure` method.

        Returns:
        fig, axd -- tuple of figure handle and axes dictionary
        """
        if not svgpath.endswith('.svg'):
            svgpath += '.svg'
        if label is None:
            label = os.path.basename(os.path.splitext(svgpath)[0])
        self.out('Figure template: {}', svgpath)
        svg = get_svg_figinfo(svgpath)
        figsize = svg['figsize'] if figsize is None else figsize
        figkw.update(figsize=figsize, label=label, clear=True)
        f = self.figure(**figkw)
        axdict = {key: f.add_axes(rect) for key, rect in svg['axes'].items()}
        return f, axdict

    def set_static_figures(self, b=None):
        """Set or toggle whether figure windows are static across calls."""
        old = self._staticfigs

        if b is None:
            self._staticfigs = not self._staticfigs
        else:
            self._staticfigs = bool(b)

        if old != self._staticfigs:
            self._save()

    def figure(self, label=None, clear=False, title=None, handle=None, **kwds):
        """Create or clear a labeled figure window.

        Remaining keywords are passed to `plt.figure`.
        """
        if label is None:
            if self._lastfig:
                label = self._lastfig
            else:
                raise ValueError('no current figure')

        if handle is not None:
            self._figures[label] = handle

        try:
            fig = self._figures[label]
        except KeyError:
            if self._staticfigs:
                kwds.update(num=label)
            fig = self._figures[label] = plt.figure(**kwds)
        else:
            if 'figsize' in kwds:
                fig.set_size_inches(kwds.pop('figsize'), forward=True)
            fig.set(**kwds)
        finally:
            if clear:
                fig.clear()
                fig.subplotpars.left = None
                fig.subplotpars.right = None
                fig.subplotpars.bottom = None
                fig.subplotpars.top = None
                fig.subplotpars.wspace = None
                fig.subplotpars.hspace = None
            self._lastfig = label
            if title is not None:
                fig.suptitle(title)

        return fig

    def savefig(self, label=None, basepath=None, tag=None, tight_padding=None,
        closeafter=False, **savefig):
        """Save an open figure as an image file.

        Arguments:
        label -- optional, label of figure to be saved (default, last figure)
        basepath -- optional, save path for image file up to base name
        tag -- optional, extra text post-pended to file base name
        tight_padding -- optional, padding in inches for tight bounds

        Remaining keywords are passed to `Figure.savefig()`.
        """
        label = label or self._lastfig
        if label not in self._figures:
            self.out('Figure does not exist: {}', label, error=True)
            return

        # Set label-based path if optional base path not specified
        stem = self.path(label) if basepath is None else basepath
        if tag:
            stem += '-{}'.format(tag)

        # Generate unique path with the given format extension
        ext = savefig.pop('format', self._figfmt)
        if ext == 'mpl':
            ext = mpl.rcParams['savefig.format']
        path = uniquify(stem, ext=ext, fmt='%s-%02d')
        parent, img = os.path.split(path)
        if not os.path.isdir(parent):
            os.makedirs(parent)

        if tight_padding is not None:
            savefig.update(bbox_inches='tight', pad_inches=tight_padding)

        # Persist the current figure save settings
        savefig.update(format=ext)
        self._savefig.update(savefig)

        self._figures[label].savefig(path, **self._savefig)
        self.out('Saved image: {}', path)
        self._savefig_path = path

        if closeafter:
            self.closefig(label=label)

    def openfig(self):
        """Open the last image file saved by using `savefig`."""
        if self._savefig_path is None:
            self.out('There is no previously saved image to open.', error=True)
            return
        if sys.platform != 'darwin':
            self.out('Image viewing only available on OS X.', error=True)
            return
        if subprocess.call(['open', self._savefig_path]) != 0:
            self.out('Error opening file: {}', self._savefig_path, error=True)
        else:
            self.out('Opened image: {}', self._savefig_path)

    def closefig(self, label=None):
        """Close an open figure."""
        if label is None:
            label = self._lastfig

        if label not in self._figures:
            self.out('Figure does not exist: {}', label, error=True)
            return

        plt.close(self._figures[label])
        if not self._holdfigs:
            del self._figures[label]

        if label == self._lastfig:
            self._lastfig = None

    def save_figures(self, **savefig):
        """Save the currently open figures as image files.

        Keyword arguments are passed to `plt.Figure.savefig`:
        """
        for key in list(self._figures.keys()):
            self.savefig(key, **savefig)

    def close_figures(self):
        """Close all open figure windows."""
        self._holdfigs = True
        for key in self._figures:
            self.closefig(key)
        self._holdfigs = False
        self._figures.clear()

    def set_figfmt(self, fmt, save=False):
        """Set the image format for saving figures."""
        assert fmt in ('png','pdf','svg','ps','eps','mpl'), \
                'bad figure format: {}'.format(fmt)
        self._figfmt = fmt
        if save:
            self._save()
