"""
Development tests for adaptive binning algorithm
"""

from __future__ import division

import os

import numpy as np
import matplotlib.pyplot as plt
from scipy import signal

from spc.lib import motion, cluster, time, parsers
from spc.tools import maps, kernels


############## Plotting state and functions ###############

cur = 0
curfig = 99
figsize = (11, 8.5)
rows, cols = (4, 5)
f = None
flist = []
axlist = None

def next_page():
    global f, axlist, cur, curfig
    cur = 0
    curfig += 1
    plt.close(curfig)
    f, axlist = plt.subplots(nrows=rows, ncols=cols,
            sharex=True, sharey=True, num=curfig, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    flist.append(f)
    return f

def next_axis():
    global cur
    ax = axlist[cur]
    cur += 1
    return ax

plt.ioff()


############## Testing code to plot figures ###############

testdir = '/Users/joe/Dropbox/Source/spc/tests'
datadir = os.path.join(testdir, 'adaptive-binning-test')

S_ID = 27
TTC = (7,1)
SPEED = 'fast'
BINS = 15
KBINS = 256
FS = 30.0
PMIN, PMAX = 0, 100

def get_map_data():
    global FS, duration, n_spikes

    print "Loading session data from database..."
    session = parsers.parse_session(S_ID)
    grp = session['group']

    mdata = motion.MotionData(grp)
    cdata = cluster.ClusterData(grp, TTC)
    FS = mdata.Fs

    mix = getattr(mdata, '%s_index' % SPEED)
    six = time.select_from(cdata.spikes, getattr(mdata, '%s_intervals' % SPEED))

    Pfull = mdata.x, mdata.y
    P = mdata.x[mix], mdata.y[mix]
    S = mdata.F('x', cdata.spikes[six]), mdata.F('y', cdata.spikes[six])

    duration = mix.sum() / FS
    n_spikes = six.sum()
    return S, P, Pfull

def mapshow(M, ax, size=100, cmap='CMRmap', cmax=None):
    cmax = cmax is None and M[np.isfinite(M)].max() or cmax
    return ax.imshow(np.flipud(M.T), cmap=cmap,
                vmin=0.0, vmax=cmax,
                interpolation='nearest',
                extent=(0, size, 0, size), aspect='equal')

def plot_spatial_map(M, label, **kwargs):
    ax = next_axis()
    im = mapshow(M, ax, **kwargs)
    ax.text(60.0, 105.0, label, size='small', ha='center', va='bottom')
    plt.colorbar(im, ax=ax, ticks=None, pad=0.025, aspect=15,
            shrink=0.7, fraction=0.2)

def kernel_masking(x, y, bins, smoothing=0.08):
    """Generate masking matrix based on a smoothed histogram."""
    H = np.histogram2d(x, y, bins=bins, range=[[0,PMAX],[0,PMAX]])[0]
    H = signal.medfilt2d((H>0).astype('d'), kernel_size=3)

    M = np.zeros_like(H, 'i')
    r, c = H.shape
    for i in xrange(r):
        j = 0
        while j < c and H[i,j] == 0.0:
            M[i,j] += 1
            j += 1
        j = c - 1
        while j >= 0 and H[i,j] == 0.0:
            M[i,j] += 1
            j -= 1
    for j in xrange(c):
        i = 0
        while i < r and H[i,j] == 0.0:
            M[i,j] += 1
            i += 1
        i = r - 1
        while i >= 0 and H[i,j] == 0.0:
            M[i,j] += 1
            i -= 1

    k = max(int(smoothing * bins), 3)
    k = (k % 2) and k or k + 1
    mask = signal.medfilt2d((M<2).astype('d'), kernel_size=k) == 0

    return H, mask


f = next_page()

nfrac = 0.04
minrad = 8.0
maxrad = 30.0
mask_smoothing = 1/15

(xs, ys), (xp, yp), (Xp, Yp) = get_map_data()
rs = lambda arr: arr.reshape(KBINS, KBINS).T


if os.path.isdir(datadir):
    print "Found data. Loading data traces now..."
    R1 = np.load(os.path.join(datadir, 'R1.npy'))
    R2 = np.load(os.path.join(datadir, 'R2.npy'))
    R3 = np.load(os.path.join(datadir, 'R3.npy'))
    R4 = np.load(os.path.join(datadir, 'R4.npy'))
    R5 = np.load(os.path.join(datadir, 'R5.npy'))

    Ppos = np.load(os.path.join(datadir, 'Ppos.npy'))
    radpos = np.load(os.path.join(datadir, 'radpos.npy'))

    Pspk = np.load(os.path.join(datadir, 'Pspk.npy'))
    radspk = np.load(os.path.join(datadir, 'radspk.npy'))

else:
    os.makedirs(datadir)

    # The standard histogram-based spatial map
    print "Computing a series of standard rate maps..."
    R1 = maps.spatial_ratemap(xs, ys, xp, yp, bins=BINS, min_occ=1.0, freq=FS)
    R2 = maps.spatial_ratemap(xs, ys, xp, yp, bins=2*BINS, min_occ=1.0, freq=FS)
    R3 = maps.spatial_ratemap(xs, ys, xp, yp, bins=3*BINS, min_occ=1.0, freq=FS)
    R4 = maps.spatial_ratemap(xs, ys, xp, yp, bins=4*BINS, min_occ=1.0, freq=FS)
    R5 = maps.spatial_ratemap(xs, ys, xp, yp, bins=5*BINS, min_occ=1.0, freq=FS)

    np.save(os.path.join(datadir, 'R1'), R1)
    np.save(os.path.join(datadir, 'R2'), R2)
    np.save(os.path.join(datadir, 'R3'), R3)
    np.save(os.path.join(datadir, 'R4'), R4)
    np.save(os.path.join(datadir, 'R5'), R5)

    # Kernel estimation training points
    XYpos = (xp, yp)
    XYspk = (xs, ys)

    pos_nbs = int(nfrac * xp.size)
    spk_nbs = int(nfrac * xs.size)

    print "N_pos =", xp.size, "(%d neighbors)" % pos_nbs
    print "N_spk =", xs.size, "(%d neighbors)" % spk_nbs

    # Kernel estimation testing points (evaluation grid)
    kgridx = np.linspace(0, 100, KBINS)
    kgridy = np.linspace(0, 100, KBINS)
    XYkernel = map(lambda a: a.flatten(), np.meshgrid(kgridx, kgridy))

    Kpos = kernels.AdaptiveGaussianKernel(XYpos, n_neighbors=pos_nbs)
    Kspk = kernels.AdaptiveGaussianKernel(XYspk, n_neighbors=spk_nbs)

    print "Evaluating position kernel..."
    dbg = {}
    Ppos = Kpos(XYkernel, minrad=minrad, maxrad=maxrad, debug=dbg)
    radpos = rs(dbg['rad'])
    np.save(os.path.join(datadir, 'Ppos'), Ppos)
    np.save(os.path.join(datadir, 'radpos'), radpos)

    print "Evaluating spike kernel..."
    dbg = {}
    Pspk = Kspk(XYkernel, minrad=minrad, maxrad=maxrad, debug=dbg)
    radspk = rs(dbg['rad'])
    np.save(os.path.join(datadir, 'Pspk'), Pspk)
    np.save(os.path.join(datadir, 'radspk'), radspk)


print "Plotting standard rate maps..."

plot_spatial_map(R1, 'standard ratemap : %d bins' % BINS)
plot_spatial_map(R2, 'standard ratemap : %d bins' % (2*BINS))
plot_spatial_map(R3, 'standard ratemap : %d bins' % (3*BINS))
plot_spatial_map(R4, 'standard ratemap : %d bins' % (4*BINS))
plot_spatial_map(R5, 'standard ratemap : %d bins' % (5*BINS))

print "Computing adaptive kernel mask..."

mask_hist, mask = kernel_masking(Xp, Yp, KBINS, smoothing=mask_smoothing)
Gpos = rs(Ppos) * duration
Gpos[mask] = np.nan

Gspk = rs(Pspk) * n_spikes
Grate = Gspk / Gpos
Grate[mask] = np.nan

print "Plotting adaptive spatial maps..."

plot_spatial_map(radpos, 'position radius')
plot_spatial_map(np.clip(radpos, minrad, maxrad), 'position radius (clip)',
        cmax=radpos.max())
plot_spatial_map(Gpos, 'Pr(occ) estimate')

plot_spatial_map(radspk, 'spike radius')
plot_spatial_map(np.clip(radspk, minrad, maxrad), 'spike radius (clip)',
        cmax=radspk.max())
plot_spatial_map(Gspk, 'Pr(spikes) estimate')

plot_spatial_map(Grate, 'Pr(rate) estimate')

plot_spatial_map(mask, 'kernel mask')
plot_spatial_map(mask_hist, 'mask histo', cmap='gray_r')


############## Tighten layout and draw ###############

for f in flist:
    plt.figure(f.number)
    plt.tight_layout(pad=0.4)

plt.ion()
plt.draw()
plt.show()
