#!/usr/bin/env python

"""
Generate histograms of angular motion data to check for interpolation errors.
"""

import os

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d

from spc import PROJECT_ROOT, store
from spc.output import log
from spc.lib.motion import MotionData
from spc.lib.cluster import ClusterData
from spc.tools import paths
from spc.lib.parsers import parse_ttc


############## Set up output directories ###############

outputdir = os.path.splitext(__file__)[0]


############## Plotting state and functions ###############

interactive = False

cur = 0
f = None
axlist = []
rows = 5
cols = 3
figsize = (8.5,11)
hist_bins = np.linspace(0, 360, 64)

def new_figure(rows=rows, cols=cols, figsize=figsize):
    global f, axlist, cur
    cur = 0
    f, axlist = plt.subplots(nrows=rows, ncols=cols, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        if f is not None:
            save_and_close('histos')
        new_figure()
    ax = axlist[cur]
    ax.set_axis_on()
    cur += 1
    return ax

def save_and_close(stem='figure'):
    f.set_tight_layout(True)
    if interactive:
        log('Displayed: {} ({})', stem, f.number)
        return
    stem = os.path.join(outputdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = paths.uniquify(stem, ext='pdf', fmt='%s-%02d')
    plt.savefig(fn)
    log('Saved: {}', fn)
    plt.close()

plt.ioff()


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(0.5, 1.0, title, transform=ax.transAxes,
                size='x-small', ha='center', va='top')

def plot_histo(hd, md, label='', **kwargs):
    ax = next_axis()
    hkw = dict(bins=hist_bins)
    hkw.update(kwargs)
    for data, color in [(hd, 'b'), (md, 'r')]:
        hkw.update(color=color)
        ax.hist(data, histtype='stepfilled', zorder=-1, alpha=0.3, **hkw)
        hkw.update(color='k')
        ax.hist(data, histtype='step', linewidth=1, **hkw)
    [ax.spines[s].set_visible(False) for s in ('top', 'right', 'left')]
    ax.tick_params(top=False, right=False, left=False)
    ax.set(xlim=(-40,400), xticklabels=[], yticklabels=[])
    quicktitle(ax, label)


############## Script for the generation of many maps ###############

df = store.get(True)

for rec in df.root.recordings.iterrows():
    s_id = rec['session_id']
    spath = df.root.sessions[s_id]['group']
    if df.root.sessions[s_id]['arena'] != 0:
        continue

    ttc = parse_ttc(rec)
    log('Plotting {} from session {}...', rec['ttc'], s_id)

    mdata = MotionData(spath)
    Nt = mdata.t.size
    Nhd = mdata.hd.size
    Nmd = mdata.md.size
    if Nhd != Nt or Nmd != Nt:
        log('Size mismatch ({}, {}, {})', Nhd, Nmd, Nt, error=True)
        continue

    cdata = ClusterData(spath, ttc)
    hd = mdata.F('hd', cdata.spikes)
    md = mdata.F('md', cdata.spikes)

    plot_histo(hd, md, label='[%d] %s' % (s_id, rec['ttc']))

save_and_close('histos')
store.close()


############## Draw anything left back to the screen ###############

plt.ion()
if interactive:
    plt.draw()
    plt.show()
