"""
Tests for estimating theta burst frequency.
"""

from __future__ import division

import os
import numpy as np
import matplotlib.pyplot as plt

from spc import store
from spc.lib import labels, session
from spc.tools import spikes, arrays, paths
from spc.lib.theta import burst_frequency_estimate


outdir = os.path.splitext(__file__)[0]
if not os.path.isdir(outdir):
    os.makedirs(outdir)


c_id = 176
maxlag = 0.25
bins = 128
tau_blur = 0.002
tau_blur_inc = 0.001
tau_blur_max = 0.020
theta_frac = 1.0
H = 0.75
savefig = False

df = store.get(True)
cell = df.root.recordings[c_id]
lags = df.root.theta.autocorr.lags.read()
C = getattr(df.root.theta.autocorr, 'id%04d' % c_id).read()

tag = labels.session_cell_id(c_id)
print 'Correlation for {}'.format(tag)

f_theta = 8.5
tau_theta = theta_frac / f_theta
taulim = (1 - H) * tau_theta, (1 + H) * tau_theta

posix = lags > 0
lags = lags[posix]
C = C[posix]

check = lambda mn, mx: len(mn) == 2 and len(mx) == 1 and (
        mn[0] < mx[0] < mn[1])

mins = maxs = []
tix = np.logical_and(lags>=taulim[0], lags<=taulim[1])
while not check(mins, maxs) and tau_blur <= tau_blur_max:
    print 'Trying blur = {}'.format(tau_blur)
    Cg = arrays.blur(C, pixels=(bins / (2 * maxlag)) * tau_blur, padding='same')

    mins = arrays.minima(Cg[tix])
    maxs = arrays.maxima(Cg[tix])
    print 'Found {} minima'.format(len(mins))
    print 'Found {} maxima'.format(len(maxs))

    tau_blur = round(tau_blur + tau_blur_inc, 5)

if check(mins, maxs):
    tau_MAP = lags[tix][maxs[0]]
    f_MAP = 1 / tau_MAP
    print 'Max. a posteriori f = {} Hz'.format(f_MAP)

    lagt = lags[tix]
    b = np.logical_and(lagt>=lagt[mins[0]], lagt<=lagt[mins[1]])
    W = Cg[tix][b] - Cg[tix][b].min()
    W /= W.sum()
    tau_hat = np.dot(W, lags[tix][b])
    f_hat = 1 / tau_hat
    print 'Weighted estimate f = {} Hz'.format(f_hat)

    f_comb = (f_MAP + f_hat) / 2
    tau_comb = 1 / f_comb
    print 'Combined estimate f = {} Hz'.format(f_comb)

    fig = plt.figure(10, (9,7))
    fig.clear()
    ax = fig.add_subplot(111)

    spikes.corrplot((C, lags), alpha=0.3)
    spikes.corrplot((C[tix], lags[tix]), style='lines', c='b', marker='.')
    spikes.corrplot((Cg[tix][b], lags[tix][b]), style='lines', c='b', alpha=0.6)

    ax.axvline(tau_MAP, c='r', ls=':', lw=1)
    ax.axvline(tau_hat, c='r', ls='--')
    ax.axvline(tau_comb, c='r', lw=2)

    ax.set(xlabel='Lag (s)', ylabel='count')
    ax.set_xlim(-0.5 * maxlag, maxlag)
    s = C[tix].ptp()
    ax.set_ylim(C[tix].min() - 0.5 * s, C[tix].max() + 0.5 * s)
    ax.set_title('{}: f_comb = {} Hz'.format(tag, f_comb))

    if savefig:
        savefn = paths.uniquify(os.path.join(outdir, 'corrplot'),
            ext='png', fmt='%s-%03d')
        plt.savefig(savefn)
        print 'Saved: {}'.format(savefn)

    plt.draw()
    plt.show()
else:
    print 'Unable to find theta peak'

print '--- spc.lib.theta.burst_frequency_estimate ---'

f_combined = burst_frequency_estimate((C, lags), maxlag=maxlag, bins=bins,
        debug=True)
print 'Combined estimate f = {} Hz'.format(f_combined)
