#!/usr/bin/env python

"""
Test generation of phase maps with inverted multi-output kernel functions.
"""

import os

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import ticker

from spc import store
from spc.output import log
from spc.lib import cell
from spc.tools import maps, radians
from spc.tools.adaptive import NBR_FRAC, MINRAD, MAXRAD
from spc.tools.paths import uniquify

testdir = os.path.splitext(__file__)[0]

c_id = 445
desc = '%g-%d-%d' % (NBR_FRAC, int(MINRAD), int(MAXRAD))
levels = 16
plot_phase_components = False

C = cell.record(c_id)
S = cell.session_for(c_id, res=128)

plt.ioff()
mpl.rc('ytick', labelsize=22.0)

radfmt = radians.RadianFormatter(inline=False)
radinline = radians.RadianFormatter(inline=True)
threeticks = ticker.LinearLocator(numticks=3)

cbargs = {'pad':0.0, 'fraction':0.07, 'format':'%.1f', 'ticks':threeticks}
pcbargs = cbargs.copy()
pcbargs.update({'format':radfmt})
mkw = dict(cbar=True, cbargs=cbargs)
pmkw = dict(cbar=True, cbargs=pcbargs)
ikw = dict(mask_color='0.9')
mkw.update(ikw)
pmkw.update(ikw)

R = S.ratemap(C, bursts=False)
P = S.phasemap(C, bursts=False)

nrows = 2 if plot_phase_components else 1
height = 8.1125 if plot_phase_components else 3.6875

fig, ax = plt.subplots(nrows, 2, figsize=(9,height))
ax = ax.flatten()
[_.set_axis_off() for _ in ax]
if plot_phase_components:
    axa, axb, axc, axd = ax
    maps.plot(P[0], ax=axa, levels=levels, circular=True, **pmkw)
    maps.plot(P[1], ax=axb, cmap='gray', **mkw)
else:
    axd, axc = ax

maps.plot(P, ax=axc, levels=levels, satnorm=True, **pmkw)
maps.plot(R, ax=axd, levels=levels, cmin=0.0, **mkw)

pmap_fn = uniquify(os.path.join(testdir, 'phasemap-%d-%s' % (c_id, desc)),
        fmt='%s-%d', ext='png')
rmap_fn = uniquify(os.path.join(testdir, 'ratemap-%d-%s' % (c_id, desc)),
        fmt='%s-%d', ext='png')
maps.save(P, pmap_fn, levels=levels, satnorm=True, **ikw)
maps.save(R, rmap_fn, levels=levels, cmin=0.0, **ikw)
log('Saved: {}', pmap_fn)

fn = uniquify(os.path.join(testdir, 'allmaps-%d-%s' % (c_id, desc)),
        fmt='%s-%d', ext='pdf')
fig.savefig(fn, dpi=300)
log('Saved: {}', fn)
plt.close(fig.number)


plt.ion()
plt.show()

mpl.rc_file_defaults()
store.close()
