"""
Test kernel estimates of firing rate given spike train data.
"""

from __future__ import division

import os

import numpy as np
from sklearn.neighbors import KernelDensity
import matplotlib.pyplot as plt
import seaborn as sns
sns.reset_orig()

from spc import store
from spc.lib import session, labels
from spc.tools import paths
from spc.tools.spikes import FiringRateEstimate


outdir = os.path.splitext(__file__)[0]
if not os.path.isdir(outdir):
    os.makedirs(outdir)


c_id = 108
dur = 1000.0
width_eff = 0.1
Ft = 60.0  # Hz
saveimg = True

df = store.get(True)
cell = df.root.recordings[c_id]
s_id = cell['session_id']
S = session.RecordingSession(s_id)
tag = labels.session_cell_id(c_id)

cdata = S.get_cluster(cell)
st = cdata.spikes[S.speed_filter(cdata.spikes)]
t0 = st[0]
t1 = t0 + dur
st = st[st < t1]
avg_rate = st.size / dur
print 'Found {} spikes: {}'.format(len(st), tag)
print 'Spike train has average rate = {} Hz'.format(avg_rate)

spikes = np.atleast_2d(st).T
sigma = width_eff / np.sqrt(12)
t_plot = np.linspace(t0, t1, int(Ft * dur))
test = np.atleast_2d(t_plot).T

kernel = KernelDensity(bandwidth=sigma, kernel='gaussian')
kernel.fit(spikes)
pf = np.exp(kernel.score_samples(test)).squeeze()
f_hat = avg_rate * pf / pf.mean()
f_hat_max = f_hat.max()

# --- validate library class ---

f_est = FiringRateEstimate(st, dur)
f_atspikes = f_est(st)
f_full = f_est(t_plot[::2]) # decimate by 2
print 'Mean of full rate estimate = {} Hz'.format(f_full.mean())

# --- plotting ---

fig = plt.figure(11, (10,6))
fig.clear()
ax = fig.add_subplot(111)

ax.plot(t_plot, f_hat, 'r-', lw=2, alpha=0.7)
sns.rugplot(st, height=f_hat_max / 5, c='k')
ax.plot(st, f_atspikes, 'r+')

ax.axhline(c='k')
ax.set(xlabel='Time (s)', ylabel='Est. rate')
ax.set_xlim(t0, t0+40.0)
ax.set_ylim(-f_hat_max, 2*f_hat_max)
ax.set_title('{}: {} Hz (peak)'.format(tag, f_hat_max))

if saveimg:
    savefn = paths.uniquify(os.path.join(outdir, 'rate-estimate'),
        ext='png', fmt='%s-%03d')
    plt.savefig(savefn)
    print 'Saved: {}'.format(savefn)

plt.draw()
plt.show()

store.close()
