"""
Test implementation of spike autocorrelation with dask.array.dot.
"""

from __future__ import division

import time

import numpy as np
import dask.array as da
import matplotlib.pyplot as plt
from ipyparallel import Client

from spc import store
from spc.output import anybar
from spc.lib import labels, session
from spc.tools import spikes


def xcorr_smart(a, b, maxlag=1.0, bins=128):
    """Unrolled implementation of spike train correlogram."""
    na, nb = a.size, b.size
    if bins % 2 == 0:
        bins += 1
    start = end = 0
    H = np.zeros(bins)
    edges = np.linspace(-maxlag, maxlag, bins + 1)
    for t in a:
        while b[start] < t - maxlag:
            start += 1
        while end < nb and b[end] <= t + maxlag:
            end += 1
        H += np.histogram(b[start:end] - t, bins=edges)[0]
    centers = (edges[:-1] + edges[1:]) / 2
    return H, centers

def _xcorr(maxlag, bins):
    a, b, edges = xcorr_spiketrain, xcorr_reference, xcorr_edges
    nb = b.size
    start = end = 0
    H = np.zeros(bins)
    for t in a:
        while b[start] < t - maxlag:
            start += 1
        while end < nb and b[end] <= t + maxlag:
            end += 1
        H += np.histogram(b[start:end] - t, bins=edges)[0]
    return H

def pxcorr(a, b, maxlag=1.0, bins=128):
    """Parallel unrolled implementation of spike train correlogram."""
    if bins % 2 == 0:
        bins += 1
    edges = np.linspace(-maxlag, maxlag, bins + 1)

    rc = Client()
    dview = rc[:]
    dview.execute('import numpy as np')
    dview.scatter('xcorr_spiketrain', a)
    dview['xcorr_reference'] = b
    dview['xcorr_edges'] = edges
    ar = dview.apply_async(_xcorr, maxlag, bins)
    H = np.array(ar.get()).sum(axis=0)
    centers = (edges[:-1] + edges[1:]) / 2
    return H, centers

def xcorr(a, b, blocked=False, chunksize=1000, maxlag=1.0, bins=128):
    """Compute the spike train correlogram."""
    na, nb = a.size, b.size

    if blocked:
        ad = da.from_array(a.reshape(1, na), chunks=(1,chunksize))
        bd = da.from_array(b.reshape(nb, 1), chunks=(chunksize,1))
        D = (bd - ad).compute()
    else:
        D = b.reshape(nb,1) - a.reshape(1,na)

    if bins % 2 == 0:
        bins += 1
    H, edges = np.histogram(D, range=(-maxlag, maxlag), bins=bins, density=False)
    centers = edges[:-1] + np.diff(edges)/2
    return H.astype('d'), centers


df = store.get(True)

c_id = 99
tlimit = None  # 2000.0
lag = 0.5
bins = 72

cell = df.root.recordings[c_id]
S = session.get(cell['session_id'])
st = S.spike_filter(cell, speed='fast')

if tlimit is not None:
    st = st[st > S.tend - tlimit]

N = st.size
tag = labels.session_cell(c_id)
print 'Spike train for {}: {} spikes'.format(tag, N)

anybar.set_color('green')
t0 = time.time()
C1, lags = spikes.xcorr(st, st, maxlag=lag, bins=bins)
print 'Xcorr: {} seconds'.format(time.time() - t0)
anybar.set_color('white')

anybar.set_color('purple')
t0 = time.time()
C2, lags = spikes.xcorr(st, st, maxlag=lag, bins=bins, parallel=True)
print 'Parallel xcorr: {} seconds'.format(time.time() - t0)
anybar.set_color('white')

C1 = np.ma.masked_where(np.isclose(lags, 0), C1, False)
C2 = np.ma.masked_where(np.isclose(lags, 0), C2, False)

if np.all(np.isclose(C1, C2)):
    print 'Correlations match!'
else:
    print 'Warning: correlation value mismatch!'

fig =  plt.figure(1, figsize=(8,6))
fig.clear()
ax = fig.add_subplot(111)
spikes.corrplot((C1, lags), style='verts')
spikes.corrplot((C2, lags), c='r', style='lines')
#ax.plot(lags, C1, 'k-', lw=0.5, drawstyle='steps-mid')
#ax.plot(lags, C2, 'r--', drawstyle='steps-mid')
ax.set(xlabel='Time (s)', ylabel='count')
plt.draw()
