"""
Develop spatial information measure for firing phase.
"""

import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec

from spc.lib import cell
from spc.lib.motion import NORM_POS_RANGE
from spc.lib.lfp import PHASE_MIN, PHASE_MAX
from spc.tools import maps, binned, paths, plots

XMIN, XMAX = NORM_POS_RANGE[0]
YMIN, YMAX = NORM_POS_RANGE[1]

outdir = os.path.splitext(__file__)[0]
if not os.path.isdir(outdir):
    os.makedirs(outdir)


c_id = 97
xbins = ybins = 15
pbins = binned.DEFAULT_PHASE_BINS
shuffle = False
savefig = False

tag, st = cell.tag_and_spikes(c_id)
print 'Computing phase information for {}'.format(tag)

C = cell.record(c_id)
S = cell.session_for(c_id)
phi = S.lfp.F('phase', st)
if shuffle:
    np.random.shuffle(phi)
xs = S.motion.F('x', st)
ys = S.motion.F('y', st)
sample = np.c_[xs, ys, phi]
N = sample.shape[0]

print 'Computing distributions...'
xedges = np.linspace(XMIN, XMAX, xbins + 1)
yedges = np.linspace(YMIN, YMAX, ybins + 1)
pedges = np.linspace(PHASE_MIN, PHASE_MAX, pbins + 1)
edges = [xedges, yedges, pedges]

H, _ = np.histogramdd(sample, bins=edges)

pdf = lambda h: h / h.sum()

Hr = H.reshape((-1, pbins))
Px = pdf(Hr.sum(axis=-1))
Pp = pdf(Hr.sum(axis=0))
Pxp = pdf(Hr)
Pxpind = pdf(Px.reshape(-1,1) * Pp)

valid = np.logical_and(Pxp != 0.0, Pxpind != 0.0)
joint = Pxp[valid]
marginal = Pxpind[valid]
I = np.sum(joint * np.log(joint / marginal))

print 'Plotting joint and marginal distributions...'
fig = plt.figure(14, (12.7125, 8.1375))
fig.clear()
if shuffle:
    fig.suptitle('{}\nI = {:.4f} bits/spike (shuffled!)'.format(tag, I))
else:
    # Compare with library implementation...
    print 'Session method with p-value...'
    #Ilib = binned.phaseinfo(xs, ys, phi)
    Ip = S.phase_space_info(C, pvalue=True)

    fig.suptitle('{}\nI = {:.4f} bits/spike (I = {:.4f}, p < {:g})'.format(
        tag, I, Ip[0], Ip[1]))

gs = gridspec.GridSpec(2,3, width_ratios=(3,3,1), height_ratios=(1,3))
ax = plt.subplot(gs[1,1])
xax = plt.subplot(gs[1,2], sharey=ax)
pax = plt.subplot(gs[0,1], sharex=ax)
iax = plt.subplot(gs[1,0], sharex=ax, sharey=ax)

totbins = xbins*ybins
iax.imshow(Pxpind, origin='lower', interpolation='nearest', cmap='afmhot',
        extent=[PHASE_MIN, PHASE_MAX, 0, totbins], aspect='auto')
iax.set(ylabel='Spike position bin')
iax.set(xlim=(PHASE_MIN, PHASE_MAX), ylim=(0, totbins))
iax.set_title('P(x) * P(phi)')

ax.imshow(Pxp, origin='lower', interpolation='nearest', cmap='afmhot',
        extent=[PHASE_MIN, PHASE_MAX, 0, totbins], aspect='auto')
ax.set(xlabel='Spike phase (rad)')
ax.set(xlim=(PHASE_MIN, PHASE_MAX), ylim=(0, totbins))
ax.set_title('P(x,phi)')

xax.plot(Px, np.arange(totbins), 'k-')
xax.set_xlabel('P(x)')
xax.set_xticklabels([])
plots.despine(xax, bottom=False)

pax.plot((pedges[1:] + pedges[:-1])/2, Pp, 'k-')
pax.set_ylabel('P(phi)')
pax.set_yticklabels([])
plots.despine(pax, left=False)

if savefig:
    savefn = paths.uniquify(os.path.join(outdir,
        'phase-info-%04d' % c_id), ext='png', fmt='%s-%02d')
    plt.savefig(savefn)
    print 'Saved: {}'.format(savefn)

plt.draw()
plt.show()
