#!/usr/bin/env python

"""
Generate histograms of spike theta-phase for pre/post interpolation fix.
"""

import os

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d

from spc import PROJECT_ROOT, store
from spc.output import log
from spc.lib.lfp import ContinuousData
from spc.lib.cluster import ClusterData
from spc.lib.parsers import parse_ttc
from spc.tools import maps, paths


############## Set up output directories ###############

testdir = os.path.split(__file__)[0]
outputdir = os.path.join(testdir, 'phase-interp-histos')


############## Plotting state and functions ###############

interactive = False

cur = 0
f = None
axlist = []
rows = 5
cols = 3
figsize = (8.5,11)
hist_bins = np.linspace(-np.pi, np.pi, 64)

def new_figure(rows=rows, cols=cols, figsize=figsize):
    global f, axlist, cur
    cur = 0
    f, axlist = plt.subplots(nrows=rows, ncols=cols, figsize=figsize)
    axlist = axlist.flatten()
    [ax.set_axis_off() for ax in axlist]
    return f

def next_axis():
    global cur
    if cur == len(axlist):
        if f is not None:
            save_and_close('histos')
        new_figure()
    ax = axlist[cur]
    ax.set_axis_on()
    cur += 1
    return ax

def save_and_close(stem='figure'):
    f.set_tight_layout(True)
    if interactive:
        log('Displayed: {} ({})', stem, f.number)
        return
    stem = os.path.join(outputdir, stem)
    head, tail = os.path.split(stem)
    if not os.path.isdir(head):
        os.makedirs(head)
    fn = paths.uniquify(stem, ext='pdf', fmt='%s-%02d')
    plt.savefig(fn)
    log('Saved: {}', fn)
    plt.close()

plt.ioff()


############## Helper functions ###############

def quicktitle(ax, title):
    if title:
        ax.text(0.5, 1.0, title, transform=ax.transAxes,
                size='small', ha='center', va='bottom')

def plot_histo(pre, post, label='', **kwargs):
    ax = next_axis()
    hkw = dict(bins=hist_bins)
    hkw.update(kwargs)
    for data, color in [(pre, 'b'), (post, 'r')]:
        hkw.update(color=color)
        ax.hist(data, histtype='stepfilled', zorder=-1, alpha=0.3, **hkw)
        hkw.update(color='k')
        ax.hist(data, histtype='step', linewidth=1, **hkw)
    [ax.spines[s].set_visible(False) for s in ('top', 'right', 'left')]
    ax.tick_params(top=False, right=False, left=False)
    ax.set(xlim=(-4,4), xticklabels=[], yticklabels=[])
    quicktitle(ax, label)

def circinterp(cdata, lfp, **ikw):
    uphase = np.unwrap(lfp.phase)
    Fu = interp1d(lfp.t, uphase, **ikw)
    sphase_fixed = Fu(cdata.spikes) % (2 * np.pi)
    sphase_fixed[sphase_fixed > np.pi] -= 2 * np.pi
    return sphase_fixed


############## Script for the generation of many maps ###############

df = store.get(True)

for rec in df.root.recordings.iterrows():
    s_id = rec['session_id']
    spath = df.root.sessions[s_id]['group']

    ttc = parse_ttc(rec)
    log('Plotting {} from session {}...', rec['ttc'], s_id)

    lfp = ContinuousData(spath)
    cdata = ClusterData(spath, ttc)
    ikw = dict(copy=False, bounds_error=False, fill_value=np.nan)
    F_phase = interp1d(lfp.t, lfp.phase, **ikw)
    sphase = F_phase(cdata.spikes)

    sphase_fixed = lfp.F('phase', cdata.spikes)
    #sphase_fixed = circinterp(cdata, lfp, **ikw)

    plot_histo(sphase, sphase_fixed,
        label='[%d] %s' % (s_id, rec['ttc']))

save_and_close('histos')
store.close()


############## Draw anything left back to the screen ###############

plt.ion()
if interactive:
    plt.draw()
    plt.show()
