#!/usr/bin/env python

"""
Convert SVG Tiny 1.1 document with rectangles to an Axes list

Written by Joe Monaco, 03-01-2012.
Updated, 05-20-2015.
Copyright (c) 2012-2015 Johns Hopkins University. All rights reserved.
"""

import os
import sys
import time
import subprocess

import numpy as np


DPI = 72
SCRIPT = sys.argv[0].split(os.path.sep)[-1]
USAGE = """Usage: %s <svg_file> [<output_file>]

This script creates a boilerplate core.figure.BaseFigure subclass based on
plot axes whose rectangles are specified by a SVG file (Tiny 1.1
specification). The SVG input file can be exported from Illustrator after
laying out the plot rectangles. Optionally, *name* and *projection* keywords
can be added to the <rect/> nodes in the SVG file that will cause names and
projections to be properly set up in the generated subclass. If not
specified, single-letter names are provided starting with 'A', and sequenced
according to the order of <rect/> nodes. The default projection is the
standard rectilinear axis."""%SCRIPT


HEADER = """#!/usr/bin/env python
#encoding: utf-8

\"\"\"
Create a multi-paneled matplotlib figure from SVG data.
\"\"\"

import matplotlib.pyplot as plt

"""


def parse(svg_file):
    svg = svg_file.read()

    descrs = []
    i = 0
    while i < len(svg):
        if svg[i] != "<":
            i += 1
            continue
        tag_end = svg.find(' ', i)
        tag = svg[i+1:tag_end]
        if tag[0] in '?!/':
            i += 1
            continue
        s = 'type="%s"'%tag
        i = tag_end
        inquote = False
        while svg[i] not in '/>' or inquote:
            if svg[i] == '"':
                inquote = not inquote
            if inquote and svg[i] == ' ':
                s += '_'
            elif not inquote and svg[i] in ':-':
                s += '_'
            else:
                s += svg[i]
            i += 1
        descrs.append(s.split())

    nodes = []
    for attrs in descrs:
        node = eval('dict(%s)'%(','.join(attrs)))
        for key in node:
            if node[key].endswith('px'):
                node[key] = node[key][:-2]
            try:
                num = float(node[key])
            except ValueError:
                pass
            else:
                node[key] = num
        nodes.append(node)
    return nodes

def write(nodes, out):
    # Get page dimensions
    svg = filter(lambda x: x['type'] == 'svg', nodes)[0]
    pagewidth = svg['width'] / DPI
    pageheight = svg['height'] / DPI

    # Get plot rectangles positions and dimensions
    rects = filter(lambda x: x['type'] == 'rect', nodes)
    toplefts = []
    for rect in rects:
        rect['x'] /= svg['width'] # xform to [0,1] figure coordinates...
        rect['width'] /= svg['width']
        rect['y'] /= svg['height']
        toplefts.append((rect['x'], rect['y']))
        rect['height'] /= svg['height']
        rect['y'] = 1 - rect['y'] - rect['height'] # invert, get bottom

    # Sort rectangle nodes by (top, left) corner position
    sortix = np.argsort(np.array(toplefts, dtype=np.dtype([('top', 'd'),
        ('left', 'd')])), order=('left', 'top'))
    rects = [rects[i] for i in sortix]

    # Create the position and projection definitions
    ax_str = []
    autoname = 65 # start at 'A'
    for rect in rects:
        if 'name' not in rect:
            rect['name'] = str(bytearray([autoname]))
            autoname += 1
        proj = ''
        if 'projection' in rect:
            proj = ', polar=True'
        pos = '[%(x).3f, %(y).3f, %(width).3f, %(height).3f]'%rect
        ax_str.append('ax[\'%s\'] = plt.axes(%s%s)' % (
            rect['name'], pos, proj))
    ax_str.insert(0, 'ax = {}')

    print >>out, HEADER
    print >>out, 'plt.ioff()'
    print >>out, 'fig = plt.figure(figsize=(%.1f, %.1f))\n'%(
            pagewidth, pageheight)
    print >>out, '\n'.join(ax_str)
    print >>out, '\nplt.ion()\nplt.show()'
    print >>out, '\n# prevent immediate exit in shell\nraw_input()'


if __name__ == "__main__":
    if not 1 < len(sys.argv) < 4:
        print >>sys.stderr, USAGE
        sys.exit(1)

    svg_fn = sys.argv[1]
    if not svg_fn.endswith('.svg'):
        print '%s: Input file does not end in .svg'%SCRIPT
        sys.exit(0)

    if len(sys.argv) == 3:
        out_fn = os.path.abspath(sys.argv[2])
        out_fd = file(out_fn, 'w')
    elif len(sys.argv) == 2:
        out_fn = None
        out_fd = sys.stdout

    svg_fd = file(svg_fn, 'r')

    write(parse(svg_fd), out_fd)

    svg_fd.close()
    if out_fn is not None:
        out_fd.close()
        subprocess.call(['chmod', '+x', out_fn])
