"""
Tests for quantifying theta rhythmicity (as opposed to locking/modulation).
"""

from __future__ import division

import os
import numpy as np
import matplotlib.pyplot as plt

from spc import store
from spc.lib import labels, session
from spc.tools import spikes, arrays, paths
from spc.lib.theta import rhythmicity_estimate


outdir = os.path.splitext(__file__)[0]
if not os.path.isdir(outdir):
    os.makedirs(outdir)


c_id = 22
maxlag = 0.25
bins = 128
tau_blur = 0.002
tau_blur_inc = 0.001
tau_blur_max = 0.020
theta_frac = 0.75
H = 2/3

df = store.get(True)
cell = df.root.recordings[c_id]
lags = df.root.theta.autocorr.lags.read()
C = getattr(df.root.theta.autocorr, 'id%04d' % c_id).read()

tag = labels.session_cell_id(c_id)
print 'Correlation for {}'.format(tag)

f_theta = 8.0
tau_theta = theta_frac / f_theta
taulim = (1 - H) * tau_theta, (1 + H) * tau_theta

posix = lags > 0
lags = lags[posix]
C = C[posix]

check = lambda mn, mx: (len(mn) == 1 and len(mx) == 1) and mn[0] < mx[0]

mins = maxs = []
tix = np.logical_and(lags>=taulim[0], lags<=taulim[1])
while not check(mins, maxs) and tau_blur <= tau_blur_max:
    print 'Trying blur = {}'.format(tau_blur)
    Cg = arrays.blur(C, pixels=(bins/(2*maxlag))*tau_blur, padding='same')

    mins = arrays.minima(Cg[tix])
    maxs = arrays.maxima(Cg[tix])

    print 'Found {} minima'.format(len(mins))
    print 'Found {} maxima'.format(len(maxs))

    tau_blur = round(tau_blur + tau_blur_inc, 5)

assert check(mins, maxs), 'failed to find theta trough and peak'

lag_trough = lags[tix][mins[0]]
lag_peak = lags[tix][maxs[0]]
C_trough = Cg[tix][mins[0]]
C_peak = Cg[tix][maxs[0]]
print 'Trough (t = {:.2f}) = {}'.format(lag_trough, C_trough)
print 'Peak (t = {:.2f}) = {}'.format(lag_peak, C_peak)

index = (C_peak - C_trough) / C_peak
print 'Rhythmicity index = {}'.format(index)

fig = plt.figure(10, (9,7))
fig.clear()
ax = fig.add_subplot(111)
spikes.corrplot((C, lags), alpha=0.3)
spikes.corrplot((C[tix], lags[tix]), style='lines', c='b', marker='.')
spikes.corrplot((Cg[tix], lags[tix]), style='lines', c='r', alpha=0.6)

ax.axvline(lag_trough, c='r', lw=1.5)
ax.axvline(lag_peak, c='r', lw=1.5)
ax.set_xlim(-0.5 * maxlag, maxlag)
s = C[tix].ptp()
ax.set_ylim(C[tix].min() - s, C[tix].max() + s)
ax.set_title('{}: Index = {}'.format(tag, index))

savefn = paths.uniquify(os.path.join(outdir, 'corrplot'), ext='png', fmt='%s-%03d')
plt.savefig(savefn)
print 'Saved: {}'.format(savefn)

plt.draw()
plt.show()

print '--- spc.lib.theta.rhythmicity_estimate ---'

index = rhythmicity_estimate((C, lags), maxlag=maxlag, bins=bins)
print 'Rhythmicity index = {}'.format(index)
