"""
Test method for reclocking phase based on TT9 cell in Theta11 dataset.
"""

import os

import numpy as np
from scipy.signal import fftconvolve, firwin, filtfilt, hilbert
import matplotlib.pyplot as plt
import seaborn as sns
sns.reset_orig()

from spc import store
from spc.lib import cell, labels
from spc.tools import spikes, paths

def safe_hilbert(x):
    N = x.size
    M = pow(2, int(np.ceil(np.log2(N))))
    return hilbert(np.r_[x, np.zeros(M - N)])[:N]


outdir = os.path.splitext(__file__)[0]
if not os.path.isdir(outdir):
    os.makedirs(outdir)


t9label = '110900'
rec_id = 0
Fs_est = 60.0
Fs_corr = 250.0
bp_ntaps = 69
savefig = False

df = store.get()
t9recs = df.root.recordings.get_where_list('cell_id=="%s"' % t9label)
assert rec_id < len(t9recs), 'only %s TT9 recordings' % len(t9recs)

c_id = t9recs[rec_id]
C = df.root.recordings[c_id]
S = cell.session_for(c_id)
tag = labels.session_cell_id(c_id)
st = S.get_cluster(C).spikes

tlim = S.tstart, S.tstart + 500.0  #S.tend
duration = tlim[1] - tlim[0]
st = st[np.logical_and(st >= tlim[0], st <= tlim[1])]

ratefn = os.path.join(outdir, 'r_hat-%s-%02d.npy' % (t9label, rec_id))
if os.path.isfile(ratefn):
    print 'Loading previous {} estimates...'.format(t9label)
    t_est, r_hat = np.load(ratefn)
else:
    print 'Computing {} estimates...'.format(t9label)
    t_est = np.linspace(tlim[0], tlim[1], Fs_est * duration)
    r_est = spikes.FiringRateEstimate(st, duration, width=0.075, Fs_norm=0.0)
    r_hat = r_est(t_est).squeeze()
    r_hat /= r_hat.max()
    np.save(ratefn, np.c_[t_est, r_hat].T)
    print 'Saved estimate: {}'.format(ratefn)

acorrfn = os.path.join(outdir, 'acorr-%s-%02d.npy' % (t9label, rec_id))
if os.path.isfile(acorrfn):
    print 'Loading previous {} autocorr...'.format(t9label)
    autocorr, lags = np.load(acorrfn)
else:
    print 'Computing {} autocorr...'.format(t9label)
    maxlag = 3.0
    corrbins = 2 * maxlag * Fs_corr
    autocorr, lags = spikes.acorr(st, maxlag=maxlag, bins=corrbins)
    baseline = autocorr[lags >= maxlag - 1.0].mean()
    autocorr -= baseline
    autocorr /= autocorr.sum()
    np.save(acorrfn, np.c_[autocorr, lags].T)
    print 'Saved autocorr: {}'.format(acorrfn)

print 'Computing FFT convolution...'
bins = round(Fs_corr * duration)
stbin = np.histogram(st, bins=bins, range=tlim)[0]
t_match = np.linspace(tlim[0], tlim[1], bins)
y = fftconvolve(stbin, autocorr, mode='same')
y -= y.min()
y /= y.max()
y -= y.mean()

# Create a theta bandpass filter
print 'Performing theta bandpass filter...'
b = firwin(bp_ntaps, [4, 12], pass_zero=False, nyq=Fs_corr/2)
yf = filtfilt(b, [1], y)

print 'Computing Hilbert transform for phase...'
H = safe_hilbert(yf)
phase = np.arctan2(np.imag(H), yf)

# --- library validation ---

print 'Running library method validation...'
phi_est = spikes.FiringPhaseEstimate(st, tlim=tlim)
ts, libphase = phi_est.compute()

# --- plotting ---

print 'Plotting...'
plt.ioff()

fig = plt.figure(15, (10,6))
fig.clear()
ax = fig.add_subplot(111)

ax.plot(t_est, r_hat, 'r-', lw=1.2, alpha=0.8)
ax.plot(t_match, y, 'b-', lw=0.8, alpha=0.4)
ax.plot(t_match, yf, 'b-', lw=1.4, alpha=0.7)
ax.plot(t_match, phase/(8*np.pi) - 0.5, 'g-', lw=1.5)
ax.plot(ts, libphase/(8*np.pi) - 0.5, 'm--', lw=1.2)

ax.axhline(c='k')
ax.set(xlabel='Time (s)', ylabel='Est. rate / Conv / BP / Phase')
ax.set_xlim(112, 120)
ax.set_ylim(-0.7, 1.0)
ax.set_title('Reference: {}'.format(tag))

if savefig:
    savefn = paths.uniquify(os.path.join(outdir, 'tt9-reclock'),
        ext='png', fmt='%s-%03d')
    plt.savefig(savefn)
    print 'Saved: {}'.format(savefn)

fig = plt.figure(16, (8,8))
fig.clear()
ax = fig.add_subplot(111)
spikes.corrplot((autocorr, lags), colors='b')

plt.draw()
plt.show()

store.close()
